unit abcda;

(*------------------------------------------------------------------------------

Advanced Business Components
ABC for Delphi
Version 6

Copyright  1995, 2001 Objective Software Technology Pty Limited ACN 068 640 353
Functionality derived from Delphi VCL Source Code
is Copyright  1995, 2001 Borland Software Corporation

------------------------------------------------------------------------------*)

{ Dataset Adapter base class and Class Factory for ABC }

{
Note for implementers of third party dataset adapters:

  A descendant class of TabcDatasetAdapter is created for each dataset descendant.
  e.g. TabcTableAdapter, TabcQueryAdapter

  You must override TabcDatasetAdapter
    class function DatasetClass;

  You must register each adapter class using
    abcDSAdapterFactory.RegisterAdapter(AnAdapterClass: TabcDatasetAdapterClass);
  e.g. abcDSAdapterFactory.RegisterAdapter(TabcTableAdapter);

  You must unregister each adapter class using
    abcDSAdapterFactory.UnRegisterAdapter(AnAdapterClass: TabcDatasetAdapterClass);
  e.g. abcDSAdapterFactory.UnRegisterAdapter(TabcTableAdapter);

  You can override any other virtual methods in TabcDatasetAdapter.

  An adapter does not have to override all methods, the controls that use the
  adapter are responsible for querying the interface to see if a capability is supported.

  See TabcTableAdapter and TabcQueryAdapter for examples.

  If you want to register adapters for intermediate dataset classes,
    e.g. a TabcBDEDatasetAdapter
  you must register it before any descendant dataset class adapter,
    i.e. before registering TabcTableAdapter.
  This is because the class factory does a linear search on known dataset classes
  when creating an adapter and returns the most recently registered adapter class
  where dataset IS (adapter dataset class).

}

{
Note for component writers using adapters:

  The adapters are lightweight objects created by a component for abstracting access to
  TDataset descendant class methods.

  Adapters are created on a dataset instance using the class factory:
    MyAdapter := abcDSAdapterFactory.CreateAdapter(MyDataset);

  Important:
    The factory does not maintain notifications about datasets and adapters.
    The component creating the adapter must free the adapter when it is no longer required.

  A typical component usage would be:

    Dataset property.
    FDataset, FAdapter fields.
    SetDataset method frees FAdapter, then creates new FAdapter using class factory
    or sets FAdapter to nil.
    Notification override detects when dataset is freed and calls SetDataset.
    Other component methods use FDataset and FAdapter, checking whether FDataset is assigned.

  The class factory will always return a default TabcDatasetAdapter
    if there is no concrete adapter class for the dataset class.

}

{$I abcdefs.inc}

interface

uses Classes, Db;

type

  TabcDatasetAdapterReg = class(TComponent);

  TabcDatasetClass = class of TDataset;

  { TabcDatasetAdapter base class for TDataset }

  TabcSearchMethod = (asmFindKey, asmLocate);
  TabcFilterFieldDelimiter = (affNone, affSquareBracket);
  TabcFilterPartialMatch = (afpNone, afpAsterisk);

  TabcDatasetAdapter = class(TObject)
  private
    FDataset: TDataset;
  protected
    procedure AddStrIndent(const Str: string; Indent: Integer; const List: TStrings);

    { Dataset Descriptive Name to Show User }
    { e.g. TableName or similar }
    function GetDescription: string; virtual;

    { Empty Table }
    function IsEmptyTableEnabled: Boolean; virtual;

    { SQL Transactions }
    function IsTransactionEnabled: Boolean; virtual;
    function IsStartTransactionEnabled: Boolean; virtual;
    function IsCommitEnabled: Boolean; virtual;
    function IsRollbackEnabled: Boolean; virtual;

    { Indexes }
    function IsIndexEnabled: Boolean; virtual;
    function GetIndexDefs: TIndexDefs; virtual;
    procedure SetIndexDefs(const Value: TIndexDefs); virtual;
    function GetIndexField(Index: Integer): TField; virtual;
    procedure SetIndexField(Index: Integer; const Value: TField); virtual;
    function GetIndexName: string; virtual;
    procedure SetIndexName(const Value: string); virtual;
    function GetIndexFieldNames: string; virtual;
    procedure SetIndexFieldNames(const Value: string); virtual;
    function GetIndexFieldCount: Integer; virtual;
    function GetIndexValue: string; virtual;

    { Range Filters }
    function IsRangeEnabled: Boolean; virtual;
    function GetKeyExclusive: Boolean; virtual;
    procedure SetKeyExclusive(Value: Boolean); virtual;

    { Find }
    function IsFindKeyEnabled: Boolean; virtual;
    function IsFindNearestEnabled: Boolean; virtual;
    function IsLocateEnabled: Boolean; virtual;

    function GetSearchMethod: TabcSearchMethod; virtual;

    { Query Params }
    function IsParamsEnabled: Boolean; virtual;
    function GetParams: TParams; virtual;
    procedure SetParams(Value: TParams); virtual;
    function GetParamCount: Integer; virtual;
    { Dataset Filters }
    function IsFilterEnabled: Boolean; virtual;
    function IsSingleFilter: Boolean; virtual;
    function GetFilterText: string; virtual;
    function GetFiltered: Boolean; virtual;
    function GetFilterGoToFirst: Boolean; virtual;
    function GetFilterOptions: TFilterOptions; virtual;
    function GetFilterPriority: Word; virtual;
    function GetFilterFieldDelimiter: TabcFilterFieldDelimiter; virtual;
    function GetFilterPartialMatch: TabcFilterPartialMatch; virtual;
    function GetOnActivateFilter: TNotifyEvent; virtual;
    function GetOnDeactivateFilter: TNotifyEvent; virtual;
    function GetOnFilterRecord: TFilterRecordEvent; virtual;
    procedure SetFilterText(Value: string); virtual;
    procedure SetFiltered(Value: Boolean); virtual;
    procedure SetFilterGoToFirst(Value: Boolean); virtual;
    procedure SetFilterOptions(const Value: TFilterOptions); virtual;
    procedure SetFilterPriority(const Value: Word); virtual;
    procedure SetOnActivateFilter(const Value: TNotifyEvent); virtual;
    procedure SetOnDeactivateFilter(const Value: TNotifyEvent); virtual;
    procedure SetOnFilterRecord(const Value: TFilterRecordEvent); virtual;

    {Table/Query ReadOnly}
    function IsReadOnlyEnabled: Boolean; virtual;
    function GetReadOnly: Boolean; virtual;
    procedure SetReadOnly(Value: Boolean); virtual;

    {NetUserName}
    function GetNetUserName: string; virtual;

  public
    constructor Create(const ADataset: TDataset); virtual;
    class function DatasetClass: TabcDatasetClass; virtual;

    { Debugging Information }
    procedure GetDatasetInfo(const Strings: TStrings); virtual;
    procedure GetFieldInfo(const Strings: TStrings); virtual;
    procedure GetIndexInfo(const Strings: TStrings); virtual;
    procedure GetParamInfo(const Strings: TStrings); virtual;
    procedure GetDatabaseInfo(const Strings: TStrings); virtual;
    procedure GetSessionInfo(const Strings: TStrings); virtual;
    procedure GetDBEngineInfo(const Strings: TStrings); virtual;

    { Empty Table }
    procedure EmptyTable; virtual;

    { SQL Transactions }
    procedure StartTransaction; virtual;
    procedure Commit; virtual;
    procedure Rollback; virtual;

    { Indexes }
    procedure GetIndexNames(const Strings: TStrings); virtual;

    { Range Filters }
    procedure ApplyRange; virtual;
    procedure CancelRange; virtual;
    procedure EditRangeStart; virtual;
    procedure EditRangeEnd; virtual;
    procedure SetRange(const StartValues, EndValues: Variant); virtual;
    procedure SetRangeStart; virtual;
    procedure SetRangeEnd; virtual;

    { Find Key }
    procedure EditKey; virtual;
    function FindKey(const KeyValues: Variant): Boolean; virtual;
    procedure FindNearest(const KeyValues: Variant); virtual;
    procedure GoToKey; virtual;
    procedure GoToNearest; virtual;
    procedure SetKey; virtual;

    { Locate }
    function Locate(const KeyFields: string; const KeyValues: Variant;
      Options: TLocateOptions): Boolean; virtual;


    { Dataset Filter Buffer Params }
    function FilterParamByName(const FieldName: string): TParam; virtual;

    { Update Buffer }
    procedure FlushBuffers; virtual;

    { Properties }

    property CanEmptyTable: Boolean read IsEmptyTableEnabled;
    property CanUseTransactions: Boolean read IsTransactionEnabled;
    property CanStartTransaction: Boolean read IsStartTransactionEnabled;
    property CanCommit: Boolean read IsCommitEnabled;
    property CanRollback: Boolean read IsRollbackEnabled;
    property CanIndex: Boolean read IsIndexEnabled;
    property CanSetRange: Boolean read IsRangeEnabled;
    property CanFilter: Boolean read IsFilterEnabled;
    property CanFindKey: Boolean read IsFindKeyEnabled;
    property CanFindNearest: Boolean read IsFindNearestEnabled;
    property CanLocate: Boolean read IsLocateEnabled;
    property CanSetParams: Boolean read IsParamsEnabled;
    property CanSetReadOnly: Boolean read IsReadOnlyEnabled;
    property SearchMethod: TabcSearchMethod read GetSearchMethod;
    property SingleFilter: Boolean read IsSingleFilter;
    property Description: string read GetDescription;
    property Filter: string read GetFilterText write SetFilterText;
    property Filtered: Boolean read GetFiltered write SetFiltered;
    property FilterGoToFirst: Boolean read GetFilterGoToFirst write SetFilterGoToFirst;
    property FilterOptions: TFilterOptions read GetFilterOptions write SetFilterOptions;
    property FilterPriority: Word read GetFilterPriority write SetFilterPriority;
    property FilterFieldDelimiter: TabcFilterFieldDelimiter read GetFilterFieldDelimiter;
    property FilterPartialMatch: TabcFilterPartialMatch read GetFilterPartialMatch;
    property IndexDefs: TIndexDefs read GetIndexDefs write SetIndexDefs;
    property IndexFieldNames: string read GetIndexFieldNames write SetIndexFieldNames;
    property IndexFieldCount: Integer read GetIndexFieldCount;
    property IndexFields[Index: Integer]: TField read GetIndexField write SetIndexField;
    property IndexName: string read GetIndexName write SetIndexName;
    property IndexValue: string read GetIndexValue;
    property KeyExclusive: Boolean read GetKeyExclusive write SetKeyExclusive;
    property NetUserName: string read GetNetUserName;
    property Params: TParams read GetParams write SetParams;
    property ParamCount: Integer read GetParamCount;
    property ReadOnly: Boolean read GetReadOnly write SetReadOnly;
    property OnActivateFilter: TNotifyEvent read GetOnActivateFilter write SetOnActivateFilter;
    property OnDeactivateFilter: TNotifyEvent read GetOnDeactivateFilter write SetOnDeactivateFilter;
    property OnFilterRecord: TFilterRecordEvent read GetOnFilterRecord write SetOnFilterRecord;
  end;

  TabcDatasetAdapterClass = class of TabcDatasetAdapter;

  { TabcDSAdapterFactory - single instance global factory to create adapters for datasets}

  TabcDSAdapterFactory = class(TObject)
  private
    FAdapters: TList;
    FClassTypes: TList;
  public
    constructor Create;
    destructor Destroy; override;
    procedure RegisterAdapter(AnAdapterClass: TabcDatasetAdapterClass);
    procedure UnRegisterAdapter(AnAdapterClass: TabcDatasetAdapterClass);
    function GetAdapterClass(ADatasetClass: TabcDatasetClass): TabcDatasetAdapterClass;
    function GetAdapterClassDefault(ADatasetClass: TabcDatasetClass): TabcDatasetAdapterClass;
    function CreateAdapter(ADataset: TDataset): TabcDatasetAdapter;  //NB Objects are not destroyed, creator must free them when dataset is dereferenced.
  end;

function abcDSAdapterFactory: TabcDSAdapterFactory;

implementation

uses abcnpcon, SysUtils;

 { TabcDatasetAdapter }

constructor TabcDatasetAdapter.Create(const ADataset: TDataset);
begin
  FDataset := ADataset;
end;

class function TabcDatasetAdapter.DatasetClass: TabcDatasetClass;
begin
  Result := TDataset;
end;

function TabcDatasetAdapter.GetDescription: string;
begin
  Result := FDataset.Name;
end;

{ Debugging Information }

procedure TabcDatasetAdapter.GetDatasetInfo(const Strings: TStrings);
var
  Indent: Integer;
  I: integer;
  LongestFieldName: Integer;
  TempStr: string;
begin
  Indent := 0;
  AddStrIndent(sabcnpiDS,Indent,Strings);
  Inc(Indent);
  with FDataSet do begin
    AddStrIndent(sabcnpiNameEq + Name, Indent, Strings);
    AddStrIndent(sabcnpiClassNameEq + ClassName, Indent, Strings);
    AddStrIndent(sabcnpiActive + ABC_BOOLEANS[Active], Indent, Strings);
    AddStrIndent(sabcnpiAutoCalc + ABC_BOOLEANS[AutoCalcFields], Indent, Strings);
    AddStrIndent(sabcnpiBOF + ABC_BOOLEANS[BOF], Indent, Strings);
    AddStrIndent(sabcnpiEOF + ABC_BOOLEANS[EOF], Indent, Strings);
    AddStrIndent(sabcnpiCanModify + ABC_BOOLEANS[CanModify], Indent, Strings);
    if DataSource = nil then
      AddStrIndent(sabcnpiDSNil, Indent, Strings)
    else
      AddStrIndent(sabcnpiDSEq + DataSource.Name, Indent, Strings);
    AddStrIndent(sabcnpiDefaultFields + ABC_BOOLEANS[DefaultFields], Indent, Strings);
    AddStrIndent(sabcnpiFieldCount + IntToStr(FieldCount), Indent, Strings);
    AddStrIndent(sabcnpiModified + ABC_BOOLEANS[Modified], Indent, Strings);
    AddStrIndent(sabcnpiRecordCount + IntToStr(RecordCount), Indent, Strings);
    AddStrIndent(sabcnpiRecordSize + IntToStr(RecordSize), Indent, Strings);
    AddStrIndent(sabcnpiState + ABC_DATASETSTATES[State],Indent,Strings);
    AddStrIndent('',0,Strings);

    if FieldCount > 0 then begin
      AddStrIndent(sabcnpiFieldSummary, Indent, Strings);
      Inc(Indent);
      LongestFieldName := 0;
      for I := 0 to FieldCount - 1 do
        with Fields[I] do
          if Length(FieldName) > LongestFieldName then
            LongestFieldName := Length(FieldName);
      for I := 0 to FieldCount - 1 do
        with Fields[I] do begin
          TempStr := FieldName;
          while (Length(TempStr) < LongestFieldName) do
            TempStr := TempStr + sabcnpiSpace;
          TempStr := TempStr + sabcnpiSpace + ABC_FMTFIELDTYPES[DataType]
            + Format('%7d', [Size]);
          if Required then
            TempStr := TempStr + sabcnpiReqd
          else
            TempStr := TempStr + sabcnpiNotReqd;
          if ReadOnly then
            TempStr := TempStr + sabcnpiRO
          else
            TempStr := TempStr + sabcnpiNotRO;
          if IsIndexField then
            TempStr := TempStr + sabcnpiIndexed;
          AddStrIndent(TempStr, Indent, Strings);
        end;
      AddStrIndent('',0,Strings);
    end;

  end;
  AddStrIndent('',0,Strings);

end;

procedure TabcDatasetAdapter.GetFieldInfo(const Strings: TStrings);
var
  Indent: Integer;
  I: Integer;
begin
  Indent := 0;
  AddStrIndent(sabcnpiFields,Indent,Strings);
  AddStrIndent('',0,Strings);
  Inc(Indent);
  with FDataSet do begin
    AddStrIndent(sabcnpiFieldCount + InttoStr(FieldCount), Indent, Strings);
    AddStrIndent('',0,Strings);
    Inc(Indent);
    for I := 0 to FieldCount - 1 do
      with Fields[I] do begin
        AddStrIndent(sabcnpiAsterisk, Indent, Strings);
        AddStrIndent(sabcnpiAsFieldName + FieldName, Indent, Strings);
        AddStrIndent(sabcnpiAsFieldNo + IntToStr(FieldNo), Indent, Strings);
        AddStrIndent(sabcnpiAsDataType + ABC_FIELDTYPES[DataType], Indent, Strings);
        AddStrIndent(sabcnpiAsClass + ClassName, Indent, Strings);
        AddStrIndent(sabcnpiAsSize + IntToStr(Size), Indent, Strings);
        AddStrIndent(sabcnpiAsterisk, Indent, Strings);
        Inc(Indent);
        AddStrIndent(sabcnpiAlignment + ABC_ALIGNMENTS[Alignment], Indent, Strings);
        AddStrIndent(sabcnpiCalculated + ABC_BOOLEANS[Calculated], Indent, Strings);
        AddStrIndent(sabcnpiCanModify + ABC_BOOLEANS[CanModify], Indent, Strings);
        AddStrIndent(sabcnpiDataSize + IntToStr(DataSize), Indent, Strings);
        AddStrIndent(sabcnpiDisplayLabel + DisplayLabel, Indent, Strings);
        AddStrIndent(sabcnpiDisplayName + DisplayName, Indent, Strings);
        AddStrIndent(sabcnpiDisplayText + DisplayText, Indent, Strings);
        AddStrIndent(sabcnpiDisplayWidth + IntToStr(DisplayWidth), Indent, Strings);
        AddStrIndent(sabcnpiEditMask + EditMask, Indent, Strings);
        AddStrIndent(sabcnpiIndex + IntToStr(Index), Indent, Strings);
        AddStrIndent(sabcnpiIsIndexField + ABC_BOOLEANS[IsIndexField], Indent, Strings);
        AddStrIndent(sabcnpiReadOnly + ABC_BOOLEANS[ReadOnly], Indent, Strings);
        AddStrIndent(sabcnpiRequired + ABC_BOOLEANS[Required], Indent, Strings);
        AddStrIndent(sabcnpiVisible + ABC_BOOLEANS[Visible], Indent, Strings);
        Dec(Indent);
        AddStrIndent('',0,Strings);
      end;
  end;
  AddStrIndent('',0,Strings);
end;

procedure TabcDatasetAdapter.GetIndexInfo(const Strings: TStrings);
begin
end;

procedure TabcDatasetAdapter.GetParamInfo(const Strings: TStrings);
begin
end;

procedure TabcDatasetAdapter.GetDatabaseInfo(const Strings: TStrings);
begin
end;

procedure TabcDatasetAdapter.GetSessionInfo(const Strings: TStrings);
begin
end;

procedure TabcDatasetAdapter.GetDBEngineInfo(const Strings: TStrings);
begin
end;

{ Empty Table }

function TabcDatasetAdapter.IsEmptyTableEnabled: Boolean;
begin
  Result := False;
end;

procedure TabcDatasetAdapter.EmptyTable;
begin
end;

{ SQL Transactions }

function TabcDatasetAdapter.IsTransactionEnabled: Boolean;
begin
  Result := False;
end;

function TabcDatasetAdapter.IsStartTransactionEnabled: Boolean;
begin
  Result := False;
end;

function TabcDatasetAdapter.IsCommitEnabled: Boolean;
begin
  Result := False;
end;

function TabcDatasetAdapter.IsRollbackEnabled: Boolean;
begin
  Result := False;
end;

procedure TabcDatasetAdapter.StartTransaction;
begin
end;

procedure TabcDatasetAdapter.Commit;
begin
end;

procedure TabcDatasetAdapter.Rollback;
begin
end;

{ Indexes }

function TabcDatasetAdapter.IsIndexEnabled: Boolean;
begin
  Result := False;
end;

function TabcDatasetAdapter.GetIndexDefs: TIndexDefs;
begin
  Result := nil;
end;

procedure TabcDatasetAdapter.SetIndexDefs(const Value: TIndexDefs);
begin
end;

function TabcDatasetAdapter.GetIndexField(Index: Integer): TField;
begin
  Result := nil;
end;

procedure TabcDatasetAdapter.SetIndexField(Index: Integer; const Value: TField);
begin
end;

procedure TabcDatasetAdapter.GetIndexNames(const Strings: TStrings);
begin
end;

function TabcDatasetAdapter.GetIndexName: string;
begin
  Result := '';
end;

procedure TabcDatasetAdapter.SetIndexName(const Value: string);
begin
end;

function TabcDatasetAdapter.GetIndexFieldNames: string;
begin
  Result := '';
end;

procedure TabcDatasetAdapter.SetIndexFieldNames(const Value: string);
begin
end;

function TabcDatasetAdapter.GetIndexFieldCount: Integer;
begin
  Result := 0;
end;

function TabcDatasetAdapter.GetIndexValue: string;
begin
  Result := '';
end;

{ Range Filters }

function TabcDatasetAdapter.IsRangeEnabled: Boolean;
begin
  Result := False;
end;

procedure TabcDatasetAdapter.ApplyRange;
begin
end;

procedure TabcDatasetAdapter.CancelRange;
begin
end;

procedure TabcDatasetAdapter.EditRangeStart;
begin
end;

procedure TabcDatasetAdapter.EditRangeEnd;
begin
end;

procedure TabcDatasetAdapter.SetRange(const StartValues, EndValues: Variant);
begin
end;

procedure TabcDatasetAdapter.SetRangeStart;
begin
end;

procedure TabcDatasetAdapter.SetRangeEnd;
begin
end;

function TabcDatasetAdapter.GetKeyExclusive: Boolean;
begin
  Result := False;
end;

procedure TabcDatasetAdapter.SetKeyExclusive(Value: Boolean);
begin
end;

{ Find Key }

function TabcDatasetAdapter.IsFindKeyEnabled: Boolean;
begin
  Result := False;
end;

function TabcDatasetAdapter.IsFindNearestEnabled: Boolean;
begin
  Result := False;
end;

function TabcDatasetAdapter.FindKey(const KeyValues: Variant): Boolean;
begin
  Result := False;
end;

procedure TabcDatasetAdapter.FindNearest(const KeyValues: Variant);
begin
end;

procedure TabcDatasetAdapter.EditKey;
begin
end;

procedure TabcDatasetAdapter.GoToKey;
begin
end;

procedure TabcDatasetAdapter.GoToNearest;
begin
end;

procedure TabcDatasetAdapter.SetKey;
begin
end;

{ Locate }

function TabcDatasetAdapter.IsLocateEnabled: Boolean;
begin
{$IFDEF DELPHI6}
  Result := not FDataset.IsUniDirectional;
{$ELSE}
  Result := True;
{$ENDIF}
end;

function TabcDatasetAdapter.Locate(const KeyFields: string;
  const KeyValues: Variant; Options: TLocateOptions): Boolean;
begin
  if IsLocateEnabled then
      Result := FDataset.Locate(KeyFields, KeyValues, Options)
    else
      Result := False;
end;

function TabcDatasetAdapter.GetSearchMethod: TabcSearchMethod;
begin
  Result := asmLocate;
end;

{ Query Params }

function TabcDatasetAdapter.IsParamsEnabled: Boolean;
begin
  Result := False;
end;

{$IFDEF DELPHI5}
type
  THackDataset = class(TDataset);
{$ENDIF}

function TabcDatasetAdapter.GetParams: TParams;
begin
{$IFDEF DELPHI5}
  if IsParamsEnabled then
    Result := THackDataset(FDataset).PSGetParams
  else
{$ENDIF}
    Result := nil;
end;

procedure TabcDatasetAdapter.SetParams(Value: TParams);
begin
{$IFDEF DELPHI5}
  if IsParamsEnabled then
    THackDataset(FDataset).PSSetParams(Value);
{$ENDIF}
end;

function TabcDatasetAdapter.GetParamCount: Integer;
begin
  Result := 0;
end;

procedure TabcDatasetAdapter.AddStrIndent(
  const Str: string; Indent: Integer; const List: TStrings);
var
  IndentStr: string;
  I: integer;
begin
  IndentStr := Str;
  for I := 1 to Indent do
    IndentStr := sabcnpiIndent + IndentStr;
  List.Add(IndentStr);
end;

{ Dataset Filters }

function TabcDatasetAdapter.IsFilterEnabled: Boolean;
begin
{$IFDEF DELPHI6}
  Result := not FDataset.IsUniDirectional;
{$ELSE}
  Result := True;
{$ENDIF}
end;

function TabcDatasetAdapter.IsSingleFilter: Boolean;
begin
  Result := True;
end;

function TabcDatasetAdapter.GetFiltered: Boolean;
begin
  Result := FDataset.Filtered;
end;

function TabcDatasetAdapter.GetFilterGoToFirst: Boolean;
begin
  Result := False;
end;

function TabcDatasetAdapter.GetFilterOptions: TFilterOptions;
begin
  Result := FDataset.FilterOptions;
end;

function TabcDatasetAdapter.GetFilterPriority: Word;
begin
  Result := 0;
end;

function TabcDatasetAdapter.GetFilterText: string;
begin
  Result := FDataset.Filter;
end;

function TabcDatasetAdapter.GetOnActivateFilter: TNotifyEvent;
begin
  Result := nil;
end;

function TabcDatasetAdapter.GetOnDeactivateFilter: TNotifyEvent;
begin
  Result := nil;
end;

function TabcDatasetAdapter.GetOnFilterRecord: TFilterRecordEvent;
begin
  Result := FDataset.OnFilterRecord;
end;

function TabcDatasetAdapter.GetFilterFieldDelimiter: TabcFilterFieldDelimiter;
begin
  Result := affNone;
end;

function TabcDatasetAdapter.GetFilterPartialMatch: TabcFilterPartialMatch;
begin
  Result := afpNone;
end;

procedure TabcDatasetAdapter.SetFiltered(Value: Boolean);
begin
  if IsFilterEnabled then
    FDataset.Filtered := Value;
end;

procedure TabcDatasetAdapter.SetFilterGoToFirst(Value: Boolean);
begin
end;

procedure TabcDatasetAdapter.SetFilterOptions(const Value: TFilterOptions);
begin
  if IsFilterEnabled then
    FDataset.FilterOptions := Value;
end;

procedure TabcDatasetAdapter.SetFilterPriority(const Value: Word);
begin
end;

procedure TabcDatasetAdapter.SetFilterText(Value: string);
begin
  if IsFilterEnabled then
    FDataset.Filter := Value;
end;

procedure TabcDatasetAdapter.SetOnActivateFilter(
  const Value: TNotifyEvent);
begin
end;

procedure TabcDatasetAdapter.SetOnDeactivateFilter(
  const Value: TNotifyEvent);
begin
end;

procedure TabcDatasetAdapter.SetOnFilterRecord(
  const Value: TFilterRecordEvent);
begin
  if IsFilterEnabled then
    FDataset.OnFilterRecord := Value;
end;

{ Dataset Filter Buffer Params }

function TabcDatasetAdapter.FilterParamByName(const FieldName: string): TParam;
begin
  Result := nil;
end;

{Table/Query ReadOnly}

function TabcDatasetAdapter.IsReadOnlyEnabled: Boolean;
begin
  Result := False;
end;

function TabcDatasetAdapter.GetReadOnly: Boolean;
begin
  Result := FDataset.CanModify;
end;

procedure TabcDatasetAdapter.SetReadOnly(Value: Boolean);
begin
end;

{ Update Buffer }
procedure TabcDatasetAdapter.FlushBuffers;
begin
end;

{NetUserName}
function TabcDatasetAdapter.GetNetUserName: string;
begin
  Result := '';
end;

{------------------------------------------------------------------------------}

{ TabcDSAdapterFactory }

var
  GabcDSAdapterFactory: TabcDSAdapterFactory;

function abcDSAdapterFactory: TabcDSAdapterFactory;
begin
  if GabcDSAdapterFactory = nil then
    GabcDSAdapterFactory := TabcDSAdapterFactory.Create;
  Result := GabcDSAdapterFactory;
end;

constructor TabcDSAdapterFactory.Create;
begin
  FAdapters := TList.Create;
  FClassTypes := TList.Create;
end;

destructor TabcDSAdapterFactory.Destroy;
begin
  FAdapters.Free;
  FClassTypes.Free;
  inherited;
end;

procedure TabcDSAdapterFactory.RegisterAdapter(AnAdapterClass: TabcDatasetAdapterClass);
begin
  FAdapters.Insert(0, AnAdapterClass);
  FClassTypes.Insert(0, AnAdapterClass.DatasetClass);
end;

procedure TabcDSAdapterFactory.UnRegisterAdapter(AnAdapterClass: TabcDatasetAdapterClass);
var
  i: integer;
begin
  i := FAdapters.IndexOf(AnAdapterClass);
  if i <> -1 then
  begin
    FAdapters.Delete(i);
    FClassTypes.Delete(i);
  end;
end;

function TabcDSAdapterFactory.GetAdapterClass(ADatasetClass: TabcDatasetClass): TabcDatasetAdapterClass;
var
  AncestorClass: TabcDatasetClass;
  aClass: TClass;

  function FindAdapterClass(ADatasetClass: TabcDatasetClass): TabcDatasetAdapterClass;
  var
    i: integer;
  begin
    Result := nil;
    i := FClassTypes.IndexOf(ADatasetClass);
    if i <> -1 then
      Result := TabcDatasetAdapterClass(FAdapters[i])
    else
    for i := 0 to FAdapters.Count - 1 do
      if ADatasetClass = TabcDatasetAdapterClass(FAdapters[i]).DataSetClass then
      begin
        Result := TabcDatasetAdapterClass(FAdapters[i]);
        Exit;
      end;
  end;

begin
  Result := FindAdapterClass(ADatasetClass);
  if Result = nil then
  begin
    AncestorClass := ADatasetClass;
    repeat
      aClass := AncestorClass.ClassParent;
      if aClass.InheritsFrom(TDataset) then
        AncestorClass := TabcDatasetClass(aClass)
      else
        Exit;
      Result := FindAdapterClass(AncestorClass);
    until (Result <> nil) or (AncestorClass = TDataset);
  end;
end;

function TabcDSAdapterFactory.GetAdapterClassDefault(ADatasetClass: TabcDatasetClass): TabcDatasetAdapterClass;
begin
  Result := GetAdapterClass(ADatasetClass);
  if Result = nil then
    Result := TabcDatasetAdapter;
end;

function TabcDSAdapterFactory.CreateAdapter(ADataset: TDataset): TabcDatasetAdapter;
//NB Objects are not destroyed, creator must free them when dataset is dereferenced.
begin
  Result := GetAdapterClassDefault(TabcDatasetClass(ADataset.ClassType)).Create(ADataset);
end;

initialization

finalization
  GabcDSAdapterFactory.Free;
  GabcDSAdapterFactory := nil;

end.
