unit abcdbexc;

(*------------------------------------------------------------------------------

Advanced Business Components
ABC for Delphi
Version 6

Copyright  1995, 2001 Objective Software Technology Pty Limited ACN 068 640 353
Functionality derived from Delphi VCL Source Code
is Copyright  1995, 2001 Borland Software Corporation

------------------------------------------------------------------------------*)

{$I abcdefs.inc}

interface

uses
  Windows, SysUtils, Messages, Classes, Controls, Forms, Dialogs,
  BDE, DB, DBTables, abcexcep, abcapp;

type

   {TabcDBAppException}
   {Adds a lookup table for message processing and a log table for exception log}
   {Retains the default message and log properties of TabcAppException for
    use when the database connection fails or a message lookup isn't found}

  TabcAppExcTableType = (etSQL, etParadox, etDBase);

  TabcDBAppException = class(TabcAppException)
  private
    FStreamedActive: Boolean;
    FCancel: Boolean;
    FDBExcepMsg: TabcExcepMsg;
    FMsgTable: TTable;
    FLogTable: TTable;
    FTableType: TabcAppExcTableType;
    FExtraLogInfo: TStrings;
    function GetActive: Boolean;
    procedure SetActive(Value: Boolean);
    function GetDatabaseName: TFileName;
    function GetMsgTableName: TFileName;
    function GetLogTableName: TFileName;
    procedure SetDatabaseName(Value: TFileName);
    procedure SetTableType(Value: TabcAppExcTableType);
    procedure SetMsgTableName(Value: TFileName);
    procedure SetLogTableName(Value: TFileName);
    function GetBeforePost: TDataSetNotifyEvent;
    procedure SetBeforePost(Value: TDataSetNotifyEvent);
    function GetSessionName: string;
    procedure SetSessionName(Value: string);
  protected
    procedure ApplicationExceptionHandler(Sender: TObject; E: Exception); override;
    function DBShowExceptionMsg(Sender: TObject; E: Exception; var Log: Boolean): Boolean; virtual;
    function DBLogException(Sender: TObject; E: Exception): Boolean; virtual;
    procedure Loaded; override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure AppendLogMessage(Sender: TObject; E: Exception; ExtraInfo: TStrings); override;
    procedure EmptyLogTable;
    property ExtraLogInfo: TStrings read FExtraLogInfo;
    {Runtime subcomponent properties for access to their methods}
    property MsgTable: TTable read FMsgTable;
    property LogTable: TTable read FLogTable;
  published
    property Active: Boolean read GetActive write SetActive default True;
    property MsgTableName: TFileName read GetMsgTableName write SetMsgTableName;
    property LogTableName: TFileName read GetLogTableName write SetLogTableName;
    property DatabaseName: TFileName read GetDatabaseName write SetDatabaseName;
    property TableType: TabcAppExcTableType read FTableType write SetTableType;
    property BeforePost: TDataSetNotifyEvent read GetBeforePost write SetBeforePost;
    property SessionName: string read GetSessionName write SetSessionName;
  end;

  { TabcDBQryAppException }

  TabcDBQryAppException = class(TabcAppException)
  private
    FActive: Boolean;
    FCancel: Boolean;
    FDBExcepMsg: TabcExcepMsg;
    FMsgQuery: TQuery;
    FLogQuery: TQuery;
    FExtraLogInfo: TStrings;
    function GetDatabaseName: TFileName;
    function GetMsgSQL: TStrings;
    function GetLogSQL: TStrings;
    procedure SetDatabaseName(Value: TFileName);
    procedure SetMsgSQL(Value: TStrings);
    procedure SetLogSQL(Value: TStrings);
    function GetBeforePost: TDataSetNotifyEvent;
    procedure SetBeforePost(Value: TDataSetNotifyEvent);
    function GetSessionName: string;
    procedure SetSessionName(Value: string);
  protected
    procedure ApplicationExceptionHandler(Sender: TObject; E: Exception); override;
    function DBShowExceptionMsg(Sender: TObject; E: Exception; var Log: Boolean): Boolean; virtual;
    function DBLogException(Sender: TObject; E: Exception): Boolean; virtual;
    function CreateMsgQuery: TQuery; virtual;
    function CreateLogQuery: TQuery; virtual;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure AppendLogMessage(Sender: TObject; E: Exception; ExtraInfo: TStrings); override;
    property ExtraLogInfo: TStrings read FExtraLogInfo;
    {Runtime subcomponent properties for access to their methods}
    property LogQuery: TQuery read FLogQuery;
    property MsgQuery: TQuery read FMsgQuery;
  published
    property Active: Boolean read FActive write FActive default True;
    property MsgSQL: TStrings read GetMsgSQL write SetMsgSQL;
    property LogSQL: TStrings read GetLogSQL write SetLogSQL;
    property DatabaseName: TFileName read GetDatabaseName write SetDatabaseName;
    property BeforePost: TDataSetNotifyEvent read GetBeforePost write SetBeforePost;
    property SessionName: string read GetSessionName write SetSessionName;
  end;

implementation

uses abcconst;

{ Get logical address from physical address as formatted string }

function StrLogicalAddr(PhysicalAddr: Pointer): string;
begin
  Result := Format('%p', [PhysicalAddr]);
end;

{TabcDBAppException}

constructor TabcDBAppException.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FDBExcepMsg := TabcExcepMsg.Create;
  FMsgTable := TTable.Create(Self);
  FLogTable := TTable.Create(Self);
  if not (csDesigning in ComponentState) then
    AppException := TabcAppException(Self);
  TableType := etParadox;
  FStreamedActive := True;
end;

destructor TabcDBAppException.Destroy;
begin
  FDBExcepMsg.Free;
  if FMsgTable <> nil then FMsgTable.Close;
  FMsgTable.Free;
  if FLogTable <> nil then FLogTable.Close;
  FLogTable.Free;
  if AppException = self then
    AppException := nil;
  inherited Destroy;
end;

procedure TabcDBAppException.Loaded;
begin
  inherited Loaded;
  if FStreamedActive then
    try
      FLogTable.Open;
      if MsgTableName <> '' then begin
        FMsgTable.ReadOnly := True;
        FMsgTable.Open;
        if FTableType = etDBase then
          FMsgTable.IndexName := 'PRIMARYKEY';
      end;
    except
      on E: Exception do ShowMessage('Failed to open DBAppException tables: ' + E.Message);
    end;
end;

function TabcDBAppException.GetActive: Boolean;
begin
  Result := FLogTable.Active;
end;

procedure TabcDBAppException.SetActive(Value: Boolean);
begin
  if (csLoading in ComponentState) then
    FStreamedActive := Value
  else if Value then
  try
    FLogTable.Open;
    if MsgTableName <> '' then begin
      FMsgTable.ReadOnly := True;
      FMsgTable.Open;
      if FTableType = etDBase then
        FMsgTable.IndexName := 'PRIMARYKEY';
    end;
  except
    on E: Exception do ShowMessage('Failed to open DBAppException tables: ' + E.Message);
  end
  else begin
    FLogTable.Close;
    FMsgTable.Close;
  end;
end;

procedure TabcDBAppException.ApplicationExceptionHandler(Sender: TObject; E: Exception);
var
  FLogExcep: Boolean;
  FOnException: TabcExceptionEvent;
begin
  CurrExcepAddr := StrLogicalAddr(ExceptAddr);
  FCancel := false;
  FOnException := OnException;
  if Assigned(FOnException) then OnException(Sender, E, FCancel);
  if not FCancel then DBShowExceptionMsg(Sender, E, FLogExcep);
  if FLogExcep and ExcepLog.Enabled then DBLogException(Sender, E);
  if Terminate then
    if not ConfirmTerminate or
      (MessageDlg(sabcConfirmExcepTerminate, mtConfirmation,
       [mbYes, mbNo], 0) = mrYes) then
      Application.Terminate;
end;

function TabcDBAppException.DBLogException(Sender: TObject; E: Exception): Boolean;
var
  ExtraStrings: TStringList;
  ALogExtraInfo: TabcLogAppendEvent;
begin
  Result := True;
  ExtraStrings := TStringList.Create;
  try
    try
      ALogExtraInfo := LogExtraInfo;
      if Assigned(ALogExtraInfo) then
        ALogExtraInfo(Self, ExtraStrings);
      AppendLogMessage(Sender, E, ExtraStrings);
    except
      on E: Exception do begin
        Result := False;
        ShowMessage('Error writing to error log: ' + E.Message);
        ExcepLog.AppendLogMessage(Sender, E, ExtraStrings);
      end;
    end;
  finally
    ExtraStrings.Free;
  end;
end;

procedure TabcDBAppException.AppendLogMessage(Sender: TObject; E: Exception; ExtraInfo: TStrings);
var
  NetUser: array[0..255] of char;
  LogExcepAddr: string;
begin
    with FLogTable do begin
      Append;
      {assign fields}
      FieldByName('ERR_TIME').AsDateTime := Now;
      if ExcepLog.UserName <> '' then
         FieldByName('USER_NAME').AsString := ExcepLog.UserName
      else
        {Lookup Net User Id}
        if DbiGetNetUserName(NetUser) = 0 then
          FieldByName('USER_NAME').AsString := StrPas(NetUser)
        else
          FieldByName('USER_NAME').AsString := '';
      if Assigned(E) then
      begin
        FieldByName('ERR_CLASS').AsString := E.ClassName;
        FieldByName('ERR_MSG').AsString := E.Message;
      end
      else begin
        FieldByName('ERR_CLASS').AsString := '';
        FieldByName('ERR_MSG').AsString := '';
      end;
      if ExcepLog.LogExeName then
        FieldByName('EXE_NAME').AsString := Application.ExeName
      else
        FieldByName('EXE_NAME').AsString := '';
      if ExcepLog.LogAppTitle then
        FieldByName('APP_NAME').AsString := Application.Title
      else
        FieldByName('APP_NAME').AsString := '';
      if ExcepLog.LogSender and Assigned(Sender) then begin
        FieldByName('SNDR_CLASS').AsString := Sender.ClassName;
        if Sender is TComponent then
          FieldByName('SNDR_NAME').AsString := TComponent(Sender).Name
        else
          FieldByName('SNDR_NAME').AsString := '';
      end
      else begin
        FieldByName('SNDR_CLASS').AsString := '';
        FieldByName('SNDR_NAME').AsString := '';
      end;
      if ExcepLog.LogDebug then begin
        LogExcepAddr := StrLogicalAddr(ExceptAddr);
        FieldByName('ERR_ADDR').AsString := LogExcepAddr;
        if (E is EabcOverrideSysError) then
          with EabcOverrideSysError(E) do begin
            FieldByName('SYS_CLASS').AsString := SysClass;
            FieldByName('SYS_MSG').AsString := SysMsg;
            FieldByName('SYS_ADDR').AsString := SysAddr;
          end
        else begin
          FieldByName('SYS_CLASS').AsString := '';
          FieldByName('SYS_MSG').AsString := '';
          FieldByName('SYS_ADDR').AsString := '';
        end;
      end
      else begin
        FieldByName('ERR_ADDR').AsString := '';
        FieldByName('SYS_CLASS').AsString := '';
        FieldByName('SYS_MSG').AsString := '';
        FieldByName('SYS_ADDR').AsString := '';
      end;
      FExtraLogInfo := ExtraInfo;
      {Write Record}
      Post;
      FExtraLogInfo := nil;
    end;
end;

function TabcDBAppException.DBShowExceptionMsg(Sender: TObject; E: Exception; var Log: Boolean): Boolean;
begin
  Result := True;
  Log := True;
  try
    with FMsgTable do
    if Active then
      if FindKey([E.ClassName]) then
        with FDBExcepMsg do begin
          {assign defaults from component properties}
          Beep := ExcepMsg.Beep;
          Enabled := ExcepMsg.Enabled;
          MsgBox := ExcepMsg.MsgBox;
          ShowDebug := ExcepMsg.ShowDebug;
          ShowExcMsg := ExcepMsg.ShowExcMsg;
          Title := ExcepMsg.Title;
          {assign fields from MsgTable}
          case FieldByName('MSG_TYPE').AsInteger of
            0: MsgType := mtWarning;
            1: MsgType := mtError;
            2: MsgType := mtInformation;
            3: MsgType := mtConfirmation;
            4: MsgType := mtCustom;
          end;
          Terminate := FieldByName('TERMINATE').AsBoolean;
          Log := FieldByName('LOG').AsBoolean;
          Lines.Assign(FieldByName('ERR_MSG'));
          ShowExceptionMsg(E);
        end
      else
        ExcepMsg.ShowExceptionMsg(E)
    else
      ExcepMsg.ShowExceptionMsg(E);
  except
    on E: Exception do begin
      Result := False;
      ShowMessage('Error message DB error: ' + E.Message);
      ExcepMsg.ShowExceptionMsg(E);
    end;
  end;
end;

function TabcDBAppException.GetDatabaseName: TFileName; begin Result := FMsgTable.DatabaseName; end;
function TabcDBAppException.GetMsgTableName: TFileName; begin Result := FMsgTable.TableName; end;
function TabcDBAppException.GetLogTableName: TFileName; begin Result := FLogTable.TableName; end;

procedure TabcDBAppException.SetDatabaseName(Value: TFileName);
begin
  FMsgTable.DatabaseName := Value;
  FLogTable.DatabaseName := Value;
end;

function TabcDBAppException.GetSessionName: string; begin Result := FMsgTable.SessionName; end;

procedure TabcDBAppException.SetSessionName(Value: string);
begin
  FMsgTable.SessionName := Value;
  FLogTable.SessionName := Value;
end;

procedure TabcDBAppException.SetTableType(Value: TabcAppExcTableType);
begin
  FTableType := Value;
  case Value of
    etSQL:
      begin
        FMsgTable.TableType := ttDefault;
        FLogTable.TableType := ttDefault;
      end;
    etParadox:
      begin
        FMsgTable.TableType := ttParadox;
        FLogTable.TableType := ttParadox;
      end;
    etDBase:
      begin
        FMsgTable.TableType := ttDBase;
        FLogTable.TableType := ttDBase;
      end;
  end;
end;

procedure TabcDBAppException.SetMsgTableName(Value: TFileName);
begin
  if FMsgTable.TableName <> Value then begin
    FMsgTable.TableName := Value;
  end;
end;

procedure TabcDBAppException.SetLogTableName(Value: TFileName);
begin
  if FLogTable.TableName <> Value then begin
    FLogTable.TableName := Value;
  end;
end;

procedure TabcDBAppException.EmptyLogTable;
var
  SaveActive: Boolean;
begin
  with FLogTable do begin
    SaveActive := Active;
    try
      Close;
      Exclusive := True;
      Open;
      EmptyTable;
      Close;
    finally
      Exclusive := False;
      if SaveActive then Open;
    end;
  end;
end;

function TabcDBAppException.GetBeforePost: TDataSetNotifyEvent;
begin
  if FLogTable <> nil then Result := FLogTable.BeforePost;
end;

procedure TabcDBAppException.SetBeforePost(Value: TDataSetNotifyEvent);
begin
  if FLogTable <> nil then FLogTable.BeforePost := Value;
end;

{ TabcDBQryAppException }

constructor TabcDBQryAppException.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FDBExcepMsg := TabcExcepMsg.Create;
  FMsgQuery := CreateMsgQuery;
  FLogQuery := CreateLogQuery;
  if not (csDesigning in ComponentState) then
    AppException := TabcAppException(self);
  FActive := True;
end;

function TabcDBQryAppException.CreateMsgQuery: TQuery;
begin
  Result := TQuery.Create(Self);
  with Result do begin
    SQL.Add('Select * from ERRMSG');
    SQL.Add('where ERR_CLASS = :pErrClass');
    Params.CreateParam(ftstring, 'pErrClass', ptInput);
  end;
end;

function TabcDBQryAppException.CreateLogQuery: TQuery;
begin
  Result := TQuery.Create(Self);
  with Result do begin

    SQL.Add('Insert into ERRLOG(');
    SQL.Add('ERR_TIME, USER_NAME, ERR_CLASS, ERR_MSG,');
    SQL.Add('EXE_NAME, APP_NAME, SNDR_CLASS, SNDR_NAME,');
    SQL.Add('ERR_ADDR, SYS_CLASS, SYS_MSG, SYS_ADDR)');
    SQL.Add('values(');
    SQL.Add(':pErrTime, :pUserName, :pErrClass, :pErrMsg,');
    SQL.Add(':pExeName, :pAppName, :pSndrClass, :pSndrName, ');
    SQL.Add(':pErrAddr, :pSysClass, :pSysMsg, :pSysAddr)');

    Params.CreateParam(ftDateTime, 'pErrTime', ptInput);
    Params.CreateParam(ftstring, 'pUserName', ptInput);
    Params.CreateParam(ftstring, 'pErrClass', ptInput);
    Params.CreateParam(ftstring, 'pErrMsg', ptInput);
    Params.CreateParam(ftstring, 'pExeName', ptInput);
    Params.CreateParam(ftstring, 'pAppName', ptInput);
    Params.CreateParam(ftstring, 'pSndrClass', ptInput);
    Params.CreateParam(ftstring, 'pSndrName', ptInput);
    Params.CreateParam(ftstring, 'pErrAddr', ptInput);
    Params.CreateParam(ftstring, 'pSysClass', ptInput);
    Params.CreateParam(ftstring, 'pSysMsg', ptInput);
    Params.CreateParam(ftstring, 'pSysAddr', ptInput);
  end;
end;

destructor TabcDBQryAppException.Destroy;
begin
  FDBExcepMsg.Free;
  FMsgQuery.Free;
  FLogQuery.Free;
  if AppException = self then
    AppException := nil;
  inherited Destroy;
end;

procedure TabcDBQryAppException.ApplicationExceptionHandler(Sender: TObject; E: Exception);
var
  FLogExcep: Boolean;
  FOnException: TabcExceptionEvent;
begin
  CurrExcepAddr := StrLogicalAddr(ExceptAddr);
  FCancel := false;
  FOnException := OnException;
  if Assigned(FOnException) then OnException(Sender, E, FCancel);
  if not FCancel then
    if FActive then begin
      DBShowExceptionMsg(Sender, E, FLogExcep);
      if FLogExcep and ExcepLog.Enabled then DBLogException(Sender, E);
      if Terminate then
        if not ConfirmTerminate or
          (MessageDlg(sabcConfirmExcepTerminate, mtConfirmation,
           [mbYes, mbNo], 0) = mrYes) then
          Application.Terminate
    end
    else
      raise E;
end;

function TabcDBQryAppException.DBLogException(Sender: TObject; E: Exception): Boolean;
var
  ExtraStrings: TStringList;
  ALogExtraInfo: TabcLogAppendEvent;
begin
  Result := True;
  ExtraStrings := TStringList.Create;
  try
    try
      ALogExtraInfo := LogExtraInfo;
      if Assigned(ALogExtraInfo) then
        ALogExtraInfo(Self, ExtraStrings);
      AppendLogMessage(Sender, E, ExtraStrings);
    except
      on E: Exception do begin
        Result := False;
        ShowMessage('Error writing to error log: ' + E.Message);
        ExcepLog.AppendLogMessage(Sender, E, ExtraStrings);
      end;
    end;
  finally
    ExtraStrings.Free;
  end;
end;

procedure TabcDBQryAppException.AppendLogMessage(Sender: TObject; E: Exception; ExtraInfo: TStrings);
var
  NetUser: array[0..255] of char;
begin
  with FLogQuery do begin
    {assign field params}
    ParamByName('pErrTime').AsDateTime := Now;
    if ExcepLog.UserName <> '' then
       ParamByName('pUserName').AsString := ExcepLog.UserName
    else
      {Lookup Net User Id}
      if DbiGetNetUserName(NetUser) = 0 then
        ParamByName('pUserName').AsString := StrPas(NetUser)
      else
        ParamByName('pUserName').AsString := '';
    if Assigned(E) then
    begin
      ParamByName('pErrClass').AsString := E.ClassName;
      ParamByName('pErrMsg').AsString := E.Message;
    end
    else
    begin
      ParamByName('pErrClass').AsString := '';
      ParamByName('pErrMsg').AsString := '';
    end;
    if ExcepLog.LogExeName then
      ParamByName('pExeName').AsString := Application.ExeName
    else
      ParamByName('pExeName').AsString := '';
    if ExcepLog.LogAppTitle then
      ParamByName('pAppName').AsString := Application.Title
    else
      ParamByName('pAppName').AsString := '';
    if ExcepLog.LogSender and Assigned(Sender) then begin
      ParamByName('pSndrClass').AsString := Sender.ClassName;
      if Sender is TComponent then
        ParamByName('pSndrName').AsString := TComponent(Sender).Name
      else
        ParamByName('pSndrName').AsString := '';
    end
    else begin
      ParamByName('pSndrClass').AsString := '';
      ParamByName('pSndrName').AsString := '';
    end;
    if ExcepLog.LogDebug then begin
      ParamByName('pErrAddr').AsString := CurrExcepAddr;
      if (E is EabcOverrideSysError) then
        with EabcOverrideSysError(E) do begin
          ParamByName('pSysClass').AsString := SysClass;
          ParamByName('pSysMsg').AsString := SysMsg;
          ParamByName('pSysAddr').AsString := SysAddr;
        end
      else begin
        ParamByName('pSysClass').AsString := '';
        ParamByName('pSysMsg').AsString := '';
        ParamByName('pSysAddr').AsString := '';
      end;
    end
    else begin
      ParamByName('pErrAddr').AsString := '';
      ParamByName('pSysClass').AsString := '';
      ParamByName('pSysMsg').AsString := '';
      ParamByName('pSysAddr').AsString := '';
    end;

    {Before Post}
    FExtraLogInfo := ExtraInfo;
    if Assigned(Self.BeforePost) then Self.BeforePost(FLogQuery);
    {Write Record}
    ExecSQL;
    FExtraLogInfo := nil;
  end;
end;

function TabcDBQryAppException.DBShowExceptionMsg(Sender: TObject; E: Exception; var Log: Boolean): Boolean;
begin
  Result := True;
  Log := True;
  try
    with FMsgQuery do begin
      if Active then Close;
      ParamByName('pErrClass').AsString := E.ClassName;
      Open;
      if not (BOF and EOF) then
        with FDBExcepMsg do begin
          {assign defaults from component properties}
          Beep := ExcepMsg.Beep;
          Enabled := ExcepMsg.Enabled;
          MsgBox := ExcepMsg.MsgBox;
          ShowDebug := ExcepMsg.ShowDebug;
          ShowExcMsg := ExcepMsg.ShowExcMsg;
          Title := ExcepMsg.Title;
          {assign fields from MsgTable}
          case FieldByName('MSG_TYPE').AsInteger of
            0: MsgType := mtWarning;
            1: MsgType := mtError;
            2: MsgType := mtInformation;
            3: MsgType := mtConfirmation;
            4: MsgType := mtCustom;
          end;
          Terminate := FieldByName('TERMINATE').AsBoolean;
          Log := FieldByName('LOG').AsBoolean;
          Lines.Assign(FieldByName('ERR_MSG'));
          ShowExceptionMsg(E);
        end
      else
        ExcepMsg.ShowExceptionMsg(E);
      Close;
    end;
  except
    on E: Exception do begin
      Result := False;
      ShowMessage('Error message DB error: ' + E.Message);
      ExcepMsg.ShowExceptionMsg(E);
    end;
  end;
end;

function TabcDBQryAppException.GetDatabaseName: TFileName; begin Result := FMsgQuery.DatabaseName; end;

procedure TabcDBQryAppException.SetDatabaseName(Value: TFileName);
begin
  FMsgQuery.DatabaseName := Value;
  FLogQuery.DatabaseName := Value;
end;

function TabcDBQryAppException.GetSessionName: string; begin Result := FMsgQuery.SessionName; end;

procedure TabcDBQryAppException.SetSessionName(Value: string);
begin
  FMsgQuery.SessionName := Value;
  FLogQuery.SessionName := Value;
end;

function TabcDBQryAppException.GetMsgSQL: TStrings; begin Result := MsgQuery.SQL; end;
function TabcDBQryAppException.GetLogSQL: TStrings; begin Result := LogQuery.SQL; end;

procedure TabcDBQryAppException.SetMsgSQL(Value: TStrings);
begin
  MsgQuery.SQL.Assign(Value);
end;

procedure TabcDBQryAppException.SetLogSQL(Value: TStrings);
begin
  LogQuery.SQL.Assign(Value);
end;

function TabcDBQryAppException.GetBeforePost: TDataSetNotifyEvent;
begin Result := FLogQuery.BeforePost; end;

procedure TabcDBQryAppException.SetBeforePost(Value: TDataSetNotifyEvent);
begin FLogQuery.BeforePost := Value; end;

end.
