unit abcdlg;

(*------------------------------------------------------------------------------

Advanced Business Components
ABC for Delphi
Version 6

Copyright  1995, 2001 Objective Software Technology Pty Limited ACN 068 640 353
Functionality derived from Delphi VCL Source Code
is Copyright  1995, 2001 Borland Software Corporation

------------------------------------------------------------------------------*)

{$I abcdefs.inc}

{:This unit contains ABC custom dialog components
  which display modal dialog boxes when their Execute method is called.
}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, ExtCtrls, ComCtrls,
  abcbusy, abcctl32;

type

  { TabcCustomDlg }

  {:This is a base class for descendant dialogs.
  This class contains common protected properties and methods, used by descendant classes
  <See class=TabcMessageDlg> and <See class=TabcProgressDlg>.}

  TabcCustomDlg = class(TComponent)
  private
    FCaption: string;
    FMessage: TStrings;
    FButtonCaptions: TStrings;
    FButtons: TMsgDlgButtons;
    FHelpContext: Integer;
    FForm: TForm;
    FFormMessageLabel: TLabel;
    FFormCheckBox: TabcCheckBox;
    FOnShow: TNotifyEvent;
    FOnClose: TNotifyEvent;
    FDlgType: TMsgDlgType;
    FIcon: TIcon;
    FResult: TModalResult;
    FFont: TFont;
    FDlgLeft: Integer;
    FDlgTop: Integer;
    FDlgPosition: TabcPopupPosition;
    FChecked, FWasChecked: Boolean;
    FCheckBoxVisible: Boolean;
    FCheckBoxCaption: string;
    FCheckBoxResult: TModalResult;
    FOnChecked: TNotifyEvent;

    FHorzMargin, FVertMargin, FHorzSpacing, FVertSpacing, FButtonWidth,
    FButtonHeight, FButtonSpacing, FButtonCount, FButtonGroupWidth, FButtonGroupTop,
    FIconTextWidth, FIconTextHeight, FIconWidth, FIconHeight: Integer;

    FButtonWidths: array[TMsgDlgBtn] of Integer;

    procedure ButtonCaptionsChanged(Sender: TObject);
    procedure SetButtonCaptions(const Value: TStrings);
    procedure SetMessage(const Value: TStrings);
    procedure SetIcon(const Value: TIcon);
    procedure SetFont(const Value: TFont);
    function IsFontStored: Boolean;
    procedure InitFont;
    procedure HelpButtonClick(Sender: TObject);
    procedure SetDlgLeft(const Value: Integer);
    procedure SetDlgPosition(const Value: TabcPopupPosition);
    procedure SetDlgTop(const Value: Integer);

    procedure FormShow(Sender: TObject);
    procedure FormHide(Sender: TObject);

  protected
    procedure DoCreateForm; virtual;
    procedure DoPositionForm;
    function DoExecute: TModalResult; virtual;
    procedure DoFormShow; virtual;
    procedure DoFormHide; virtual;

    {can be public in descendant}
    {::Read only reference to the form created when Execute is called.}
    property Form: TForm read FForm;
    {::Read only reference to the result of the last call to the Execute method.}
    property Result: TModalResult read FResult;
    {can be published in descendant}
    {::<p>
<code>
  TMsgDlgBtn = (mbYes, mbNo, mbOK, mbCancel, mbAbort, mbRetry, mbIgnore, mbAll, mbNoToAll, mbYesToAll, mbHelp);
  TMsgDlgButtons = set of TMsgDlgBtn;
</code>
    <p>
    This defines the set of buttons to be shown in the dialog box.
    }
    property Buttons: TMsgDlgButtons read FButtons write FButtons default mbOKCancel;
    {::This can be used to override the default button captions by defining a string for each button in the buttons property.}
    property ButtonCaptions: TStrings read FButtonCaptions write SetButtonCaptions;
    {::This is the caption of the dialog form.}
    property Caption: string read FCaption write FCaption;
    {::When CheckBoxVisible is true, this is the text displayed for the checkbox.}
    property CheckBoxCaption: string read FCheckBoxCaption write FCheckBoxCaption;
    {::When CheckBoxVisible is true, if this value is set to anything other than mrNone (0),
     the dialog box does not show when the Execute method is called.
     Instead this value is automatically returned.<P>
     This property supports the implementation of checkbox features such as
     "don't ask me again". Use the OnChecked event to set this property value.
     Your dialog execution code can then just always call Execute without
     first needing to inspect the value of the Checked property.
     }
    property CheckBoxResult: TModalResult read FCheckBoxResult write FCheckBoxResult default mrNone;
    {::Set to True to display a CheckBox.}
    property CheckBoxVisible: Boolean read FCheckBoxVisible write FCheckBoxVisible default False;
    {::Set to True if the CheckBox is checked. This property changes when the user checks or unchecks the box.}
    property Checked: Boolean read FChecked write FChecked default False;
    {::The dialog position on the monitor. Used when DlgPosition=appCustom.}
    property DlgLeft: Integer read FDlgLeft write SetDlgLeft default -1;
    {::The dialog position on the monitor. Used when DlgPosition=appCustom.}
    property DlgTop: Integer read FDlgTop write SetDlgTop default -1;

    {::<p>
<code>
TabcPopupPosition = (appScreenCenter, appMainFormCenter, appActiveFormCenter, appCustom);
</code><p>
    Options for automatically positioning the dialog box.<p>
    The default is appScreenCenter.
    }
    property DlgPosition: TabcPopupPosition read FDlgPosition write SetDlgPosition default appScreenCenter;

    {::<p>
<code>
TMsgDlgType = (mtWarning, mtError, mtInformation, mtConfirmation, mtCustom);
</code><p>
    The type of dialog box to display. This will automatically add standard a system icon for the dialog type,
    unless the icon property is used for a custom icon.<p>
    The default is mtWarning.
    }
    property DlgType: TMsgDlgType read FDlgType write FDlgType default mtWarning;
    {::The font used to display the message lines on the form.}
    property Font: TFont read FFont write SetFont stored IsFontStored;
    {::The help context of the form and help button, if used.}
    property HelpContext: Integer read FHelpContext write FHelpContext default 0;
    {::Use to display a custom icon, instead of one of the standard system icons derived from the DlgType.}
    property Icon: TIcon read FIcon write SetIcon;
    {::Lines of message text to display in the dialog box.}
    property Message: TStrings read FMessage write SetMessage;
    {::Occurs immediately after the dialog box is closed, before the Execute method finishes.}
    property OnClose: TNotifyEvent read FOnClose write FOnClose;
    {::Occurs immediately before the dialog box is shown.}
    property OnShow: TNotifyEvent read FOnShow write FOnShow;
    {::Occurs only when the checkbox checked state has been changed by the user; immediately after the OnClose event, before the Execute method finishes.}
    property OnChecked: TNotifyEvent read FOnChecked write FOnChecked;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Assign(Source: TPersistent); override;
    {::Call Execute to display the modal dialog box.}
    function Execute: TModalResult;
  published
  end;

  { TabcMessageDlg }

  {:A flexible component to replace or supplement Delphi's MessageDlg function.
  TabcMessageDlg displays a modal message dialog when its Execute method is called.<p>
  A CheckBox can be displayed using the <See property=CheckBoxVisible>,
  <See property=CheckBoxCaption> and <See property=Checked> properties.
  This feature can also be used to automatically not show the dialog box while still calling Execute,
  using the <See event=OnChecked> event and the <See property=CheckBoxResult> property.<P>

  A multiline message can be displayed using the <See property=Message> strings property.<P>

  The appearance can be customized using the <See property=Buttons>,
  <See property=ButtonCaptions>, <See property=Caption>,
  <See property=Font> and <See property=Icon> properties.
  }

  TabcMessageDlg = class(TabcCustomDlg)
  public
    property Form;
    property Result;
  published
    property Buttons;
    property ButtonCaptions;
    property Caption;
    property CheckBoxCaption;
    property CheckBoxVisible;
    property Checked;
    property CheckBoxResult;
    property DlgLeft;
    property DlgTop;
    property DlgPosition;
    property DlgType;
    property Font;
    property HelpContext;
    property Icon;
    property Message;
    property OnClose;
    property OnShow;
    property OnChecked;
  end;

  { TabcProgressDlg }

  {:This component displays a modal dialog box that measures an elapsed time interval.

  TabcProgressDlg can count up or down through a range of time intervals.<p>
  The current count can be displayed in the message text and is updated while the dialog box is visible.
  A progress bar can also be shown.<p>

  Call Execute to display the dialog box.<p>

  The dialog counts a number of intervals specified by the <See property=TimerInterval> property.<p>
  When the dialog box is shown, the current <See property=Value> is initialized
  to the <See property=Min> value, if <See property=Step> is positive,
  or to the <See property=Max> value, if <See property=Step> is negative.<p>
  On each <See property=TimerInterval> the current <See property=Value> changes by <See property=Step>,
  until the <See property=Min> or <See property=Max> is reached.<p>
  For example, the default Min, Max, Step and TimerInterval values of 0, 60, -1 and 1000 will display a
  count of remaining seconds from 60 down to 0.<p>

  If the user does not close the dialog box before the end point, it closes automatically.<P>

  The current <See property=Value> can be displayed in the <See property=Message> strings, according to the
  Delphi format function: use %d in the message to display an integer value.<p>

  Set <see property=ProgressBarVisible> to show or hide the progressbar.<p>

  The <see event=OnTimer> event occurs on each elapsed interval.<p>
  The <see event=OnTimeout> event occurs if the <See property=Value> reaches the end point (Min or Max) and
  the user has not closed the dialog box. When this occurs the dialog box closes automatically
  and a special Modal Result value abcmrTimeout is returned.<p>
  }

  TabcProgressDlg = class(TabcMessageDlg)
  private
    FTimer: TTimer;
    FProgressBar: TProgressBar;
    FProgressBarVisible: Boolean;
    FTimerInterval: Cardinal;
    FMin: Integer;
    FMax: Integer;
    FStep: Integer;
    FValue: Integer;
    FOnTimer: TNotifyEvent;
    FOnTimeout: TNotifyEvent;
    procedure TimerTick(Sender: TObject);
    procedure UpdateMessageLabel;
  protected
    procedure DoCreateForm; override;
    function DoExecute: TModalResult; override;
    procedure DoFormShow; override;
    procedure DoFormHide; override;
  public
    constructor Create(AOwner: TComponent); override;
    procedure Assign(Source: TPersistent); override;
    {::The current value, incremented or decremented in Steps between
    Min and Max, at each TimerInterval.}
    property Value: Integer read FValue;
  published
    {::Set to False to hide the progress bar.}
    property ProgressBarVisible: Boolean read FProgressBarVisible write FProgressBarVisible default True;
    {::The minimum value of the range to be counted by the timer.}
    property Min: Integer read FMin write FMin default 0;
    {::The maximum value of the range to be counted by the timer.}
    property Max: Integer read FMax write FMax default 60;
    {::The step to increment or decrement the Value by on each TimerInterval.<p>
    A negative value of Step decrements the value from Max down to Min.}
    property Step: Integer read FStep write FStep default -1;
    {::The TimerInterval in milliseconds. On each interval, the Value is incremented or decremented by Step.}
    property TimerInterval: Cardinal read FTimerInterval write FTimerInterval default 1000;
    {::Occurs on each TimerInterval.}
    property OnTimer: TNotifyEvent read FOnTimer write FOnTimer;
    {::Occurs when timing reaches the end point without the user closing the dialog box.<p>
    The dialog is hidden and then this event occurs.<p>
    A special modal result value of abcmrTimeout is also returned by the Execute function.
    }
    property OnTimeout: TNotifyEvent read FOnTimeout write FOnTimeout;
  end;

const

  {:This is a special value for TModalResult, returned by the TabcProgressDialog.Execute
    method; when the count of time intervals reaches the end point and the dialog
    box closes automatically without the user closing it.}

  abcmrTimeout = -8;

implementation

uses Consts, Dlgs, abcutilm;

{ TabcCustomDlg }

function StringsToDlgText(strs: TStrings): string;
begin
  Result := strs.Text;
  if Result <> '' then
    Result := Copy(Result, 1, Length(Result)-2);
end;

constructor TabcCustomDlg.Create(AOwner: TComponent);
begin
  inherited;
  FButtonCaptions := TStringList.Create;
  TStringList(FButtonCaptions).OnChange := ButtonCaptionsChanged;
  FMessage := TStringList.Create;
  FDlgType := mtWarning;
  FIcon := TIcon.Create;
  InitFont;
  FButtons := mbOKCancel;
  FDlgLeft := -1;
  FDlgTop := -1;
  FDlgPosition := appScreenCenter;
  FHelpContext := 0;
end;

destructor TabcCustomDlg.Destroy;
begin
  if FForm <> nil then
    FForm.Free;
  FButtonCaptions.Free;
  FMessage.Free;
  FIcon.Free;
  FFont.Free;
  inherited;
end;

procedure TabcCustomDlg.Assign(Source: TPersistent);
begin
  if Source is TabcCustomDlg then
    with TabcCustomDlg(Source) do
    begin
      Self.Buttons := Buttons;
      Self.ButtonCaptions := ButtonCaptions;
      Self.Caption := Caption;
      Self.CheckBoxVisible := CheckBoxVisible;
      Self.CheckBoxCaption := CheckBoxCaption;
      Self.Checked := Checked;
      Self.CheckBoxResult := CheckBoxResult;
      Self.FDlgLeft := DlgLeft;
      Self.FDlgTop := DlgTop;
      Self.DlgPosition := DlgPosition;
      Self.DlgType := DlgType;
      Self.Font := Font;
      Self.HelpContext := HelpContext;
      Self.Icon := Icon;
      Self.Message := Message;
      Self.OnClose := OnClose;
      Self.OnShow := OnShow;
      Self.OnChecked := OnChecked;
    end
  else
    inherited;
end;

procedure TabcCustomDlg.ButtonCaptionsChanged(Sender: TObject);
var
  b: TMsgDlgBtn;
begin
  for b := Low(TMsgDlgBtn) to High(TMsgDlgBtn) do
    FButtonWidths[b] := 0;
end;

procedure TabcCustomDlg.InitFont;
var
  NonClientMetrics: TNonClientMetrics;
begin
  FFont := TFont.Create;
  NonClientMetrics.cbSize := sizeof(NonClientMetrics);
  if SystemParametersInfo(SPI_GETNONCLIENTMETRICS, 0, @NonClientMetrics, 0) then
    Font.Handle := CreateFontIndirect(NonClientMetrics.lfMessageFont);
  FFont.OnChange := ButtonCaptionsChanged;
end;

const
  defaultCaptions: array[TMsgDlgType] of string = (SMsgDlgWarning, SMsgDlgError,
    SMsgDlgInformation, SMsgDlgConfirm, '');
  defaultIconIDs: array[TMsgDlgType] of PChar = (IDI_EXCLAMATION, IDI_HAND,
    IDI_ASTERISK, IDI_QUESTION, nil);
  defaultButtonNames: array[TMsgDlgBtn] of string = (
    'Yes', 'No', 'OK', 'Cancel', 'Abort', 'Retry', 'Ignore', 'All', 'NoToAll',
    'YesToAll', 'Help');
  defaultButtonCaptions: array[TMsgDlgBtn] of string = (
    SMsgDlgYes, SMsgDlgNo, SMsgDlgOK, SMsgDlgCancel, SMsgDlgAbort,
    SMsgDlgRetry, SMsgDlgIgnore, SMsgDlgAll, SMsgDlgNoToAll, SMsgDlgYesToAll,
    SMsgDlgHelp);
  defaultModalResults: array[TMsgDlgBtn] of Integer = (
    mrYes, mrNo, mrOk, mrCancel, mrAbort, mrRetry, mrIgnore, mrAll, mrNoToAll,
    mrYesToAll, 0);

function GetAveCharSize(Canvas: TCanvas): TPoint;
var
  I: Integer;
  Buffer: array[0..51] of Char;
begin
  for I := 0 to 25 do Buffer[I] := Chr(I + Ord('A'));
  for I := 0 to 25 do Buffer[I + 26] := Chr(I + Ord('a'));
  GetTextExtentPoint(Canvas.Handle, Buffer, 52, TSize(Result));
  Result.X := Result.X div 52;
end;

procedure TabcCustomDlg.DoCreateForm;
const
  mcFHorzMargin = 8;
  mcFVertMargin = 8;
  mcFHorzSpacing = 10;
  mcFVertSpacing = 10;
  mcButtonWidth = 50;
  mcButtonHeight = 14;
  mcButtonSpacing = 4;
  mcIconSize = 32;
  mcButtonExtraWidth = 16;
  mcCheckWidth = 21;
var
  DialogUnits: TPoint;
  X, ALeft, iButton: Integer;
  B, DefaultButton, CancelButton: TMsgDlgBtn;
  btnCap, msgText: string;
  IconID: PChar;
  TextRect, CheckBoxRect: TRect;
begin
  if FForm = nil then
  begin
    FForm := TForm.Create(self);
    with FForm do
    begin
      BiDiMode := Application.BiDiMode;
      BorderStyle := bsDialog;
      Font := Self.Font;
      Canvas.Font := Font;
      DialogUnits := GetAveCharSize(Canvas);
      FHorzMargin := MulDiv(mcFHorzMargin, DialogUnits.X, 4);
      FVertMargin := MulDiv(mcFVertMargin, DialogUnits.Y, 8);
      FHorzSpacing := MulDiv(mcFHorzSpacing, DialogUnits.X, 4);
      FVertSpacing := MulDiv(mcFVertSpacing, DialogUnits.Y, 8);
      FButtonWidth := MulDiv(mcButtonWidth, DialogUnits.X, 4);

      iButton := -1;
      for B := Low(TMsgDlgBtn) to High(TMsgDlgBtn) do
      begin
        if B in Buttons then
        begin
          iButton := iButton + 1;
          if FButtonWidths[B] = 0 then
          begin
            TextRect := Rect(0,0,0,0);
            if iButton < FButtonCaptions.Count then
              btnCap := FButtonCaptions[iButton]
            else
              btnCap := '';
            if btnCap = '' then
              btnCap := defaultButtonCaptions[B];
            DrawText(Canvas.Handle,
              PChar(btnCap), -1,
              TextRect, DT_CALCRECT or DT_LEFT or DT_SINGLELINE or
              DrawTextBiDiModeFlagsReadingOnly);
            with TextRect do
              FButtonWidths[B] := Right - Left + mcButtonExtraWidth;
          end;
          if FButtonWidths[B] > FButtonWidth then
            FButtonWidth := FButtonWidths[B];
        end;
      end;
      FButtonCount := iButton + 1;

      FButtonHeight := MulDiv(mcButtonHeight, DialogUnits.Y, 8);
      FButtonSpacing := MulDiv(mcButtonSpacing, DialogUnits.X, 4);

      //calc text rects for wordwrapped message and checkbox caption

      SetRect(TextRect, 0, 0, Screen.Width div 2, 0);
      msgText := StringsToDlgText(Self.Message);
      DrawText(Canvas.Handle, PChar(msgText), Length(msgText)+1, TextRect,
        DT_EXPANDTABS or DT_CALCRECT or DT_WORDBREAK or
        DrawTextBiDiModeFlagsReadingOnly);

      if CheckBoxVisible then
      begin
        SetRect(CheckBoxRect, 0, 0, Screen.Width div 2, 0);
        DrawText(Canvas.Handle, PChar(FCheckBoxCaption), -1, CheckBoxRect,
          DT_EXPANDTABS or DT_CALCRECT or DT_WORDBREAK or
          DrawTextBiDiModeFlagsReadingOnly);

        CheckBoxRect.Right := CheckBoxRect.Right + mcCheckWidth;
        if CheckBoxRect.Right > TextRect.Right then
        begin
          //resize TextRect
          TextRect.Right := CheckBoxRect.Right;
          TextRect.Bottom := 0;
          DrawText(Canvas.Handle, PChar(msgText), Length(msgText)+1, TextRect,
            DT_EXPANDTABS or DT_CALCRECT or DT_WORDBREAK or
            DrawTextBiDiModeFlagsReadingOnly);
        end
        else
        begin
          //resize CheckBoxRect
          CheckBoxRect.Right := TextRect.Right - mcCheckWidth;
          CheckBoxRect.Bottom := 0;
          DrawText(Canvas.Handle, PChar(FCheckBoxCaption), -1, CheckBoxRect,
            DT_EXPANDTABS or DT_CALCRECT or DT_WORDBREAK or
            DrawTextBiDiModeFlagsReadingOnly);
          CheckBoxRect.Right := CheckBoxRect.Right + mcCheckWidth;
        end;
        FIconTextWidth := abcMax(TextRect.Right, CheckBoxRect.Right);
      end
      else
        FIconTextWidth := TextRect.Right;

      IconID := defaultIconIDs[DlgType];
      FIconTextHeight := TextRect.Bottom;

      if not Self.Icon.Empty or (IconID <> nil) then
      begin
        if not Self.Icon.Empty then
        begin
          FIconWidth := Self.Icon.Width;
          FIconHeight := Self.Icon.Height;
        end
        else
        begin
          FIconWidth := mcIconSize;
          FIconHeight := mcIconSize;
        end;
        Inc(FIconTextWidth, FIconWidth + FHorzSpacing);
        if FIconTextHeight < FIconHeight then
          FIconTextHeight := FIconHeight;
      end
      else
      begin
        FIconWidth := 0;
        FIconHeight := 0;
      end;

      FButtonGroupWidth := 0;
      if FButtonCount <> 0 then
        FButtonGroupWidth := FButtonWidth * FButtonCount +
          FButtonSpacing * (FButtonCount - 1);

      ClientWidth := abcMax(FIconTextWidth, FButtonGroupWidth) + FHorzMargin * 2;
      ClientHeight := FIconTextHeight + FButtonHeight + FVertSpacing +
        FVertMargin * 2;

      if Self.Caption <> '' then
        Caption := Self.Caption
      else if DlgType <> mtCustom then
        Caption := defaultCaptions[DlgType]
      else
        Caption := Application.Title;

      if not Self.Icon.Empty or (IconID <> nil) then
        with TImage.Create(FForm) do
        begin
          Name := 'Image';
          Parent := FForm;
          if not Self.Icon.Empty then
            Picture.Icon := Self.Icon
          else
            Picture.Icon.Handle := LoadIcon(0, IconID);
          SetBounds(FHorzMargin, FVertMargin, FIconWidth, FIconHeight);
        end;

      FFormMessageLabel := TLabel.Create(FForm);
      with FFormMessageLabel do
      begin
        Name := 'Message';
        Parent := FForm;
        WordWrap := True;
        Caption := msgText;
        BoundsRect := TextRect;
        BiDiMode := FForm.BiDiMode;
        ALeft := FHorzMargin;
        if FIconWidth <> 0 then
          ALeft := ALeft + FIconWidth + FHorzSpacing;

        if UseRightToLeftAlignment then
          ALeft := FForm.ClientWidth - ALeft - Width;
        SetBounds(ALeft, FVertMargin,
          TextRect.Right, TextRect.Bottom);
      end;

      FButtonGroupTop := FIconTextHeight + FVertMargin + FVertSpacing;

      if CheckBoxVisible then
      begin
        FFormCheckBox := TabcCheckBox.Create(FForm);
        with FFormCheckBox do
        begin
          Name := 'CheckBox';
          Parent := FForm;
          Caption := CheckBoxCaption;
          Checked := Self.Checked;
          BiDiMode := FForm.BiDiMode;
          CheckBoxRect.Bottom := abcMax(CheckBoxRect.Bottom, Height);
          SetBounds(FFormMessageLabel.Left, FButtonGroupTop, CheckBoxRect.Right, CheckBoxRect.Bottom);
          FButtonGroupTop := FButtonGroupTop + Height + FVertSpacing;
          FForm.ClientHeight := FForm.ClientHeight + Height + FVertSpacing;
        end;
      end;

      if mbOk in Buttons then
        DefaultButton := mbOk
      else if mbYes in Buttons then
        DefaultButton := mbYes
      else
        DefaultButton := mbRetry;
      if mbCancel in Buttons then
        CancelButton := mbCancel
      else if mbNo in Buttons then
        CancelButton := mbNo
      else
        CancelButton := mbOk;

      X := (ClientWidth - FButtonGroupWidth) div 2;

      iButton := -1;

      for B := Low(TMsgDlgBtn) to High(TMsgDlgBtn) do
        if B in Buttons then
          with TButton.Create(FForm) do
          begin
            Name := defaultButtonNames[B];
            Parent := FForm;
            iButton := iButton + 1;
            TabOrder := iButton;
            if iButton < FButtonCaptions.Count then
              Caption := FButtonCaptions[iButton]
            else
              Caption := '';
            if Caption = '' then
              Caption := defaultButtonCaptions[B];
            ModalResult := defaultModalResults[B];
            if B = DefaultButton then
              Default := True;
            if B = CancelButton then
              Cancel := True;
            SetBounds(X, FButtonGroupTop,
              FButtonWidth, FButtonHeight);
            Inc(X, FButtonWidth + FButtonSpacing);
            if B = mbHelp then
              OnClick := HelpButtonClick;
          end;

      HelpContext := Self.HelpContext;

      OnShow := FormShow;
      OnHide := FormHide;
    end;
  end;
end;

procedure TabcCustomDlg.DoPositionForm;
var
  calcDlgPosition: TabcPopupPosition;
  newLeft, newTop: Integer;
begin
  calcDlgPosition := DlgPosition;

  if ((calcDlgPosition = appActiveFormCenter) and (Screen.ActiveForm = nil)) then
    calcDlgPosition := appMainFormCenter;
  if ((calcDlgPosition = appMainFormCenter) and (Application.MainForm = nil)) then
    calcDlgPosition := appScreenCenter;

  case calcDlgPosition of
    appScreenCenter:
      begin
        newLeft := (Screen.Width - FForm.Width) div 2;
        newTop := (Screen.Height - FForm.Height) div 2;
      end;
    appMainFormCenter:
      with Application.MainForm do
      begin
        newLeft := Left + (Width - FForm.Width) div 2;
        newTop := Top + (Height - FForm.Height) div 2;
      end;
    appActiveFormCenter:
      with Screen.ActiveForm do
      begin
        newLeft := Left + (Width - FForm.Width) div 2;
        newTop := Top + (Height - FForm.Height) div 2;
      end;
  else {appCustom}
    begin
      newLeft := DlgLeft;
      newTop := DlgTop;
    end;
  end;

  with FForm do SetBounds(newLeft, newTop, Width, Height);
end;

procedure TabcCustomDlg.DoFormShow;
begin
  if Assigned(OnShow) then
    OnShow(Form);
  if CheckBoxVisible then
    FWasChecked := FFormCheckBox.Checked
  else
    FWasChecked := FChecked;
end;

procedure TabcCustomDlg.DoFormHide;
begin
  if CheckBoxVisible then
    FChecked := FFormCheckBox.Checked;
  if Assigned(OnClose) then
    OnClose(Form);
  if (FWasChecked <> FChecked) and Assigned(OnChecked) then
    OnChecked(Self);
end;

procedure TabcCustomDlg.HelpButtonClick(Sender: TObject);
begin
  Application.HelpContext(HelpContext);
end;

function TabcCustomDlg.DoExecute: TModalResult;
begin
  if CheckBoxVisible and (CheckBoxResult <> mrNone) then
    Result := CheckBoxResult
  else
  begin
    DoCreateForm;
    DoPositionForm;
    try
      FResult := mrNone;
      FResult := FForm.ShowModal;
      Result := FResult;
    finally
      FForm.Free;
      FForm := nil;
      FFormMessageLabel := nil;
      FFormCheckBox := nil;
    end;
  end;
end;

function TabcCustomDlg.Execute: TModalResult;
begin
  Result := DoExecute;
end;

procedure TabcCustomDlg.FormHide(Sender: TObject);
begin
  DoFormHide;
end;

procedure TabcCustomDlg.FormShow(Sender: TObject);
begin
  DoFormShow;
end;

function TabcCustomDlg.IsFontStored: Boolean;
var
  NonClientMetrics: TNonClientMetrics;
  msgFont: TFont;
begin
  msgFont := TFont.Create;
  try
    NonClientMetrics.cbSize := sizeof(NonClientMetrics);
    if SystemParametersInfo(SPI_GETNONCLIENTMETRICS, 0, @NonClientMetrics, 0) then
      msgFont.Handle := CreateFontIndirect(NonClientMetrics.lfMessageFont);
    Result :=
      (msgFont.Charset <> Font.Charset) or
      (msgFont.Color <> Font.Color) or
      (msgFont.Name <> Font.Name) or
      (msgFont.Size <> Font.Size) or
      (msgFont.Style <> Font.Style);
  finally
    msgFont.Free;
  end;
end;

procedure TabcCustomDlg.SetButtonCaptions(const Value: TStrings);
begin
  if Value <> FButtonCaptions then
    FButtonCaptions.Assign(Value);
end;

procedure TabcCustomDlg.SetFont(const Value: TFont);
begin
  if Value <> FFont then
    FFont.Assign(Value);
end;

procedure TabcCustomDlg.SetIcon(const Value: TIcon);
begin
  if Value <> FIcon then
    FIcon.Assign(Value);
end;

procedure TabcCustomDlg.SetMessage(const Value: TStrings);
begin
  if Value <> FMessage then
    FMessage.Assign(Value);
end;

procedure TabcCustomDlg.SetDlgLeft(const Value: Integer);
begin
  if Value <> FDlgLeft then
  begin
    FDlgLeft := Value;
    FDlgPosition := appCustom;
  end;
end;

procedure TabcCustomDlg.SetDlgPosition(const Value: TabcPopupPosition);
begin
  if Value <> FDlgPosition then
  begin
    FDlgPosition := Value;
    if Value <> appCustom then
    begin
      FDlgLeft := -1;
      FDlgTop := -1;
    end;
  end;
end;

procedure TabcCustomDlg.SetDlgTop(const Value: Integer);
begin
  if Value <> FDlgTop then
  begin
    FDlgTop := Value;
    FDlgPosition := appCustom;
  end;
end;

{ TabcProgressDlg }

constructor TabcProgressDlg.Create(AOwner: TComponent);
begin
  inherited;
  FProgressBarVisible := True;
  FMin := 0;
  FMax := 60;
  FStep := -1;
  FTimerInterval := 1000;
end;

procedure TabcProgressDlg.Assign(Source: TPersistent);
begin
  inherited;
  if Source is TabcProgressDlg then
    with TabcProgressDlg(Source) do
    begin
      Self.ProgressBarVisible := ProgressBarVisible;
      Self.Min := Min;
      Self.Max := Max;
      Self.Step := Step;
      Self.TimerInterval := TimerInterval;
      Self.OnTimer := OnTimer;
      Self.OnTimeout := OnTimeout;
    end;
end;

procedure TabcProgressDlg.DoCreateForm;
var
  adjustHeight, i: integer;
begin
  inherited;
  FTimer := TTimer.Create(Form);
  with FTimer do
  begin
    Enabled := False;
    Interval := TimerInterval;
    OnTimer := TimerTick;
  end;
  if Step > 0 then
    FValue := Min
  else
    FValue := Max;

  UpdateMessageLabel;

  if ProgressBarVisible then
  begin
    FProgressBar := TProgressBar.Create(Form);
    with FProgressBar do
    begin
      Name := 'ProgressBar';
      Parent := Form;
      Min := Self.Min;
      Max := Self.Max;
      Step := Self.Step;
      Position := Self.Value;
      Smooth := True;
      SetBounds(FHorzMargin, FButtonGroupTop, Form.ClientWidth - 2 * FHorzMargin, Height);
      adjustHeight := Height + FVertSpacing;
      FButtonGroupTop := FButtonGroupTop + adjustHeight;
    end;
    with Form do
    begin
      //move buttons down
      for i := 0 to ComponentCount - 1 do
        if Components[i] is TButton then
          with TButton(Components[i]) do
            Top := FButtonGroupTop;
      //resize form
      Height := Height + adjustHeight;
    end;
  end;
end;

function TabcProgressDlg.DoExecute: TModalResult;
begin
  Result := inherited DoExecute;
  if (Result = abcmrTimeout) and Assigned(OnTimeout) then
    OnTimeout(Self);
end;

procedure TabcProgressDlg.DoFormShow;
begin
  inherited;
  FTimer.Enabled := True;
end;

procedure TabcProgressDlg.DoFormHide;
begin
  FTimer.Enabled := False;
  inherited;
end;

procedure TabcProgressDlg.TimerTick(Sender: TObject);
begin
  FValue := FValue + Step;
  UpdateMessageLabel;
  if ProgressBarVisible then
    FProgressBar.StepIt;
  if Assigned(OnTimer) then
    OnTimer(Self);
  if ((Step > 0) and (Value >= Max)) or (Value <= Min) then
    Form.ModalResult := abcmrTimeout;
end;

procedure TabcProgressDlg.UpdateMessageLabel;
begin
  FFormMessageLabel.Caption := Format(StringsToDlgText(Message), [Value]);
end;

end.
