unit abcdsexc;

(*------------------------------------------------------------------------------

Advanced Business Components
ABC for Delphi
Version 6

Copyright  1995, 2001 Objective Software Technology Pty Limited ACN 068 640 353
Functionality derived from Delphi VCL Source Code
is Copyright  1995, 2001 Borland Software Corporation

------------------------------------------------------------------------------*)

{$I abcdefs.inc}

interface

uses
  Windows, SysUtils, Messages, Classes, Controls, Forms, Dialogs,
  DB, abcexcep, abcapp, abcda;

type

   {TabcDSAppException}
   { uses virtual TDataset interface to external datasets }
   {Adds a lookup table for message processing and a log table for exception log}
   {Retains the default message and log properties of TabcAppException for
    use when the database connection fails or a message lookup isn't found}

  TabcDSExcLogDatalink = class;
  TabcDSExcMsgDatalink = class;

  TabcDSAppException = class(TabcAppException)
  private
    FStreamedActive: Boolean;
    FCancel: Boolean;
    FDBExcepMsg: TabcExcepMsg;
    FLogDataLink: TabcDSExcLogDatalink;
    FLogDataset: TDataset;
    FLogAdapter: TabcDatasetAdapter;
    FMsgDataLink: TabcDSExcMsgDatalink;
    FMsgDataset: TDataset;
    FMsgAdapter: TabcDatasetAdapter;
    FExtraLogInfo: TStrings;
    function GetActive: Boolean;
    procedure SetActive(Value: Boolean);
    function GetLogSource: TDataSource;
    procedure SetLogSource(Value: TDataSource);
    function GetMsgSource: TDataSource;
    procedure SetMsgSource(Value: TDataSource);
  protected
    procedure LogActiveChanged;
    procedure MsgActiveChanged;
    procedure ApplicationExceptionHandler(Sender: TObject; E: Exception); override;
    function DBShowExceptionMsg(Sender: TObject; E: Exception; var Log: Boolean): Boolean; virtual;
    function DBLogException(Sender: TObject; E: Exception): Boolean; virtual;
    procedure Loaded; override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure AppendLogMessage(Sender: TObject; E: Exception; ExtraInfo: TStrings); override;
    property ExtraLogInfo: TStrings read FExtraLogInfo;
  published
    property Active: Boolean read GetActive write SetActive default True;
    property LogDataSource: TDataSource read GetLogSource write SetLogSource;
    property MsgDataSource: TDataSource read GetMsgSource write SetMsgSource;
  end;

  { TabcDSExcLogDatalink }

  TabcDSExcLogDatalink = class(TDataLink)
  private
    FAppException: TabcDSAppException;
  protected
    procedure ActiveChanged; override;
  public
    constructor Create(AppException: TabcDSAppException);
    destructor Destroy; override;
  end;

  { TabcDSExcMsgDatalink }

  TabcDSExcMsgDatalink = class(TDataLink)
  private
    FAppException: TabcDSAppException;
  protected
    procedure ActiveChanged; override;
  public
    constructor Create(AppException: TabcDSAppException);
    destructor Destroy; override;
  end;

implementation

uses abcconst;

{ Get logical address from physical address as formatted string }

function StrLogicalAddr(PhysicalAddr: Pointer): string;
begin
  Result := Format('%p', [PhysicalAddr]);
end;

{TabcDSAppException}

constructor TabcDSAppException.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FDBExcepMsg := TabcExcepMsg.Create;
  FLogDataLink := TabcDSExcLogDatalink.Create(Self);
  FMsgDataLink := TabcDSExcMsgDatalink.Create(Self);
  if not (csDesigning in ComponentState) then
    AppException := TabcAppException(Self);
  FStreamedActive := True;
end;

destructor TabcDSAppException.Destroy;
begin
  FDBExcepMsg.Free;
  FLogAdapter.Free;
  FLogAdapter := nil;
  FMsgAdapter.Free;
  FMsgAdapter := nil;
  FLogDataLink.Free;
  FLogDataLink := nil;
  FMsgDataLink.Free;
  FMsgDataLink := nil;
  if AppException = Self then
    AppException := nil;
  inherited Destroy;
end;

procedure TabcDSAppException.Loaded;
begin
  inherited Loaded;
  LogActiveChanged;
  MsgActiveChanged;
end;

procedure TabcDSAppException.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited;
  if (Operation = opRemove) and (AComponent is TDataSource) then
  begin
    if (FLogDataLink <> nil) and (AComponent = LogDataSource) then LogDataSource := nil;
    if (FMsgDataLink <> nil) and (AComponent = MsgDataSource) then MsgDataSource := nil;
  end;
end;

function TabcDSAppException.GetActive: Boolean;
begin
  Result := FLogDataLink.Active and FMsgDataLink.Active;
end;

procedure TabcDSAppException.SetActive(Value: Boolean);
begin
  if (csLoading in ComponentState) then
    FStreamedActive := Value
  else if Value then
  try
  except
    on E: Exception do ShowMessage('Failed to open DBAppException tables: ' + E.Message);
  end
  else begin
    FLogDataset.Close;
    FMsgDataset.Close;
  end;
end;

function TabcDSAppException.GetLogSource: TDataSource;
begin
  Result := FLogDataLink.DataSource;
end;

procedure TabcDSAppException.SetLogSource(Value: TDataSource);
begin
  FLogDataLink.DataSource := Value;
  if not (csLoading in ComponentState) then
    LogActiveChanged;
end;

function TabcDSAppException.GetMsgSource: TDataSource;
begin
  Result := FMsgDataLink.DataSource;
end;

procedure TabcDSAppException.SetMsgSource(Value: TDataSource);
begin
  FMsgDataLink.DataSource := Value;
  if not (csLoading in ComponentState) then
    MsgActiveChanged;
end;

procedure TabcDSAppException.LogActiveChanged;
begin
  if FLogDataset <> FLogDataLink.Dataset then
  begin
    FLogAdapter.Free;
    FLogAdapter := nil;
    FLogDataset := FLogDataLink.Dataset;
    if FLogDataset <> nil then
      FLogAdapter := abcDSAdapterFactory.CreateAdapter(FLogDataset);
  end;
end;

procedure TabcDSAppException.MsgActiveChanged;
begin
  if FMsgDataset <> FMsgDataLink.Dataset then
  begin
    FMsgAdapter.Free;
    FMsgAdapter := nil;
    FMsgDataset := FMsgDataLink.Dataset;
    if FMsgDataset <> nil then
      FMsgAdapter := abcDSAdapterFactory.CreateAdapter(FMsgDataset);
  end;
end;

procedure TabcDSAppException.ApplicationExceptionHandler(Sender: TObject; E: Exception);
var
  FLogExcep: Boolean;
  FOnException: TabcExceptionEvent;
begin
  CurrExcepAddr := StrLogicalAddr(ExceptAddr);
  FCancel := false;
  FOnException := OnException;
  if Assigned(FOnException) then OnException(Sender, E, FCancel);
  if not FCancel then DBShowExceptionMsg(Sender, E, FLogExcep);
  if FLogExcep and ExcepLog.Enabled then DBLogException(Sender, E);
  if Terminate then
    if not ConfirmTerminate or
      (MessageDlg(sabcConfirmExcepTerminate, mtConfirmation,
       [mbYes, mbNo], 0) = mrYes) then
      Application.Terminate;
end;

function TabcDSAppException.DBLogException(Sender: TObject; E: Exception): Boolean;
var
  ExtraStrings: TStringList;
  ALogExtraInfo: TabcLogAppendEvent;
begin
  Result := True;
  ExtraStrings := TStringList.Create;
  try
    try
      ALogExtraInfo := LogExtraInfo;
      if Assigned(ALogExtraInfo) then
        ALogExtraInfo(Self, ExtraStrings);
      AppendLogMessage(Sender, E, ExtraStrings);
    except
      on E: Exception do begin
        Result := False;
        ShowMessage('Error writing to error log: ' + E.Message);
        ExcepLog.AppendLogMessage(Sender, E, ExtraStrings);
      end;
    end;
  finally
    ExtraStrings.Free;
  end;
end;

procedure TabcDSAppException.AppendLogMessage(Sender: TObject; E: Exception; ExtraInfo: TStrings);
var
  LogExcepAddr: string;
begin
  if FLogAdapter <> nil then
    with FLogDataset, FLogAdapter do
    begin
      Append;
      {assign fields}
      FieldByName('ERR_TIME').AsDateTime := Now;
      if ExcepLog.UserName <> '' then
        FieldByName('USER_NAME').AsString := ExcepLog.UserName
      else
        FieldByName('USER_NAME').AsString := NetUserName;
      if Assigned(E) then
      begin
        FieldByName('ERR_CLASS').AsString := E.ClassName;
        FieldByName('ERR_MSG').AsString := E.Message;
      end
      else begin
        FieldByName('ERR_CLASS').AsString := '';
        FieldByName('ERR_MSG').AsString := '';
      end;
      if ExcepLog.LogExeName then
        FieldByName('EXE_NAME').AsString := Application.ExeName
      else
        FieldByName('EXE_NAME').AsString := '';
      if ExcepLog.LogAppTitle then
        FieldByName('APP_NAME').AsString := Application.Title
      else
        FieldByName('APP_NAME').AsString := '';
      if ExcepLog.LogSender and Assigned(Sender) then begin
        FieldByName('SNDR_CLASS').AsString := Sender.ClassName;
        if Sender is TComponent then
          FieldByName('SNDR_NAME').AsString := TComponent(Sender).Name
        else
          FieldByName('SNDR_NAME').AsString := '';
      end
      else begin
        FieldByName('SNDR_CLASS').AsString := '';
        FieldByName('SNDR_NAME').AsString := '';
      end;
      if ExcepLog.LogDebug then begin
        LogExcepAddr := StrLogicalAddr(ExceptAddr);
        FieldByName('ERR_ADDR').AsString := LogExcepAddr;
        if (E is EabcOverrideSysError) then
          with EabcOverrideSysError(E) do begin
            FieldByName('SYS_CLASS').AsString := SysClass;
            FieldByName('SYS_MSG').AsString := SysMsg;
            FieldByName('SYS_ADDR').AsString := SysAddr;
          end
        else begin
          FieldByName('SYS_CLASS').AsString := '';
          FieldByName('SYS_MSG').AsString := '';
          FieldByName('SYS_ADDR').AsString := '';
        end;
      end
      else begin
        FieldByName('ERR_ADDR').AsString := '';
        FieldByName('SYS_CLASS').AsString := '';
        FieldByName('SYS_MSG').AsString := '';
        FieldByName('SYS_ADDR').AsString := '';
      end;
      FExtraLogInfo := ExtraInfo;
      {Write Record}
      Post;
      FExtraLogInfo := nil;
    end;
end;

function TabcDSAppException.DBShowExceptionMsg(Sender: TObject; E: Exception; var Log: Boolean): Boolean;
var
  recLocated: Boolean;
begin
  Result := True;
  Log := True;
  try
    if FMsgAdapter <> nil then
      with FMsgDataset, FMsgAdapter do
        if Active then
        begin
          if SearchMethod = asmFindKey then
            recLocated := FindKey(E.ClassName)
          else
            recLocated := Locate('ERR_CLASS', E.ClassName, []);
          if recLocated then
            with FDBExcepMsg do begin
              {assign defaults from component properties}
              Beep := ExcepMsg.Beep;
              Enabled := ExcepMsg.Enabled;
              MsgBox := ExcepMsg.MsgBox;
              ShowDebug := ExcepMsg.ShowDebug;
              ShowExcMsg := ExcepMsg.ShowExcMsg;
              Title := ExcepMsg.Title;
              {assign fields from MsgTable}
              case FieldByName('MSG_TYPE').AsInteger of
                0: MsgType := mtWarning;
                1: MsgType := mtError;
                2: MsgType := mtInformation;
                3: MsgType := mtConfirmation;
                4: MsgType := mtCustom;
              end;
              Terminate := FieldByName('TERMINATE').AsBoolean;
              Log := FieldByName('LOG').AsBoolean;
              Lines.Assign(FieldByName('ERR_MSG'));
              ShowExceptionMsg(E);
            end
          else
            ExcepMsg.ShowExceptionMsg(E);
        end
        else
          ExcepMsg.ShowExceptionMsg(E)
    else
      ExcepMsg.ShowExceptionMsg(E);
  except
    on E: Exception do begin
      Result := False;
      ShowMessage('Error message DB error: ' + E.Message);
      ExcepMsg.ShowExceptionMsg(E);
    end;
  end;
end;

{ TabcDSExcLogDatalink }

constructor TabcDSExcLogDatalink.Create(AppException: TabcDSAppException);
begin
  inherited Create;
  FAppException := AppException;
end;

destructor TabcDSExcLogDatalink.Destroy;
begin
  FAppException := nil;
  inherited Destroy;
end;

procedure TabcDSExcLogDatalink.ActiveChanged;
begin
  if FAppException <> nil then FAppException.LogActiveChanged;
end;

{ TabcDSExcMsgDatalink }

constructor TabcDSExcMsgDatalink.Create(AppException: TabcDSAppException);
begin
  inherited Create;
  FAppException := AppException;
end;

destructor TabcDSExcMsgDatalink.Destroy;
begin
  FAppException := nil;
  inherited Destroy;
end;

procedure TabcDSExcMsgDatalink.ActiveChanged;
begin
  if FAppException <> nil then FAppException.MsgActiveChanged;
end;


end.
