unit abcexcep;

(*------------------------------------------------------------------------------

Advanced Business Components
ABC for Delphi
Version 6

Copyright  1995, 2001 Objective Software Technology Pty Limited ACN 068 640 353
Functionality derived from Delphi VCL Source Code
is Copyright  1995, 2001 Borland Software Corporation

------------------------------------------------------------------------------*)

{$I abcdefs.inc}

interface

uses
  Windows, SysUtils, Messages, Classes, Controls, Forms, Dialogs, abcapp;

type

{EabcOverrideSysError}
(*------------------------------------------------------------------------------
This class is designed to allow easy override of system errors, while
retaining original information for debugging purposes.

This would typically be used in your own exception processing to
raise a new exception of class EabcOverrideSysError using the original Exception.

The following optional additional string fields are provided in addition to the regular
Exception class properties:

  SysClass:   string  If an unhandled system error has occured, this can be used to
                      store the original class of the error.
  SysMsg:     string  If an unhandled system error has already occured, this can be
                      used to record the original message
  SysAddr:    Logical address of the system error for debugging.

These fields are also used by error logging functions in this unit.

Following is an example of using the constructor:

try
  a := 0;
  b := b div a;
except
  on E:Exception do
    raise EabcOverrideSysError.Create(E, 'Your message', ExceptAddr);
end;

------------------------------------------------------------------------------*)

  EabcOverrideSysError = class(Exception)
  private
    FSysClass: string;
    FSysMsg: string;
    FSysAddr: string;
  public
    constructor Create(const pSysException: Exception;
      const pAppMsg: string; ExcAddr: Pointer);
    property SysAddr: string read FSysAddr write FSysAddr;
    property SysClass: string read FSysClass write FSysClass;
    property SysMsg: string read FSysMsg write FSysMsg;
    {Also inherits public properties and methods of Exception class}
  end;

  TabcErrMsgBoxType = (mbDelphi, mbWinAPI);

{TabcExcepMsg}

type
  TabcExcepMsg = class(TPersistent)
  private
    FEnabled: Boolean;
    FBeep: Boolean;
    FMsgBox: TabcErrMsgBoxType;
    FTitle: string;
    FLines: TStrings;
    FShowDebug: Boolean;
    FShowExcMsg: Boolean;
    FMsgType: TMsgDlgType;
    procedure SetLines(Value: TStrings);
    procedure DoBeep;
  protected
    procedure ShowExceptionDelphi(E: Exception); virtual;
    procedure ShowExceptionWinAPI(E: Exception); virtual;
  public
    constructor Create; virtual;
    destructor Destroy; override;
    procedure ShowExceptionMsg(E: Exception); virtual;
  published
    property Beep: Boolean read FBeep write FBeep;
    property Enabled: Boolean read FEnabled write FEnabled default True;
    property MsgType: TMsgDlgType read FMsgType write FMsgType default mtWarning;
    property Lines: TStrings read FLines write SetLines;
    property MsgBox: TabcErrMsgBoxType read FMsgBox write FMsgBox default mbWinAPI;
    property ShowDebug: Boolean read FShowDebug write FShowDebug;
    property ShowExcMsg: Boolean read FShowExcMsg write FShowExcMsg default True;
    property Title: string read FTitle write FTitle;
  end;

{TabcExceptionEvent}

type
  TabcExceptionEvent = procedure (Sender: TObject; E: Exception; var Cancel: Boolean) of object;

{TabcLogAppendEvent}

type
  TabcLogAppendEvent = procedure (Sender: TObject; Fields: TStrings) of object;

{TabcExcepLog}

type
  TabcExcepLog = class(TPersistent)
  private
    FEnabled: Boolean;
    FFileName: TFileName;
    FUserName: string;
    FLogExeName: Boolean;
    FLogAppTitle: Boolean;
    FLogSender: Boolean;
    FLogDebug: Boolean;
    FLogExtraInfo: TabcLogAppendEvent;
  public
    constructor Create; virtual;
    procedure LogException(Sender: TObject; E: Exception); virtual;
    procedure AppendLogMessage(Sender: TObject; E: Exception; ExtraInfo: TStrings); virtual;
  published
    property Enabled: Boolean read FEnabled write FEnabled default True;
    property FileName: TFileName read FFileName write FFileName;
    property UserName: string read FUserName write FUserName;
    property LogExeName: Boolean read FLogExeName write FLogExeName default True;
    property LogAppTitle: Boolean read FLogAppTitle write FLogAppTitle default True;
    property LogSender: Boolean read FLogSender write FLogSender default True;
    property LogDebug: Boolean read FLogDebug write FLogDebug default True;
    property LogExtraInfo: TabcLogAppendEvent read FLogExtraInfo write FLogExtraInfo;
  end;

  {TabcAppException}

type
  TabcAppException = class(TComponent)
  private
    FCancel: Boolean;
    FExcepMsg: TabcExcepMsg;
    FExcepLog: TabcExcepLog;
    FTerminate: Boolean;
    FConfirmTerminate: Boolean;
    FOnException: TabcExceptionEvent;
    FCurrExcepAddr: string;
    function GetLogExtraInfo: TabcLogAppendEvent;
    procedure SetLogExtraInfo(Value: TabcLogAppendEvent);
  protected
    procedure ApplicationExceptionHandler(Sender: TObject; E: Exception); virtual;
    property CurrExcepAddr: string read FCurrExcepAddr write FCurrExcepAddr;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure AppendLogMessage(Sender: TObject; E: Exception; ExtraInfo: TStrings); virtual;
  published
    property ExcepMsg: TabcExcepMsg read FExcepMsg write FExcepMsg;
    property ExcepLog: TabcExcepLog read FExcepLog write FExcepLog;
    property Terminate: Boolean read FTerminate write FTerminate;
    property ConfirmTerminate: Boolean read FConfirmTerminate write FConfirmTerminate default False;
    property OnException: TabcExceptionEvent read FOnException write FOnException;
    property LogExtraInfo: TabcLogAppendEvent read GetLogExtraInfo write SetLogExtraInfo;
  end;

var
  {global instance}
  AppException: TabcAppException;

implementation

uses abcconst;

{ Get logical address from physical address as formatted string }

function StrLogicalAddr(PhysicalAddr: Pointer): string;
begin
  Result := Format('%p', [PhysicalAddr]);
end;

{EabcOverrideSysError}

constructor EabcOverrideSysError.Create(const pSysException: Exception;
  const pAppMsg: string; ExcAddr: Pointer);
begin
  inherited Create(pAppMsg);
  SysMsg := pSysException.Message;
  SysClass := pSysException.ClassName;
  SysAddr := StrLogicalAddr(ExcAddr);
end;

{TabcExcepMsg}

constructor TabcExcepMsg.Create;
begin
  FEnabled := True;
  FMsgBox := mbWinAPI;
  FLines := TStringList.Create;
  FMsgType := mtWarning;
  FShowExcMsg := True;
end;

destructor TabcExcepMsg.Destroy;
begin
  FLines.Free;
  inherited Destroy;
end;

procedure TabcExcepMsg.SetLines(Value: TStrings);
begin
  FLines.Assign(Value);
end;

procedure TabcExcepMsg.ShowExceptionMsg(E: Exception);
begin
  if FEnabled then begin
    if FBeep then DoBeep;
    case FMsgBox of
      mbDelphi: ShowExceptionDelphi(E);
      mbWinAPI: ShowExceptionWinAPI(E);
    end;
  end;
end;

procedure TabcExcepMsg.DoBeep;
var
  BeepType: word;
begin
  case FMsgType of
    mtWarning:
      BeepType := MB_ICONEXCLAMATION;
    mtError:
      BeepType := MB_ICONSTOP;
    mtInformation:
      BeepType := MB_ICONINFORMATION;
    mtConfirmation:
      BeepType := MB_ICONQUESTION;
    mtCustom:
      BeepType := MB_OK;
  else
    BeepType := MB_ICONEXCLAMATION;
  end;
  MessageBeep(BeepType);
end;

procedure TabcExcepMsg.ShowExceptionDelphi(E: Exception);
{Shows Delphi style message box(es)}
var
  i: integer;
begin
  for i := 0 to FLines.Count - 1 do
    MessageDlg(FLines.Strings[i],FMsgType, [mbOK], 0);
  if FShowExcMsg then
    MessageDlg(E.Message, FMsgType, [mbOK], 0);
  if FShowDebug then begin
    MessageDlg('Exception occured at: ' + AppException.CurrExcepAddr, FMsgType,
      [mbOK], 0);
    if (E is EabcOverrideSysError) then
      with E as EabcOverrideSysError do begin
        if SysClass <> '' then MessageDlg('SysClass: ' + SysClass, FMsgType, [mbOK], 0);
        if SysMsg <> '' then MessageDlg('SysMsg: ' + SysMsg, FMsgType, [mbOK], 0);
        if SysAddr <> '' then MessageDlg('SysAddr: ' + SysAddr, FMsgType, [mbOK], 0);
      end;
  end;
end;

procedure TabcExcepMsg.ShowExceptionWinAPI(E: Exception);
{Shows a Windows API style message box}
var
  MessageLines: pChar;
  MessageText: pChar;
  MessageTitle: array[0..255] of char;
  TempChar: array[0..255] of char;
  MessageFlags: word;
  BufSize: integer;
begin
  MessageLines := PChar(FLines.Text);
  BufSize := StrLen(MessageLines) + 1;
  if FShowExcMsg then BufSize := BufSize + Length(E.Message);
  if FShowDebug then
    BufSize := BufSize + 257 * 4;
  {Alloc & Fill buffer with message text}
  MessageText := StrAlloc(BufSize);
  try
    try
      StrCopy(MessageText, MessageLines);
      if FShowExcMsg then begin
        if E.Message <> '' then begin
          StrPCopy(TempChar,E.Message);
          StrCat(MessageText,TempChar);
        end;
      end;
      if FShowDebug then begin
        StrPCopy(TempChar, 'Address: ' + AppException.CurrExcepAddr);
        StrCat(MessageText, #13);
        StrCat(MessageText, #10);
        StrCat(MessageText, TempChar);
        if (E is EabcOverrideSysError) then
          with E as EabcOverrideSysError do begin
            if SysClass <> '' then begin
              StrPCopy(TempChar, 'SysClass: ' + SysClass);
              StrCat(MessageText, #13);
              StrCat(MessageText, #10);
              StrCat(MessageText, TempChar);
            end;
            if SysMsg <> '' then begin
              StrPCopy(TempChar, 'SysMsg: ' + SysMsg);
              StrCat(MessageText, #13);
              StrCat(MessageText, #10);
              StrCat(MessageText, TempChar);
            end;
            if SysAddr <> '' then begin
              StrPCopy(TempChar, 'SysAddr: ' + SysAddr);
              StrCat(MessageText, #13);
              StrCat(MessageText, #10);
              StrCat(MessageText, TempChar);
            end;
          end;
      end;
      if FTitle = '' then
        StrPCopy(MessageTitle,Application.Title)
      else
        StrPCopy(MessageTitle,FTitle);

      case FMsgType of
        mtWarning:
          MessageFlags := MB_OK + MB_ICONEXCLAMATION;
        mtError:
          MessageFlags := MB_OK + MB_ICONSTOP;
        mtInformation:
          MessageFlags := MB_OK + MB_ICONINFORMATION;
        mtConfirmation:
          MessageFlags := MB_OK + MB_ICONQUESTION;
        mtCustom:
          MessageFlags := MB_OK;
      else
        MessageFlags := MB_OK;
      end;
      {default message}
      if BufSize = 1 then
        StrPCopy(MessageText,'An exception occured in your application.');
      Application.MessageBox(MessageText, MessageTitle, MessageFlags);
    except
      on Exception do {nothing, to avoid circular error};
    end;
  finally
    StrDispose(MessageText);
  end;
end;

{TabcExcepLog}

constructor TabcExcepLog.Create;
begin
  FEnabled := True;
  FLogExeName := True;
  FLogAppTitle := True;
  FLogSender := True;
  FLogDebug := True;
end;

procedure TabcExcepLog.LogException(Sender: TObject; E: Exception);
var
  ExtraStrings: TStringList;
begin
  ExtraStrings := TStringList.Create;
  try
    if Assigned(FLogExtraInfo) then
      FLogExtraInfo(Self, ExtraStrings);
    AppendLogMessage(Sender, E, ExtraStrings);
  finally
    ExtraStrings.Free;
  end;
end;

procedure TabcExcepLog.AppendLogMessage(Sender: TObject; E: Exception; ExtraInfo: TStrings);
var
  F: TextFile;
  ErrorFileName: TFileName;
  SenderName: string;
  I: integer;
  LogExcepAddr: string;
begin
  if FileName = '' then begin
    ErrorFileName := Application.ExeName;
    ErrorFileName := Copy(ErrorFileName,1, Length(ErrorFileName)-4) + '.LOG';
  end
  else
    ErrorFileName := FileName;
  AssignFile(F, ErrorFileName);
  try
    Append(F);
  except
    on EInOutError do begin
      Rewrite(F);
      {Write heading strings}
      Writeln(F, 'DateTime,User,Exception Class,Exception Message,'
        + 'EXE,Application,Sender Class, Sender ComponentName,'
        + 'Exception Address, System Class, System Msg, System Address');
    end;
  end;
  Write(F, DateTimeToStr(Now) + ',');
  if FUserName <> '' then  Write(F, FUserName);
  Write(F,',');
  if Assigned(E) then
  begin
    Write(F, E.ClassName + ',');
    Write(F, E.Message + ',');
  end
  else
    Write(F,',,');
  if FLogExeName then Write(F, Application.ExeName);
  Write(F,',');
  if FLogAppTitle then  Write(F, Application.Title);
  Write(F,',');
  if FLogSender and Assigned(Sender) then begin
    Write(F, Sender.ClassName + ',');
    if Sender is TComponent then begin
      SenderName := (Sender as TComponent).Name;
      Write(F, SenderName);
    end;
    Write(F,',');
  end
  else
    Write(F,',,');
  if FLogDebug then begin
    LogExcepAddr := StrLogicalAddr(ExceptAddr);
    Write(F, LogExcepAddr + ',');
    if (E is EabcOverrideSysError) then
      with E as EabcOverrideSysError do begin
        Write(F, SysClass + ',');
        Write(F, SysMsg + ',');
        Write(F, SysAddr);
      end
    else
      Write(F, ',,');
    end
  else
    Write(F,',,,');

  if Assigned(ExtraInfo) then
    with ExtraInfo do
      for I := 0 to Count - 1 do
        Write(F,',' + Strings[I]);

  Writeln(F, '');
  CloseFile(F);
end;

{TabcAppException}

constructor TabcAppException.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FExcepLog := TabcExcepLog.Create;
  FExcepMsg := TabcExcepMsg.Create;
  if not (csDesigning in ComponentState) then
  begin
    AppException := self; {global instance}
    abcApplicationEvents.RegisterOnException(ApplicationExceptionHandler);
  end;
end;

destructor TabcAppException.Destroy;
begin
  FExcepMsg.Free;
  FExcepLog.Free;
  if AppException = self then
    AppException := nil;
  if not (csDesigning in ComponentState) then
    abcApplicationEvents.UnRegisterOnException(ApplicationExceptionHandler);
  inherited Destroy;
end;

procedure TabcAppException.ApplicationExceptionHandler(Sender: TObject; E: Exception);
begin
  CurrExcepAddr := StrLogicalAddr(ExceptAddr);
  FCancel := false;
  if Assigned(FOnException) then OnException(Sender, E, FCancel);
  if not FCancel then begin
    FExcepMsg.ShowExceptionMsg(E);
    FExcepLog.LogException(Sender, E);
    if FTerminate then
      if not ConfirmTerminate or
        (MessageDlg(sabcConfirmExcepTerminate, mtConfirmation,
         [mbYes, mbNo], 0) = mrYes) then
        Application.Terminate;
  end;
end;

function TabcAppException.GetLogExtraInfo: TabcLogAppendEvent;
begin
  if FExcepLog <> nil then Result := FExcepLog.LogExtraInfo;
end;

procedure TabcAppException.SetLogExtraInfo(Value: TabcLogAppendEvent);
begin
  if FExcepLog <> nil then FExcepLog.LogExtraInfo := Value;
end;

procedure TabcAppException.AppendLogMessage(Sender: TObject; E: Exception; ExtraInfo: TStrings);
begin
  FExcepLog.AppendLogMessage(Sender, E, ExtraInfo);
end;

end.
