unit abcgreg;

(*------------------------------------------------------------------------------

Advanced Business Components
ABC for Delphi
Version 6

Copyright  1995, 2001 Objective Software Technology Pty Limited ACN 068 640 353
Functionality derived from Delphi VCL Source Code
is Copyright  1995, 2001 Borland Software Corporation

------------------------------------------------------------------------------*)

{$I abcdefs.inc}

{
Contains a public version of the private file type registration in Graphics.pas.
Used to implement LoadFromStream functions for TPicture where the file type
is detected by reading the stream header.
See also abcjpeg.pas
}

interface

uses Windows, Classes, Graphics;

type
  PabcFileFormat = ^TabcFileFormat;
  TabcFileFormat = record
    GraphicClass: TGraphicClass;
    Extension: string;
    Description: string;
    DescResID: Integer;
  end;

  TabcFileFormatsList = class(TList)
  public
    constructor Create;
    destructor Destroy; override;
    procedure Add(const Ext, Desc: String; DescID: Integer; const AClass: TGraphicClass);
    procedure Remove(const AClass: TGraphicClass);
    function FindExt(const Ext: string): TGraphicClass;
    function FindClassName(const Classname: string): TGraphicClass;
    procedure BuildFilterStrings(const GraphicClass: TGraphicClass;
      var Descriptions, Filters: string);
  end;

  TabcParadoxGraphicHeader = record
    Count: Word;                { Fixed at 1 }
    HType: Word;                { Fixed at $0100 }
    Size: Longint;              { Size not including header }
  end;

const
  C_ABC_PDOX_COUNT = 1;
  C_ABC_PDOX_HTYPE = $0100;

function abcFileFormatsList: TabcFileFormatsList;

procedure abcRegisterGraphicFileFormat(const AExtension,
  ADescription: string; const AGraphicClass: TGraphicClass);

procedure abcRegisterGraphicFileFormatRes(const AExtension: String;
  ADescriptionResID: Integer; const AGraphicClass: TGraphicClass);

procedure abcUnRegisterGraphicClass(const AClass: TGraphicClass);

function abcGetGraphicClassFromFileExt(const Ext: string): TGraphicClass;

function abcGetGraphicFileExtFromStream(const Stream: TStream): string;

function abcGetGraphicClassFromStream(const Stream: TStream): TGraphicClass;

procedure abcSkipParadoxBlobHeader(const Stream: TStream);

implementation

uses SysUtils, Consts;

var
  gabcFileFormatsList: TabcFileFormatsList;

function abcFileFormatsList: TabcFileFormatsList;
begin
  if gabcFileFormatsList = nil then
    gabcFileFormatsList := TabcFileFormatsList.Create;
  Result := gabcFileFormatsList;
end;

constructor TabcFileFormatsList.Create;
begin
  inherited Create;
  Add('wmf', SVMetafiles, 0, TMetafile);
  Add('emf', SVEnhMetafiles, 0, TMetafile);
  Add('ico', SVIcons, 0, TIcon);
  Add('bmp', SVBitmaps, 0, TBitmap);
end;

destructor TabcFileFormatsList.Destroy;
var
  I: Integer;
begin
  for I := 0 to Count-1 do
    Dispose(PabcFileFormat(Items[I]));
  inherited Destroy;
  gabcFileFormatsList := nil;
end;

procedure TabcFileFormatsList.Add(const Ext, Desc: String; DescID: Integer;
  const AClass: TGraphicClass);
var
  NewRec: PabcFileFormat;
begin
  New(NewRec);
  with NewRec^ do
  begin
    Extension := AnsiLowerCase(Ext);
    GraphicClass := AClass;
    Description := Desc;
    DescResID := DescID;
  end;
  inherited Add(NewRec);
end;

function TabcFileFormatsList.FindExt(const Ext: string): TGraphicClass;
var
  I: Integer;
  tmpStr: string;
begin
  tmpStr := AnsiLowerCase(Ext);
  for I := Count-1 downto 0 do
    with PabcFileFormat(Items[I])^ do
      if Extension = tmpStr then
      begin
        Result := GraphicClass;
        Exit;
      end;
  Result := nil;
end;

function TabcFileFormatsList.FindClassName(const ClassName: string): TGraphicClass;
var
  I: Integer;
begin
  for I := Count-1 downto 0 do
  begin
    Result := PabcFileFormat(Items[I])^.GraphicClass;
    if Result.ClassName = Classname then Exit;
  end;
  Result := nil;
end;

procedure TabcFileFormatsList.Remove(const AClass: TGraphicClass);
var
  I: Integer;
  P: PabcFileFormat;
begin
  for I := Count-1 downto 0 do
  begin
    P := PabcFileFormat(Items[I]);
    if P^.GraphicClass.InheritsFrom(AClass) then
    begin
      Dispose(P);
      Delete(I);
    end;
  end;
end;

procedure TabcFileFormatsList.BuildFilterStrings(const GraphicClass: TGraphicClass;
  var Descriptions, Filters: string);
var
  C, I: Integer;
  P: PabcFileFormat;
begin
  Descriptions := '';
  Filters := '';
  C := 0;
  for I := Count-1 downto 0 do
  begin
    P := PabcFileFormat(Items[I]);
    if P^.GraphicClass.InheritsFrom(GraphicClass) and (P^.Extension <> '') then
      with P^ do
      begin
        if C <> 0 then
        begin
          Descriptions := Descriptions + '|';
          Filters := Filters + ';';
        end;
        if (Description = '') and (DescResID <> 0) then
          Description := LoadStr(DescResID);
        FmtStr(Descriptions, '%s%s (*.%s)|*.%2:s', [Descriptions, Description, Extension]);
        FmtStr(Filters, '%s*.%s', [Filters, Extension]);
        Inc(C);
      end;
  end;
  if C > 1 then
    FmtStr(Descriptions, '%s (%s)|%1:s|%s', [sAllFilter, Filters, Descriptions]);
end;

procedure abcRegisterGraphicFileFormat(const AExtension,
  ADescription: string; const AGraphicClass: TGraphicClass);
begin
  abcFileFormatsList.Add(AExtension, ADescription, 0, AGraphicClass);
end;

procedure abcRegisterGraphicFileFormatRes(const AExtension: String;
  ADescriptionResID: Integer; const AGraphicClass: TGraphicClass);
begin
  abcFileFormatsList.Add(AExtension, '', ADescriptionResID, AGraphicClass);
end;

procedure abcUnRegisterGraphicClass(const AClass: TGraphicClass);
begin
  if abcFileFormatsList <> nil then abcFileFormatsList.Remove(AClass);
end;

function abcGetGraphicClassFromFileExt(const Ext: string): TGraphicClass;
begin
  Result := abcFileFormatsList.FindExt(Ext);
end;

{ Paradox graphic BLOB header }

type

  TabcCommonImageHeader = packed record
    case Integer of
      0: ({Bitmap}
        bfType: Word);  {expecting $4D42, i.e. 'BM'}
      1: ({JPEG}
        jpgSignature: Word); {expecting $D8FF, note bytes reversed actually FFD8}
      2: ({GIF}
        gifSignature: array[0..2] of Char);  {expecting 'GIF'}
      3: ({PNG}
        pngSignatureA: DWORD;     {expecting 8 bytes hex $ 89  50  4e  47  0d  0a  1a  0a}
        pngSignatureB: DWORD);
      4: ({EMF}
        emfFiller: array[0..39] of Byte;
        emfSignature: DWORD);  {expecting $464D4520}
      5: ({WMF}
        wmfKey: DWORD);      {expecting $9AC6CDD7, note there should be a valid checksum also}
      6: ({ICO}
        Reserved: Word;
        wType: Word);          {expecting 0 or 1}
      7: ({Paradox graphic blob header used with TBlobField}
        pdoxCount: Word;       { Fixed at 1 }
        pdoxHType: Word;       { Fixed at $0100 }
        pdoxSize: Longint);    { Size not including header }
  end;

const
  C_ABC_BMP_SIG = $4D42;   {Word value is reversed}
  C_ABC_JPG_SIG = $D8FF;   {Word value is reversed}
  C_ABC_GIF_SIG = 'GIF';
  C_ABC_PNG_SIG_A = $89504E47;  {check in case byte order is reversed}
  C_ABC_PNG_SIG_B = $0D0A1A0A;  {check in case byte order is reversed}
  C_ABC_EMF_SIG = $464D4520;
  C_ABC_WMF_SIG = $9AC6CDD7;
  C_ABC_ICO_SIG_ALTA = 0;
  C_ABC_ICO_SIG_ALTB = 1;

function abcGetGraphicFileExtFromStream(const Stream: TStream): string;
var
  Header: TabcCommonImageHeader;
begin
  Result := '';
  if Stream.Size < (SizeOf(TabcParadoxGraphicHeader) + SizeOf(TabcCommonImageHeader)) then
    Exit;
  Stream.Seek(0,0);
  Stream.ReadBuffer(Header, SizeOf(TabcCommonImageHeader));
  Stream.Seek(0,0);
  {detect paradox graphic blob header used with TBlobField}
  with Header do
    if (pdoxCount = C_ABC_PDOX_COUNT) and (pdoxHType = C_ABC_PDOX_HTYPE) then
    begin
      {reread header from offset}
      Stream.Seek(SizeOf(TabcParadoxGraphicHeader),0);
      Stream.ReadBuffer(Header, SizeOf(TabcCommonImageHeader));
    end;

  with Header do
    if bfType = C_ABC_BMP_SIG then
      Result := 'bmp'
    else if jpgSignature = C_ABC_JPG_SIG then
      Result := 'jpg'
    else if gifSignature = C_ABC_GIF_SIG then
      Result := 'gif'
    else if (pngSignatureA = C_ABC_PNG_SIG_A) and (pngSignatureB = C_ABC_PNG_SIG_B) then
      Result := 'png'
    else if emfSignature = C_ABC_EMF_SIG then
      Result := 'emf'
    else if wmfKey = C_ABC_WMF_SIG then
      Result := 'wmf'
    else if wType in [C_ABC_ICO_SIG_ALTA, C_ABC_ICO_SIG_ALTB] then
      Result := 'ico'
    else
      Result := '';
end;

function abcGetGraphicClassFromStream(const Stream: TStream): TGraphicClass;
var
  Ext: string;
begin
  Ext := abcGetGraphicFileExtFromStream(Stream);
  if Ext = '' then
    Result := nil
  else
    Result := abcGetGraphicClassFromFileExt(Ext);
end;

procedure abcSkipParadoxBlobHeader(const Stream: TStream);
var
  ParadoxBlobHeader: TabcParadoxGraphicHeader;
  Pos: Integer;
begin
  Pos := Stream.Position;
  if (Stream.Size - Pos) < SizeOf(TabcParadoxGraphicHeader) then
    Exit;
  Stream.ReadBuffer(ParadoxBlobHeader, SizeOf(TabcParadoxGraphicHeader));
  with ParadoxBlobHeader do
    if (Count <> C_ABC_PDOX_COUNT) or (HType <> C_ABC_PDOX_HTYPE) then
      {no blob header so reset to previous not current position}
      Stream.Seek(Pos, 0);
end;

initialization

finalization
  gabcFileFormatsList.Free;

end.
