unit abcidle;

(*------------------------------------------------------------------------------

Advanced Business Components
ABC for Delphi
Version 6

Copyright  1995, 2001 Objective Software Technology Pty Limited ACN 068 640 353
Functionality derived from Delphi VCL Source Code
is Copyright  1995, 2001 Borland Software Corporation

------------------------------------------------------------------------------*)

{$I abcdefs.inc}

{:This unit contains the abcIdleTimeout component for detecting
when your application is inactive and optionally terminating automatically
after a period of inactivity.
}

interface

uses Messages, Windows, SysUtils, Classes, Graphics, Controls, Forms, ExtCtrls,
     abcutil, abcdlg;

type

  { TabcIdleTimeout }

  {::Timeout event for the <See class=TabcIdleTimeout> component.<br>
   Set Terminate to False to cancel automatic termination.}

  TabcIdleTimeoutEvent = procedure(Sender: TObject; var Terminate: Boolean) of object;

  {:A global component for detecting when the application has been inactive for a period of time.
  <p>
  AbcIdleTimeout detects application activity in two ways:<br>
  1. by application-wide monitoring of keyboard and mouse events.<br>
  2. by measuring process execution times (Kernel and User time).<br>
  ** Note process time measures are only possible on NT and Windows 2000 platforms.<P><P>

  This is a global singleton component, only one instance is required or permitted in the application.<P><P>

  When <See property=Enabled>, the component times out after a specified Interval of inactivity.<P>
  Interval timing is restarted by calling the Reset method; whenever any mouse or keyboard activity is detected; or when at the end of an Interval there is a measurable change in the process execution times (NT only).<P>
  If the Interval times out, then a sequence of termination steps occur, allowing other opportunities to cancel the termination:<P>
  If the <See property=Dialog> property is assigned to a <See class=TabcProgressDlg> component, then that dialog will be shown according to its properties.
  This may allow a visible countdown of another interval with an appropriate warning message to the user.<P>
  If <See property=Dialog>  is not assigned, or is shown and not cancelled by the user, the <See event=OnTimeout> event occurs.<br>
  This event has a parameter to cancel the shutdown, allowing the program to continue running.<P>
  If not cancelled, the component terminates the application using one of the shutdown methods in the <See property=TerminateBy> property.
  }

  TabcIdleTimeout = class(TComponent)
  private
    FTimer: TTimer;
    FEnabled: Boolean;
    FInterval: Cardinal;
    FOnTimeout: TabcIdleTimeoutEvent;
    FTerminateBy: TabcTerminateMethod;
    FKernelTime, FUserTime: TFileTime;
    FDialog: TabcProgressDlg;
    FTicking: Boolean;
    procedure TimerTick(Sender: TObject);
    procedure SetEnabled(const Value: Boolean);
    procedure SetInterval(const Value: Cardinal);
    procedure InitTiming;
    procedure EndTiming;
  protected
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    function ProcessTimingChanged: Boolean;
    procedure UpdateProcessTimes;
    function ConfirmTerminate: Boolean;
    procedure TerminateApp;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    {:Resets the time interval.}
    procedure Reset;
  published
    {::Link to another component to display a timed dialog before terminating the application.}
    property Dialog: TabcProgressDlg read FDialog write FDialog;
    {::If enabled is true, the component monitors activity and may terminate the application.<br>
     If false, monitoring is suspended.<br>}
    property Enabled: Boolean read FEnabled write SetEnabled default True;
    {::The period of inactivity before terminating, in milliseconds.<br>
     The count is reset when any user activity occurs in the application, or if the interval expires and the process execution times have changed.}
    property Interval: Cardinal read FInterval write SetInterval default 600000;
    {::The method used to terminate the application.<P>
    <code>
TabcTerminateMethod = (atmAppTerminate, atmCloseMainForm, atmExitProcess, atmHalt, atmTerminateProcess);
    </code><p>
    The options are:<UL>
    <li>atmAppTerminate - calls Application.Terminate method. This is the default.
    <li>atmCloseMainForm - calls Application.MainForm.Close.
    <li>atmExitProcess - calls Win32 ExitProcess API, which gracefully terminates an application.
    <li>atmHalt - executes the Pascal Halt instruction.
    <li>atmTerminateProcess - calls Win32 TerminateProcess API which forces an application to terminate.
    </UL>
    }
    property TerminateBy: TabcTerminateMethod read FTerminateBy write FTerminateBy default atmAppTerminate;
    {::This event occurs when the component times out by not detecting any activity for the duration of the Interval.<br>
     Set the Terminate parameter to False to cancel automatic termination.}
    property OnTimeout: TabcIdleTimeoutEvent read FOnTimeout write FOnTimeout;
  end;

implementation

uses abcconst;

var
  GlobalTimeout: TabcIdleTimeout;
  KeyboardHook, MouseHook: HHook;

{ Windows Hook Procs }

function KeyboardProc(Code: Integer; WParam: Cardinal; LParam: Longint): Longint; stdcall;
begin
  GlobalTimeout.Reset;
  Result := CallNextHookEx(KeyboardHook, Code, WParam, LParam);
end;

function MouseProc(Code: Integer; WParam: Cardinal; LParam: Longint): Longint; stdcall;
begin
  GlobalTimeout.Reset;
  Result := CallNextHookEx(MouseHook, Code, WParam, LParam);
end;

{ TabcIdleTimeout }

constructor TabcIdleTimeout.Create(AOwner: TComponent);
begin
  if GlobalTimeout <> nil then
    raise EComponentError.Create(sabcErrorSingleInstanceIdleTimeout);
  inherited;
  GlobalTimeout := Self;
  FEnabled := True;
  FInterval := 600000;
  FTerminateBy := atmAppTerminate;
  if not (csDesigning in ComponentState) then
    InitTiming;
end;

procedure TabcIdleTimeout.InitTiming;
begin
  FTimer := TTimer.Create(Self);
  with FTimer do
  begin
    Interval := FInterval;
    OnTimer := TimerTick;
    Enabled := FEnabled;
  end;
  //activity hooks
  KeyboardHook := SetWindowsHookEx(WH_KEYBOARD, @KeyboardProc, MainInstance, MainThreadID);
  MouseHook := SetWindowsHookEx(WH_MOUSE, @MouseProc, MainInstance, MainThreadID);
  UpdateProcessTimes;
end;

procedure TabcIdleTimeout.EndTiming;
begin
  FTimer.Enabled := False;
  //remove activity hooks
  if KeyboardHook <> 0 then
    UnhookWindowsHookEx(KeyboardHook);
  KeyboardHook := 0;
  if MouseHook <> 0 then
    UnhookWindowsHookEx(MouseHook);
  MouseHook := 0;
  FTimer.Free;
  FTimer := nil;
end;

destructor TabcIdleTimeout.Destroy;
begin
  if GlobalTimeout = Self then
  begin
    if not (csDesigning in ComponentState) and (FTimer <> nil) then
      EndTiming;
    GlobalTimeout := nil;
  end;
  inherited;
end;

procedure TabcIdleTimeout.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited;
  if (Operation = opRemove) and (AComponent = FDialog) and not (csDestroying in ComponentState) then
    FDialog := nil;
end;

procedure TabcIdleTimeout.Reset;
begin
  UpdateProcessTimes;
  if not FTicking and Enabled and (FTimer <> nil) then
  begin
    FTimer.Enabled := False;
    FTimer.Enabled := True;
  end;
end;

procedure TabcIdleTimeout.SetEnabled(const Value: Boolean);
begin
  FEnabled := Value;
  if FTimer <> nil then
  begin
    FTimer.Enabled := Value;
    if Value then UpdateProcessTimes;
  end;
end;

procedure TabcIdleTimeout.SetInterval(const Value: Cardinal);
begin
  FInterval := Value;
  if FTimer <> nil then
  begin
    FTimer.Interval := Value;
    Reset;
  end;
end;

procedure TabcIdleTimeout.TerminateApp;
begin
  EndTiming;  //make sure hook procs are removed before stopping application
  case TerminateBy of
    atmAppTerminate:
      begin
        Application.ShowMainForm := False;
        Application.Terminate;
      end;
    atmCloseMainForm:
      if Application.MainForm <> nil then
        Application.MainForm.Close;
    atmExitProcess:
      ExitProcess(0);
    atmHalt:
      Halt;
    atmTerminateProcess:
      TerminateProcess(GetCurrentProcess, 0);
  end;
end;

procedure TabcIdleTimeout.TimerTick(Sender: TObject);
begin
  if not FTicking then
  begin
    FTicking := True;
    if ProcessTimingChanged or not ConfirmTerminate then
    begin
      Reset;
      FTicking := False;
    end
    else
      TerminateApp;
  end;
end;

function TabcIdleTimeout.ConfirmTerminate: Boolean;
begin
  Result := True;
  if Assigned(OnTimeout) then
    OnTimeout(Self, Result);
  if Result and (FDialog <> nil) then
    case FDialog.Execute of
      abcmrTimeout, mrOk, mrYes:
        Result := True;
    else
      Result := False;
    end;
end;

function TabcIdleTimeout.ProcessTimingChanged: Boolean;
var
  lastKernelTime, lastUserTime: TFileTime;
begin
  if ((Win32Platform = VER_PLATFORM_WIN32_NT) or
      ((Win32Platform = VER_PLATFORM_WIN32_WINDOWS) and
       (Win32MajorVersion > 4))) then
  begin
    lastKernelTime := FKernelTime;
    lastUserTime := FUserTime;
    UpdateProcessTimes;
    Result :=
      (Int64(lastKernelTime) <> Int64(FKernelTime)) or
      (Int64(lastUserTime) <> Int64(FUserTime));
  end
  else
    Result := False;
end;

procedure TabcIdleTimeout.UpdateProcessTimes;
var
  creationTime, exitTime: TFileTime;
begin
  if ((Win32Platform = VER_PLATFORM_WIN32_NT) or
      ((Win32Platform = VER_PLATFORM_WIN32_WINDOWS) and
       (Win32MajorVersion > 4))) then
    GetProcessTimes(GetCurrentProcess, creationTime, exitTime, FKernelTime, FUserTime);
end;

end.
