unit abcmem;

(*------------------------------------------------------------------------------

Advanced Business Components
ABC for Delphi
Version 6

Copyright  1995, 2001 Objective Software Technology Pty Limited ACN 068 640 353
Functionality derived from Delphi VCL Source Code
is Copyright  1995, 2001 Borland Software Corporation

------------------------------------------------------------------------------*)

{$I abcdefs.inc}

interface

uses
  SysUtils, Classes, Windows;

type

  { TabcPersistentMemory }

  TabcPersistentMemory = class(TPersistent)
  private
    FMemory: TMemoryStream;
    FModified: Boolean;
    FOnChange: TNotifyEvent;
    FOnChanging: TNotifyEvent;
    procedure ReadBuffer(Buffer: Pointer; Count: Longint);
    procedure ReadStream(Stream: TStream; Count: Longint);
    procedure SetModified(Value: Boolean);
    procedure WriteStream(Stream: TStream; WriteSize: Boolean);
  protected
    procedure Changed; dynamic;
    procedure Changing; dynamic;
    procedure CopyToClipboardFormat(AFormat: Word);
    procedure DefineProperties(Filer: TFiler); override;
    function GetEmpty: Boolean;
    function GetMemory: Pointer;
    function GetSize: Longint;
    procedure PasteFromClipboardFormat(AFormat: Word);
    procedure ReadData(Stream: TStream);
    procedure ValidateDataFormat(Stream: TMemoryStream); virtual;
    procedure WriteData(Stream: TStream);
  public
    constructor Create; virtual;
    destructor Destroy; override;
    procedure Assign(Source: TPersistent); override;
    procedure Clear;
    procedure CopyToClipboard; virtual;
    procedure LoadFromFile(const Filename: string);
    procedure LoadFromResource(const ResourceDLL, ResourceName: string);
    procedure LoadFromStream(Stream: TStream);
    procedure PasteFromClipboard; virtual;
    procedure SaveToFile(const Filename: string);
    procedure SaveToStream(Stream: TStream);
    property Empty: Boolean read GetEmpty;
    property Modified: Boolean read FModified write SetModified;
    property Memory: Pointer read GetMemory;
    property Size: Longint read GetSize;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
    property OnChanging: TNotifyEvent read FOnChanging write FOnChanging;
  end;

  { TabcResourceData }

  TabcResourceData = class(TPersistent)
  private
    FModuleHandle: THandle; { can be user assigned or from ResourceModule property }
    FResourceHandle: THandle; { internal from LoadRes function }

    FResourceId: Integer;
    FResourceModule: string;
    FResourceName: string;
    FData: TabcPersistentMemory;
    FTempFile: string;
    FFileName: string;
    FOnChange: TNotifyEvent;
    FOnChanging: TNotifyEvent;
    FResourceType: string;
    procedure SetData(const Value: TabcPersistentMemory);
    procedure SetFileName(const Value: string);
    procedure SetResourceID(const Value: Integer);
    procedure SetResourceName(const Value: string);
    procedure SetModuleHandle(const Value: THandle);
  protected
    procedure Changed; virtual;
    procedure Changing; virtual;
    function GetEmpty: Boolean; virtual;
    function IsResource: Boolean; virtual;
    function GetResourceHandle: THandle; virtual;
  public
    constructor Create; virtual;
    destructor Destroy; override;
    procedure Assign(Source: TPersistent); override;
    procedure Clear; virtual;
    function CopyDataToTempFile: string;
    procedure DeleteTempFile;
    function LoadRes: THandle;
    property Empty: Boolean read GetEmpty;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
    property OnChanging: TNotifyEvent read FOnChanging write FOnChanging;
    property ModuleHandle: THandle read FModuleHandle write SetModuleHandle;
    property ResourceHandle: THandle read GetResourceHandle;
    property TempFile: string read FTempFile;
  published
    property Data: TabcPersistentMemory read FData write SetData;
    property FileName: string read FFileName write SetFileName;
    property ResourceModule: string read FResourceModule write FResourceModule;
    property ResourceName: string read FResourceName write SetResourceName;
    property ResourceID: Integer read FResourceId write SetResourceID default 0;
    property ResourceType: string read FResourceType write FResourceType;
  end;

implementation

uses Clipbrd, abcconst, abcutil, abcutilw;

{ TabcPersistentMemory }

constructor TabcPersistentMemory.Create;
begin
  FMemory := TMemoryStream.Create;
end;

destructor TabcPersistentMemory.Destroy;
begin
  FMemory.Free;
  inherited Destroy;
end;

procedure TabcPersistentMemory.Assign(Source: TPersistent);
begin
  if (Source = nil) or (Source is TabcPersistentMemory) then begin
    if Source <> nil then
      ReadBuffer(TabcPersistentMemory(Source).Memory, TabcPersistentMemory(Source).Size)
    else
      Clear;
    Exit;
  end;
  inherited Assign(Source);
end;

procedure TabcPersistentMemory.Changed;
begin
  if Assigned(FOnChange) then FOnChange(Self);
end;

procedure TabcPersistentMemory.Changing;
begin
  FModified := True;
  if Assigned(FOnChanging) then FOnChanging(Self);
end;

procedure TabcPersistentMemory.Clear;
begin
  Changing;
  FMemory.Clear;
  Changed;
end;

procedure TabcPersistentMemory.CopyToClipboard;
begin
  {does nothing since we don't know the data format
   method exists for use in descendant classes}
end;

procedure TabcPersistentMemory.PasteFromClipboard;
begin
  {does nothing since we don't know the data format
   method exists for use in descendant classes}
end;

procedure TabcPersistentMemory.CopyToClipboardFormat(AFormat: Word);
var
  ClipHandle: THandle;
  ClipData: Pointer;
begin
  if not Empty then begin
    ClipHandle := GlobalAlloc(GMEM_MOVEABLE, Size);
    try
      ClipData := GlobalLock(ClipHandle);
      try
        CopyMemory(ClipData, Memory, Size);
      finally
        GlobalUnLock(ClipHandle);
      end;
        Clipboard.SetAsHandle(AFormat, ClipHandle);
    except
      GlobalFree(ClipHandle);
    end;
  end;
end;

procedure TabcPersistentMemory.DefineProperties(Filer: TFiler);
begin
  Filer.DefineBinaryProperty('Data', ReadData, WriteData, not Empty);
end;

function TabcPersistentMemory.GetEmpty: Boolean;
begin
  Result := (Size = 0);
end;

function TabcPersistentMemory.GetMemory: Pointer;
begin
  Result := FMemory.Memory;
end;

function TabcPersistentMemory.GetSize: Longint;
begin
  Result := FMemory.Size;
end;

procedure TabcPersistentMemory.LoadFromFile(const Filename: string);
var
  Stream: TStream;
begin
  Stream := TFileStream.Create(Filename, fmOpenRead);
  try
    LoadFromStream(Stream);
  finally
    Stream.Free;
  end;
end;

procedure TabcPersistentMemory.LoadFromResource(const ResourceDLL, ResourceName: string);
{copies RES of type RCDATA into stream}
var
  hResDLL, hResInfo, hRes: THandle;
  pRes: Pointer;
  ErrorMsg: string;
begin
  if ResourceName = '' then Exit;

  hResDLL := 0;
  hRes := 0;
  pRes := nil;

  try
    if ResourceDLL <> '' then begin
      hResDLL := LoadLibrary(PChar(ResourceDLL));
      if hResDLL < HINSTANCE_ERROR then begin
        ErrorMsg := Format(sabcErrResDllLoad, [hResDLL, ResourceDLL]);
        hResDLL := 0;
        raise EResNotFound.Create(ErrorMsg);
      end;
      hResInfo := FindResource(hResDLL, PChar(ResourceName), RT_RCDATA);
    end
    else
      hResInfo := FindResource(HInstance, PChar(ResourceName), RT_RCDATA);
    if hResInfo = 0 then
      raise EResNotFound.Create(Format(sabcErrRCDATAResNotFound, [ResourceName]));
    if hResDLL <> 0 then
      hRes := LoadResource(hResDLL, hResInfo)
    else
      hRes := LoadResource(HInstance, hResInfo);
    if hRes = 0 then
      raise EResNotFound.Create(Format(sabcErrResLoad, [ResourceName]))
    else begin
      {now we have the resource, so lock and read it}
      pRes := LockResource(hRes);
      if pRes = nil then
        raise EResNotFound.Create(Format(sabcErrResLoad, [ResourceName]))
      else
        ReadBuffer(pRes, SizeofResource(hResDLL, hResInfo));
    end;
  finally
    if pRes <> nil then
      UnlockResource(hRes);
    if hRes <> 0 then
      FreeResource(hRes);
    if hResDLL <> 0 then
      FreeLibrary(hResDLL);
  end;
end;

procedure TabcPersistentMemory.LoadFromStream(Stream: TStream);
begin
  ReadStream(Stream, Stream.Size - Stream.Position);
end;

procedure TabcPersistentMemory.PasteFromClipboardFormat(AFormat: Word);
var
  ClipHandle: THandle;
  ClipSize: Longint;
  ClipData: Pointer;
begin
  ClipHandle := Clipboard.GetAsHandle(AFormat);
  if ClipHandle <> 0 then
    try
      ClipData := GlobalLock(ClipHandle);
      ClipSize := GlobalSize(ClipHandle);
      ReadBuffer(ClipData, ClipSize);
    finally
      GlobalUnlock(ClipHandle);
    end;
end;

procedure TabcPersistentMemory.ReadBuffer(Buffer: Pointer; Count: Longint);
var
  TempStream: TMemoryStream;
begin
  if Count = 0 then
    Clear
  else begin
    TempStream := TMemoryStream.Create;
    try
      TempStream.SetSize(Count);
      TempStream.Write(Buffer^, Count);
      ValidateDataFormat(TempStream);
      TempStream.Seek(0,0);
      FMemory.Seek(0,0);
      Changing;
      FMemory.SetSize(Count);
      TempStream.SaveToStream(FMemory);
      Changed;
    finally
      TempStream.Free;
    end;
  end;
end;

procedure TabcPersistentMemory.ReadData(Stream: TStream);
var
  lSize: Longint;
begin
  Stream.Read(lSize, SizeOf(lSize));
  ReadStream(Stream, lSize);
end;

procedure TabcPersistentMemory.ReadStream(Stream: TStream; Count: Longint);
var
  TempStream: TMemoryStream;
begin
  if Count = 0 then
    Clear
  else begin
    TempStream := TMemoryStream.Create;
    try
      TempStream.SetSize(Count);
      Stream.ReadBuffer(TempStream.Memory^, Count);
      ValidateDataFormat(TempStream);
      TempStream.Seek(0,0);
      FMemory.Seek(0,0);
      Changing;
      FMemory.SetSize(Count);
      TempStream.SaveToStream(FMemory);
      Changed;
    finally
      TempStream.Free;
    end;
  end;
end;

procedure TabcPersistentMemory.SaveToFile(const Filename: string);
var
  Stream: TStream;
begin
  Stream := TFileStream.Create(Filename, fmCreate);
  try
    SaveToStream(Stream);
  finally
    Stream.Free;
  end;
end;

procedure TabcPersistentMemory.SaveToStream(Stream: TStream);
begin
  WriteStream(Stream, False);
end;

procedure TabcPersistentMemory.SetModified(Value: Boolean);
begin
  if Value then begin
    Changing;
    Changed;
  end
  else
    FModified := False;
end;

procedure TabcPersistentMemory.WriteData(Stream: TStream);
begin
  WriteStream(Stream, True);
end;

procedure TabcPersistentMemory.ValidateDataFormat(Stream: TMemoryStream);
begin
  {no validation in this class}
end;

procedure TabcPersistentMemory.WriteStream(Stream: TStream; WriteSize: Boolean);
var
  lSize: Longint;
begin
  lSize := Size;
  if WriteSize then Stream.WriteBuffer(lSize, SizeOf(lSize));
  if lSize <> 0 then Stream.WriteBuffer(Memory^, lSize);
end;

{ TabcResourceData }

procedure TabcResourceData.Assign(Source: TPersistent);
begin
  if (Source = nil) or (Source is TabcResourceData) then begin
    Changing;
    if Source <> nil then
    begin
      if not TabcResourceData(Source).Data.Empty then
        Data.Assign(TabcResourceData(Source).Data);
      FFileName := TabcResourceData(Source).FileName;
      FResourceModule := TabcResourceData(Source).ResourceModule;
      FResourceName := TabcResourceData(Source).ResourceName;
      FResourceId := TabcResourceData(Source).ResourceId;
    end
    else
      Clear;
    Changed;
    Exit;
  end;
  inherited Assign(Source);
end;

procedure TabcResourceData.Changed;
begin
  if Assigned(FOnChanging) then
    FOnChanging(Self);
end;

procedure TabcResourceData.Changing;
begin
  if Assigned(FOnChange) then
    FOnChange(Self);
end;

procedure TabcResourceData.Clear;
begin
  FResourceHandle := 0;
  Data.Clear;
  FFileName := '';
  {FResourceModule := '';}
  FResourceName := '';
  FResourceID := 0;
end;

function TabcResourceData.CopyDataToTempFile: string;
begin
  if Data.Empty then
    Result := ''
  else
  begin
    if FTempFile <> '' then DeleteTempFile;
    Result := abcGetWinTempFile;
    Data.SaveToFile(Result);
    FTempFile := Result;
  end;
end;

procedure TabcResourceData.DeleteTempFile;
begin
  SysUtils.DeleteFile(FTempFile);
  FTempFile := '';
end;

constructor TabcResourceData.Create;
begin
  FData := TabcPersistentMemory.Create;
end;

destructor TabcResourceData.Destroy;
begin
  FData.Free;
  DeleteTempFile;
  inherited Destroy;
end;

function TabcResourceData.GetEmpty: Boolean;
begin
  Result :=
    Data.Empty and
    (FileName = '') and
    (ResourceName = '') and
    (ResourceID = 0);
end;

function TabcResourceData.GetResourceHandle: THandle;
begin
  Result := FResourceHandle;
  if (Result = 0) and IsResource then
    Result := LoadRes;
end;

function TabcResourceData.IsResource: Boolean;
begin
  Result :=
    (FResourceName <> '') or
    (FResourceID <> 0);
end;

function TabcResourceData.LoadRes: THandle;
var
  foundRes: THandle;
begin
  Result := 0;
  if FResourceHandle <> 0 then
    Result := FResourceHandle
  else if IsResource then
  begin
    if (FModuleHandle = 0) and (ResourceModule <> '') then
    begin
      FModuleHandle := LoadLibrary(PChar(FResourceModule));
      if FModuleHandle < 32 then
{$IFDEF DELPHI6}
        RaiseLastOSError;
{$ELSE}
        RaiseLastWin32Error;
{$ENDIF}
    end;

    {load resource name or resource ID}

    if ResourceName <> '' then
    begin
      foundRes := FindResource(FModuleHandle, PChar(FResourceName), PChar(FResourceType));
      if foundRes = 0 then
{$IFDEF DELPHI6}
        RaiseLastOSError;
{$ELSE}
        RaiseLastWin32Error;
{$ENDIF}
      FResourceHandle := LoadResource(FModuleHandle, foundRes);
      Result := FResourceHandle;
    end
    else if ResourceID <> 0 then
    begin
      foundRes := FindResource(FModuleHandle, MAKEINTRESOURCE(FResourceID), PChar(FResourceType));
      if foundRes = 0 then
{$IFDEF DELPHI6}
        RaiseLastOSError;
{$ELSE}
        RaiseLastWin32Error;
{$ENDIF}
      FResourceHandle := LoadResource(FModuleHandle, foundRes);
      Result := FResourceHandle;
    end
    else
      Result := 0;
  end;
end;

procedure TabcResourceData.SetData(const Value: TabcPersistentMemory);
begin
  if Value <> FData then
  begin
    Changing;
    Clear;
    FData.Assign(Value);
    Changed;
  end;
end;

procedure TabcResourceData.SetFileName(const Value: string);
begin
  if Value <> FFileName then
  begin
    Changing;
    Clear;
    FFileName := Value;
    Changed;
  end;
end;

procedure TabcResourceData.SetResourceID(const Value: Integer);
begin
  if Value <> FResourceId then
  begin
    Changing;
    Clear;
    FResourceId := Value;
    Changed;
  end;
end;

procedure TabcResourceData.SetResourceName(const Value: string);
begin
  if Value <> FResourceName then
  begin
    Changing;
    Clear;
    FResourceName := Value;
    Changed;
  end;
end;

procedure TabcResourceData.SetModuleHandle(const Value: THandle);
begin
  FModuleHandle := Value;
  FResourceModule := '';
  FResourceHandle := 0;
end;

end.
