unit abcshapf;

(*------------------------------------------------------------------------------

Advanced Business Components
ABC for Delphi
Version 6

Copyright  1995, 2001 Objective Software Technology Pty Limited ACN 068 640 353
Functionality derived from Delphi VCL Source Code
is Copyright  1995, 2001 Borland Software Corporation

------------------------------------------------------------------------------*)

{$I abcdefs.inc}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics,
  Controls, Forms, abcedge;

type

  { TabcShapeForm }

  TabcShapeForm = class(TGraphicControl)
  private
    FBitmap: TBitmap;
    FObjectInstance: TFarProc;
    FOwnerForm: TForm;
    FOwnerHandle: HWND;
    FOwnerWindowProc: TFarProc;
    FDragAnywhere: Boolean;
    FTolerance: TColor;
    FDragControl: TControl;
    FPicture: TPicture;
    FDrawing: Boolean;
    procedure MessageProc(var Message: TMessage);
    procedure SubClassOwnerForm;
    procedure CMEnabledChanged(var Message: TMessage); message CM_ENABLEDCHANGED;
    procedure SetPicture(const Value: TPicture);
    procedure PictureChanged(Sender: TObject);
    procedure SetDragControl(const Value: TControl);
    function GetCanvas: TCanvas;
  protected
    procedure Loaded; override;
    function GetPalette: HPALETTE; override;
    procedure Paint; override;
    function DoProcessMessage(var Message: TMessage): Boolean; dynamic;
    procedure DoShapeForm(Shaped: Boolean); dynamic;
    function MakeRegion: HRgn;
    property OwnerForm: TForm read FOwnerForm;
    property OwnerHandle: HWND read FOwnerHandle;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Assign(Source: TPersistent); override;
    procedure SizeFormToPicture;
    property Canvas: TCanvas read GetCanvas;
  published
    property DragAnywhere: Boolean read FDragAnywhere write FDragAnywhere default False;
    property DragControl: TControl read FDragControl write SetDragControl;
    property Enabled;
    property ParentShowHint;
    property Picture: TPicture read FPicture write SetPicture;
    property PopupMenu;
    property ShowHint;
    property Tolerance: TColor read FTolerance write FTolerance default $101010;
    property Visible;
    property OnClick;
    property OnDblClick;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
  end;

implementation

uses abcconst;

  { TabcShapeForm }

constructor TabcShapeForm.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  if (AOwner is TForm) then
    FOwnerForm := TForm(AOwner)
  else
    raise EComponentError.Create(sabcErrorOwnerFormExpected);
  FPicture := TPicture.Create;
  FPicture.OnChange := PictureChanged;
  FBitmap := TBitmap.Create;
  FTolerance := $101010;
  Align := alClient;
end;

destructor TabcShapeForm.Destroy;
begin
  Destroying;
  if not (csDestroying in FOwnerForm.ComponentState) and (FOwnerForm.Handle = FOwnerHandle) then
    {restore form's original WndProc}
    SetWindowLong(FOwnerHandle, GWL_WNDPROC, Longint(FOwnerWindowProc));
  {$IFDEF DELPHI6}Classes.{$ENDIF}FreeObjectInstance(FObjectInstance);
  FBitmap.Free;
  FPicture.Free;
  inherited Destroy;
end;

procedure TabcShapeForm.Assign(Source: TPersistent);
begin
  if Source is TabcShapeForm then
  begin
    with TabcShapeForm(Source) do
    begin
      Self.DragAnywhere := DragAnywhere;
      Self.Tolerance := Tolerance;
      Self.Picture := Picture;
    end;
  end
  else
    inherited Assign(Source);
end;

procedure TabcShapeForm.Loaded;
begin
  inherited Loaded;
  if (OwnerHandle = 0) then
    if not (csDesigning in ComponentState) then
    begin
      SubClassOwnerForm;
      if Enabled then
        DoShapeForm(True);
    end;
end;

procedure TabcShapeForm.CMEnabledChanged(var Message: TMessage);
begin
  inherited;
  DoShapeForm(Enabled);
end;

function TabcShapeForm.DoProcessMessage(var Message: TMessage): Boolean;
begin
  Result := False;
  if (Message.Msg = WM_NCHITTEST) then begin
    if DragAnywhere or
        ((DragControl <> nil) and
         PtInRect(DragControl.BoundsRect,
          OwnerForm.ScreenToClient(SmallPointToPoint(TWMNCHitTest(Message).Pos))))
    then
    begin
      Message.Result := HTCAPTION;
      Result := True;
    end;
  end;
end;

procedure TabcShapeForm.DoShapeForm(Shaped: Boolean);
var
  rgn: HRGN;
begin
  if [csLoading, csDesigning] * ComponentState = [] then
  begin
    OwnerForm.HandleNeeded;
    if Shaped then
      rgn := MakeRegion
    else
      rgn := 0;
    SetWindowRgn(OwnerForm.Handle, rgn, OwnerForm.Visible);
  end;
end;

function TabcShapeForm.GetCanvas: TCanvas;
begin
  Result := FBitmap.Canvas;
end;

function TabcShapeForm.GetPalette: HPALETTE;
begin
  Result := 0;
  if FPicture.Graphic <> nil then
    Result := FPicture.Graphic.Palette;
end;

function TabcShapeForm.MakeRegion: HRgn;
begin
  Result := abcGetRegionFromBitmap(FBitmap, FTolerance, Point(0,0));
end;

procedure TabcShapeForm.MessageProc(var Message: TMessage);
begin
  if not DoProcessMessage(Message) then
    {pass message to original WndProc}
    with Message do
      Result := CallWindowProc(FOwnerWindowProc, FOwnerHandle, Msg, WParam, LParam);
end;

procedure TabcShapeForm.Paint;
var
  Save: Boolean;
begin
  if csDesigning in ComponentState then
    with inherited Canvas do
    begin
      Pen.Style := psDash;
      Brush.Style := bsClear;
      Rectangle(0, 0, Width, Height);
    end;
  Save := FDrawing;
  FDrawing := True;
  try
    with inherited Canvas do
      Draw(0, 0, FBitmap);
  finally
    FDrawing := Save;
  end;
end;

procedure TabcShapeForm.PictureChanged(Sender: TObject);
begin
  {copy to FBitmap}
  with FBitmap do
  begin
    Width := 0;
    Height := 0;
    Height := Picture.Height;
    Width := Picture.Width;
    Canvas.Draw(0,0,FPicture.Graphic);
    TransparentMode := tmAuto;
  end;

  if not (csDestroying in ComponentState) then
  begin
    SizeFormToPicture;
    if Enabled then
      DoShapeForm(True)
    else if not FDrawing then
      Invalidate;
  end;
end;

procedure TabcShapeForm.SetDragControl(const Value: TControl);
begin
  FDragControl := Value;
  if Value <> nil then
    DragAnywhere := False;
end;

procedure TabcShapeForm.SetPicture(const Value: TPicture);
begin
  FPicture.Assign(Value);
end;

procedure TabcShapeForm.SizeFormToPicture;
begin
  if (Picture.Graphic <> nil) and not Picture.Graphic.Empty then
    if (OwnerForm.ClientWidth <> Picture.Width) or (OwnerForm.ClientHeight <> Picture.Height) then
    begin
      OwnerForm.ClientWidth := Picture.Width;
      OwnerForm.ClientHeight := Picture.Height;
    end;
end;

procedure TabcShapeForm.SubClassOwnerForm;
begin
  FOwnerHandle := FOwnerForm.Handle;
  FObjectInstance := {$IFDEF DELPHI6}Classes.{$ENDIF}MakeObjectInstance(MessageProc);
  FOwnerWindowProc := Pointer(GetWindowLong(FOwnerHandle, GWL_WNDPROC));
  SetWindowLong(FOwnerHandle, GWL_WNDPROC, Longint(FObjectInstance));
end;

end.
