unit abcspr;

(*------------------------------------------------------------------------------

Advanced Business Components
ABC for Delphi
Version 6

Copyright  1995, 2001 Objective Software Technology Pty Limited ACN 068 640 353
Functionality derived from Delphi VCL Source Code
is Copyright  1995, 2001 Borland Software Corporation

------------------------------------------------------------------------------*)

{$I abcdefs.inc}

{:Contains all base classes for ABC Sprite Components.
This unit defines the framework for ABC sprite components and includes the main animation engine classes in TabcSpriteList and TabcSpriteSurface.<br>
It contains the ancestor classes for all sprites (in TabcSprite) and sprite factories (in TabcSpriteFactory).
}

interface

uses
  SysUtils, Windows, Messages, Classes, Graphics, Controls, Forms,
  MMSystem, abcapp, abcutilg, abctileb;

type
  TabcSprite = class;

  TabcSpriteList = class;

    {::For front to back ordering of sprites.<br>
    0 is at the front.<br>
    When drawing, sprites are drawn from back to front.<br>
    When animating, sprites pass under/over depending on ZOrder.<br>
    Collisions between sprites only occur when they are at the same ZOrder.<br>
    The <See Class=TabcSpriteList> also allows events to be restricted to a ZOrderLimit.
    }

  TabcZOrder = 0..MaxInt;

    {::For events that occur on a single sprite. The Sender is normally the SpriteList.}

  TabcSpriteEvent = procedure(const Sender: TObject; const Sprite: TabcSprite) of object;

    {::For collision events involving two sprites. The Sender is the SpriteList.}

  TabcSpriteCollisionEvent = procedure(const Sender: TObject; const Sprite1, Sprite2: TabcSprite) of object;

    {::Occurs on every sprite for each frame when the SpriteSurface is Active.}

  TabcSpriteAnimateEvent = procedure(const Sprite: TabcSprite; Frame: Integer) of object;

    {::For automatic animation of sprites using the TabcSpriteMover:<UL>
    <li>asmNone - the sprite does not move.
    <li>asmVelocity - move using VelocityX, VelocityY and AccelerationX/Y properties.
    <li>asmPoints - move using the Points list.
    <li>asmPointsReversed - moves backward through the Points list.}

  TabcSpriteMove = (asmNone, asmVelocity, asmPoints, asmPointsReversed);

    {::For automatic animation of sprites using the TabcSpriteMover.<P>
    Defines how the movement changes when the sprite reaches the edge of the surface (when moving by amsVelocity),
    or when the sprite is positioned at the last point in the Points list (when moving by asmPoints),
    or when the sprite is positioned at the first point in the Points list (when moving by asmPointsReversed).<P><UL>
    <li>aseHide - the sprite is hidden.
    <li>aseBounce - the sprite's VelocityX and VelocityY and AccelerationX/Y properties are reversed, or if moving through a Points list the direction is reversed.
    <li>aseWrap - wraps the sprite to the opposite edge of the surface, or to the other end of the Points list.
    <li>aseStop - stops moving. i.e. sets move to asmNone.}

  TabcSpriteEdgeMove = (aseHide, aseBounce, aseWrap, aseStop);

    {:A TList containing TPoint records.
     Each item in the list is a PPoint = ^TPoint.<br>
     This list allocates TPoints in memory.<br>
     For memory-safe access use only the AddPoint, DeletePoint and Clear methods to change the list.
    }

  TabcPointList = class (TList)
  public
        {::Allocates a new TPoint record and adds it to the list. Returns the index of the added item.}
    function AddPoint(const Pt: TPoint): Integer;
        {::Frees the TPoint record pointed to by the index and deletes the item from the list.}
    procedure DeletePoint(Index: Integer);
        {::Frees all allocated TPoint records and deletes all items from the list.}
    procedure Clear; override;
  end;

    {::Used with TabcSpriteMover.MakePoints method.<br>
    Generates a range of intermediate points to move a sprite along a straight line:<UL>
    <li>ampEven - points are evenly spaced in a straight line from start to end.
    <li>ampAccelerate - distance between points increases towards the end.
    <li>ampDecelerate - distance between points decreases towards the end.
    <li>ampAccelMid - distance between points increases in the middle and narrows at each end.
    }

  TabcMakePointsStyle = (ampEven, ampAccelerate, ampDecelerate, ampAccelMid);

    {: A class to automate sprite movement.
    This class is used by the SpriteList to automatically move a sprite on each frame of an animation.<P>
    Sprites can move in a straight line, with acceleration, or through a list of Points.<Br>
    Special actions can occur when a sprite reaches the edge of the surface or the end of the list of points.<P>
    For custom frame-by-frame movement the sprite OnAnimate event can also be used to change the sprite's position.
    }

  TabcSpriteMover = class(TPersistent)
  private
    FSprite: TabcSprite;
    FVelocityX: Integer;
    FVelocityY: Integer;
    FAccelerationX: Integer;
    FAccelerationY: Integer;
    FMoveBy: TabcSpriteMove;
    FMoveAtEdge: TabcSpriteEdgeMove;
    FPoints: TabcPointList;
    FPointIndex: Integer;
    FMovedToEdge: Boolean;
    procedure ReadPoints(Reader: TReader);
    procedure WritePoints(Writer: TWriter);
    function GetPointCount: Integer;
    function GetPoint(Index: Integer): TPoint;
    procedure SetPoint(Index: Integer; const Value: TPoint);
    procedure SetAccelerationX(const Value: Integer);
    procedure SetAccelerationY(const Value: Integer);
    procedure SetPoints(const Value: TabcPointList);
    procedure SetPointIndex(const Value: Integer);
    procedure SetVelocityX(const Value: Integer);
    procedure SetVelocityY(const Value: Integer);
  protected
    procedure DefineProperties(Filer: TFiler); override;
    procedure ClearPoints;
    property MovedToEdge: Boolean read FMovedToEdge write FMovedToEdge;
  public
      {::Creates the mover for a sprite.}
    constructor Create(ASprite: TabcSprite);
      {::Destroys the mover.}
    destructor Destroy; override;
	    {::Copies the properties of one mover object to another.}
    procedure Assign(Source: TPersistent); override;
	    {::Updates the sprite's position for the next frame}
    procedure MoveSprite;
	    {::Adds point to the Points list}
    function AddPoint(const Pt: TPoint): Integer;
	    {::Removes a point from the Points list}
    function RemovePoint(const Pt: TPoint): Integer;
	    {::Deletes a point from the Points list}
    procedure DeletePoint(index: integer);
	    {::Replaces the Points list with <i>Steps</i> number of values in a straight line from <i>FromPt</i> to <i>ToPt</i>
	     using <i>Style</i> to adjust the distance between intermediate points.}
    procedure MakePoints(const FromPt, ToPt: TPoint; Steps: Word; Style: TabcMakePointsStyle = ampEven);
	    {::The number of items in the Points list. These can be referenced as Point[0..PointCount-1].}
    property PointCount: Integer read GetPointCount;
	    {::Point[0..PointCount-1] references a point in the Points list.}
    property Point[Index: Integer]: TPoint read GetPoint write SetPoint;
  published
        {::Velocity in the X axis, in pixels. Used when MoveBy=asmVelocity.}
    property VelocityX: Integer read FVelocityX write SetVelocityX default 0;
        {::Velocity in the Y axis, in pixels. Used when MoveBy=asmVelocity.}
    property VelocityY: Integer read FVelocityY write SetVelocityY default 0;
        {::Acceleration in the X axis, in pixels. Used when MoveBy=asmVelocity.}
    property AccelerationX: Integer read FAccelerationX write SetAccelerationX default 0;
        {::Acceleration in the Y axis, in pixels. Used when MoveBy=asmVelocity.}
    property AccelerationY: Integer read FAccelerationY write SetAccelerationY default 0;
        {::Defines how the movement changes when the sprite reaches the edge of the surface (when moving by amsVelocity),
        or when the sprite is positioned at the last point in the Points list (when moving by asmPoints),
        or when the sprite is positioned at the first point in the Points list (when moving by asmPointsReversed).<P><UL>
        <li>aseHide - the sprite is hidden.
        <li>aseBounce - the sprite's VelocityX and VelocityY and AccelerationX/Y properties are reversed, or if moving through a Points list the direction is reversed.
        <li>aseWrap - wraps the sprite to the opposite edge of the surface, or to the other end of the Points list.
        <li>aseStop - stops moving. i.e. sets move to asmNone.}
    property MoveAtEdge: TabcSpriteEdgeMove read FMoveAtEdge write FMoveAtEdge default aseHide;
        {::For automatic animation of sprites:<UL>
        <li>asmNone - the sprite does not move.
        <li>asmVelocity - move using VelocityX, VelocityY and AccelerationX/Y properties.
        <li>asmPoints - move using the Points list and PointIndex.
        <li>asmPointsReversed - moves backward through the Points list.}
    property MoveBy: TabcSpriteMove read FMoveBy write FMoveBy default asmNone;
    	  {::List of points to move through. The current Point is at PointIndex.<p>
	       Used when MoveBy=asmPoints or asmPointsReversed.<p>
	       Points can be populated using the <See Method=MakePoints> method.}
    property Points: TabcPointList read FPoints write SetPoints
               stored False; {values stored by DefineProperties}
        {::The current point when animating a sprite with the Points list.}
    property PointIndex: Integer read FPointIndex write SetPointIndex default 0;
  end;

    {: This is the base class for all ABC Sprite components.
    Sprites are components (i.e. they have an owner) and are not directly visible (i.e. not controls).<P>
    Sprites exist in conjunction with a <See Class=TabcSpriteList> and <See Class=TabcSpriteSurface>.<P>
    Each descendant class defines properties and methods to draw the sprite onto a Canvas at runtime.<P>
    Sprites receive event notifications at runtime from their SpriteList.<P>
    To create sprites at designtime, use the component editor of <See Class=TabcSpriteList>.<P>
    New sprite classes can be registered in the designer. See unit <See Unit=abcpspr> for details.
    }

  TabcSprite = class (TComponent)
  private
    FDrawInNextFrame: Boolean;
    FBoundsRect: TRect;
    FCollisionRect: TRect;
    FDropFrames: Integer;
    FHeight: Integer;
    FLeft: Integer;
    FOnAnimate: TabcSpriteAnimateEvent;
    FOnMovedToEdge: TNotifyEvent;
    FOnChange: TNotifyEvent;
    FOnClick: TNotifyEvent;
    FOnDblClick: TNotifyEvent;
    FSpriteList: TabcSpriteList;
    FTop: Integer;
    FVisible: Boolean;
    FWidth: Integer;
    FZOrder: TabcZOrder;
    FMover: TabcSpriteMover;
    FOnMouseUp: TMouseEvent;
    FOnMouseDown: TMouseEvent;
    FOnMouseEnter: TNotifyEvent;
    FOnMouseLeave: TNotifyEvent;
    FOnMouseMove: TMouseMoveEvent;
    FOnAnimateBegin: TNotifyEvent;
    FOnAnimateEnd: TNotifyEvent;
    FAnimating: Boolean;
    FListRemove: Boolean;
    FListFree: Boolean;
    FCursor, FSaveCursor: TCursor;
    FCursorStored: Boolean;
    procedure SetCursor(Value: TCursor);
    procedure SetCursorStored(Value: Boolean);
    function IsCursorStored: Boolean;
    function GetActive: Boolean;
    function GetPosition: TPoint;
    procedure SetBoundsRect(const Rect: TRect);
    procedure SetPosition(Point: TPoint);
    procedure SetLeft(Value: Integer);
    procedure SetTop(Value: Integer);
    procedure SetWidth(Value: Integer);
    procedure SetHeight(Value: Integer);
    function GetCursor: TCursor;
  protected
    FNeverDraw: Boolean;
    FCheckCollision: Boolean;
    procedure DefineProperties(Filer: TFiler); override;
    procedure ReadState(Reader: TReader); override;
    procedure SetParentComponent(AParent: TComponent); override;
    procedure CheckCollision;
    procedure Animate(Frame: Integer; var oldBounds, newBounds: TRect); virtual;
    procedure EndFrame; virtual;
    procedure AnimateBegin; virtual;
    procedure AnimateEnd; virtual;
    procedure EraseBounds; virtual;
    procedure Click(Pos: TPoint); dynamic;
    procedure DblClick(Pos: TPoint); dynamic;
    procedure MouseEnter; dynamic;
    procedure MouseLeave; dynamic;
    procedure MouseMove(Shift: TShiftState; Pos: TPoint); dynamic;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; Pos: TPoint); dynamic;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; Pos: TPoint); dynamic;
    procedure Draw(ACanvas: TCanvas); virtual;
    procedure Change(Sender: TObject); dynamic;
    procedure SetZOrder(Value: TabcZOrder); virtual;
    procedure SetCollisionRect(const Rect: TRect); virtual;
    procedure SetMover(Value: TabcSpriteMover); virtual;
    procedure SetSpriteList(Value: TabcSpriteList); virtual;
    procedure SetVisible(Value: Boolean); virtual;
    procedure Hide;
    procedure SetDrawInNextFrame(Value: Boolean); virtual;
    {::This is used to draw the sprite in the next animation frame.<p>
    This property is not for publishing, but is used internally when the sprite changes.}
    property DrawInNextFrame: Boolean read FDrawInNextFrame write SetDrawInNextFrame;
    {::The height of the sprite.<br>
    This is published if the sprite can be resized.<br>
    Note the height of some sprites is derived automatically, e.g. from the height of an image.}
    property Height: Integer read FHeight write SetHeight;
    {::The width of the sprite.<br>
    This is published if the sprite can be resized.<br>
    Note the width of some sprites is derived automatically, e.g. from the width of an image.}
    property Width: Integer read FWidth write SetWidth;
    {::Whether the sprite is visible at runtime.<br>
    This is published in all descendants except TabcInputSprite.}
    property Visible: Boolean read FVisible write SetVisible default True;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
	{::In a form, sprites are stored within their SpriteList.}
    function GetParentComponent: TComponent; override;
	{::In a form, sprites are stored within their SpriteList.}
    function HasParent: Boolean; override;
	{::Returns true if APoint is within the bounds of the Sprite.}
    function PtInSprite(const APoint: TPoint): Boolean; virtual;
	{::Returns true if ARect intersects with the Sprite.}
    function RectInSprite(const ARect: TRect): Boolean; virtual;
	{::Ensure the sprite is inside ARect and move it if necessary.<br>
	Returns true if the sprite was moved.
	}
    function MoveToRect(const ARect: TRect): Boolean;
	{::Ensure the sprite is inside the surface and move it if necessary.<br>
	Returns true if the sprite was moved.<br>
	Returns false if Sprite.SpriteList.Surface is not assigned.
	}
    function MoveToSurface: Boolean;
	{::Adjust the bounds of the Sprite. All dimensions are screen pixels.<br>
        Updates the Left, Top, Width and Height properties. (In some classes, Width and Height may be overridden or not published).<P>
	See also <See Property=BoundsRect> and <See Property=CollisionRect> properties.<br>
	Changing the <See Property=BoundsRect> updates the <See Property=CollisionRect>.<br>
	The default <See Property=CollisionRect> is 80% of the <See Property=BoundsRect> dimensions.}
    procedure SetBounds(ALeft, ATop, AWidth, AHeight: Integer); virtual;
	{::Returns the value of Sprite.SpriteList.Surface.Active<br>
	Returns false if Sprite.SpriteList.Surface is not assigned.
	}
    property Active: Boolean read GetActive;
	{::The bounding rectangle of the sprite, in pixels.<br>
	Changing this property also updates the <See Property=CollisionRect>.<br>
	The default <See Property=CollisionRect> is 80% of the BoundsRect dimensions.<br>
	Changing this property calls the <See Method=SetBounds> method<br>.}
    property BoundsRect: TRect read FBoundsRect write SetBoundsRect;
	{::The collision rectangle of the sprite, in pixels.
	This is used to make sure that sprites appear to touch when colliding.<p>
	Changing the <See Property=BoundsRect> updates the CollisionRect.<br>
	The default CollisionRect is 80% of the <See Property=BoundsRect> dimensions.<br>
	You can modify this property to set a custom CollisionRect.}
    property CollisionRect: TRect read FCollisionRect write SetCollisionRect;
	{::This is used to automate sprite animation by velocity, or using a list of points.}
    property Mover: TabcSpriteMover read FMover write SetMover;
	{::The current (Left,Top) position of the sprite, in pixels.}
    property Position: TPoint read GetPosition write SetPosition;
	{::The SpriteList to which the sprite belongs.<P>
	All sprite drawing and animation events are controlled by the SpriteList.
	}
    property SpriteList: TabcSpriteList read FSpriteList write SetSpriteList;
  published
        {::Can be used to define a custom cursor when the mouse moves over the sprite.}
    property Cursor: TCursor read GetCursor write SetCursor
      stored IsCursorStored default crDefault;
        {::Whether the Cursor is stored.}
    property CursorStored: Boolean read FCursorStored write SetCursorStored default False;
	{::Position of the sprite on the X-Axis, in pixels.}
    property Left: Integer read FLeft write SetLeft;
	{::Position of the sprite on the Y-Axis, in pixels.}
    property Top: Integer read FTop write SetTop;
	{::Use to slow the animation of a sprite, relative to the overall frame animation rate.<p>
	Set to a value of 2 or higher to only animate the sprite every Nth frame, starting with the first frame.<br>
	(All sprites are animated on the first frame to ensure consistent drawing.)<br>
	This property is ignored if the value is less than 2.
	}
    property DropFrames: Integer read FDropFrames write FDropFrames default 0;
        {::For front to back ordering of sprites.<br>
        0 is at the front.<br>
        When drawing, sprites are drawn from back to front.<br>
        When animating, sprites pass under/over depending on ZOrder.<br>
        Collisions between sprites only occur when they are at the same ZOrder.<br>
        <See Class=TabcSpriteList> also allows events to be restricted to a ZOrderLimit.
        }
    property ZOrder: TabcZOrder read FZOrder write SetZOrder default 0;
	{::Occurs prior to drawing each frame. This event can be used to move, resize or hide a sprite.}
    property OnAnimate: TabcSpriteAnimateEvent read FOnAnimate write FOnAnimate;
	{::Occurs when the <See Class=TabcSpriteSurface> commences animation}
    property OnAnimateBegin: TNotifyEvent read FOnAnimateBegin write FOnAnimateBegin;
	{::Occurs when the <See Class=TabcSpriteSurface> stops animation}
    property OnAnimateEnd: TNotifyEvent read FOnAnimateEnd write FOnAnimateEnd;
	{::Occurs when a property of the sprite changes.}
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
	{::Occurs when the mouse is clicked on the sprite.<br>
	See <See Class=TabcSpriteList> for details of mouse event handling.}
    property OnClick: TNotifyEvent read FOnClick write FOnClick;
	{::Occurs when the mouse is double-clicked on the sprite.<br>
	See <See Class=TabcSpriteList> for details of mouse event handling.}
    property OnDblClick: TNotifyEvent read FOnDblClick write FOnDblClick;
	{::Occurs when a mouse button is pushed down over the sprite.<br>
	See <See Class=TabcSpriteList> for details of mouse event handling.}
    property OnMouseDown: TMouseEvent read FOnMouseDown write FOnMouseDown;
	{::Occurs when the mouse first moves over the sprite.<br>
	See <See Class=TabcSpriteList> for details of mouse event handling.}
    property OnMouseEnter: TNotifyEvent read FOnMouseEnter write FOnMouseEnter;
	{::Occurs when the mouse was previously moving over the sprite and first moves away from the sprite.<br>
	See <See Class=TabcSpriteList> for details of mouse event handling.}
    property OnMouseLeave: TNotifyEvent read FOnMouseLeave write FOnMouseLeave;
	{::Occurs when the mouse is moved over the sprite.<br>
	See <See Class=TabcSpriteList> for details of mouse event handling.}
    property OnMouseMove: TMouseMoveEvent read FOnMouseMove write FOnMouseMove;
	{::Occurs when a mouse button is released over the sprite.<br>
	See <See Class=TabcSpriteList> for details of mouse event handling.}
    property OnMouseUp: TMouseEvent read FOnMouseUp write FOnMouseUp;
	{::Occurs after a sprite has moved using the Mover object and the new position overlaps or is completely off the edge of the surface.}
    property OnMovedToEdge: TNotifyEvent read FOnMovedToEdge write FOnMovedToEdge;
  end;

    {::Class type for use in sprite factories and the designtime editor.}

  TabcSpriteClass = class of TabcSprite;

    {:A TList containing TRect records.
     This class is mainly used for internal calculation of drawing rectangles.<br>
     It is also passed as an argument to some SpriteList events to allow additional rectangles to be added to the repaint list.<P>
     Each item in the list is a PRect = ^TRect.<br>
     This list allocates TRects in memory.<br>
     For memory-safe access use only the AddRect, AddRectNotEmpty, DeleteRect and Clear methods to change the list.<P>
     The protected Merge method sorts and compacts the list by intersecting rects.
    }

  TabcRectList = class (TList)
  protected
    procedure Merge;
  public
    function AddRect(const R: TRect): Integer;
    function AddRectNotEmpty(const R: TRect): Integer;
    procedure DeleteRect(Index: Integer);
    procedure Clear; override;
  end;

    {::Defines which events can be processed and/or passed on to Sprites in a SpriteList:<UL>
    <li> aseClick - Sprites can receive OnClick events
    <li> aseDblClick - Sprites can receive OnDblClick events.
    <li> aseUpDown - Sprites can receive OnMouseUp and OnMouseDown events.
    <li> aseEnterLeave - Sprites can receive OnMouseEnter and OnMouseLeave events.
    <li> aseMove - Sprites can receive OnMouseMove events.
    <li> aseInvisible - Invisible sprites can receive events as well as visible sprites.
    <li> aseCollisions - Collision events can be detected (this event occurs on the SpriteList not the Sprite).
    }

  TabcSpriteListEvent = (aseClick, aseDblClick, aseUpDown, aseEnterLeave, aseMove, aseInvisible, aseCollisions);

    {::Defines the set of events which can be processed and/or passed on to Sprites in a SpriteList.<P>
     Processing mouse events and collisions requires some CPU time to locate sprites.<br>
     While performance is typically not an issue, unwanted events can be filtered for optimum frame drawing performance when drawing many sprites or using very fast animation.
    }

  TabcSpriteListEvents = set of TabcSpriteListEvent;

    {::Occurs at the beginning of a frame, when animating sprites.<br>
     Sender is the SpriteList.<br>
     Sprites is a locked threaded list of TabcSprite components. Display properties of Sprites can be changed during this event.<br>
     Frame is the current frame number.<br>
     DirtyRects is a list of rectangles to be redrawn in the frame. Additional rectangles can be added to this list using its AddRect method.
    }

  TabcSpriteListAnimateEvent = procedure(const Sender: TObject; const Sprites: TList; Frame: Integer; const DirtyRects: TabcRectList) of object;

    {::This event can be used for custom drawing on the sprite surface canvas.<br>
     It is dispatched by the SpriteList at two times:
     OnDrawBegin, before any sprites are drawn to the canvas and OnDrawEnd after all sprites are drawn to the canvas.<br>
     Sender is the SpriteList.<br>
     The dirtyRect parameter must be updated with the Rect drawn on.<br>
    }

  TabcSpriteListDrawEvent = procedure (const Sender: TObject; const Canvas: TCanvas; var dirtyRect: TRect) of object;

  TabcSpriteSurface = class;

    {:This component is a container list and controller of <See Class=TabcSprite> components.
    <br>
    The main functions of this model and controller class are to:<UL>
    <li>Define a set of <See Class=TabcSprite> components to be drawn and moved on a <See Class=TabcSpriteSurface>.
    <li>Marshall animation and drawing events between the SpriteSurface and a (private) animation drawing/timing thread.
    <li>Dispatch frame animation and drawing methods of each sprite.
    <li>Control which mouse events are passed from the SpriteSurface to a sprite.
    <li>Detect collisions between sprites.
    </UL>
    <P>
    A SpriteList is able to be drawn by connecting it to the <See Property=TabcSpriteSurface.Sprites> property.<br>
    Multiple SpriteLists can be used for separate animations to be drawn on a surface at different times.
    <P><P>
    <Strong>Design and Runtime Properties:</Strong>
    <P>
    The <See Property=Background> property can be used to set a gradient, image or tiled background behind all sprites.
    (The <See Property=TabcSpriteSurface.Background> property can also be used to set a common background). If this property is not empty it takes precedence over the <See Property=TabcSpriteSurface.Background>.<P>
    The <See Property=Events>, <See Property=ZOrderLimitCollisions>, and <See Property=ZOrderLimitMouse> properties can be used to control mouse and collision event processing. With a very large number of sprites, these can also be used to optimize performance by limiting events. In most cases, event handling performance is not a limitation.
    <P>
    <P>
    <Strong>Runtime Properties:</Strong><P>
    <See Property=Count> indicates the number of sprites in the list.<br>
    <See Property=Sprites>[0..Count-1] references a sprite in the list.<br>
    Note the index of a sprite may not be constant. Sprites are sorted internally by the list processing.<br>
    You can set <See Property=Sorted> to true to sort the sprites by their ZOrder.<br>
    <See Property=Surface> returns the <See Class=TabcSpriteSurface> that the list is connected to.
    <P>
    <P>
    <Strong>Runtime Methods:</Strong><P>
    <P>
    <See Method=BeginUpdate>..<See Method=EndUpdate> methods are used when making a series of changes to the list. This defers change events which might otherwise require recalculation or repainting, until EndUpdate is called.<br>
    <See Method=Add> adds a sprite to the list.<br>
    <See Method=Remove> removes a sprite from the list. If the sprite is to be destroyed, use <See Method=RemoveAndFree>. RemoveAndFree frees the sprite only after animation and drawing are completed.<br>
    <See Method=Clear> removes all sprites from the list. If all sprites are to be destroyed, use <See Method=ClearAndFree>. ClearAndFree frees the sprites only after animation and drawing are completed.<br>
    <P>
    <Strong>Runtime Events:</Strong><P>
    <See Event=OnCollision> occurs when two sprites collide. Sprites can only collide if they move over each other's CollisionRect and both have the same ZOrder.<br>
    <See Event=OnAdd> occurs when a sprite is added to the list. This might be useful when a <See Class=TabsSpriteFactory> is used to add sprites.<br>
    <See Event=OnRemove> occurs when a sprite is removed from the list.<br>
    <See Event=OnAnimateBegin> occurs at the start of the first frame, when the SpriteSurface commences animation.<br>
    <See Event=OnAnimateEnd> occurs after the end of the last frame, when the SpriteSurface stops animation.<br>
    <See Event=OnFrameBegin> occurs for every animation frame, before any sprites are animated or drawn.<br>
    <See Event=OnDrawBegin> occurs for every animation frame, before any sprites are drawn.<br>
    <See Event=OnDrawEnd> occurs for every animation frame, after all sprites are drawn.<br>
    <See Event=OnFrameEnd> occurs for every animation frame, after the frame has been drawn to the screen.<br>
    <P>
    <P>
    }

  TabcSpriteList = class (TComponent)
  private
    FSprites: TThreadList;
    FEvents: TabcSpriteListEvents;
    FZOrderLimitCollisions: Integer;
    FZOrderLimitMouse: Integer;
    FOnAdd: TabcSpriteEvent;
    FOnFrameBegin: TabcSpriteListAnimateEvent;
    FOnFrameEnd: TNotifyEvent;
    FOnChange: TNotifyEvent;
    FOnCollision: TabcSpriteCollisionEvent;
    FOnRemove: TabcSpriteEvent;
    FSorted: Boolean;
    FRemovedRects: TabcRectList;
    FSurface: TabcSpriteSurface;
    FOnAnimateBegin: TNotifyEvent;
    FOnAnimateEnd: TNotifyEvent;
    FChanged: Boolean;
    FUpdateCount: Integer;
    FLastSpriteEntered, FLastSpriteLeftDown: TabcSprite;
    FMouseEntered: Boolean;
    FListLockCount: Integer;
    FList: TList;
    FTileBack: TabcTileBack;
    FOnDrawEnd: TabcSpriteListDrawEvent;
    FOnDrawBegin: TabcSpriteListDrawEvent;
    procedure TileBackChanged(Sender: TObject);
    procedure SetBackground(Value: TabcTileBack);
    function LockSprites: TList;
    procedure UnlockSprites;
    procedure UnlockandRemoveSprites;
    function GetCount: Integer;
    function GetSprites(Index: Integer): TabcSprite;
    procedure SetSorted(Value: Boolean);
    procedure SetMouseEntered(Value: Boolean);
    procedure SpriteEnterLeave(SpriteMoved: TabcSprite);
    procedure AddIntersectingRects(const SurfaceBounds: TRect; const DirtyRects: TabcRectList);
  protected
    procedure InternalRemove(Sprite: TabcSprite);
    procedure InternalRemoveSprites;
    procedure GetChildren(Proc: TGetChildProc; Root: TComponent); override;
    procedure Loaded; override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure HideSprite(Sprite: TabcSprite);
    procedure AnimateSprites(Frame: Integer; const DirtyRects: TabcRectList);
    procedure AnimateBegin;
    procedure AnimateEnd;
    procedure DetectCollisions;
    procedure DrawSprites(const ACanvas: TCanvas; DrawAll: Boolean; const DirtyRects: TabcRectList);
    procedure EndFrame;
    procedure SpriteClick(Pos: TPoint);
    procedure SpriteDblClick(Pos: TPoint);
    procedure SpriteMouseMove(Shift: TShiftState; Pos: TPoint);
    procedure SpriteMouseDown(Button: TMouseButton; Shift: TShiftState; Pos: TPoint);
    procedure SpriteMouseUp(Button: TMouseButton; Shift: TShiftState; Pos: TPoint);
    property MouseEntered: Boolean read FMouseEntered write SetMouseEntered;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
	{::Adds a sprite to the list.}
    procedure Add(Sprite: TabcSprite);
	{::Begins a batch of updates to the list, deferring changes. Each call MUST be followed by calling EndUpdate. Calls can be nested.}
    procedure BeginUpdate;
	{::Called for the closure of BeginUpdate.}
    procedure EndUpdate;
	{::Normally only called by a sprite method to signify that a change has occured that could require sorting or repainting the spritelist.}
    procedure Change;
	{::Removes all sprites from the list.}
    procedure Clear;
	{::Removes all sprites from the list and frees them after any pending animation or drawing events.}
    procedure ClearAndFree;
	{::Erases a rectangle on the surface, if the spritelist is visible. Normally only called by sprite methods.}
    procedure EraseRect(const R: TRect);
	{::Returns the topmost sprite in the ZOrder located at APoint.}
    function GetSpriteAtPos(APoint: TPoint; IncludeInvisible: Boolean = False; ZOrderLimit: TabcZOrder = Maxint): TabcSprite;
	{::Removes a sprite from the list.}
    procedure Remove(Sprite: TabcSprite);
	{::Removes a sprite from the list and frees it after any pending animation or drawing events.}
    procedure RemoveAndFree(Sprite: TabcSprite);
	{::Sorts the sprites by ZOrder.}
    procedure SortSprites;
	{::Returns the number of sprites in the list.}
    property Count: Integer read GetCount;
	{::Occurs when the list contents, or a sprite in the list changes.}
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
	{::Set to True to sort the sprites by ZOrder.}
    property Sorted: Boolean read FSorted write SetSorted;
	{::Returns sprites in the list, indexed from [0..Count-1].}
    property Sprites[Index: Integer]: TabcSprite read GetSprites;
	{::Returns the surface the spritelist is connected to.}
    property Surface: TabcSpriteSurface read FSurface;
  published
        {::Allows a gradient, image or tiled background to be drawn behind all sprites.}
    property Background: TabcTileBack read FTileBack write SetBackground;
        {::Controls which mouse events and collisions are processed.<P>
        The default is all events [aseClick, aseDblClick, aseUpDown, aseEnterLeave, aseMove, aseInvisible, aseCollisions].}
    property Events: TabcSpriteListEvents read FEvents write FEvents
        default [aseClick, aseDblClick, aseUpDown, aseEnterLeave, aseMove, aseInvisible, aseCollisions];
        {::Limits the ZOrder depth to which collisions are tested on sprites.}
    property ZOrderLimitCollisions: Integer read FZOrderLimitCollisions write
        FZOrderLimitCollisions default Maxint;
        {::Limits the ZOrder depth to which mouse events are passed on to sprites.}
    property ZOrderLimitMouse: Integer read FZOrderLimitMouse write
        FZOrderLimitMouse default Maxint;
        {::Occurs when a sprite is added to the list.}
    property OnAdd: TabcSpriteEvent read FOnAdd write FOnAdd;
        {::Occurs during each frame, after all sprites are animated and before any sprites are drawn.<br>
        This allows custom drawing on the canvas.}
    property OnDrawBegin: TabcSpriteListDrawEvent read FOnDrawBegin write FOnDrawBegin;
        {::Occurs during each frame, after all sprites are animated and after all sprites are drawn.<br>
        This allows custom drawing on the canvas.}
    property OnDrawEnd: TabcSpriteListDrawEvent read FOnDrawEnd write FOnDrawEnd;
        {::Occurs at the beginning of each frame, before any sprites are animated or drawn.}
    property OnFrameBegin: TabcSpriteListAnimateEvent read FOnFrameBegin write FOnFrameBegin;
        {::Occurs at the end of each frame, after sprites have been drawn to the screen.}
    property OnFrameEnd: TNotifyEvent read FOnFrameEnd write FOnFrameEnd;
        {::Occurs at the start of the first frame, when the SpriteSurface commences animation.}
    property OnAnimateBegin: TNotifyEvent read FOnAnimateBegin write FOnAnimateBegin;
        {::Occurs after the end of the last frame, when the SpriteSurface stops animation.}
    property OnAnimateEnd: TNotifyEvent read FOnAnimateEnd write FOnAnimateEnd;
        {::Occurs when two sprites collide. Sprites can only collide if they move over each other's CollisionRect and both have the same ZOrder.}
    property OnCollision: TabcSpriteCollisionEvent read FOnCollision write FOnCollision;
        {::Occurs when a sprite is removed from the list.}
    property OnRemove: TabcSpriteEvent read FOnRemove write FOnRemove;
  end;

{$IFNDEF T2H}  //this internal class is not documented.

  TabcSpriteThread = class (TThread)
  private
    FSurface: TabcSpriteSurface;
    FDirtyRects: TabcRectList;
    FFrame, FFramesLastSec: Integer;
    FStartNextFrame, FStartThisFrame, FStartThisSec,
    FFrameDuration, FSleepFor: DWORD;
  protected
    procedure Execute; override;
    procedure AnimateBegin;
    procedure AnimateEnd;
    procedure AnimateFrame;
    procedure AnimateSprites;
    procedure DetectCollisions;
    procedure DrawMemory;
    procedure DrawScreen;
    procedure EndFrame;
    procedure StartFrame;
  public
    constructor Create(ASurface: TabcSpriteSurface);
    destructor Destroy; override;
    procedure Reset;
  end;
{$ENDIF} //{DEF T2H} this internal class is not documented.


    {::Number of FramesPerSec to draw on a <See Class=TabcSpriteSurface>.}

  TabcFrameRate = 0..MaxInt;

    {::Duration of each frame in milliseconds to draw on a <See Class=TabcSpriteSurface>.}

  TabcFrameDuration = 1..MaxInt;

    {:This control is the visible surface that a <See class=TabcSpriteList> is drawn onto.
    The SpriteSurface is a graphic control that bounds the visible drawing area for a SpriteList.
    It allows Sprites in the SpriteList to be drawn on the canvas and moved using sequential frame animation.
    <p>
    Set the <See Property=Sprites> property to the SpriteList to be drawn.<p>
    Sprite frame animation is set to a millisecond time interval using the
    <See Property=FramesPerSec> or <See Property=FrameDuration> properties.
    In most cases a rate of more than 40 FramesPerSec will not result in any visible difference.
    This timing is the desired maximum animation rate and should be achieved with most animations.<p>

    Animation is turned on or off by the <See Property=Active> and <See Property=SuspendOnAppInactive> properties;
    or using the <See Method=Start>, <See Method=Stop>, <See Method=Suspend>, and <See Method=Resume> methods.
    Animation is only performed at runtime.
    The <See Property=EraseFrames> property can be turned off to allow trails of moving sprites to draw across the surface.
    <p>

    Painting of the control's background appearance is controlled by the
    <See Property=Color>, <See Property=Background> and <See Property=Transparent> properties.
    When Transparent, a copy is made of the image of all controls behind this control.
    The <See Method=RefreshBackground> method can be called to update this snapshot.<P>

    Sprites are usually only drawn when Active or at designtime, unless the <See Property=DrawWhenInactive> property is set.
    If this property is true, a buffered image of the sprites is used when painting.
    This buffer is invalidated if the sprites change when the surface is not Active.
    The <See Method=ClearSprites> and <See Method=DrawSprites> methods can be called to repaint the surface.<p>

    Runtime mouse events occur on the surface and are also passed to adjacent sprites by <See Class=TabcSpriteList>.<p>

    Animation drawing is precisely timed and synchronized using an internal thread.
    This thread interacts with the Surface and the SpriteList.
    The thread and SpriteList optimize frame drawing by only painting changed sprites
    and invalid rectangles within the surface; and by buffering all screen painting.<p>

    Runtime animation performance can be monitored by the <See Property=ShowPerformance>,
    <See Property=RealFramesPerSec>, <See Property=AvgFramesPerSec>, and <See Property=FrameCount> properties.
    Performance depends on many factors: the machine speed, video, version of Windows,
    number of moving sprites and total number and size of sprites, number of other threads, etc.
    and should be tested with your application.<p>

    For optimal performance, Animation drawing is not synchronized with parent window repainting.
    Any TWinControl on top of the surface is not affected, but any other TControl on top of the surface may not repaint correctly when animating.<p>
    }

  TabcSpriteSurface = class (TGraphicControl)
  private
    FActive: Boolean;
    FAfterFrame: TNotifyEvent;
    FBeforeFrame: TNotifyEvent;
    FDrawWhenInactive: Boolean;
    FScreenBuffer: TBitmap;
    FEraseFrames: Boolean;
    FFrameCount: Integer;
    FFrameDuration: TabcFrameDuration;
    FStoreFrameDuration: Boolean;
    FFramesPerSec: TabcFrameRate;
    FTileBack: TabcTileBack;
    FBackgroundBitmap: TBitmap;
    FRealFramesPerSec: Integer;
    FSprites: TabcSpriteList;
    FSpriteThread: TabcSpriteThread;
    FTransparent: Boolean;
    FNoPainting: Boolean;
    FNeverPainted: Boolean;
    FCreateThreadOnPaint: Boolean;
    FSuspendOnAppInactive: Boolean;
    FOnMouseEnter: TNotifyEvent;
    FOnMouseLeave: TNotifyEvent;
    FMouseOver: Boolean;
    FWasSuspended: Boolean;
    FActiveSeconds, FActiveSecsFrameCount: Integer;
    FPerfFont: TFont;
    FShowPerformance: Boolean;
    FPerfTextR, FPrevPerfTextR: TRect;
    FPerformanceText: string;
    procedure TileBackChanged(Sender: TObject);
    procedure SetBackground(Value: TabcTileBack);
    function MouseIsOver: Boolean;
    procedure UpdateMouseOver;
    function GetCanvas: TCanvas;
    procedure SetActive(Value: Boolean);
    procedure SetDrawWhenInactive(Value: Boolean);
    procedure SetSprites(Value: TabcSpriteList);
    procedure SetTransparent(Value: Boolean);
    procedure CMColorChanged(var Message: TMessage); message CM_COLORCHANGED;
    procedure CMEnabledChanged(var Message: TMessage); message CM_ENABLEDCHANGED;
    procedure CMMouseEnter(var Message: TMessage); message CM_MOUSEENTER;
    procedure CMMouseLeave(var Message: TMessage); message CM_MOUSELEAVE;
    procedure CMParentColorChanged(var Message: TMessage); message CM_PARENTCOLORCHANGED;
    procedure CMVisibleChanged(var Message: TMessage); message CM_VISIBLECHANGED;
    procedure WMPaint(var Message: TWMPaint); message WM_PAINT;
    procedure ApplicationActivated(Sender: TObject);
    procedure ApplicationDeactivated(Sender: TObject);
    function IsFrameDurationStored: Boolean;
    function IsFramesPerSecStored: Boolean;
    procedure SetFrameDuration(const Value: TabcFrameDuration);
    procedure SetFramesPerSec(const Value: TabcFrameRate);
    function GetAvgFramesPerSec: Integer;
  protected
    procedure DoInactiveDrawing;
    procedure UpdateBackgroundBitmap;
    procedure DrawBackground;
    procedure Click; override;
    procedure DblClick; override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;
    procedure Loaded; override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure Paint; override;
    procedure MouseEnter; dynamic;
    procedure MouseLeave; dynamic;
    procedure SpriteListChanged(Sender: TObject);
    procedure CanCreateThread;
    procedure CanTerminateThread;
    procedure CanResetThread;
    function GetInheritedCanvas: TCanvas;
    function DrawPerformanceText: TRect;
    procedure UpdatePerformanceText;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure SetBounds(ALeft, ATop, AWidth, AHeight: Integer); override;
        {::Starts frame animation, if not already Active. This is the same as setting Active to true.}
    procedure Start;
        {::Stops frame animation, if Active. This is the same as setting Active to false.}
    procedure Stop;
        {::Suspends frame animation, if Active. This has no effect if not Active.}
    procedure Suspend;
        {::Resumes frame animation, if Suspended. This has no effect if not Suspended.}
    procedure Resume;
        {::Clears the buffered image of Sprites, when inactive.}
    procedure ClearSprites;
        {::Draws the buffered image of Sprites, when inactive.}
    procedure DrawSprites;
        {::Refreshes the background image of the control.
        This can be used when <See Property=Transparent> is true and the image behind the control changes.}
    procedure RefreshBackground;
        {::This returns the drawing canvas of the control which is an internal buffer, not the parent window's canvas.}
    property Canvas: TCanvas read GetCanvas;
        {::This defaults to true. If this is set to false,
        sprites that move will leave trails on the surface.}
    property EraseFrames: Boolean read FEraseFrames write FEraseFrames default True;
        {::The current frame that is being drawn, when Active.}
    property FrameCount: Integer read FFrameCount;
        {::Returns true when the mouse is over the surface.}
    property MouseOver: Boolean read FMouseOver;
        {::Returns the number of frames drawn in the last second.}
    property RealFramesPerSec: Integer read FRealFramesPerSec;
        {::Returns the average number of frames drawn per second, since animation started.}
    property AvgFramesPerSec: Integer read GetAvgFramesPerSec;
  published
        {::Set to true to start animation.
        When set at designtime, the control will begin animation at runtime once
        it is Visible and showing on the screen. This has no effect if the Sprites property is not assigned.}
    property Active: Boolean read FActive write SetActive default False;
        {::Contains a gradient, image or tiled image background that is drawn behind all sprites.<br>
        If the <See Class=TabcSpriteList> contains a background the spritelist's background takes precedence.}
    property Background: TabcTileBack read FTileBack write SetBackground;
        {::This defaults to false. Set to true for the sprites to be drawn when not Active, i.e. like the control paints at designtime.}
    property DrawWhenInactive: Boolean read FDrawWhenInactive write SetDrawWhenInactive default False;
        {::Use to define frame animation duration in milliseconds.<p>
        Both this property and <See Property=FramesPerSec> can be used to adjust the timing.<p>
        Use this property to set a more accurate duration in case of rounding, or for slow rates of less than one frame per second.}
    property FrameDuration: TabcFrameDuration read FFrameDuration write SetFrameDuration stored IsFrameDurationStored;
        {::Use to define frame rate in number of frames per second.<p>
        Setting this property changes <See Property=FrameDuration>.<p>
        Use this property to set a typical frame rate, such as 40 frames per second.<p>
        Use <See Property=FrameDuration> to set a more accurate duration in case of rounding, or for slow rates of less than one frame per second.}
    property FramesPerSec: TabcFrameRate read FFramesPerSec write SetFramesPerSec stored IsFramesPerSecStored;
        {::If true, animation is suspended whenever the application is not the Windows foreground application.}
    property SuspendOnAppInactive: Boolean read FSuspendOnAppInactive write FSuspendOnAppInactive default True;
        {::Connects the SpriteList to be drawn on the surface.}
    property Sprites: TabcSpriteList read FSprites write SetSprites;
        {::Set to true if the parent window or other control images behind the surface must be painted.}
    property Transparent: Boolean read FTransparent write SetTransparent default False;
        {::Occurs after all processing of a frame is completed and the animation has finished drawing to the screen.}
    property AfterFrame: TNotifyEvent read FAfterFrame write FAfterFrame;
        {::Occurs immediately after the FrameCount is incremented, but before any other processing of the next animation frame.}
    property BeforeFrame: TNotifyEvent read FBeforeFrame write FBeforeFrame;
    property Align;
    property Anchors;
    property Color;
    property Constraints;
    property DragCursor;
    property DragKind;
    property DragMode;
    property Enabled;
    property ParentColor;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
        {::Set to true for a text display of the average and current Frames Per Second.}
    property ShowPerformance: Boolean read FShowPerformance write FShowPerformance default False;
    property Visible;
    property OnClick;
{$IFDEF DELPHI5}
    property OnContextPopup;
{$ENDIF}
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDock;
    property OnEndDrag;
    property OnMouseDown;
        {::Occurs when the mouse first moves over the control.}
    property OnMouseEnter: TNotifyEvent read FOnMouseEnter write FOnMouseEnter;
        {::Occurs when the mouse first moves off the control.}
    property OnMouseLeave: TNotifyEvent read FOnMouseLeave write FOnMouseLeave;
    property OnMouseMove;
    property OnMouseUp;
    property OnStartDock;
    property OnStartDrag;
  end;

    {: This is an abstract class for factory components which can create multiple sprites.
    Like a cookie cutter, a sprite factory is used to create multiple instances
    of sprite components, based on a common template.<p>
    The sprite factory defines some common properties and an interface with the SpriteList class.<p>
    This class contains abstract methods which must be overridden in each descendant class.<p>

    <See Class=TabcBitmapSpriteFactory> is an example of a descendant class.
    Its purpose is to create multiple instances of bitmap sprites.
    The bitmap sprite class it creates is optimized to use a single shared bitmap image that belongs to the factory.
    }

  TabcSpriteFactory = class (TComponent)
  private
    FChanged: Boolean;
    FUpdateCount: Integer;
    FOnChange: TNotifyEvent;
    FMover: TabcSpriteMover;
    FSpriteList: TabcSpriteList;
    procedure SetMover(Value: TabcSpriteMover);
  protected
    procedure BeginUpdate;
    procedure EndUpdate;
    procedure Change; dynamic;
    procedure SetSpriteList(const Value: TabcSpriteList); virtual;
    {can be published}
    property Mover: TabcSpriteMover read FMover write SetMover;
    property SpriteList: TabcSpriteList read FSpriteList write SetSpriteList;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
        {::You must override this method to return the class type of sprites created by the factory class.
        }
    class function SpriteClass: TabcSpriteClass; virtual;
        {::By default, this function returns a new instance of SpriteClass that is owned by the factory and
        copies the Mover property (protected, can be published) to the new sprite.
        If the SpriteList (protected, can be published) property is assigned, it adds the new sprite to the list.<p>
        You can replace or supplement this if some other constructor method is required.
        For consistency, it is best if sprites created by factories are owned by the factory.
        }
    function CreateSprite: TabcSprite; virtual;
  end;

implementation

uses dialogs;

{
-------------------------------------- TabcPointList ---------------------------
}

function TabcPointList.AddPoint(const Pt: TPoint): Integer;
var
  newPt: PPoint;
begin
  New(newPt);
  newPt^.X := Pt.X;
  NewPt^.Y := Pt.Y;
  Result := Add(newPt);
end;

procedure TabcPointList.DeletePoint(Index: Integer);
var
  Pt: PPoint;
begin
  Pt := Items[Index];
  Dispose(Pt);
  Delete(Index);
end;

procedure TabcPointList.Clear;
var
  i: integer;
begin
  for i := 0 to Count - 1 do
    Dispose(PPoint(Items[i]));
  inherited Clear;
end;

{
-------------------------------------- TabcRectList ----------------------------
}

function TabcRectList.AddRect(const R: TRect): Integer;
var
  newRect: PRect;
begin
  New(newRect);
  CopyRect(newRect^, R);
  Result := Add(newRect);
end;

function TabcRectList.AddRectNotEmpty(const R: TRect): Integer;
begin
  if IsRectEmpty(R) then
    Result := -1
  else
    Result := AddRect(R);
end;

procedure TabcRectList.DeleteRect(Index: Integer);
var
  pR: PRect;
begin
  pR := Items[Index];
  Dispose(pR);
  Delete(Index);
end;

procedure TabcRectList.Clear;
var
  i: integer;
begin
  for i := 0 to Count - 1 do
    Dispose(PRect(Items[i]));
  inherited Clear;
end;

procedure TabcRectList.Merge;
var
  i,j: integer;
  IntersectR, R1, R2: TRect;
label BeginMerge;
begin
BeginMerge:
  for i := 0 to Count - 2 do
  begin
    R1 := TRect(Items[i]^);
    for j := i+1 to Count-1 do
    begin
      R2 := TRect(Items[j]^);
      if IntersectRect(IntersectR, R1, R2) then
      begin
        UnionRect(TRect(Items[i]^), R1, R2);
        DeleteRect(j);
        goto BeginMerge;
      end;
    end;
  end;
end;

{
---------------------------- TabcSpriteMover -----------------------------------
}

constructor TabcSpriteMover.Create(ASprite: TabcSprite);
begin
  inherited Create;
  FSprite := ASprite;
  FPoints := TabcPointList.Create;
end;

destructor TabcSpriteMover.Destroy;
begin
  ClearPoints;
  FPoints.Free;
  inherited;
end;

procedure TabcSpriteMover.Assign(Source: TPersistent);
begin
  if Source is TabcSpriteMover then
    with TabcSpriteMover(Source) do
    begin
      Self.FVelocityX := VelocityX;
      Self.FVelocityY := VelocityY;
      Self.FAccelerationX := AccelerationX;
      Self.FAccelerationY := AccelerationY;
      Self.FMoveAtEdge := MoveAtEdge;
      Self.FMoveBy := MoveBy;
      Self.Points := Points;
      Self.FPointIndex := PointIndex;
    end
  else
    inherited;
end;

procedure TabcSpriteMover.ClearPoints;
begin
  FPoints.Clear;
end;

procedure TabcSpriteMover.MoveSprite;
var
  NewSpriteBounds, EdgeRect, EdgeUnion: TRect;
  offsetX, offsetY, midX, midY: integer;
begin
//  TabcSpriteMove = (asmNone, asmVelocity, asmPoints, asmPointsReversed);
//  TabcSpriteEdgeMove = (aseHide, aseBounce, aseWrap, aseStop);
  if (FSprite = nil) or (MoveBy = asmNone) then
  begin
    FMovedToEdge := False;
    Exit;
  end;
  if MoveBy = asmVelocity then
  begin
    if (VelocityX <> 0) or (VelocityY <> 0) then
    begin
      CopyRect(NewSpriteBounds, FSprite.BoundsRect);
      OffsetRect(NewSpriteBounds, VelocityX, VelocityY);
      CopyRect(EdgeRect, FSprite.SpriteList.Surface.ClientRect);
      UnionRect(EdgeUnion, NewSpriteBounds, EdgeRect);
      FMovedToEdge := not EqualRect(EdgeUnion, EdgeRect);
      if FMovedToEdge then
        case MoveAtEdge of
          aseHide:
            FSprite.Visible := False;
          aseBounce:
          begin
            //change velocity and acceleration depending on edge overlapped
            if (NewSpriteBounds.Left < EdgeRect.Left) or (NewSpriteBounds.Right > EdgeRect.Right) then
            begin
              VelocityX := - VelocityX;
              AccelerationX := - AccelerationX;
            end;
            if (NewSpriteBounds.Top < EdgeRect.Top) or (NewSpriteBounds.Bottom > EdgeRect.Bottom) then
            begin
              VelocityY := - VelocityY;
              AccelerationY := - AccelerationY;
            end;
          end;
          aseWrap:
          begin
            midX := (NewSpriteBounds.Left + NewSpriteBounds.Right) div 2;
            midY := (NewSpriteBounds.Top + NewSpriteBounds.Bottom) div 2;
            if midX < EdgeRect.Left then
              offsetX := EdgeRect.Right
            else if midX > EdgeRect.Right then
              offsetX := - EdgeRect.Right
            else
              offsetX := 0;
            if midY < EdgeRect.Top then
              offsetY := EdgeRect.Bottom
            else if midY > EdgeRect.Bottom then
              offsetY := - EdgeRect.Bottom
            else
              offsetY := 0;
            OffsetRect(NewSpriteBounds, offsetX, offsetY);
          end;
          aseStop:
            MoveBy := asmNone;
        end;
      FSprite.BoundsRect := NewSpriteBounds;
    end;
    VelocityX := VelocityX + AccelerationX;
    VelocityY := VelocityY + AccelerationY;
  end
  else
  begin
    if PointCount < 2 then Exit;
    if MoveBy = asmPoints then
    begin
      FMovedToEdge := (PointIndex = PointCount - 1);
      if FMovedToEdge then
        case MoveAtEdge of
          aseHide:
            FSprite.Visible := False;
          aseBounce:
            begin
              MoveBy := asmPointsReversed;
              PointIndex := PointIndex - 1;
            end;
          aseWrap:
            PointIndex := 0;
          aseStop:
            MoveBy := asmNone;
        end
      else
        PointIndex := PointIndex + 1;
    end
    else //if MoveBy = asmPointsReversed then
    begin
      FMovedToEdge := (PointIndex = 0);
      if FMovedToEdge then
        case MoveAtEdge of
          aseHide:
            FSprite.Visible := False;
          aseBounce:
            begin
              MoveBy := asmPoints;
              PointIndex := 1;
            end;
          aseWrap:
            PointIndex := PointCount - 1;
          aseStop:
            MoveBy := asmNone;
        end
      else
        PointIndex := PointIndex - 1;
    end;
  end;
end;

function TabcSpriteMover.AddPoint(const Pt: TPoint): Integer;
begin
  Result := FPoints.AddPoint(Pt);
end;

function DivInterval(const PtA, PtB: TPoint; ratioM, ratioN: integer): TPoint;
begin
  Result.X := (ratioM * PtB.X + ratioN * PtA.X) div (ratioM + ratioN);
  Result.Y := (ratioM * PtB.Y + ratioN * PtA.Y) div (ratioM + ratioN);
end;

procedure TabcSpriteMover.MakePoints(const FromPt, ToPt: TPoint; Steps: Word; Style: TabcMakePointsStyle = ampEven);
var
  i, Left, Right, Divs: integer;
begin
  Left := 0;
  Right := 0;
  Divs := 0;
  ClearPoints;
  AddPoint(FromPt);
  if Style in [ampAccelerate, ampDecelerate] then
  begin
    for i := 1 to Steps do
      Divs := Divs + i;
    if Style = ampAccelerate then
      Left := 0
    else
      Right := Divs;
  end;
  //calculate intermediate points along the line
  if Steps > 1 then
    for i := 1 to Steps - 1 do
    begin
      case Style of
        ampEven:
          begin
            Left := i;
            Right := Steps-i;
          end;
        ampAccelerate:
          begin
            Left := Left+i;
            Right := Divs-Left;
          end;
        ampDecelerate:
          begin
            Right := Right-(Steps-i+1);
            Left := Divs-Right;
          end;
        ampAccelMid:
          begin
            Left := i*i;
            Right := (Steps-i)*(Steps-i);
          end;
      end;
      AddPoint(DivInterval(FromPt, ToPt, Left, Right));
    end;
  AddPoint(ToPt);
end;

procedure TabcSpriteMover.DefineProperties(Filer: TFiler);

  function PointsEqual(List1, List2: TabcPointList): Boolean;
  var
    i: integer;
    Pt1, Pt2: TPoint;
  begin
    Result := List1.Count = List2.Count;
    if Result then
      for i := 0 to List1.Count - 1 do
      begin
        Pt1 := PPoint(List1.Items[i])^;
        Pt2 := PPoint(List2.Items[i])^;
        if (Pt1.X <> Pt2.X) or (Pt1.Y <> Pt2.Y) then
        begin
          Result := False;
          Break;
        end;
      end;
  end;

  function DoWritePoints: Boolean;
  begin
    if Filer.Ancestor <> nil then
      Result := not PointsEqual(FPoints, TabcSpriteMover(Filer.Ancestor).FPoints)
    else
      Result := (FPoints.Count > 0);
  end;

begin
  inherited;
  Filer.DefineProperty('PointsValues', ReadPoints, WritePoints, DoWritePoints);
end;

procedure TabcSpriteMover.DeletePoint(index: integer);
begin
  FPoints.DeletePoint(index);
end;

function TabcSpriteMover.GetPoint(Index: Integer): TPoint;
begin
  Result := PPoint(FPoints.Items[Index])^;
end;

procedure TabcSpriteMover.SetPoint(Index: Integer; const Value: TPoint);
begin
  if (Index >= 0) and (Index < FPoints.Count) then
    with PPoint(FPoints.Items[Index])^ do
    begin
      X := Value.X;
      Y := Value.Y;
    end;
end;

function TabcSpriteMover.GetPointCount: Integer;
begin
  Result := FPoints.Count;
end;

function TabcSpriteMover.RemovePoint(const Pt: TPoint): Integer;
var
  i: integer;
begin
  with FPoints do
    for i := 0 to Count - 1 do
      with PPoint(Items[i])^ do
        if (X = Pt.X) and (Y = Pt.Y) then
        begin
          Result := i;
          DeletePoint(i);
          Exit;
        end;
  Result := -1;
end;

procedure TabcSpriteMover.SetAccelerationX(const Value: Integer);
begin
  if Value <> FAccelerationX then
  begin
    FAccelerationX := Value;
    if (FSprite <> nil) and not (csLoading in FSprite.ComponentState) then
      MoveBy := asmVelocity;
  end;
end;

procedure TabcSpriteMover.SetAccelerationY(const Value: Integer);
begin
  if Value <> FAccelerationY then
  begin
    FAccelerationY := Value;
    if (FSprite <> nil) and not (csLoading in FSprite.ComponentState) then
      MoveBy := asmVelocity;
  end;
end;

procedure TabcSpriteMover.SetPoints(const Value: TabcPointList);
var
  i: integer;
begin
  if Value <> FPoints then
  begin
    ClearPoints;
    if Value <> nil then
      for i := 0 to Value.Count - 1 do
        AddPoint(PPoint(Value.Items[i])^);
  end;
end;

procedure TabcSpriteMover.SetPointIndex(const Value: Integer);
begin
  if Value <> FPointIndex then
  begin
    FPointIndex := Value;
    if (FSprite <> nil) and not (csLoading in FSprite.ComponentState) then
      if (Value >= 0) and (Value < PointCount) then
      begin
        if not (MoveBy in [asmPoints, asmPointsReversed]) then
          MoveBy := asmPoints;
        if (FSprite <> nil) then
          with FSprite, PPoint(FPoints[Value])^ do
            SetBounds(X, Y, Width, Height);
      end;
  end;
end;

procedure TabcSpriteMover.SetVelocityX(const Value: Integer);
begin
  if Value <> FVelocityX then
  begin
    FVelocityX := Value;
    if (FSprite <> nil) and not (csLoading in FSprite.ComponentState) then
      MoveBy := asmVelocity;
  end;
end;

procedure TabcSpriteMover.SetVelocityY(const Value: Integer);
begin
  if Value <> FVelocityY then
  begin
    FVelocityY := Value;
    if (FSprite <> nil) and not (csLoading in FSprite.ComponentState) then
      MoveBy := asmVelocity;
  end;
end;

procedure TabcSpriteMover.ReadPoints(Reader: TReader);
var
  Pt: TPoint;
begin
  with Reader do
  begin
    ReadListBegin;
    while not EndOfList do
    begin
      Pt.X := ReadInteger;
      Pt.Y := ReadInteger;
      AddPoint(Pt);
    end;
    ReadListEnd;
  end;
end;

procedure TabcSpriteMover.WritePoints(Writer: TWriter);
var
  i: Integer;
begin
  with Writer do
  begin
    WriteListBegin;
    with FPoints do
    begin
      for i := 0 to Count - 1 do
        with PPoint(Items[i])^ do
        begin
          WriteInteger(X);
          WriteInteger(Y);
        end;
    end;
    WriteListEnd;
  end;
end;

{
---------------------------- TabcSprite ----------------------------------------
}

constructor TabcSprite.Create(AOwner: TComponent);
begin
  inherited;
  FVisible := True;
  FMover := TabcSpriteMover.Create(Self);
end;

destructor TabcSprite.Destroy;
begin
  if FSpriteList <> nil then
    FSpriteList.Remove(Self);
  FMover.Free;
  inherited;
end;

procedure TabcSprite.Animate(Frame: Integer; var oldBounds, newBounds: TRect);
begin
  FAnimating := True;
  CopyRect(oldBounds, BoundsRect);
  if (DropFrames > 1) and (Frame mod DropFrames <> 1) then Exit;
  Mover.MoveSprite;
  if Assigned(FOnAnimate) then
    FOnAnimate(Self, Frame);
  CopyRect(newBounds, BoundsRect);
  if not EqualRect(oldBounds, newBounds) then
  begin
    DrawInNextFrame := True;
    CheckCollision;
  end;
  FAnimating := False;
end;

procedure TabcSprite.EndFrame;
begin
  if Mover.MovedToEdge then
  begin
    if Assigned(FOnMovedToEdge) then
      FOnMovedToEdge(Self);
    Mover.MovedToEdge := False;
  end;
end;

procedure TabcSprite.AnimateBegin;
begin
  if Assigned(FOnAnimateBegin) then
    FOnAnimateBegin(self);
end;

procedure TabcSprite.AnimateEnd;
begin
  if Assigned(FOnAnimateEnd) then
    FOnAnimateEnd(self);
end;

procedure TabcSprite.Change(Sender: TObject);
begin
  if Visible then
    DrawInNextFrame := True;
  if [csLoading, csDestroying] * ComponentState = [] then
  begin
    if FSpriteList <> nil then
      FSpriteList.Change;
    if Assigned(FOnChange) then
      FOnChange(Self);
  end;
end;

procedure TabcSprite.CheckCollision;
begin
  if SpriteList <> nil then
    with SpriteList do
     if ((aseCollisions in Events) and Assigned(FOnCollision) and
         (ZOrderLimitCollisions >= Self.ZOrder)) and Self.Visible then
       Self.FCheckCollision := True;
end;

procedure TabcSprite.Hide;
begin
  if [csLoading, csDestroying] * ComponentState = [] then
    if FSpriteList <> nil then
      FSpriteList.HideSprite(Self);
end;

procedure TabcSprite.Click(Pos: TPoint);
begin
  if Assigned(FOnClick) then
    FOnClick(Self);
end;

procedure TabcSprite.DblClick(Pos: TPoint);
begin
  if Assigned(FOnDblClick) then
    FOnDblClick(Self);
end;

procedure TabcSprite.MouseDown(Button: TMouseButton; Shift: TShiftState;
  Pos: TPoint);
var
  ClientPos: TPoint;
begin
  if Assigned(FOnMouseDown) then
  begin
    ClientPos.X := Pos.X - Left;
    ClientPos.Y := Pos.Y - Top;
    FOnMouseDown(Self, Button, Shift, ClientPos.X, ClientPos.Y);
  end;
end;

procedure TabcSprite.MouseEnter;
begin
  if CursorStored then
    if (SpriteList <> nil) and (SpriteList.Surface <> nil) then
    begin
      FSaveCursor := SpriteList.Surface.Cursor;
      SpriteList.Surface.Cursor := FCursor;
    end;
  if Assigned(FOnMouseEnter) then FOnMouseEnter(Self);
end;

procedure TabcSprite.MouseLeave;
begin
  if CursorStored then
    if (SpriteList <> nil) and (SpriteList.Surface <> nil) then
      SpriteList.Surface.Cursor := FSaveCursor;
  if Assigned(FOnMouseLeave) then FOnMouseLeave(Self);
end;

procedure TabcSprite.MouseMove(Shift: TShiftState; Pos: TPoint);
var
  ClientPos: TPoint;
begin
  if Assigned(FOnMouseMove) then
  begin
    ClientPos.X := Pos.X - Left;
    ClientPos.Y := Pos.Y - Top;
    FOnMouseMove(Self, Shift, ClientPos.X, ClientPos.Y);
  end;
end;

procedure TabcSprite.MouseUp(Button: TMouseButton; Shift: TShiftState;
  Pos: TPoint);
var
  ClientPos: TPoint;
begin
  if Assigned(FOnMouseUp) then
  begin
    ClientPos.X := Pos.X - Left;
    ClientPos.Y := Pos.Y - Top;
    FOnMouseUp(Self, Button, Shift, ClientPos.X, ClientPos.Y);
  end;
end;

procedure TabcSprite.DefineProperties(Filer: TFiler);
begin
  { The call to inherited DefinedProperties is omitted since the Left and
    Top special properties are redefined with real properties }
end;

function TabcSprite.GetParentComponent: TComponent;
begin
  if SpriteList <> nil then
    Result := SpriteList
  else
    Result := inherited GetParentComponent;
end;

function TabcSprite.HasParent: Boolean;
begin
  if SpriteList <> nil then
    Result := True
  else
    Result := inherited HasParent;
end;

function TabcSprite.PtInSprite(const APoint: TPoint): Boolean;
begin
  Result := PtInRect(BoundsRect, APoint);
end;

function TabcSprite.RectInSprite(const ARect: TRect): Boolean;
var
  R: TRect;
begin
  Result := IntersectRect(R, ARect, BoundsRect);
end;

function TabcSprite.MoveToRect(const ARect: TRect): Boolean;
var
  R: TRect;
  NewPos: TPoint;
begin
  UnionRect(R, ARect, BoundsRect);
  Result :=
    not EqualRect(R, ARect) and
    (Width <= (ARect.Right - ARect.Left)) and
    (Height <= (ARect.Bottom - ARect.Top));
  if Result then
  begin
    if Left < ARect.Left then
      NewPos.X := ARect.Left
    else if (Left + Width) > ARect.Right then
      NewPos.X := ARect.Right - Width
    else
      NewPos.X := Left;
    if Top < ARect.Top then
      NewPos.Y := ARect.Top
    else if (Top + Height) > ARect.Bottom then
      NewPos.Y := ARect.Bottom - Height
    else
      NewPos.Y := Top;
    Position := NewPos;
  end;
end;

function TabcSprite.MoveToSurface: Boolean;
begin
  if (SpriteList <> nil) and (SpriteList.Surface <> nil) then
    Result := MoveToRect(SpriteList.Surface.ClientRect)
  else
    Result := False;
end;

procedure TabcSprite.SetParentComponent(AParent: TComponent);
begin
  if not (csLoading in ComponentState) and (AParent is TabcSpriteList) then
    TabcSpriteList(AParent).Add(Self);
end;

procedure TabcSprite.ReadState(Reader: TReader);
begin
  inherited ReadState(Reader);
  if Reader.Parent is TabcSpriteList then
    TabcSpriteList(Reader.Parent).Add(Self);
end;

procedure TabcSprite.Draw(ACanvas: TCanvas);
begin
end;

function TabcSprite.GetActive: Boolean;
begin
  Result := (SpriteList <> nil) and
            (SpriteList.Surface <> nil) and
            SpriteList.Surface.Active;
end;

function TabcSprite.GetPosition: TPoint;
begin
  Result.x := Left;
  Result.y := Top;
end;

procedure TabcSprite.SetBounds(ALeft, ATop, AWidth, AHeight: Integer);
begin
  if Visible and not FAnimating then
    EraseBounds;
  FLeft := ALeft;
  FTop := ATop;
  FWidth := AWidth;
  FHeight := AHeight;
  with FBoundsRect do
  begin
    Left := ALeft;
    Top := ATop;
    Right := ALeft + AWidth;
    Bottom := ATop + AHeight;
  end;
  CopyRect(FCollisionRect, FBoundsRect);
  InflateRect(FCollisionRect, - AWidth div 8,  - AHeight div 8);
  if not FAnimating then
  begin
    CheckCollision;
    Change(Self);
  end;
end;

procedure TabcSprite.EraseBounds;
begin
  if not FNeverDraw and (SpriteList <> nil) then
    SpriteList.EraseRect(BoundsRect);
end;

procedure TabcSprite.SetDrawInNextFrame(Value: Boolean);
begin
  FDrawInNextFrame := Value;
end;

procedure TabcSprite.SetBoundsRect(const Rect: TRect);
begin
  with Rect do
    SetBounds(Left, Top, Right-Left, Bottom-Top);
end;

procedure TabcSprite.SetPosition(Point: TPoint);
begin
  with Point do
    SetBounds(X, Y, Width, Height);
end;

procedure TabcSprite.SetCollisionRect(const Rect: TRect);
begin
  CopyRect(FCollisionRect, Rect);
end;

procedure TabcSprite.SetHeight(Value: Integer);
begin
  if Value <> FHeight then
    SetBounds(Left, Top, Width, Value);
end;

procedure TabcSprite.SetLeft(Value: Integer);
begin
  if Value <> FLeft then
    SetBounds(Value, Top, Width, Height);
end;

procedure TabcSprite.SetMover(Value: TabcSpriteMover);
begin
  FMover.Assign(Value);
end;

procedure TabcSprite.SetSpriteList(Value: TabcSpriteList);
begin
  if Value <> FSpriteList then
  begin
    if FSpriteList <> nil then
      FSpriteList.Remove(Self);
    FSpriteList := Value;
    if Value <> nil then
      Value.Add(Self);
  end;
end;

procedure TabcSprite.SetTop(Value: Integer);
begin
  if Value <> FTop then
    SetBounds(Left, Value, Width, Height);
end;

procedure TabcSprite.SetVisible(Value: Boolean);
begin
  if Value <> FVisible then
  begin
    FVisible := Value;
    if not Value and (FSpriteList <> nil) then
      FSpriteList.HideSprite(Self);
    if Value then
      CheckCollision;
    Change(Self);
  end;
end;

procedure TabcSprite.SetWidth(Value: Integer);
begin
  if Value <> FWidth then
    SetBounds(Left, Top, Value, Height);
end;

procedure TabcSprite.SetZOrder(Value: TabcZOrder);
begin
  if Value <> FZOrder then
  begin
    FZOrder := Value;
    if FSpriteList <> nil then
    begin
      FSpriteList.Sorted := False;
      CheckCollision;
    end;
    Change(Self);
  end;
end;

function TabcSprite.GetCursor: TCursor;
begin
  if CursorStored then
    Result := FCursor
  else if (SpriteList <> nil) and (SpriteList.Surface <> nil) then
    Result := SpriteList.Surface.Cursor
  else
    Result := crDefault;
end;

function TabcSprite.IsCursorStored: Boolean;
begin
  Result := CursorStored;
end;

procedure TabcSprite.SetCursor(Value: TCursor);
begin
  if Value <> FCursor then
  begin
    CursorStored := True;
    FCursor := Value;
  end;
end;

procedure TabcSprite.SetCursorStored(Value: Boolean);
begin
  FCursorStored := Value;
  if not Value then
    FCursor := GetCursor;
end;

{
---------------------------- TabcSpriteList ------------------------------------
}

function SpriteListZOrderSortCompare(Item1, Item2: Pointer): Integer;
begin
  Result := TabcSprite(Item1).ZOrder - TabcSprite(Item2).ZOrder;
end;

constructor TabcSpriteList.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FSprites := TThreadList.Create;
  FRemovedRects := TabcRectList.Create;
  FEvents := [aseClick, aseDblClick, aseUpDown, aseEnterLeave, aseMove, aseInvisible, aseCollisions];
  FZOrderLimitCollisions := Maxint;
  FZOrderLimitMouse := Maxint;
  FTileBack := TabcTileBack.Create;
  FTileBack.OnChange := TileBackChanged;
end;

destructor TabcSpriteList.Destroy;
begin
  if FSurface <> nil then
    FSurface.Sprites := nil;
  while Count > 0 do
    Sprites[Count - 1].Free;
  FSprites.Free;
  FRemovedRects.Free;
  FTileBack.Free;
  inherited Destroy;
end;

procedure TabcSpriteList.Loaded;
begin
  inherited;
  Change;
end;

procedure TabcSpriteList.Add(Sprite: TabcSprite);
begin
  FSprites.Add(Sprite);
  Sprite.FSpriteList := Self;
  FSorted := False;
  if Assigned(OnAdd) then
    OnAdd(Self, Sprite);
  Change;
end;

procedure TabcSpriteList.AnimateBegin;
var
  i: integer;
begin
  if Assigned(OnAnimateBegin) then
    OnAnimateBegin(Self);
  Sorted := True;
  LockSprites;
  try
    with FList do
      for i := Count - 1 downto 0 do
        TabcSprite(Items[i]).AnimateBegin;
  finally
    UnlockandRemoveSprites;
  end;
end;

function TabcSpriteList.LockSprites: TList;
begin
  Inc(FListLockCount);
  if FListLockCount = 1 then
    FList := FSprites.LockList;
  Result := FList;
end;

procedure TabcSpriteList.UnlockSprites;
begin
  if FListLockCount > 0 then
  begin
    Dec(FListLockCount);
    if FListLockCount = 0 then
    begin
      FSprites.UnlockList;
      FList := nil;
    end;
  end;
end;

procedure TabcSpriteList.UnlockandRemoveSprites;
begin
  try
    InternalRemoveSprites;
  finally
    UnlockSprites;
  end;
end;

procedure TabcSpriteList.AnimateEnd;
var
  i: integer;
begin
  if (csDestroying in ComponentState) or
     (Surface = nil) or
     (csDestroying in Surface.ComponentState) or
     Application.Terminated then Exit;

  Sorted := True;
  LockSprites;
  try
    with FList do
      for i := Count - 1 downto 0 do
        TabcSprite(Items[i]).AnimateEnd;
  finally
    UnlockandRemoveSprites;
  end;

  if Assigned(OnAnimateEnd) then
    OnAnimateEnd(Self);
end;

procedure TabcSpriteList.EndFrame;
var
  i: integer;
begin
  if (csDestroying in ComponentState) or
     (Surface = nil) or
     (csDestroying in Surface.ComponentState) or
     Application.Terminated then Exit;

  Sorted := True;
  LockSprites;
  try
    with FList do
      for i := Count - 1 downto 0 do
        TabcSprite(Items[i]).EndFrame;
  finally
    UnlockandRemoveSprites;
  end;

  if Assigned(OnFrameEnd) then
    OnFrameEnd(Self);
end;

procedure TabcSpriteList.AddIntersectingRects(const SurfaceBounds: TRect; const DirtyRects: TabcRectList);
//need to draw any sprites that intersect with dirty rects
var
  i, j, lastRect, saveCount: integer;
  rTest, rTemp: TRect;
begin
  lastRect := 0;
  repeat
    saveCount := DirtyRects.Count;
    for i := DirtyRects.Count - 1 downto lastRect do
    begin
      rTest := PRect(DirtyRects[i])^;
      with FList do
        for j := Count - 1 downto 0 do
          with TabcSprite(Items[j]) do
            if Visible and not DrawInNextFrame and not FNeverDraw and
               IntersectRect(rTemp, surfaceBounds, BoundsRect) then
              if IntersectRect(rTemp, rTest, BoundsRect) then
              begin
                DrawInNextFrame := True;
                DirtyRects.AddRect(BoundsRect);
              end;
    end;
    lastRect := saveCount - 1;
  until DirtyRects.Count = saveCount;
end;

procedure TabcSpriteList.AnimateSprites(Frame: Integer; const DirtyRects: TabcRectList);
var
  i: integer;
  oldBounds, newBounds, surfaceBounds: TRect;
begin
  Sorted := True;
  LockSprites;
  try
    if Assigned(FOnFrameBegin) then
      FOnFrameBegin(Self, FList, Frame, DirtyRects);

    Sorted := True;

    with FRemovedRects do
    begin
      for i := 0 to Count - 1 do
        DirtyRects.AddRect(TRect(Items[i]^));
      Clear;
    end;

    with FList do
    begin
      CopyRect(surfaceBounds, Surface.ClientRect);
      //animate each sprite
      for i := Count - 1 downto 0 do
        with TabcSprite(Items[i]) do
        begin
          if not Visible then Continue;
          SetRectEmpty(oldBounds);
          SetRectEmpty(newBounds);
          Animate(Frame, oldBounds, newBounds);
          if DrawInNextFrame then
          begin
            IntersectRect(oldBounds, surfaceBounds, oldBounds);
            IntersectRect(newBounds, surfaceBounds, newBounds);
            if not IsRectEmpty(oldBounds) then
              DirtyRects.AddRect(oldBounds);
            if IsRectEmpty(newBounds) then
              DrawInNextFrame := False
            else if not EqualRect(oldBounds, newBounds) then
              DirtyRects.AddRect(newBounds);
          end;
        end;

      if (Frame <> 1) and (DirtyRects.Count > 0) then
        AddIntersectingRects(surfaceBounds, DirtyRects);
    end;
  finally
    UnlockandRemoveSprites;
  end;
end;

procedure TabcSpriteList.DetectCollisions;
var
  i,j: integer;
  Sprite1, Sprite2: TabcSprite;
  R: TRect;
begin
  if ((ZOrderLimitCollisions >= 0) and (aseCollisions in Events) and Assigned(FOnCollision)) then
  begin
    Sorted := True;
    LockSprites;
    try
      with FList do
        for i := 0 to Count - 1 do
        begin
          Sprite1 := TabcSprite(Items[i]);
          if Sprite1.ZOrder > ZOrderLimitCollisions then
            Break;
          if not Sprite1.FCheckCollision then
            Continue
          else
          begin
            for j := 0 to Count - 1 do
              if j = i then
                Continue
              else
              begin
                Sprite2 := TabcSprite(Items[j]);
                if Sprite2.ZOrder > Sprite1.ZOrder then
                  Break
                else
                  if (Sprite2.ZOrder = Sprite1.ZOrder) and
                    IntersectRect(R,Sprite1.CollisionRect, Sprite2.CollisionRect) then
                  begin
                    FOnCollision(Self, Sprite1, Sprite2);
                    if Sprite2.FCheckCollision then
                      Sprite2.FCheckCollision := False;
                  end;
              end;
            Sprite1.FCheckCollision := False;
          end;
        end;
    finally
      UnlockandRemoveSprites;
    end;
  end;
end;

procedure TabcSpriteList.DrawSprites(const ACanvas: TCanvas; DrawAll: Boolean; const DirtyRects: TabcRectList);
var
  i: integer;
  dirtyRect: TRect;
begin
  if ACanvas.TryLock then
  try
    if Assigned(OnDrawBegin) then
    begin
      dirtyRect := Rect(0,0,0,0);
      OnDrawBegin(Self, ACanvas, dirtyRect);
      if DirtyRects <> nil then
        DirtyRects.AddRectNotEmpty(dirtyRect);
    end;
    Sorted := True;
    LockSprites;
    try
      with FList do
        for i := Count - 1 downto 0 do
          with TabcSprite(Items[i]) do
            if Visible and (DrawAll or DrawInNextFrame) then
            begin
              Draw(ACanvas);
              DrawInNextFrame := False;
            end;
    finally
      UnlockSprites;
    end;
    if Assigned(OnDrawEnd) then
    begin
      dirtyRect := Rect(0,0,0,0);
      OnDrawEnd(Self, ACanvas, dirtyRect);
      if DirtyRects <> nil then
        DirtyRects.AddRectNotEmpty(dirtyRect);
    end;
  finally
    ACanvas.Unlock;
  end;
end;

procedure TabcSpriteList.BeginUpdate;
begin
  Inc(FUpdateCount);
end;

procedure TabcSpriteList.EndUpdate;
begin
  if FUpdateCount > 0 then Dec(FUpdateCount);
  if FChanged then
  begin
    FChanged := False;
    Change;
  end;
end;

procedure TabcSpriteList.Change;
begin
  if [csLoading, csDestroying] * ComponentState = [] then
  begin
    FChanged := True;
    if FUpdateCount > 0 then Exit;
    if Assigned(FOnChange) then
      FOnChange(Self);
  end;
end;

procedure TabcSpriteList.TileBackChanged(Sender: TObject);
begin
  if not (csLoading in ComponentState) and (FSurface <> nil) then
    FSurface.RefreshBackground;
end;

procedure TabcSpriteList.SetBackground(Value: TabcTileBack);
begin
  FTileBack.Assign(Value);
end;

procedure TabcSpriteList.Clear;
var
  i: integer;
begin
  if (Surface <> nil) and Surface.Active then
  begin
    LockSprites;
    try
      with FList do
        for i := Count - 1 downto 0 do
          TabcSprite(Items[i]).FListRemove := True;
    finally
      UnlockSprites;
    end;
  end
  else
  begin
    BeginUpdate;
    LockSprites;
    try
      InternalRemoveSprites;
      with FList do
        for i := Count - 1 downto 0 do
        begin
          TabcSprite(Items[i]).FSpriteList := nil;
          Delete(i);
        end;
    finally
      UnlockSprites;
    end;
    FSorted := True;
    if Surface <> nil then
      FRemovedRects.AddRect(Surface.ClientRect);
    EndUpdate;
  end;
end;

procedure TabcSpriteList.ClearAndFree;
var
  i: integer;
  Sprite: TabcSprite;
begin
  if (Surface <> nil) and Surface.Active then
  begin
    LockSprites;
    try
      with FList do
        for i := Count - 1 downto 0 do
          with TabcSprite(Items[i]) do
          begin
            FListRemove := True;
            FListFree := True;
            Destroying;
        end;
    finally
      UnlockSprites;
    end;
  end
  else
  begin
    BeginUpdate;
    LockSprites;
    try
      InternalRemoveSprites;
      with FList do
        for i := Count - 1 downto 0 do
        begin
          Sprite := TabcSprite(Items[i]);
          Sprite.FSpriteList := nil;
          Delete(i);
          Sprite.Free;
        end;
    finally
      UnlockSprites;
    end;
    FSorted := True;
    if Surface <> nil then
      FRemovedRects.AddRect(Surface.ClientRect);
    EndUpdate;
  end;
end;

procedure TabcSpriteList.SpriteClick(Pos: TPoint);
var
  Sprite: TabcSprite;
begin
  if (not (csDestroying in ComponentState) and (ZOrderLimitMouse >= 0) and (aseClick in Events)) then
  begin
    Sprite := GetSpriteAtPos(Pos, (aseInvisible in Events), ZOrderLimitMouse);
    if (Sprite <> nil) and (Sprite = FLastSpriteLeftDown) then
      Sprite.Click(Pos);
  end;
end;

procedure TabcSpriteList.SpriteDblClick(Pos: TPoint);
var
  Sprite: TabcSprite;
begin
  if (not (csDestroying in ComponentState) and (ZOrderLimitMouse >= 0) and (aseDblClick in Events)) then
  begin
    Sprite := GetSpriteAtPos(Pos, (aseInvisible in Events), ZOrderLimitMouse);
    if Sprite <> nil then
      Sprite.DblClick(Pos);
  end;
end;

procedure TabcSpriteList.SpriteMouseMove(Shift: TShiftState; Pos: TPoint);
var
  Sprite: TabcSprite;
begin
  if (not (csDestroying in ComponentState) and (ZOrderLimitMouse >= 0) and
    (Events * [aseEnterLeave, aseMove] <> []) ) then
  begin
    Sprite := GetSpriteAtPos(Pos, (aseInvisible in Events), ZOrderLimitMouse);
    if (Sprite <> nil) and (aseMove in Events) then
      Sprite.MouseMove(Shift, Pos);
    if MouseEntered then
      SpriteEnterLeave(Sprite);
  end;
end;

procedure TabcSpriteList.SpriteEnterLeave(SpriteMoved: TabcSprite);
var
  SaveLastSpriteEntered: TabcSprite;
begin
  if SpriteMoved <> FLastSpriteEntered then
  begin
    if (FLastSpriteEntered <> nil) then
    begin
      SaveLastSpriteEntered := FLastSpriteEntered;
      FLastSpriteEntered := nil;
      if (aseEnterLeave in Events) and
         ((aseInvisible in Events) or (SaveLastSpriteEntered.Visible))  then
        SaveLastSpriteEntered.MouseLeave;
    end;
    if SpriteMoved <> nil then
    begin
      if (aseEnterLeave in Events) then
        SpriteMoved.MouseEnter;
      FLastSpriteEntered := SpriteMoved;
    end;
  end;
end;

procedure TabcSpriteList.SpriteMouseDown(Button: TMouseButton; Shift: TShiftState; Pos: TPoint);
var
  Sprite: TabcSprite;
begin
  if (not (csDestroying in ComponentState) and (ZOrderLimitMouse >= 0) and (aseUpDown in Events)) then
  begin
    Sprite := GetSpriteAtPos(Pos, (aseInvisible in Events), ZOrderLimitMouse);
    if (Button = mbLeft) then
      FLastSpriteLeftDown := Sprite;
    if Sprite <> nil then
      Sprite.MouseDown(Button, Shift, Pos);
  end;
end;

procedure TabcSpriteList.SpriteMouseUp(Button: TMouseButton; Shift: TShiftState; Pos: TPoint);
var
  Sprite: TabcSprite;
begin
  if (not (csDestroying in ComponentState) and (ZOrderLimitMouse >= 0) and (aseUpDown in Events)) then
    if (Button = mbLeft) and
      (FLastSpriteLeftDown <> nil) then
    begin
      FLastSpriteLeftDown.MouseUp(Button, Shift, Pos);
      FLastSpriteLeftDown := nil;
    end
    else
    begin
      Sprite := GetSpriteAtPos(Pos, (aseInvisible in Events), ZOrderLimitMouse);
      if Sprite <> nil then
        Sprite.MouseUp(Button, Shift, Pos);
    end;
end;

procedure TabcSpriteList.GetChildren(Proc: TGetChildProc; Root: TComponent);
var
  i: Integer;
  Sprite: TabcSprite;
begin
  Sorted := True;
  LockSprites;
  try
    InternalRemoveSprites;
    with FList do
      for i := 0 to Count - 1 do
      begin
        Sprite := TabcSprite(Items[i]);
        if Sprite.Owner = Root then Proc(Sprite);
      end;
  finally
    UnlockSprites;
  end;
end;

procedure TabcSpriteList.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited;
  if (Operation = opRemove) and (AComponent is TabcSprite) then
    Remove(TabcSprite(AComponent));
end;

function TabcSpriteList.GetSpriteAtPos(APoint: TPoint; IncludeInvisible: Boolean = False; ZOrderLimit: TabcZOrder = Maxint): TabcSprite;
var
  i: integer;
begin
  Result := nil;
  Sorted := True;
  LockSprites;
  try
    with FList do
      for i := 0 to Count - 1 do
        with TabcSprite(Items[i]) do
        begin
          if ZOrder > ZOrderLimit then
            Exit;
          if (Visible or IncludeInvisible) and PtInRect(BoundsRect, APoint) then
            if PtInSprite(APoint) then
            begin
              Result := TabcSprite(Items[i]);
              Exit;
            end;
        end;
  finally
    UnlockSprites;
  end;
end;

procedure TabcSpriteList.Remove(Sprite: TabcSprite);
begin
  if (FListLockCount = 0) then
  begin
    LockSprites;
    try
      InternalRemove(Sprite);
    finally
      UnlockSprites;
    end;
  end
  else
    Sprite.FListRemove := True;
end;

procedure TabcSpriteList.RemoveAndFree(Sprite: TabcSprite);
begin
  if (FListLockCount = 0) then
  begin
    Remove(Sprite);
    Sprite.Free;
  end
  else with Sprite do
  begin
    FListRemove := True;
    FListFree := True;
    Destroying;
  end;
end;

procedure TabcSpriteList.InternalRemove(Sprite: TabcSprite);
begin
  if (FList.Remove(Sprite) >= 0) then
  begin
    with Sprite do
    begin
      FSpriteList := nil;
      if not (csDestroying in Self.ComponentState) then
        if Visible then
          HideSprite(Sprite);
    end;
    if FLastSpriteEntered = Sprite then
      FLastSpriteEntered := nil;
    if Assigned(OnRemove) then
      OnRemove(Self, Sprite);
    if not (csDestroying in Self.ComponentState) then
      Change;
  end;
end;

procedure TabcSpriteList.InternalRemoveSprites;
var
  i: integer;
  Sprite: TabcSprite;
begin
  with FList do
    for i := Count - 1 downto 0 do
      if TabcSprite(Items[i]).FListRemove then
      begin
        Sprite := TabcSprite(Items[i]);
        InternalRemove(Sprite);
        Sprite.FListRemove := False;
        if Sprite.FListFree then
          Sprite.Free;
      end;
end;

procedure TabcSpriteList.HideSprite(Sprite: TabcSprite);
begin
  if FLastSpriteEntered = Sprite then
    SpriteEnterLeave(nil);
  EraseRect(Sprite.BoundsRect);
end;

procedure TabcSpriteList.EraseRect(const R: TRect);
begin
  if (FSurface <> nil) then
    if FSurface.Active then
      FRemovedRects.AddRect(R)
    else
      Change;
end;

procedure TabcSpriteList.SortSprites;
begin
  LockSprites;
  with FList do
  try
    Sort(SpriteListZOrderSortCompare);
    FSorted := True;
  finally
    UnlockSprites;
  end;
end;

function TabcSpriteList.GetCount: Integer;
begin
  LockSprites;
  try
    Result := FList.Count;
  finally
    UnlockSprites;
  end;
end;

procedure TabcSpriteList.SetSorted(Value: Boolean);
begin
  if Value <> FSorted then
  begin
    FSorted := Value;
    if Value then SortSprites;
  end;
end;

function TabcSpriteList.GetSprites(Index: Integer): TabcSprite;
begin
  LockSprites;
  try
    Result := TabcSprite(FList.Items[Index]);
  finally
    UnlockSprites;
  end;
end;

procedure TabcSpriteList.SetMouseEntered(Value: Boolean);
begin
  if Value <> FMouseEntered then
  begin
    FMouseEntered := Value;
    SpriteEnterLeave(nil);
  end;
end;

{
---------------------------- TabcSpriteSurface ---------------------------------
}

constructor TabcSpriteSurface.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  ControlStyle := ControlStyle + [csOpaque];
  FTileBack := TabcTileBack.Create;
  FTileBack.OnChange := TileBackChanged;
  FScreenBuffer := TBitmap.Create;
  FScreenBuffer.PixelFormat := pfDevice;
  FBackgroundBitmap := TBitmap.Create;
  FBackgroundBitmap.PixelFormat := pfDevice;
  FEraseFrames := True;
  FFramesPerSec := 30;
  FFrameDuration := 33;
  FNeverPainted := True;
  Width := 200;
  Height := 200;
  abcApplicationEvents.RegisterOnActivate(ApplicationActivated);
  abcApplicationEvents.RegisterOnDeactivate(ApplicationDeactivated);
  FSuspendOnAppInactive := True;
  FPerfFont := TFont.Create;
  with FPerfFont do
  begin
    Name := 'Arial';
    Size := 14;
    Color := clBlue;
  end;
end;

destructor TabcSpriteSurface.Destroy;
begin
  Active := False;
  abcApplicationEvents.UnRegisterOnActivate(ApplicationActivated);
  abcApplicationEvents.UnRegisterOnDeactivate(ApplicationDeactivated);
  Sprites := nil;
  FScreenBuffer.Free;
  FBackgroundBitmap.Free;
  FTileBack.Free;
  FPerfFont.Free;
  inherited Destroy;
end;

procedure TabcSpriteSurface.ApplicationActivated(Sender: TObject);
begin
  if SuspendOnAppInactive and Active and not FWasSuspended then
    Resume;
end;

procedure TabcSpriteSurface.ApplicationDeactivated(Sender: TObject);
begin
  if SuspendOnAppInactive and Active and (FSpriteThread <> nil) then
  begin
    FWasSuspended := FSpriteThread.Suspended;
    Suspend;
  end;
end;

procedure TabcSpriteSurface.Click;
var
  P: TPoint;
begin
  inherited;
  GetCursorPos(P);
  P := ScreenToClient(P);
  if Sprites <> nil then
    Sprites.SpriteClick(P);
end;

procedure TabcSpriteSurface.DblClick;
var
  P: TPoint;
begin
  inherited;
  GetCursorPos(P);
  P := ScreenToClient(P);
  if Sprites <> nil then
    Sprites.SpriteDblClick(P);
end;

function TabcSpriteSurface.GetAvgFramesPerSec: Integer;
begin
  if FActiveSeconds > 0 then
    Result := FActiveSecsFrameCount div FActiveSeconds
  else
    Result := 0;
end;

procedure TabcSpriteSurface.MouseDown(Button: TMouseButton; Shift: TShiftState;
  X, Y: Integer);
begin
  inherited;
  if Sprites <> nil then
    Sprites.SpriteMouseDown(Button, Shift, Point(X,Y));
end;

procedure TabcSpriteSurface.MouseMove(Shift: TShiftState; X, Y: Integer);
begin
  inherited;
  if not FMouseOver then
    UpdateMouseOver;
  if Sprites <> nil then
    Sprites.SpriteMouseMove(Shift, Point(X,Y));
end;

procedure TabcSpriteSurface.MouseUp(Button: TMouseButton; Shift: TShiftState;
  X, Y: Integer);
begin
  inherited;
  if Sprites <> nil then
    Sprites.SpriteMouseUp(Button, Shift, Point(X,Y));
  if MouseIsOver then
    Perform(CM_MOUSEENTER,0,0)
  else
    Perform(CM_MOUSELEAVE,0,0);
end;

function TabcSpriteSurface.GetCanvas: TCanvas;
begin
  Result := FScreenBuffer.Canvas;
end;

function TabcSpriteSurface.GetInheritedCanvas: TCanvas;
begin
  Result := inherited Canvas;
end;

procedure TabcSpriteSurface.Loaded;
begin
  inherited Loaded;
  if not Transparent then
    if FNeverPainted then
    begin
      FNeverPainted := False;
      RefreshBackground;
      if Active then
      begin
        FCreateThreadOnPaint := True;
        CanCreateThread;
      end;
    end;
end;

procedure TabcSpriteSurface.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
end;

procedure TabcSpriteSurface.CMColorChanged(var Message: TMessage);
begin
  if not (Transparent or (csLoading in ComponentState)) then
    if not Active or (csDesigning in ComponentState) then
      DoInactiveDrawing
    else
    begin
      DrawBackground;
      Repaint;
    end;
end;

procedure TabcSpriteSurface.CMParentColorChanged(var Message: TMessage);
begin
  inherited;
  if Transparent and not (FNeverPainted or (csLoading in ComponentState)) then
    RefreshBackground;
end;

procedure TabcSpriteSurface.WMPaint(var Message: TWMPaint);
begin
  if FNoPainting then Exit;
  if FNeverPainted then
  begin
    FNeverPainted := False;
    RefreshBackground;
    if Active then
      CanCreateThread;
    Exit;
  end
  else if FCreateThreadOnPaint then
  begin
    FCreateThreadOnPaint := False;
    CanCreateThread;
  end;
  inherited;
end;

procedure TabcSpriteSurface.Paint;
begin
  inherited Canvas.Draw(0,0, FScreenBuffer);
  if (csDesigning in ComponentState) then
    with inherited Canvas do
    begin
      with Pen do
      begin
        Color := clBlack;
        Mode := pmCopy;
        Style := psDash;
        Width := 1;
      end;
      Brush.Style := bsClear;
      Rectangle(0, 0, Width, Height);
    end;
end;

procedure TabcSpriteSurface.DoInactiveDrawing;
begin
  if FNeverPainted then Exit;
  if [csLoading, csDestroying] * ComponentState = [] then
    if (csDesigning in ComponentState) or
       (not Active and Visible and DrawWhenInactive) then
        DrawSprites
      else
        ClearSprites;
end;

procedure TabcSpriteSurface.ClearSprites;
begin
  if Active and not (csDesigning in ComponentState) then Exit;
  DrawBackground;
  Repaint;
end;

procedure TabcSpriteSurface.DrawBackground;
begin
  BitBlt(FScreenBuffer.Canvas.Handle, 0, 0, Width, Height,
    FBackgroundBitmap.Canvas.Handle,0,0,SRCCOPY);
end;

function TabcSpriteSurface.DrawPerformanceText: TRect;
begin
  with FScreenBuffer.Canvas do
  begin
    Font := FPerfFont;
    Brush.Style := bsClear;
    TextRect(FPerfTextR, FPerfTextR.Left, FPerfTextR.Top, FPerformanceText);
  end;
  Result := FPerfTextR;
end;

procedure TabcSpriteSurface.UpdatePerformanceText;
var
  sizeText: TSize;
begin
  FPerformanceText := 'Avg ' + inttostr(AvgFramesPerSec) + ' Now ' + inttostr(RealFramesPerSec) + ' fps';
  //update size
  if ShowPerformance then
    with FScreenBuffer.Canvas do
    begin
      Font := FPerfFont;
      Brush.Style := bsClear;
      sizeText := TextExtent(FPerformanceText);
      FPrevPerfTextR := FPerfTextR;
      with FPerfTextR do
      begin
        Left := 4;
        Right := Left + sizeText.cx + 10;
        Top := Self.Height - 4 - sizeText.cy;
        Bottom := Top + sizeText.cy;
      end;
    end;
end;

procedure TabcSpriteSurface.DrawSprites;
begin
  if Active and not (csDesigning in ComponentState) then Exit;
  DrawBackground;
  if (Sprites <> nil) then
    Sprites.DrawSprites(FScreenBuffer.Canvas, True, nil);
  Repaint;
end;

procedure TabcSpriteSurface.SetBounds(ALeft, ATop, AWidth, AHeight: Integer);
begin
  inherited SetBounds(ALeft, ATop, AWidth, AHeight);
  with FScreenBuffer do
  begin
    Width := AWidth;
    Height := AHeight;
  end;
  if FNeverPainted then Exit;

  FNoPainting := True;
  try
    Application.ProcessMessages;
  finally
    FNoPainting := False;
  end;

  UpdateBackgroundBitmap;

  if (csDesigning in ComponentState) or not Active then
    DoInactiveDrawing
  else if Active then
    CanResetThread;
end;

procedure TabcSpriteSurface.SpriteListChanged(Sender: TObject);
begin
  if not Active or (csDesigning in ComponentState) then
    DoInactiveDrawing;
end;

procedure TabcSpriteSurface.SetActive(Value: Boolean);
begin
  if Value <> FActive then
  begin
    FActive := Value;
    if [csLoading, csDesigning] * ComponentState = [] then
      if Value then
        CanCreateThread
      else
        CanTerminateThread;
  end;
end;

procedure TabcSpriteSurface.SetDrawWhenInactive(Value: Boolean);
begin
  if Value <> FDrawWhenInactive then
  begin
    FDrawWhenInactive := Value;
    if not (csDesigning in ComponentState) then
      DoInactiveDrawing;
  end;
end;

procedure TabcSpriteSurface.SetSprites(Value: TabcSpriteList);
begin
  if Value <> FSprites then
  begin
    if FSprites <> nil then
    begin
      FSprites.FSurface := nil;
      FSprites.OnChange := nil;
    end;
    if Value = nil then
    begin
      CanTerminateThread;
      FSprites := Value;
    end
    else
    begin
      CanTerminateThread;
      FSprites := Value;
      Value.FSurface := Self;
      Value.OnChange := SpriteListChanged;
      CanCreateThread;
    end;
    if not (csDestroying in ComponentState) then
    begin
      RefreshBackground;
      SpriteListChanged(Self);
    end;
  end;
end;

procedure TabcSpriteSurface.RefreshBackground;
begin
  SetBounds(Left, Top, Width, Height);
end;

procedure TabcSpriteSurface.TileBackChanged(Sender: TObject);
begin
  if not (csLoading in ComponentState) then
    RefreshBackground;
end;

procedure TabcSpriteSurface.SetBackground(Value: TabcTileBack);
begin
  FTileBack.Assign(Value);
end;

procedure TabcSpriteSurface.SetTransparent(Value: Boolean);
begin
  if Value <> FTransparent then
  begin
    FTransparent := Value;
    if not (csLoading in ComponentState) then
      RefreshBackground;
  end;
end;

procedure TabcSpriteSurface.SetFrameDuration(
  const Value: TabcFrameDuration);
begin
  if Value <> FFrameDuration then
  begin
    FFrameDuration := Value;
    FFramesPerSec := 1000 div Value;
    FStoreFrameDuration := True;
  end;
end;

procedure TabcSpriteSurface.SetFramesPerSec(
  const Value: TabcFrameRate);
begin
  if Value <> FFramesPerSec then
  begin
    FFramesPerSec := Value;
    if Value <> 0 then
    begin
      FFrameDuration := 1000 div Value;
      FStoreFrameDuration := False;
    end
    else
    begin
      FFrameDuration := 1001;
      FStoreFrameDuration := True;
    end;
  end;
end;

function TabcSpriteSurface.IsFrameDurationStored: Boolean;
begin
  Result := FStoreFrameDuration;
end;

function TabcSpriteSurface.IsFramesPerSecStored: Boolean;
begin
  Result := not FStoreFrameDuration and (FFramesPerSec <> 30);
end;

procedure TabcSpriteSurface.CMVisibleChanged(var Message: TMessage);
begin
  if not (csDesigning in ComponentState) then
  begin
    if Visible then
    begin
      if Active then
        CanCreateThread
      else
        DoInactiveDrawing;
      if not Transparent then Repaint;
    end
    else
    begin
      CanTerminateThread;
      inherited;
    end;
  end;
end;

procedure TabcSpriteSurface.CanCreateThread;
begin
  if Active and (FSpriteThread = nil) and (Sprites <> nil) and
     Visible and
     (Parent <> nil) and Parent.HandleAllocated and
     not (csDesigning in ComponentState) then
    FSpriteThread := TabcSpriteThread.Create(Self);
end;

procedure TabcSpriteSurface.CanTerminateThread;
begin
  if FSpriteThread <> nil then
    with FSpriteThread do
    begin
      Suspended := False;
      Terminate;
      WaitFor;
      FSpriteThread.Free;
      FSpriteThread := nil;
      if not DrawWhenInactive then
        ClearSprites;
    end;
end;

procedure TabcSpriteSurface.CanResetThread;
begin
  if (ComponentState * [csLoading, csDestroying] = []) then
    if (FSpriteThread <> nil) then
      FSpriteThread.Reset;
end;

procedure TabcSpriteSurface.Start;
begin
  Active := True;
end;

procedure TabcSpriteSurface.Stop;
begin
  Active := False;
end;

procedure TabcSpriteSurface.Suspend;
begin
  if FSpriteThread <> nil then
    FSpriteThread.Suspended := True;
end;

procedure TabcSpriteSurface.Resume;
begin
  if FSpriteThread <> nil then
    FSpriteThread.Suspended := False;
end;

procedure TabcSpriteSurface.CMEnabledChanged(var Message: TMessage);
begin
  if not (csDestroying in ComponentState) then
    UpdateMouseOver;
end;

procedure TabcSpriteSurface.CMMouseEnter(var Message: TMessage);
begin
  inherited;
  if not ForegroundTask then Exit;
  if not FMouseOver and Enabled and (DragMode <> dmAutomatic) and (GetCapture = 0) then
  begin
    FMouseOver := True;
    MouseEnter;
  end;
end;

procedure TabcSpriteSurface.CMMouseLeave(var Message: TMessage);
begin
  inherited;
  FMouseOver := False;
  if Enabled and (DragMode <> dmAutomatic) and (GetCapture = 0) then
    MouseLeave;
end;

procedure TabcSpriteSurface.MouseEnter;
begin
  if Sprites <> nil then
    Sprites.MouseEntered := True;
  if Assigned(FOnMouseEnter) then FOnMouseEnter(Self);
end;

procedure TabcSpriteSurface.MouseLeave;
begin
  if Sprites <> nil then
    Sprites.MouseEntered := False;
  if Assigned(FOnMouseLeave) then FOnMouseLeave(Self);
end;

function TabcSpriteSurface.MouseIsOver: Boolean;
var
  P: TPoint;
begin
  GetCursorPos(P);
  Result := (FindDragTarget(P, True) = Self);
end;

procedure TabcSpriteSurface.UpdateMouseOver;
begin
  if MouseIsOver then
    Perform(CM_MOUSEENTER, 0, 0)
  else
    Perform(CM_MOUSELEAVE, 0, 0);
end;

procedure TabcSpriteSurface.UpdateBackgroundBitmap;
var
  parentBorderOffset: Integer;
begin
  with FBackgroundBitmap do
  try
    Width := Self.Width;
    Height := Self.Height;
    FNoPainting := True;
    Canvas.Lock;

    if self.Transparent then
    begin
      if GetWindowLong(Parent.Handle, GWL_EXSTYLE) and WS_EX_CLIENTEDGE <> 0 then
        parentBorderOffset := 2
      else if GetWindowLong(Parent.Handle, GWL_STYLE) and WS_BORDER <> 0 then
        parentBorderOffset := 1
      else
        parentBorderOffset := 0;
      Parent.PaintTo(Canvas.Handle, -Left-parentBorderOffset, -Top-parentBorderOffset);
    end
    else
      with Canvas do
      begin
        //paint color
        Brush.Style := bsSolid;
        Brush.Color := Self.Color;
        FillRect(Rect(0,0,Width,Height));
      end;
    if (Sprites <> nil) and not Sprites.Background.Empty then
      Sprites.Background.Draw(Canvas, Rect(0,0,Width,Height))
    else if not Background.Empty then
      Background.Draw(Canvas, Rect(0,0,Width,Height));
  finally
    Canvas.Unlock;
    FNoPainting := False;
  end;
end;

{
---------------------------- TabcSpriteThread ----------------------------------
}

constructor TabcSpriteThread.Create(ASurface: TabcSpriteSurface);
begin
  FSurface := ASurface;
  FDirtyRects := TabcRectList.Create;
  timeBeginPeriod(1);
  inherited Create(
    ASurface.SuspendOnAppInactive and not Application.Active);
end;

destructor TabcSpriteThread.Destroy;
begin
  FDirtyRects.Clear;
  FDirtyRects.Free;
  timeEndPeriod(1);
  inherited;
end;

procedure TabcSpriteThread.AnimateBegin;
begin
  if (FSurface <> nil) then
    with FSurface do
    begin
      FActiveSeconds := 0;
      FActiveSecsFrameCount := 0;
      if (Sprites <> nil) then
        Sprites.AnimateBegin;
    end;
end;

procedure TabcSpriteThread.AnimateEnd;
begin
  if (FSurface <> nil) and (FSurface.Sprites <> nil) then
    FSurface.Sprites.AnimateEnd;
end;

procedure TabcSpriteThread.AnimateSprites;
begin
  FDirtyRects.Clear;
  with FSurface do
  begin
    if FDirtyRects.AddRectNotEmpty(FPrevPerfTextR) <> -1 then
      FPrevPerfTextR := Rect(0,0,0,0);
    if FDirtyRects.AddRectNotEmpty(FPerfTextR) <> -1 then
      if not ShowPerformance then
        FPerfTextR := Rect(0,0,0,0);
    Sprites.AnimateSprites(FFrame, FDirtyRects);
  end;
end;

procedure TabcSpriteThread.DetectCollisions;
begin
  FSurface.Sprites.DetectCollisions;
end;

procedure TabcSpriteThread.DrawMemory;
var
  i: integer;
  DestDC, SrcDC: HDC;
begin
  with FSurface do
  begin
    if (Parent = nil) or not Parent.HandleAllocated then
      Exit;

    //draw background or parts of it to screen buffer
    if FFrame = 1 then
      DrawBackground
    else if FEraseFrames then
    begin
      FScreenBuffer.Canvas.Lock;
      FBackgroundBitmap.Canvas.Lock;
      try
      DestDC := FScreenBuffer.Canvas.Handle;
      SrcDC := FBackgroundBitmap.Canvas.Handle;
      with FDirtyRects do
        for i := 0 to Count - 1 do
          with PRect(Items[i])^ do
            BitBlt(
              DestDC, Left, Top, Right-Left,Bottom-Top,
              SrcDC, Left, Top, SRCCOPY);
      finally
        FScreenBuffer.Canvas.Unlock;
        FBackgroundBitmap.Canvas.Unlock;
      end;
    end;
    if Terminated then Exit;
    //draw sprites to buffer
    Sprites.DrawSprites(FScreenBuffer.Canvas, (FFrame = 1), FDirtyRects);
    if Terminated then Exit;
    if ShowPerformance then
      FDirtyRects.AddRect(DrawPerformanceText);
  end;
end;

procedure TabcSpriteThread.DrawScreen;
var
  i: integer;
  DestDC, SrcDC: HDC;
  Dest: TCanvas;
begin
  with FSurface do
  begin
    if (Parent = nil) or not Parent.HandleAllocated then
      Exit;

    //draw dirty rects from buffer to screen

    FDirtyRects.Merge;

    Dest := GetInheritedCanvas;
    Dest.Lock;
    try
      DestDC := Dest.Handle;
      FScreenBuffer.Canvas.Lock;
      try
        SrcDC := FScreenBuffer.Canvas.Handle;
        if (FFrame <> 1) then
          with FDirtyRects do
            for i := 0 to Count - 1 do
              with PRect(Items[i])^ do
                BitBlt(
                  DestDC, Left, Top, Right-Left,Bottom-Top,
                  SrcDC, Left, Top, SRCCOPY)
        else
          BitBlt(
            DestDC, 0, 0, Width, Height,
            SrcDC, 0, 0, SRCCOPY);
      finally
        FScreenBuffer.Canvas.Unlock;
      end;
    finally
      Dest.Unlock;
    end;
  end;
end;

{
Begin Animation:
  SpriteList --> OnAnimateBegin
  for each Sprite --> OnAnimateBegin

while Surface is Active:
  Animate Frame:
    Start Frame:
      Update FrameCount
      Surface --> BeforeFrame
    Animate Sprites:
      SpriteList AnimateSprites:
        SpriteList --> OnFrameBegin
        for each Sprite Move or change image:
          Sprite class internal animation using Mover and other properties
          Sprite --> OnAnimate
      Detect Collisions:
        for each pair of colliding sprites SpriteList --> OnCollision
    Draw Memory:
      BeforeDraw:
        Surface --> BeforeDraw
      SpriteList.DrawSprites:
        Draw each sprite needing to be drawn to Buffer
      AfterDraw:
        Surface --> AfterDraw

    Draw Buffer to Screen

    End Frame:
      Sprites.EndFrame:
        for each Sprite moved to an edge or end point --> OnMovedToEdge
        SpriteList --> OnFrameEnd
      Surface --> AfterFrame
      Update timing

  wait until next frame

End Animation:
  for each Sprite OnAnimateEnd event
  SpriteList OnAnimateEnd event
}

procedure TabcSpriteThread.AnimateFrame;
begin
  StartFrame;
  if Terminated then Exit;
  AnimateSprites;
  if Terminated then Exit;
  DetectCollisions;
  if Terminated then Exit;
  DrawMemory;
  if Terminated then Exit;
  DrawScreen;
  GDIFlush;  //??
  if Terminated then Exit;
  EndFrame;
end;

procedure TabcSpriteThread.StartFrame;
begin
  //init timing
  FStartThisFrame := timeGetTime;
  FFrameDuration := FSurface.FrameDuration;
  FStartNextFrame := FStartThisFrame + FFrameDuration;
  FFrame := FFrame + 1;
  FSurface.FFrameCount := FFrame;

  with FSurface do
    if Assigned(FBeforeFrame) then
      FBeforeFrame(FSurface);
end;

procedure TabcSpriteThread.EndFrame;
var
  endThisFrame: DWORD;
begin
  with FSurface do
  begin
    Sprites.EndFrame;
    if Assigned(FAfterFrame) then
      FAfterFrame(FSurface);
    //end timing
    endThisFrame := timeGetTime;
    if FStartNextFrame < endThisFrame then
      FSleepFor := 0
    else
      FSleepFor := FStartNextFrame - endThisFrame;
    if (FStartNextFrame - FStartThisSec) >= 1000 then
    begin
      FRealFramesPerSec := FFrame - FFramesLastSec;
      Inc(FActiveSeconds);
      FActiveSecsFrameCount := FActiveSecsFrameCount + FRealFramesPerSec;
      UpdatePerformanceText;
      FFramesLastSec := FFrame;
      FStartThisSec := FStartNextFrame;
    end;
  end;
end;

procedure TabcSpriteThread.Execute;
begin
  Synchronize(AnimateBegin);
  while not Terminated do
  begin
    Synchronize(AnimateFrame);
    if not Terminated then
      Sleep(FSleepFor);
  end;
  Synchronize(AnimateEnd);
end;

procedure TabcSpriteThread.Reset;
begin
  FFrame := 0;
  FFramesLastSec := 0;
  FStartThisSec := FStartNextFrame;
end;

{
---------------------------- TabcSpriteFactory ---------------------------------
}

constructor TabcSpriteFactory.Create(AOwner: TComponent);
begin
  inherited;
  FMover := TabcSpriteMover.Create(nil);
end;

destructor TabcSpriteFactory.Destroy;
begin
  FMover.Free;
  inherited;
end;

class function TabcSpriteFactory.SpriteClass: TabcSpriteClass;
begin
  Result := TabcSprite;
end;

function TabcSpriteFactory.CreateSprite: TabcSprite;
begin
  Result := SpriteClass.Create(Self);
  Result.Mover := Mover;
  if SpriteList <> nil then
    SpriteList.Add(Result);
end;

procedure TabcSpriteFactory.BeginUpdate;
begin
  Inc(FUpdateCount);
end;

procedure TabcSpriteFactory.EndUpdate;
begin
  if FUpdateCount > 0 then Dec(FUpdateCount);
  if FChanged then
  begin
    FChanged := False;
    Change;
  end;
end;

procedure TabcSpriteFactory.Change;
var
  I: Integer;
begin
  if [csLoading, csDestroying] * ComponentState = [] then
  begin
    FChanged := True;
    if FUpdateCount > 0 then Exit;
    if FSpriteList <> nil then
      FSpriteList.BeginUpdate;
    for I := 0 to ComponentCount - 1 do
      if Components[I] is TabcSprite then
        TabcSprite(Components[I]).Change(Self);
    if FSpriteList <> nil then
      FSpriteList.EndUpdate;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TabcSpriteFactory.SetMover(Value: TabcSpriteMover);
begin
  FMover.Assign(Value);
end;

procedure TabcSpriteFactory.SetSpriteList(const Value: TabcSpriteList);
begin
  FSpriteList := Value;
end;

end.
