unit abcsprfx;

(*------------------------------------------------------------------------------

Advanced Business Components
ABC for Delphi
Version 6

Copyright  1995, 2001 Objective Software Technology Pty Limited ACN 068 640 353
Functionality derived from Delphi VCL Source Code
is Copyright  1995, 2001 Borland Software Corporation

------------------------------------------------------------------------------*)

{$I abcdefs.inc}

{:Contains sprite component classes that display images with transition effects.
This unit contains sprite classes that display a picture graphic, or imagelist images with a caption.<br>
The transition effects allow a blend, slide, wipe or other shaped effect to play in multiple animation frames.<br>
The transition occurs automatically when the Picture or Caption changes.
}

interface

uses
  SysUtils, Windows, Messages, Classes, Graphics, Controls, Imglist, StdCtrls, ExtCtrls,
  Forms, abcuspr, abcspr, abcutilg, abcfx, abcufx, abcbtn, abclabel;

type

  { TabcCustomEffectSprite }

  {:A picture sprite with transition effects between images.
  This is the base class with common properties and methods for effect sprites,
  extended in <See Class=TabcEffectSprite> and <See Class=TabcRichEffectSprite>.<P>
  These classes draw blend, fade, wipe or shaped effects whenever the Picture or Caption changes.<P>
  Transition effects are multithreaded, so they do not block user interface events.<P>
  Transition effect steps are overlaid on the Sprite Surface animation frames.
  }

  TabcCustomEffectSprite = class (TabcSprite)
  private
    FChangeLink: TChangeLink;
    FCallEffectEnd: Boolean;
    FBitmap, FMask, FPictureBitmap: TBitmap;
    FThreadBitmap, FThreadMask: TBitmap;
    FPicture: TPicture;
    FThread: TabcTransitionThread;
    FEffectSelIndex: Integer;
    FOnEffectStep: TabcCancelEvent;
    FEffectSize: TabcEffectSize;
    FEffect: TabcTransitionEffect;
    FEffectSelection: TList;
    FOnEffectEnd: TNotifyEvent;
    FEffectSpeed: Word;
    FEffectPause: Word;
    FNeverDrawn: Boolean;
    FThreadBitmapWasEmpty, FWasTransparent: Boolean;

    FPictureHasChanged: Boolean;
    FPictureIsChanging: Boolean;
    FPictureSize: TabcPictureSize;
    FPaintPictureSize: TabcPictureSize;

    FAnimationStopped: Boolean;
    FImageIndex: TImageIndex;
    FImageList: TImageList;
    FOnAnimateNext: TabcCancelEvent;
    FPlaying, FLoop: Boolean;
    FPlayIndexes: TList;
    FPlayItem: integer;
    FPlayItemChanging: Boolean;

    FShowPicture: Boolean;
    FShowText: Boolean;
    FTransparent: Boolean;
    FColor: TColor;

    FUpdateCount: Integer;
    FUpdateNeeded: Boolean;
    FThreadNeeded: Boolean;
    FInTerminate: Boolean;
    
    procedure AdjustAutoSize;
    function CalcPictureRect: TRect;
    procedure CancelEffectStep(Sender: TObject; var Cancel: Boolean);
    procedure SetPicture(Value: TPicture);
    procedure ReadEffectSelections(Reader: TReader);
    procedure WriteEffectSelections(Writer: TWriter);
    function GetEffectSel(Index: Integer): TabcTransitionEffect;
    function GetEffectSelCount: Integer;
    procedure SetEffectSel(Index: Integer;
      const Value: TabcTransitionEffect);
    procedure SetEffectSelection(const Value: TList);
    procedure SetEffectSelIndex(const Value: Integer);
    procedure SetImageIndex(const Value: TImageIndex);
    procedure SetImageList(const Value: TImageList);

    procedure SetTransparent(const Value: Boolean);
    procedure SetShowPicture(const Value: Boolean);
    procedure SetShowText(const Value: Boolean);
    procedure SetColor(const Value: TColor);

    procedure SetPictureSize(Value: TabcPictureSize);
    procedure CheckAnimationStop;
    procedure DrawBackground;
    procedure InitThreadBitmaps;
    procedure ImageListChanged(Sender: TObject);
  protected
    procedure DefineProperties(Filer: TFiler); override;
    procedure Loaded; override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;

    procedure Animate(Frame: Integer; var oldBounds, newBounds: TRect); override;
    procedure AnimateBegin; override;
    procedure AnimateEnd; override;
    procedure EndFrame; override;

    procedure Draw(ACanvas: TCanvas); override;
    procedure SetVisible(Value: Boolean); override;

    procedure DrawPicture(R: TRect); virtual;
    procedure DrawImageText(R: TRect); virtual;

    procedure CanCreateThread; virtual;
    procedure TerminateThread; virtual;
    procedure ThreadTerminated(Sender: TObject); virtual;
    procedure PictureChanged(Sender: TObject); virtual;
    function DoEdit: Boolean; virtual;
    procedure DoClear; virtual;
    procedure DoEffectEnd; virtual;
    function HasText: Boolean; virtual;
    function IsEmpty: Boolean; virtual;
    property Bitmap: TBitmap read FBitmap;
    {:: Published in descendant classes.<br>
     Set to false to hide the Picture image.}
    property ShowPicture: Boolean read FShowPicture write SetShowPicture default True;
    {:: Published in descendant classes.<br>
     Set to false to hide the caption text.}
    property ShowText: Boolean read FShowText write SetShowText default True;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function PtInSprite(const APoint: TPoint): Boolean; override;
    procedure SetBounds(ALeft, ATop, AWidth, AHeight: Integer); override;
    {::Add a transition effect to the list of selected effects.<br>
    The selected effects are published in the <See Property=EffectSelection> TList property.
    There is a design-time property editor for this list.<br>
    At runtime the list elements are referenced by properties <See Property=EffectSel> and <See Property=EffectSelCount>.<br>
    They are used in transitions when the <See Property=Effect> is atfSelectionRandom or atfSelectionSequence.<br>
    <See Property=EffectSelIndex> is used to iterate through the selected effects with Effect atfSelectionSequence.<p>
    <strong>See Also</strong><p>
    <A HRef="TabcTransitionEffect_Type">TabcTransitionEffect</A>
    }
    procedure AddSelectedEffect(Effect: TabcTransitionEffect);
    {::Remove a transition effect from the list of selected effects.<br>
    The selected effects are published in the <See Property=EffectSelection> TList property.
    There is a design-time property editor for this list.<br>
    At runtime the list elements are referenced by properties <See Property=EffectSel> and <See Property=EffectSelCount>.<br>
    They are used in transitions when the <See Property=Effect> is atfSelectionRandom or atfSelectionSequence.<br>
    <See Property=EffectSelIndex> is used to iterate through the selected effects with Effect atfSelectionSequence.<p>
    <strong>See Also</strong><p>
    <A HRef="TabcTransitionEffect_Type">TabcTransitionEffect</A>
    }
    procedure RemoveSelectedEffect(Effect: TabcTransitionEffect);
    {::Returns true if a transition effect is in the list of selected effects.}
    function IsEffectSelected(Effect: TabcTransitionEffect): Boolean;
    {::Clear the sprite Picture and Caption properties.}
    procedure Clear;
    {::Plays a sequence of images from an image list.
     When each transition effect finishes, the next image is loaded.<P>
     <strong>See Related Methods</strong><br>
     <See Method=PlayRange>, <See Method=PlaySelection>, <See Method=Stop>, <See Method=Suspend>, <See Method=Resume><P>
     <strong>See Related Events</strong><br>
     <See Event=OnAnimateNext>
    }
    procedure Play(Loop: Boolean);
    {::Plays a range of images from an image list.
     When each transition effect finishes, the next image is loaded.<P>
     <strong>See Related Methods</strong><br>
     <See Method=Play>, <See Method=PlaySelection>, <See Method=Stop>, <See Method=Suspend>, <See Method=Resume><P>
     <strong>See Related Events</strong><br>
     <See Event=OnAnimateNext>
    }
    procedure PlayRange(FromIndex, ToIndex: TImageIndex; Loop: Boolean);
    {::Plays a selection of images from an image list.
     When each transition effect finishes, the next image is loaded.<P>
     <strong>See Related Methods</strong><br>
     <See Method=Play>, <See Method=PlayRange>, <See Method=Stop>, <See Method=Suspend>, <See Method=Resume><P>
     <strong>See Related Events</strong><br>
     <See Event=OnAnimateNext>
    }
    procedure PlaySelection(const Selection: array of TImageIndex; Loop: Boolean);
    {::Stops any current image transition, and any looped animation using a Play method.}
    procedure Stop;
    {::Suspends the current transition effect.}
    procedure Suspend;
    {::Resumes the current transition effect.}
    procedure Resume;
    {::Called before changing multiple properties that affect the appearance of the sprite.<br>
     Must be matched with a call to EndUpdate.}
    procedure BeginUpdate;
    {::Called after changing multiple properties that affect the appearance of the sprite.}
    procedure EndUpdate;
    {::The count of items in the <See Property=EffectSelection> list.<br>
    At runtime these can be referenced as their true type using <See Property=EffectSel>[index].}
    property EffectSelCount: Integer read GetEffectSelCount;
    {::The selected effects are in a TList <See Property=EffectSelection> indexed 0..<See Property=EffectSelCount>-1.<br>
    At runtime this property references the true type of the <See Property=EffectSelection> items.<br>
    <See Method=AddSelectedEffect> and <See Method=RemoveSelectedEffect> methods can be used to change items.<br>
    These effects are used in transitions when the <See Property=Effect> is atfSelectionRandom or atfSelectionSequence.<br>
    <See Property=EffectSelIndex> is used to iterate through the selected effects with Effect atfSelectionSequence.<p>
    <strong>See Also</strong><p>
    <A HRef="TabcTransitionEffect_Type">TabcTransitionEffect</A>
    }
    property EffectSel[Index: Integer]: TabcTransitionEffect read GetEffectSel write SetEffectSel;
  published
    property Mover;
    property Width;
    property Height;
    property Visible;
    {::The rectangular image background is filled with this color if <See Property=Transparent> is false.}
    property Color: TColor read FColor write SetColor default clBtnFace;
    {::A blend, fade, wipe or shaped effect used to transition from one image to another.<P>
    Transition effects are multithreaded, so they do not block user interface events.<P>
    <strong>See Also</strong><p>
    <A HRef="TabcTransitionEffect_Type">TabcTransitionEffect</A>
    }
    property Effect: TabcTransitionEffect read FEffect write FEffect default atfNone;
    {::This is a published TList of selected transition effects.<br>
    There is a design-time property editor for this list.<br>
    At runtime the list elements are referenced by properties <See Property=EffectSel> and <See Property=EffectSelCount>;
    and by methods <See Method=AddSelectedEffect> and <See Method=RemoveSelectedEffect>.<br>
    The selected effects are used in transitions when the <See Property=Effect> is atfSelectionRandom or atfSelectionSequence.<br>
    <See Property=EffectSelIndex> is used to iterate through the selected effects with Effect atfSelectionSequence.<p>
    <strong>See Also</strong><p>
    <A HRef="TabcTransitionEffect_Type">TabcTransitionEffect</A>
    }
    property EffectSelection: TList read FEffectSelection write SetEffectSelection
             stored False; {values stored by DefineProperties}
    {::Used to iterate through the selected effects in <See Property=EffectSel> with <See Property=Effect> = atfSelectionSequence.<p>
    <strong>See Also</strong><p>
    <A HRef="TabcTransitionEffect_Type">TabcTransitionEffect</A>
    }
    property EffectSelIndex: Integer read FEffectSelIndex write SetEffectSelIndex default -1;
    {::A size parameter used to adjust the granularity and speed of transition effects.<br>
    For large images, use a larger value. For small images use a smaller value.<br>
    Typical values in the range of 2..32 are adequate for most images, depending on the speed of your system.<br>
    The use of this property depends on the particular transition effect.
    }
    property EffectSize: TabcEffectSize read FEffectSize write FEffectSize default 8;
    {::Speed of the transition effect, in milliseconds.<br>
    This is the minimum time the effect lasts for.<br>
    Actual completion could take longer with a slow system, complex image or complex effect.<P>
    Transition effects are multithreaded, so they do not block user interface events.
    }
    property EffectSpeed: Word read FEffectSpeed write FEffectSpeed default 500;
    {::A delay after the transition effect, in milliseconds.<br>
    This can be used to pause between images when using the Play methods,
    or to insert a delay before the OnEffectEnd event,
    to ensure the final image is visible for some time.
    }
    property EffectPause: Word read FEffectPause write FEffectPause default 0;
    {::Index of the current image to be displayed from the ImageList.}
    property ImageIndex: TImageIndex read FImageIndex write SetImageIndex default -1;
    {::Use to display a single image indexed by ImageIndex;
     or use with the <See method=Play> methods to iterate through images in the ImageList.}
    property ImageList: TImageList read FImageList write SetImageList;
    {::Use to display a single graphic image.<br>
    When the graphic changes, a transition effect is drawn from the old image to the new image.
    }
    property Picture: TPicture read FPicture write SetPicture;
    {::Use to scale or align the picture inside the sprite bounds.<P>
    <strong>See Also</strong><p>
    <A HRef="TabcPictureSize_Type">TabcPictureSize</A>
    }
    property PictureSize: TabcPictureSize read FPictureSize write SetPictureSize default apsCenterBestFit;
    {::If true, the Picture image is transparently masked and the sprite is drawn without
    a background color.}
    property Transparent: Boolean read FTransparent write SetTransparent default False;
    {::This event occurs before the next image, during the <See Method=Play> methods.<br>
     It can be used to cancel a loop.}
    property OnAnimateNext: TabcCancelEvent read FOnAnimateNext write FOnAnimateNext;
    {::This event occurs periodically during a transition effect and can be used to cancel the effect.}
    property OnEffectStep: TabcCancelEvent read FOnEffectStep write FOnEffectStep;
    {::This event occurs after a transition effect ends, and after any <See Property=EffectPause> delay has timed out.}
    property OnEffectEnd: TNotifyEvent read FOnEffectEnd write FOnEffectEnd;
  end;

  { TabcEffectSprite }

  {:A picture sprite with transition effects between images.
  This class draws blend, fade, wipe or shaped effects whenever the Picture or ImageIndex or caption text changes.<p>
  The sprite can display either an image or text or both using the <See Property=ShowPicture> and <See Property=ShowText> properties.<P>
  A <See Property=Caption> is drawn over any Picture or Imagelist image. The text is clipped to the picture dimensions.<p>
  <See Property=CaptionStyle>, <See Property=Font>, <See Property=Alignment>, <See Property=Layout> and <See Property=WordWrap> adjust the text appearance.<p>
  Transition effects are multithreaded, so they do not block user interface events.<P>
  Transition effect steps are overlaid on the Sprite Surface animation frames.<P>
  This is the main published class for effect sprites, that have a Picture or ImageList and a text Caption.<P>
  For sprites with a Rich Text (RTF) caption, see class <See Class=TabcRichEffectSprite>.<P>
  For non-sprite transition effect components, see classes
  <A HRef="TabcEffectsImage_Component">TabcEffectsImage</A>,
  <A HRef="TabcRichEffectsImage_Component">TabcRichEffectsImage</A>, and
  <A HRef="TabcDBEffectsImage_Component">TabcDBEffectsImage</A>.<P>
  }

  TabcEffectSprite = class (TabcCustomEffectSprite)
  private
    FCaption: string;
    FCaptionStyle: TabcCaptionStyle;
    FFont: TFont;
    FAlignment: TAlignment;
    FLayout: TTextLayout;
    FWordWrap: Boolean;
    FBiDiMode: TBiDiMode;
    procedure SetAlignment(Value: TAlignment);
    procedure SetCaptionStyle(Value: TabcCaptionStyle);
    procedure SetLayout(const Value: TTextLayout);
    procedure SetWordWrap(const Value: Boolean);
    procedure SetCaption(const Value: string);
    procedure SetFont(const Value: TFont);
    function IsFontStored: Boolean;
    function GetTextFlags: Integer;
    procedure SetBiDiMode(const Value: TBiDiMode);
  protected
    procedure DrawImageText(R: TRect); override;
    procedure DoClear; override;
    procedure DoDrawText(Canvas: TCanvas; var Rect: TRect; Flags: Longint); virtual;
    procedure CaptionStyleChanged(Sender: TObject); virtual;
    function HasText: Boolean; override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  published
    property ShowPicture;
    property ShowText;
    {::Horizontal alignment of the Caption property.}
    property Alignment: TAlignment read FAlignment write SetAlignment
      default taCenter;
    {::BiDiMode of the Caption property.}
    property BiDiMode: TBiDiMode read FBiDiMode write SetBiDiMode default bdLeftToRight;
    {::A text caption that is drawn over any Picture or Imagelist image.<p>
    The text is clipped to the picture dimensions.<p>
    <See Property=CaptionStyle>, <See Property=Font>, <See Property=Alignment>, <See Property=Layout> and <See Property=WordWrap> also adjust the text appearance.
    }
    property Caption: string read FCaption write SetCaption;
    {::<br><A HRef="TabcCaptionStyle_Component">TabcCaptionStyle</A> adds raised, lowered, shadowed or embossed text effects to the caption.}
    property CaptionStyle: TabcCaptionStyle read FCaptionStyle write SetCaptionStyle;
    {::The font used to draw the caption.}
    property Font: TFont read FFont write SetFont stored IsFontStored;
    {::Vertical alignment of the caption.}
    property Layout: TTextLayout read FLayout write SetLayout default tlCenter;
    {::Whether the caption can wrap to multiple lines or is clipped to a single line.}
    property WordWrap: Boolean read FWordWrap write SetWordWrap default True;
  end;

implementation

{ TabcCustomEffectSprite }

constructor TabcCustomEffectSprite.Create(AOwner: TComponent);
begin
  inherited;
  FBitmap := TBitmap.Create;
  FBitmap.PixelFormat := pfDevice;
  FMask := TBitmap.Create;
  FMask.Monochrome := True;
  FPictureBitmap := TBitmap.Create;
  FPictureBitmap.PixelFormat := pfDevice;
  FPicture := TPicture.Create;
  FPicture.OnChange := PictureChanged;
  FChangeLink := TChangeLink.Create;
  FChangeLink.OnChange := ImageListChanged;
  FColor := clBtnFace;
  FEffect := atfNone;
  FEffectSelection := TList.Create;
  FEffectSize := 8;
  FEffectSpeed := 500;
  FImageIndex := -1;
  FPictureSize := apsCenterBestFit;
  FPaintPictureSize := apsCenter;
  FShowPicture := True;
  FShowText := True;
  FNeverDrawn := True;
end;

destructor TabcCustomEffectSprite.Destroy;
begin
  BeginUpdate;
  FPicture.OnChange := nil;
  Destroying;
  TerminateThread;
  FBitmap.Free;
  FMask.Free;
  if FThreadBitmap <> nil then
    FThreadBitmap.Free;
  if FThreadMask <> nil then
    FThreadMask.Free;
  FPictureBitmap.Free;
  FPicture.Free;
  FEffectSelection.Free;
  FPlayIndexes.Free;
  FChangeLink.Free;
  inherited;
end;

procedure TabcCustomEffectSprite.Loaded;
begin
  inherited Loaded;
  PictureChanged(nil);
  if ImageList <> nil then
    SetImageIndex(ImageIndex);
end;

procedure TabcCustomEffectSprite.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited;
  if (Operation = opRemove) and (AComponent = FImageList) then
    SetImageList(nil);
end;

procedure TabcCustomEffectSprite.DefineProperties(Filer: TFiler);
begin
  inherited;
  Filer.DefineProperty('EffectSelections', ReadEffectSelections, WriteEffectSelections, (FEffectSelection.Count > 0));
end;

procedure TabcCustomEffectSprite.ReadEffectSelections(Reader: TReader);
begin
  with Reader do
  begin
    ReadListBegin;
    FEffectSelection.Clear;
    while not EndOfList do FEffectSelection.Add(Pointer(Reader.ReadInteger));
    ReadListEnd;
  end;
end;

procedure TabcCustomEffectSprite.WriteEffectSelections(Writer: TWriter);
var
  i: integer;
begin
  with Writer do
  begin
    WriteListBegin;
    for i := 0 to FEffectSelection.Count - 1 do
      WriteInteger(Integer(FEffectSelection[i]));
    WriteListEnd;
  end;
end;

procedure TabcCustomEffectSprite.AddSelectedEffect(
  Effect: TabcTransitionEffect);
begin
  if not IsEffectSelected(Effect) then
    FEffectSelection.Add(Pointer(Effect));
end;

procedure TabcCustomEffectSprite.AdjustAutoSize;
begin
  if (PictureSize = apsCenterBestFit) then
    if (Picture.Width > Width) or (Picture.Height > Height) then
      FPaintPictureSize := apsBestFit
    else
      FPaintPictureSize := apsCenter
  else
    FPaintPictureSize := PictureSize;
  if ((PictureSize = apsAutoSize) or
      ((csDesigning in ComponentState) and (Width = 0) and (Height = 0))) and
      (Picture.Width > 0) and
      (Picture.Height > 0) and
      ((Picture.Width <> Width) or (Picture.Height > Height)) then
    SetBounds(Left, Top, Picture.Width, Picture.Height);
end;

procedure TabcCustomEffectSprite.Animate(Frame: Integer; var oldBounds,
  newBounds: TRect);
begin
  inherited;
  if (FThread <> nil) and not FInTerminate then
  begin
    FThread.Resume;  //draw next frame
    DrawInNextFrame := True;
  end;
end;

procedure TabcCustomEffectSprite.EndFrame;
begin
  inherited;
  if FCallEffectEnd then
    DoEffectEnd;
end;

procedure TabcCustomEffectSprite.AnimateBegin;
begin
  inherited;
  if FNeverDrawn or not SpriteList.Surface.DrawWhenInactive then
  begin
    FPictureHasChanged := True;
    CanCreateThread;
  end;
end;

procedure TabcCustomEffectSprite.AnimateEnd;
begin
  TerminateThread;
  inherited;
end;

function TabcCustomEffectSprite.CalcPictureRect: TRect;
var
  PicRatio, ImageRatio: Double;
begin
  Result := Rect(0,0,Width,Height);
  if (Picture.Width > 0) and
     (Picture.Height > 0) and
     (Width > 0) then
    with Result do
      case FPaintPictureSize of
        apsNone, apsAutoSize, apsCenterBestFit, apsStretch:
          {no adjustment};
        apsCenter:
        begin
          Left := (Width - Picture.Width) div 2;
          Right := Left + Picture.Width;
          Top := (Height - Picture.Height) div 2;
          Bottom := Top + Picture.Height;
        end;
        apsBestFit:
        begin
          PicRatio := Picture.Height / Picture.Width;
          ImageRatio := Height / Width;
          if PicRatio > ImageRatio then
          begin
            Right := Trunc(Bottom / PicRatio);
            Left := (Width - Right) div 2;
            Right := Right + Left;
          end
          else
          begin
            Bottom := Trunc(Right * PicRatio);
            Top := (Height - Bottom) div 2;
            Bottom := Top + Bottom;
          end;
        end;
      end;
end;

procedure TabcCustomEffectSprite.Draw(ACanvas: TCanvas);
begin
  if IsEmpty or (csDestroying in ComponentState) then Exit;
  if FThread <> nil then
  begin
    if FThreadBitmap.Canvas.TryLock then
    try
      if Transparent or FThreadBitmapWasEmpty or FWasTransparent or FNeverDrawn then
      begin
        //need to update the mask for each frame when another thread is also drawing to it
        abcCreateSpriteMask(FThreadBitmap,FThreadMask);
        abcTransparentBlt(
          ACanvas.Handle, Left, Top, Width, Height,
          FThreadBitmap.Canvas.Handle, 0, 0,
          FThreadMask);
      end
      else
        BitBlt(ACanvas.Handle, Left, Top, Width, Height,
          FThreadBitmap.Canvas.Handle, 0, 0, SRCCOPY);
    finally
      FThreadBitmap.Canvas.Unlock;
    end;
  end
  else if not FBitmap.Empty then
  begin
    ACanvas.Lock;
    FBitmap.Canvas.Lock;
    try
      abcTransparentBlt(
        ACanvas.Handle, Left, Top, Width, Height,
        FBitmap.Canvas.Handle, 0, 0,
        FMask);
    finally
      ACanvas.Unlock;
      FBitmap.Canvas.Unlock;
    end;
  end;
  FNeverDrawn := False;
end;

function TabcCustomEffectSprite.GetEffectSel(
  Index: Integer): TabcTransitionEffect;
begin
  if (Index > -1) and (Index < FEffectSelection.Count) then
    Result := TabcTransitionEffect(FEffectSelection.Items[Index])
  else
    Result := atfNone;
end;

function TabcCustomEffectSprite.GetEffectSelCount: Integer;
begin
  Result := FEffectSelection.Count;
end;

function TabcCustomEffectSprite.IsEffectSelected(
  Effect: TabcTransitionEffect): Boolean;
begin
  Result := FEffectSelection.IndexOf(Pointer(Effect)) <> -1;
end;

procedure TabcCustomEffectSprite.InitThreadBitmaps;
begin
  if Effect <> atfNone then
  begin
    if FThreadBitmap = nil then
    begin
      FThreadBitmap := TBitmap.Create;
      FThreadBitmap.PixelFormat := pfDevice;
      with FThreadBitmap do
      begin
        FThreadBitmapWasEmpty := True;
        Assign(FBitmap);
        with Canvas do
        begin
          Lock;
          Brush.Color := ABC_TRANS_COLOR;
          Brush.Style := bsSolid;
          FillRect(Rect(0,0,Width,Height));
          Unlock;
        end;
        TransparentColor := ABC_TRANS_COLOR;
      end;
      FThreadMask := TBitmap.Create;
      FThreadMask.Monochrome := True;
    end;
    FThreadBitmap.Assign(FBitmap);
  end;
end;

procedure TabcCustomEffectSprite.DrawImageText(R: TRect);
begin
end;

procedure TabcCustomEffectSprite.DrawPicture(R: TRect);
var
  G: TGraphic;
begin
  {copy picture graphic to internal bitmap - for jpeg, wmf, ico conversion into a bitmap}
  with FPictureBitmap do
  begin
    {erase}
    Height := 0;
    Width := 0;
    {resize}
    Height := Picture.Height;
    Width := Picture.Width;

    {draw}
    G := FPicture.Graphic;
    if (G <> nil) and not ((G is TBitmap) or (G is TMetaFile) or (G is TIcon)) then
    begin
      G.Transparent := Self.Transparent;
    end;
    Canvas.Brush.Color := ABC_TRANS_COLOR;
    Canvas.Draw(0,0,FPicture.Graphic);
    Transparent := Self.Transparent;
  end;

  {draw picture to buffer}
  with FBitmap do
  begin
    if FPaintPictureSize in [apsNone, apsAutoSize, apsCenter] then
      Canvas.Draw(R.Left, R.Top, FPictureBitmap)
    else
      Canvas.StretchDraw(R, FPictureBitmap);
    TransparentColor := ABC_TRANS_COLOR;
    Transparent := Self.Transparent;
  end;
end;

procedure TabcCustomEffectSprite.DrawBackground;
begin
  with FBitmap.Canvas do
  begin
    if Self.Transparent then
      Brush.Color := ABC_TRANS_COLOR
    else
      Brush.Color := Self.Color;
    Brush.Style := bsSolid;
    FillRect(Rect(0,0,Width,Height));
  end;
  if FThreadBitmap <> nil then
    if FThreadBitmap.Empty then
      with FThreadBitmap do
      begin
        FThreadBitmapWasEmpty := True;
        Assign(FBitmap);
        with Canvas do
        begin
          Lock;
          Brush.Color := ABC_TRANS_COLOR;
          Brush.Style := bsSolid;
          FillRect(Rect(0,0,Width,Height));
          Unlock;
        end;
        TransparentColor := ABC_TRANS_COLOR;
      end
    else
      FThreadBitmapWasEmpty := False;
end;

procedure TabcCustomEffectSprite.PictureChanged(Sender: TObject);
var
  PicR: TRect;
begin
  if FUpdateCount > 0 then
  begin
    FUpdateNeeded := True;
    Exit;
  end;
  if FPictureIsChanging or ([csLoading, csDestroying] * ComponentState <> []) then
    Exit;
  FPictureIsChanging := True;
  TerminateThread;
  InitThreadBitmaps;
  AdjustAutoSize;
  FPictureHasChanged := True;
  DrawBackground;
  PicR := CalcPictureRect;
  if ShowPicture then
    DrawPicture(PicR);
  if ShowText and HasText then
    DrawImageText(PicR);
  abcCreateSpriteMask(FBitmap, FMask);
  CanCreateThread;
  Change(Self);
  FUpdateNeeded := False;
  FPictureIsChanging := False;
end;

procedure TabcCustomEffectSprite.BeginUpdate;
begin
  Inc (FUpdateCount);
end;

procedure TabcCustomEffectSprite.EndUpdate;
begin
  Dec (FUpdateCount);
  if (FUpdateCount = 0) then
    if FUpdateNeeded then
      PictureChanged(Self)
    else if FThreadNeeded then
      CanCreateThread;
end;

procedure TabcCustomEffectSprite.SetBounds(ALeft, ATop, AWidth, AHeight: Integer);
var
  wasResized: Boolean;
begin
  inherited;
  wasResized := (FBitmap.Width <> Width) or (FBitmap.Height <> Height);
  if wasResized then
  begin
    FBitmap.Width := Width;
    FBitmap.Height := Height;
  end;
  if not FPictureIsChanging and wasResized then
    PictureChanged(nil);
end;

procedure TabcCustomEffectSprite.Play(Loop: Boolean);
begin
  if Assigned(ImageList) and (ImageList.Count > 0) then
    PlayRange(0, ImageList.Count - 1, Loop);
end;

procedure TabcCustomEffectSprite.PlayRange(FromIndex, ToIndex: TImageIndex; Loop: Boolean);
var
  i: integer;
begin
  FAnimationStopped := False;
  TerminateThread;
  if Assigned(ImageList) and (ImageList.Count > 0) then
  begin
    if FPlayIndexes = nil then
      FPlayIndexes := TList.Create
    else
      FPlayIndexes.Clear;
    if FromIndex < ToIndex then
      for i := FromIndex to ToIndex do
        FPlayIndexes.Add(Pointer(i))
    else
        for i := FromIndex downto ToIndex do
          FPlayIndexes.Add(Pointer(i));
    FPlaying := True;
    FLoop := Loop;
    FPlayItem := 0;
    FPlayItemChanging := True;
    ImageIndex := FromIndex;
    FPlayItemChanging := False;
  end;
end;

procedure TabcCustomEffectSprite.PlaySelection(const Selection: array of TImageIndex; Loop: Boolean);
var
  i: integer;
begin
  FAnimationStopped := False;
  TerminateThread;
  if Assigned(ImageList) and (ImageList.Count > 0) then
  begin
    if FPlayIndexes = nil then
      FPlayIndexes := TList.Create
    else
      FPlayIndexes.Clear;
    for i := Low(Selection) to High(Selection) do
      FPlayIndexes.Add(Pointer(Selection[i]));
    FPlaying := True;
    FLoop := Loop;
    FPlayItem := 0;
    FPlayItemChanging := True;
    ImageIndex := TImageIndex(FPlayIndexes[0]);
    FPlayItemChanging := False;
  end;
end;

procedure TabcCustomEffectSprite.CheckAnimationStop;
begin
  if Assigned(OnAnimateNext) then
    OnAnimateNext(Self, FAnimationStopped);
  if Application.Terminated or (csDestroying in ComponentState) or not Visible then
    FAnimationStopped := True;
end;

function TabcCustomEffectSprite.PtInSprite(const APoint: TPoint): Boolean;
var
  clientPos: TPoint;
begin
  Result := inherited PtInSprite(APoint);
  if Result and Transparent then
  begin
    clientPos.X := APoint.X - Left;
    clientPos.Y := APoint.Y - Top;
    with clientPos, FMask do
      Result :=
        (FMask = nil) or Empty or
        ((X < Width) and (Y < Height) and
        (Canvas.Pixels[X,Y] = clBlack));
    end;
end;

procedure TabcCustomEffectSprite.RemoveSelectedEffect(
  Effect: TabcTransitionEffect);
begin
  FEffectSelection.Remove(Pointer(Effect));
end;

procedure TabcCustomEffectSprite.Resume;
begin
  if FThread <> nil then
    FThread.Resume;
end;

procedure TabcCustomEffectSprite.SetEffectSel(Index: Integer;
  const Value: TabcTransitionEffect);
begin
  if (Index > -1) and (Index < FEffectSelection.Count) then
    FEffectSelection.Items[Index] := Pointer(Value)
  else if (Index >= FEffectSelection.Count) then
    AddSelectedEffect(Value);
end;

procedure TabcCustomEffectSprite.SetEffectSelection(const Value: TList);
begin
  FEffectSelection := Value;
end;

procedure TabcCustomEffectSprite.SetEffectSelIndex(const Value: Integer);
begin
  FEffectSelIndex := Value;
end;

procedure TabcCustomEffectSprite.SetImageIndex(const Value: TImageIndex);
var
  WasTransparent: Boolean;
begin
  FImageIndex := Value;
  if not (csLoading in ComponentState) and
     Assigned(ImageList) and
     (Value > -1) and
     (Value < ImageList.Count) then
  begin
    //load imagelist image
    BeginUpdate;
    WasTransparent := Picture.Bitmap.Transparent;
    Picture.Bitmap.Transparent := False;
    Picture.Bitmap := nil;
    ImageList.GetBitmap(Value, Picture.Bitmap);
    Picture.Bitmap.Transparent := WasTransparent;
    FUpdateNeeded := True;
    EndUpdate;
  end
  else
    SetPicture(nil);
end;

procedure TabcCustomEffectSprite.SetImageList(const Value: TImageList);
begin
  if Value <> FImageList then
  begin
    if FImageList <> nil then FImageList.UnRegisterChanges(FChangeLink);
    FImageList := Value;
    if Value <> nil then
    begin
      Value.RegisterChanges(FChangeLink);
      Value.FreeNotification(Self);
    end;
    if (ComponentState * [csLoading, csDestroying] = []) then
      SetImageIndex(ImageIndex);
  end;
end;

procedure TabcCustomEffectSprite.ImageListChanged(Sender: TObject);
begin
  SetImageIndex(ImageIndex);
end;

procedure TabcCustomEffectSprite.SetPicture(Value: TPicture);
begin
  if not (csDestroying in ComponentState) then
  begin
    BeginUpdate;
    FPicture.Assign(Value);
    EndUpdate;
  end;
end;

procedure TabcCustomEffectSprite.Stop;
begin
  TerminateThread;
end;

procedure TabcCustomEffectSprite.Suspend;
begin
  if FThread <> nil then
    FThread.Suspend;
end;

procedure TabcCustomEffectSprite.CanCreateThread;
begin
  if FUpdateCount > 0 then
  begin
    FThreadNeeded := True;
    Exit;
  end;
  if (FThread = nil) and Active and FPictureHasChanged and Visible and not IsEmpty and not (csDesigning in ComponentState) then
    if (Effect <> atfNone) then
    begin
      FPictureHasChanged := False;
      FCallEffectEnd := False;
      if (FThreadBitmap = nil) then
        InitThreadBitmaps;
      FThread :=
        abcThreadTransition (
          Effect, EffectSelection, FEffectSelIndex,
          FBitmap.Canvas, FThreadBitmap.Canvas,
          0, 0, Width, Height,
          0, 0,
          EffectSize, EffectSpeed, EffectPause,
          CancelEffectStep,
          ThreadTerminated,
          True, True
          );
      FThreadNeeded := False;
    end
    else
      FCallEffectEnd := True;
end;

procedure TabcCustomEffectSprite.TerminateThread;
begin
  FInTerminate := True;
  FThreadNeeded := False;
  FCallEffectEnd := False;
  if not FPlayItemChanging then
    FPlaying := False;
  if FThread <> nil then
    with FThread do
    begin
      Suspend;
      FreeOnTerminate := False;
      OnTerminate := nil;
      Terminate;
      Resume;
      WaitFor;
      Free;
      ThreadTerminated(Self);
    end;
  FInTerminate := False;
end;

procedure TabcCustomEffectSprite.ThreadTerminated(Sender: TObject);
begin
  FThread := nil;
  DoEffectEnd;
end;

procedure TabcCustomEffectSprite.DoEffectEnd;
begin
  DrawInNextFrame := True;
  FWasTransparent := False;
  FCallEffectEnd := False;
  if not ((csDestroying in ComponentState) or Application.Terminated) and Assigned(FOnEffectEnd) then
    FOnEffectEnd(Self);
  if FPlaying then
  begin
    CheckAnimationStop;
    if FAnimationStopped then
    begin
      FPlaying := False;
      FPlayIndexes.Clear;
      Exit;
    end;
    FPlayItem := FPlayItem + 1;
    if FPlayItem >= FPlayIndexes.Count then
      if FLoop then
        FPlayItem := 0
      else
      begin
        FPlaying := False;
        FPlayIndexes.Clear;
        Exit;
      end;
    FPlayItemChanging := True;
    ImageIndex :=  TImageIndex(FPlayIndexes[FPlayItem]);
    FPlayItemChanging := False;
  end;
end;

procedure TabcCustomEffectSprite.Clear;
begin
  if not IsEmpty then
  begin
    BeginUpdate;
    DoClear;
    EndUpdate;
  end;
end;

procedure TabcCustomEffectSprite.DoClear;
begin
  Picture.Graphic := nil;
end;

function TabcCustomEffectSprite.DoEdit: Boolean;
begin
  Result := False;
end;

function TabcCustomEffectSprite.HasText: Boolean;
begin
  Result := False;
end;

function TabcCustomEffectSprite.IsEmpty: Boolean;
begin
  Result := (not HasText) and (Picture.Graphic = nil) and Transparent;
end;

procedure TabcCustomEffectSprite.CancelEffectStep(Sender: TObject;
  var Cancel: Boolean);
begin
  Cancel := Application.Terminated or (csDestroying in ComponentState) or not Active or not Visible;
  if Assigned(FOnEffectStep) then
    FOnEffectStep(Self, Cancel);
end;

procedure TabcCustomEffectSprite.SetColor(const Value: TColor);
begin
  if Value <> FColor then
  begin
    FColor := Value;
    if not (csLoading in ComponentState) and not Transparent then
      PictureChanged(nil);
  end;
end;

procedure TabcCustomEffectSprite.SetShowPicture(const Value: Boolean);
begin
  if Value <> FShowPicture then
  begin
    FShowPicture := Value;
    PictureChanged(nil);
  end;
end;

procedure TabcCustomEffectSprite.SetShowText(const Value: Boolean);
begin
  if Value <> FShowText then
  begin
    FShowText := Value;
    if HasText then
      PictureChanged(nil);
  end;
end;

procedure TabcCustomEffectSprite.SetPictureSize(Value: TabcPictureSize);
begin
  if Value <> FPictureSize then
  begin
    FPictureSize := Value;
    PictureChanged(nil);
  end;
end;

procedure TabcCustomEffectSprite.SetTransparent(const Value: Boolean);
begin
  if Value <> FTransparent then
  begin
    FTransparent := Value;
    FWasTransparent := not Value;
    PictureChanged(nil);
  end;
end;

procedure TabcCustomEffectSprite.SetVisible(Value: Boolean);
begin
  inherited;
  if Value then
    CanCreateThread
  else
    TerminateThread;
end;

{ TabcEffectSprite }

constructor TabcEffectSprite.Create(AOwner: TComponent);
begin
  inherited;
  FCaptionStyle := TabcCaptionStyle.Create;
  FCaptionStyle.OnChange := CaptionStyleChanged;
  FFont := TFont.Create;
  FFont.OnChange := CaptionStyleChanged;
  FAlignment := taCenter;
  FLayout := tlCenter;
  FWordWrap := True;
end;

destructor TabcEffectSprite.Destroy;
begin
  Destroying;
  FFont.Free;
  FCaptionStyle.Free;
  inherited;
end;

procedure TabcEffectSprite.DrawImageText(R: TRect);
var
  DrawRect, CalcRect: TRect;
  DrawFlags: Longint;

  procedure DrawOffsetText(Offset: Integer; Emboss, Shadow: Boolean);
  var
    bmpSrc, bmpMask: TBitmap;
  begin
    with Bitmap do
    begin
      if Emboss then
      begin
        bmpSrc := TBitmap.Create;
        bmpMask := TBitmap.Create;
        bmpSrc.Assign(Bitmap);
        bmpMask.Width := Width;
        bmpMask.Height := Height;
      end
      else begin
        bmpSrc := nil;
        bmpMask := nil;
      end;

      if not Shadow then begin
        OffsetRect(DrawRect, -Offset, -Offset);
        Canvas.Font.Color := FCaptionStyle.ColorHighlight;
        DoDrawText(Canvas, DrawRect, DrawFlags);
        OffsetRect(DrawRect, 2*Offset, 2*Offset);
      end
      else
        OffsetRect(DrawRect, Offset, Offset);

      Canvas.Font.Color := FCaptionStyle.ColorShadow;
      DoDrawText(Canvas, DrawRect, DrawFlags);
      OffsetRect(DrawRect, -Offset, -Offset);
      if Emboss then
      begin
        with bmpMask.Canvas do
        begin
          Brush.Style := bsClear;
          Font := Self.Font;
          Font.Color := clBlack;
        end;
        DoDrawText(bmpMask.Canvas, DrawRect, DrawFlags);
        SetBkColor(Canvas.Handle, clWhite);
        SetTextColor(Canvas.Handle, clBlack);
        BitBlt(Canvas.Handle, 0, 0, Width, Height, bmpSrc.Canvas.Handle, 0, 0, SRCINVERT);
        BitBlt(Canvas.Handle, 0, 0, Width, Height, bmpMask.Canvas.Handle, 0, 0, SRCAND);
        BitBlt(Canvas.Handle, 0, 0, Width, Height, bmpSrc.Canvas.Handle, 0, 0, SRCINVERT);
        bmpSrc.Free;
        bmpMask.Free;
      end
      else
      begin
        Canvas.Font.Color := Self.Font.Color;
        DoDrawText(Canvas, DrawRect, DrawFlags);
      end;
    end;
  end;

begin
  {draw caption to buffer}
  with Bitmap.Canvas, CaptionStyle do
  begin
    Font := Self.Font;
    Brush.Style := bsClear;
    Windows.CopyRect(DrawRect, R);
    DrawFlags := GetTextFlags;

    case Effect of
      ceNormal: {do nothing};
      ceRaised..ceEmbossLowered:
        InflateRect(DrawRect, -Offset*2, -Offset*2);
      ceShadowed:
        InflateRect(DrawRect, -Offset, -Offset);
    end;

    if FLayout <> tlTop then
    begin
      CalcRect := DrawRect;
      DoDrawText(Bitmap.Canvas, CalcRect, DrawFlags or DT_CALCRECT);
      if FLayout = tlBottom then
        OffsetRect(DrawRect, 0, R.Bottom - CalcRect.Bottom)
      else
        OffsetRect(DrawRect, 0, (R.Bottom - CalcRect.Bottom) div 2);
    end;

    case FCaptionStyle.Effect of
      ceNormal:         DoDrawText(Bitmap.Canvas, DrawRect, DrawFlags);
      ceRaised:         DrawOffsetText(Offset, False, False);
      ceLowered:        DrawOffsetText(-Offset, False, False);
      ceEmbossRaised:   DrawOffsetText(Offset, True, False);
      ceEmbossLowered:  DrawOffsetText(-Offset, True, False);
      ceShadowed:       DrawOffsetText(Offset, False, True);
    end;
  end;
end;

procedure TabcEffectSprite.DoDrawText(Canvas: TCanvas; var Rect: TRect; Flags: Longint);
begin
  DrawText(Canvas.Handle, PChar(Caption), -1, Rect, Flags);
end;

function TabcEffectSprite.HasText: Boolean;
begin
  Result := Caption <> '';
end;

function TabcEffectSprite.GetTextFlags: Integer;
const
  Alignments: array[TAlignment] of Word = (DT_LEFT, DT_RIGHT, DT_CENTER);
  WordWraps: array[Boolean] of Word = (0, DT_WORDBREAK);
begin
  Result := DT_NOPREFIX or DT_EXPANDTABS or WordWraps[FWordWrap] or Alignments[FAlignment];
  if SysLocale.MiddleEast then
  begin
    if BiDiMode = bdRightToLeft then
      if Result and DT_RIGHT = DT_RIGHT then
        Result := Result and not DT_RIGHT
      else if not (Result and DT_CENTER = DT_CENTER) then
        Result := Result or DT_RIGHT;
    if BiDiMode <> bdLeftToRight then
      Result := Result or DT_RTLREADING;
  end;
end;

function TabcEffectSprite.IsFontStored: Boolean;
begin
  with Font do
    Result :=
      (Charset <> DEFAULT_CHARSET) or
      (Color <> clWindowText) or
      (Height <> -11) or
      (Name <> 'MS Sans Serif') or
      (Style <> []);
end;

procedure TabcEffectSprite.SetAlignment(Value: TAlignment);
begin
  if Value <> FAlignment then
  begin
    FAlignment := Value;
    CaptionStyleChanged(nil);
  end;
end;

procedure TabcEffectSprite.SetBiDiMode(const Value: TBiDiMode);
begin
  if Value <> FBiDiMode then
  begin
    FBiDiMode := Value;
    CaptionStyleChanged(nil);
  end;
end;

procedure TabcEffectSprite.SetCaption(const Value: string);
begin
  if Value <> FCaption then
  begin
    FCaption := Value;
    PictureChanged(nil);
  end;
end;

procedure TabcEffectSprite.SetCaptionStyle(Value: TabcCaptionStyle);
begin
  FCaptionStyle.Assign(Value);
  CaptionStyleChanged(nil);
end;

procedure TabcEffectSprite.SetFont(const Value: TFont);
begin
  FFont.Assign(Value);
  CaptionStyleChanged(nil);
end;

procedure TabcEffectSprite.SetLayout(const Value: TTextLayout);
begin
  if Value <> FLayout then
  begin
    FLayout := Value;
    CaptionStyleChanged(nil);
  end;
end;

procedure TabcEffectSprite.SetWordWrap(const Value: Boolean);
begin
  if Value <> FWordWrap then
  begin
    FWordWrap := Value;
    CaptionStyleChanged(nil);
  end;
end;

procedure TabcEffectSprite.CaptionStyleChanged(Sender: TObject);
begin
  if ShowText and (Caption <> '') then
    PictureChanged(nil);
end;

procedure TabcEffectSprite.DoClear;
begin
  inherited;
  Caption := '';
end;

end.
