unit abcsprim;

(*------------------------------------------------------------------------------

Advanced Business Components
ABC for Delphi
Version 6

Copyright  1995, 2001 Objective Software Technology Pty Limited ACN 068 640 353
Functionality derived from Delphi VCL Source Code
is Copyright  1995, 2001 Borland Software Corporation

------------------------------------------------------------------------------*)

{$I abcdefs.inc}

{:Contains sprite component classes that display an image.
This unit contains sprite classes that display a masked bitmap, picture graphic, or imagelist image.<br>
Also contains a shape sprite and an invisible input sprite class.
}

interface

uses
  SysUtils, Windows, Messages, Classes, Graphics, Controls, imglist, ExtCtrls,
  abcspr, abcuspr, abcutilg, abcshape
  ;

type

  {$IFNDEF DELPHI5}
    TImageIndex = type Integer;
  {$ENDIF}

  {::This type is used to specify the current glyph in a bitmap containing multiple images.}

  TabcSpriteGlyphIndex = 1..MaxInt;

  { TabcCustomBitmapSprite }

  {:A common ancestor class for bitmap sprites.
   TabcCustomBitmapSprite contains common methods and properties inherited by descendant classes
   that all use a bitmap image:
   <See Class=TabcBitmapSprite>, <See Class=TabcSharedBitmapSprite> and <See Class=TabcPictureSprite>.<P>
   The Bitmap property contains the image.
   (In TabcPictureSprite this is not published and it is a bitmap copy of the Picture graphic).<P>
   The transparency mask is optional. It can be automatically created using the MaskLinked property.<P>
   The bitmap can contain multiple glyphs, sized using the GlyphsX and GlyphsY properties.<P>
   If multiple glyphs are used, they can be animated using the GlyphIndex and AnimateGlyphs properties.<P>
   Note the sprite's Height and Width are not published,
   but are derived from the Bitmap, GlyphsX and GlyphsY properties.}

  TabcCustomBitmapSprite = class (TabcSprite)
  private
    FAnimateGlyphs: Boolean;
    FGlyphIndex: TabcSpriteGlyphIndex;
    FGlyphsX, FGlyphsY: TabcSpriteGlyphIndex;
    FBitmap: TBitmap;
    FMask: TBitmap;
    FMaskLinked: Boolean;
    FSharedBitmaps: Boolean;
    procedure AdjustSize;
    procedure SetGlyphIndex(Value: TabcSpriteGlyphIndex);
    procedure SetGlyphsX(Value: TabcSpriteGlyphIndex);
    procedure SetGlyphsY(Value: TabcSpriteGlyphIndex);
    procedure SetBitmap(Value: TBitmap);
    procedure SetMask(Value: TBitmap);
    procedure SetMaskLinked(Value: Boolean);
    procedure BitmapChanged(Sender: TObject);
    function IsMaskStored: Boolean;
  protected
    procedure Animate(Frame: Integer; var oldBounds, newBounds: TRect); override;
    procedure Draw(ACanvas: TCanvas); override;
      {::Set to true when the sprite contains multiple glyphs and the GlyphIndex
       should be automatically incremented when animating the sprite.}
    property AnimateGlyphs: Boolean read FAnimateGlyphs write FAnimateGlyphs default False;
      {::For a bitmap containing multiple glyphs, this is the number of glyphs in the X axis.<br>
       i.e. each glyph will have a width of Bitmap.Width div GlyphsX.<br>
       The total number of glyphs in the Bitmap is GlyphsX * GlyphsY.}
    property GlyphsX: TabcSpriteGlyphIndex read FGlyphsX write SetGlyphsX default 1;
      {::For a bitmap containing multiple glyphs, this is the number of glyphs in the Y axis.<br>
       i.e. each glyph will have a height of Bitmap.Height div GlyphsY.<br>
       The total number of glyphs in the Bitmap is GlyphsX * GlyphsY.}
    property GlyphsY: TabcSpriteGlyphIndex read FGlyphsY write SetGlyphsY default 1;
      {::Where a Bitmap contains multiple glyphs, this is the index of the current glyph.<br>
      Glyphs are numbered from left to right and top to bottom.<br>
      <See property=AnimateGlyphs> is used to automatically increment this index on each animation frame.}
    property GlyphIndex: TabcSpriteGlyphIndex read FGlyphIndex write SetGlyphIndex default 1;
      {::This is the image used to draw the bitmap sprite.}
    property Bitmap: TBitmap read FBitmap write SetBitmap;
      {::This is the mask used to draw the bitmap sprite.<br>
       The mask pixels are black where the Bitmap is solid and white where Bitmap is transparent.<br>
       You can set this property to supply your own mask image.
       If <See property=MaskLinked> is true, the mask is automatically created and is not stored.}
    property Mask: TBitmap read FMask write SetMask stored IsMaskStored;
      {::Set to true to create a transparency mask automatically from the transparent pixels in the Bitmap.}
    property MaskLinked: Boolean read FMaskLinked write SetMaskLinked default False;
  public
    constructor Create(AOwner: TComponent); override;
    {:: This constructor is introduced for descendant class <See Class=TabcSharedBitmapSprite>.<br>
    Instead of creating its own Bitmap and Mask, these shared images are used from a class factory.<br>
    This reduces memory and makes it easy to change the image of all the sprites at once.}
    constructor CreateShared(AOwner: TComponent; ABitmap, AMask: TBitmap); virtual;
    destructor Destroy; override;
    {::Bitmap sprites override this method to allow testing against a mask.
    If the sprite has a mask, points located on a transparent pixel return false.}
    function PtInSprite(const APoint: TPoint): Boolean; override;
  end;

  { TabcSharedBitmapSprite - see TabcBitmapSpriteFactory below }

    {:A sprite that draws a shared bitmap image.
      TabcSharedBitmapSprite used a shared Bitmap and Mask to reduce memory use and
      make it easy to change the image of all sprites at once.<P>
      The Bitmap and Mask of the sprite are referenced from the <See Class=TabcBitmapSpriteFactory>
      that created the sprite.}
  TabcSharedBitmapSprite = class (TabcCustomBitmapSprite)
  protected
    procedure Change(Sender: TObject); override;
  published
    property AnimateGlyphs;
    property GlyphIndex;
    property Mover;
    property Visible;
  end;

  { TabcBitmapSprite }

  {:A sprite that draws a masked bitmap image.
   TabcBitmapSprite draws a masked bitmap image.<P>
   The transparency mask is optional. It can be automatically created using the MaskLinked property.<br>
   The bitmap can contain multiple glyphs, sized using the GlyphsX and GlyphsY properties.<br>
   If multiple glyphs are used, they can be animated using the GlyphIndex and AnimateGlyphs properties.<br>
   Note the sprite's Height and Width are not published,
   but are derived from the Bitmap, GlyphsX and GlyphsY properties.}

  TabcBitmapSprite = class (TabcCustomBitmapSprite)
  published
    property AnimateGlyphs;
    property GlyphsX;
    property GlyphsY;
    property GlyphIndex;
    property Bitmap;
    property Mask;
    property MaskLinked;
    property Mover;
    property Visible;
  end;

  { TabcPictureSprite }

  {:A sprite that draws a masked Picture image.
   TabcPictureSprite draws a masked picture image. The picture graphic is copied to a bitmap for drawing.<P>
   The transparency mask is optional. It can be automatically created using the MaskLinked property.<br>
   The picture can contain multiple glyphs, sized using the GlyphsX and GlyphsY properties.<br>
   If multiple glyphs are used, they can be animated using the GlyphIndex and AnimateGlyphs properties.<br>
   Note the sprite's Height and Width are not published,
   but are derived from the Picture.Graphic, GlyphsX and GlyphsY properties.<P>
   If the Picture graphic is a TBitmap, use the <See Class=TabcBitmapSprite> class instead.}

  TabcPictureSprite = class (TabcCustomBitmapSprite)
  private
    FPicture: TPicture;
    procedure PictureChanged(Sender: TObject);
    procedure SetPicture(Value: TPicture);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  published
    property AnimateGlyphs;
    property GlyphsX;
    property GlyphsY;
    property GlyphIndex;
    property Mover;
    property Visible;
    {::The picture containing a graphic image to be drawn.<P>
    Note Picture and Picture.Graphic are copied when assigned to another TPicture.<br>
    Picture.Graphic := nil; clears the image.}
    property Picture: TPicture read FPicture write SetPicture;
  end;

  { TabcImageListSprite }

  {:A sprite that draws images from a TImageList.
   TabcImageListSprite draws a masked image using an index on an ImageList.<P>
   Note the sprite's Height and Width are not published,
   but are derived from the ImageList size.<P>
   The ImageIndex can be automatically incremented when animating, using the AnimateGlyphs property.}

  TabcImageListSprite = class (TabcSprite)
  private
    FAnimateGlyphs: Boolean;
    FChangeLink: TChangeLink;
    FImageIndex: TImageIndex;
    FImages: TImageList;
    procedure AdjustSize;
    procedure SetImageIndex(Value: TImageIndex);
    procedure SetImages(Value: TImageList);
    procedure ImageListChanged(Sender: TObject);
  protected
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure Animate(Frame: Integer; var oldBounds, newBounds: TRect); override;
    procedure Draw(ACanvas: TCanvas); override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  published
      {::Set to true when the ImageIndex should be automatically incremented when animating the sprite.}
    property AnimateGlyphs: Boolean read FAnimateGlyphs write FAnimateGlyphs default False;
      {::The index of the sprite image in the ImageList.}
    property ImageIndex: TImageIndex read FImageIndex write SetImageIndex default 0;
      {::The TImageList containing the sprite images. These are drawn transparently if the ImageList is masked.}
    property Images: TImageList read FImages write SetImages;
    property Mover;
    property Visible;
  end;

  { TabcBitmapSpriteFactory }

  {:A factory component that creates multiple bitmap sprites using a shared image.
   TabcBitmapSpriteFactory creates multiple sprites of class <See Class=TabcSharedBitmapSprite>.<P>
   The sprite Bitmap and Mask properties are not copied or published.
   Instead each sprite uses a pointer to the images of the factory component.
   This reduces memory use and makes it easy to change the image of all sprites at once.<P>
   Each sprite can have its own GlyphIndex to draw a different part of the factory image.
  }

  TabcBitmapSpriteFactory = class (TabcSpriteFactory)
  private
    FAnimateGlyphs: Boolean;
    FGlyphIndex: TabcSpriteGlyphIndex;
    FGlyphsX, FGlyphsY: TabcSpriteGlyphIndex;
    FBitmap: TBitmap;
    FMask: TBitmap;
    FMaskLinked: Boolean;
    FSpriteLeft, FSpriteTop, FSpriteWidth, FSpriteHeight: Integer;
    procedure AdjustSize;
    procedure SetGlyphIndex(Value: TabcSpriteGlyphIndex);
    procedure SetGlyphsX(Value: TabcSpriteGlyphIndex);
    procedure SetGlyphsY(Value: TabcSpriteGlyphIndex);
    procedure SetBitmap(Value: TBitmap);
    procedure SetMask(Value: TBitmap);
    procedure SetMaskLinked(Value: Boolean);
    procedure BitmapChanged(Sender: TObject);
    function IsMaskStored: Boolean;
  public
    {::The sprite class returned by this function is <See Class=TabcSharedBitmapSprite>.}
    class function SpriteClass: TabcSpriteClass; override;
    {::This method is overridden to return a <See Class=TabcSharedBitmapSprite> using shared properties of the factory.<P>
     The sprite is owned by the factory.<P>
     If the factory's SpriteList property is assigned, the sprite is added to the list.}
    function CreateSprite: TabcSprite; override;
    {::Creates multiple sprites and adds them to the SpriteList.}
    procedure AddNSprites(numToCreate: Byte);
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  published
      {::Set to true when the bitmap contains multiple glyphs and the GlyphIndex
       should be automatically incremented when animating the sprite.<P>
       This property is copied to sprites created by the factory.<br>
       It can be changed in each sprite.}
    property AnimateGlyphs: Boolean read FAnimateGlyphs write FAnimateGlyphs default False;
      {::For a bitmap containing multiple glyphs, this is the number of glyphs in the X axis.<br>
       i.e. each glyph will have a width of Bitmap.Width div GlyphsX.<br>
       The total number of glyphs in the Bitmap is GlyphsX * GlyphsY.<P>
       This property is shared by all sprites created by the factory.}
    property GlyphsX: TabcSpriteGlyphIndex read FGlyphsX write SetGlyphsX default 1;
      {::For a bitmap containing multiple glyphs, this is the number of glyphs in the Y axis.<br>
       i.e. each glyph will have a height of Bitmap.Height div GlyphsY.<br>
       The total number of glyphs in the Bitmap is GlyphsX * GlyphsY.<P>
       This property is shared by all sprites created by the factory.}
    property GlyphsY: TabcSpriteGlyphIndex read FGlyphsY write SetGlyphsY default 1;
      {::Where a Bitmap contains multiple glyphs, this is the index of the current glyph.<br>
      Glyphs are numbered from left to right and top to bottom.<br>
      <See property=AnimateGlyphs> is used to automatically increment this index on each animation frame.<P>
       This property is copied to sprites created by the factory.<br>
       It can be changed in each sprite.}
    property GlyphIndex: TabcSpriteGlyphIndex read FGlyphIndex write SetGlyphIndex default 1;
      {::This is the image used to draw the bitmap sprite.<P>
       This property is shared by all sprites created by the factory.
       Changing this image causes all created sprites to change their appearance.}
    property Bitmap: TBitmap read FBitmap write SetBitmap;
      {::This is the mask used to draw the bitmap sprite.<br>
       The mask pixels are black where the Bitmap is solid and white where Bitmap is transparent.<br>
       You can set this property to supply your own mask image.
       If <See property=MaskLinked> is true, the mask is automatically created and is not stored.<P>
       This property is shared by all sprites created by the factory.}
    property Mask: TBitmap read FMask write SetMask stored IsMaskStored;
      {::Set to true to create a transparency mask automatically from the transparent pixels in the Bitmap.<P>
       This property is shared by all sprites created by the factory.}
    property MaskLinked: Boolean read FMaskLinked write SetMaskLinked default False;
    property Mover;
      {::This is the initial Left position of a sprite.<br>
       This property is copied to sprites created by the factory.<br>
       It can be changed in each sprite.}
    property SpriteLeft: Integer read FSpriteLeft write FSpriteLeft default 0;
    property SpriteList;
      {::This is the initial Top position of a sprite.<br>
       This property is copied to sprites created by the factory.<br>
       It can be changed in each sprite.}
    property SpriteTop: Integer read FSpriteTop write FSpriteTop default 0;
  end;

  { TabcInputSprite }

  {:An invisible sprite that is only used to detect events.
   TabcInputSprite can be used to detect mouse events in a rectangular region.<P>
   It is only visible at designtime.
  }

  TabcInputSprite = class (TabcSprite)
  protected
    procedure Draw(ACanvas: TCanvas); override;
  public
    constructor Create(AOwner: TComponent); override;
  published
    property Left;
    property Top;
    property Width;
    property Height;
    property Mover;
  end;

  { TabcShapeSprite }

  {::The shapes that can be drawn by <See Class=TabcShapeSprite>.}

  TabcShapeType = (astRectangle, astSquare, astRoundRect, astRoundSquare, astEllipse, astCircle);

  {:A sprite that draws a shape; using a brush and pen, caption, and border style.
   TabcShapeSprite draws a regular <See Property=Shape>.<P>
   The appearance is similar to Delphi's TShape component, with the optional addition of border bevels.
  }

  TabcShapeSprite = class (TabcSprite)
  private
    FPen: TPen;
    FBrush: TBrush;
    FShape: TabcShapeType;
    FWordWrap: Boolean;
    FCaption: string;
    FCornerWidth: TabcCornerWidth;
    FAlignVertical: TabcVerticalAlign;
    FAlignment: TAlignment;
    FBevelWidth: TBevelWidth;
    FBorderWidth: TBorderWidth;
    FColorHighlight: TColor;
    FColorShadow: TColor;
    FFont: TFont;
    FBevelOuter: TPanelBevel;
    FBevelInner: TPanelBevel;
    function IsFontStored: Boolean;
    procedure SetBrush(Value: TBrush);
    procedure SetPen(Value: TPen);
    procedure SetShape(Value: TabcShapeType);
    procedure SetAlignment(const Value: TAlignment);
    procedure SetAlignVertical(const Value: TabcVerticalAlign);
    procedure SetBevelInner(const Value: TPanelBevel);
    procedure SetBevelOuter(const Value: TPanelBevel);
    procedure SetBevelWidth(const Value: TBevelWidth);
    procedure SetBorderWidth(const Value: TBorderWidth);
    procedure SetCaption(const Value: string);
    procedure SetColorHighlight(const Value: TColor);
    procedure SetColorShadow(const Value: TColor);
    procedure SetCornerWidth(const Value: TabcCornerWidth);
    procedure SetFont(const Value: TFont);
    procedure SetWordWrap(const Value: Boolean);
    procedure StyleChanged(Sender: TObject);
  protected
    procedure Draw(ACanvas: TCanvas); override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure SetBounds(ALeft, ATop, AWidth, AHeight: Integer); override;
  published
    property Width;
    property Height;
    property Mover;
    property Visible;
    {::Alignment of the Caption text inside the shape.}
    property Alignment: TAlignment read FAlignment write SetAlignment default taCenter;
    {::Vertical alignment of the Caption text inside the shape.}
    property AlignVertical: TabcVerticalAlign read FAlignVertical write SetAlignVertical default vaCenter;
    {::Style of inner bevel to draw at the edges of the shape.}
    property BevelInner: TPanelBevel read FBevelInner write SetBevelInner default bvNone;
    {::Style of outer bevel to draw at the edges of the shape.}
    property BevelOuter: TPanelBevel read FBevelOuter write SetBevelOuter default bvNone;
    {::Thickness of inner and outer bevels.}
    property BevelWidth: TBevelWidth read FBevelWidth write SetBevelWidth default 1;
    {::Thickness of a border around the sprite shape.}
    property BorderWidth: TBorderWidth read FBorderWidth write SetBorderWidth default 0;
    {::The brush used to fill the sprite shape.}
    property Brush: TBrush read FBrush write SetBrush;
    {::This text is drawn inside the shape, clipped to a rectangle that fits inside the shape.<br>
    It can be sized and positioned using the <See property=Alignment>, <See property=AlignVertical>, and <See property=WordWrap> properties.}
    property Caption: string read FCaption write SetCaption;
    {::Highlight color of inner and outer bevels.}
    property ColorHighlight: TColor read FColorHighlight write SetColorHighlight default clBtnHighlight;
    {::Shadow color of inner and outer bevels.}
    property ColorShadow: TColor read FColorShadow write SetColorShadow default clBtnShadow;
    {::The width of corners when the Shape is a round rectangle or round square.}
    property CornerWidth: TabcCornerWidth read FCornerWidth write SetCornerWidth default 20;
    {::The font used to draw the Caption.}
    property Font: TFont read FFont write SetFont stored IsFontStored;
    {::The pen used to draw the border of the shape.}
    property Pen: TPen read FPen write SetPen;
    {::The shape of the sprite.<P>
    TabcShapeType = (astRectangle, astSquare, astRoundRect, astRoundSquare, astEllipse, astCircle)}
    property Shape: TabcShapeType read FShape write SetShape default astRectangle;
    {::Set to true to wrap the Caption text to multiple lines inside the shape.}
    property WordWrap: Boolean read FWordWrap write SetWordWrap default False;
  end;

implementation

{
---------------------------- TabcCustomBitmapSprite ----------------------------
}

constructor TabcCustomBitmapSprite.Create(AOwner: TComponent);
begin
  inherited;
  FBitmap := TBitmap.Create;
  FBitmap.PixelFormat := pfDevice;
  FBitmap.OnChange := BitmapChanged;
  FMask := TBitmap.Create;
  FMask.Monochrome := True;
  FGlyphIndex := 1;
  FGlyphsX := 1;
  FGlyphsY := 1;
end;

constructor TabcCustomBitmapSprite.CreateShared(AOwner: TComponent; ABitmap, AMask: TBitmap);
begin
  inherited Create(AOwner);
  FSharedBitmaps := True;
  FBitmap := ABitmap;
  FMask := AMask;
  FGlyphIndex := 1;
  FGlyphsX := 1;
  FGlyphsY := 1;
end;

destructor TabcCustomBitmapSprite.Destroy;
begin
  Destroying;
  if not FSharedBitmaps then
  begin
    FBitmap.Free;
    FMask.Free;
  end;
  inherited;
end;

procedure TabcCustomBitmapSprite.Animate(Frame: Integer; var oldBounds, newBounds: TRect);
begin
  if (DropFrames > 0) and (Frame mod DropFrames <> 1) then
  begin
    inherited;
    Exit;
  end;
  if FAnimateGlyphs and ((FGlyphsX > 1) or (FGlyphsY > 1)) then
  begin
    FGlyphIndex := FGlyphIndex + 1;
    if FGlyphIndex > FGlyphsX * FGlyphsY then
      FGlyphIndex := 1;
    DrawInNextFrame := True;
  end;
  inherited;
end;

function TabcCustomBitmapSprite.PtInSprite(const APoint: TPoint): Boolean;
var
  clientPos: TPoint;
begin
  Result := inherited PtInSprite(APoint);
  if Result then
  begin
    clientPos.X := APoint.X - Left;
    clientPos.Y := APoint.Y - Top;
    with clientPos, FMask do
      Result :=
        (FMask = nil) or Empty or
        ((X < Width) and (Y < Height) and
        (Canvas.Pixels[X,Y] = clBlack));
    end;
end;

procedure TabcCustomBitmapSprite.Draw(ACanvas: TCanvas);
var
  OffsetX, OffsetY, colX, rowY: Integer;
begin
  if not FBitmap.Empty then
  begin
    if FGlyphIndex = 1 then
    begin
      OffsetX := 0;
      OffsetY := 0;
    end
    else
    begin
      colX := FGlyphIndex mod FGlyphsX;
      rowY := FGlyphIndex div FGlyphsX;
      if colX = 0 then
        colX := FGlyphsX
      else
        rowY := rowY + 1;
      OffsetX := Width * (colX - 1);
      OffsetY := Height * (rowY - 1);
    end;
    FBitmap.Canvas.Lock;
    try

      abcTransparentBlt(
        ACanvas.Handle, Left, Top, Width, Height,
        FBitmap.Canvas.Handle, OffsetX, OffsetY,
        FMask);
        
    finally
      FBitmap.Canvas.Unlock;
    end;
  end;
end;

function TabcCustomBitmapSprite.IsMaskStored: Boolean;
begin
  Result := not MaskLinked;
end;

procedure TabcCustomBitmapSprite.AdjustSize;
begin
  SetBounds(Left, Top, Bitmap.Width div FGlyphsX, Bitmap.Height div FGlyphsY);
end;

procedure TabcCustomBitmapSprite.BitmapChanged(Sender: TObject);
begin
  AdjustSize;
  if FMaskLinked then
    abcCreateSpriteMask(FBitmap, FMask);
  Change(Self);
end;

procedure TabcCustomBitmapSprite.SetGlyphIndex(Value: TabcSpriteGlyphIndex);
begin
  if (Value <> FGlyphIndex) and (Value <= FGlyphsX * FGlyphsY) then
  begin
    FGlyphIndex := Value;
    Change(Self);
  end;
end;

procedure TabcCustomBitmapSprite.SetGlyphsX(Value: TabcSpriteGlyphIndex);
begin
  if (Value <> FGlyphsX) then
  begin
    FGlyphsX := Value;
    if FGlyphIndex > FGlyphsX * FGlyphsY then
      FGlyphIndex := 1;
    AdjustSize;
    Change(Self);
  end;
end;

procedure TabcCustomBitmapSprite.SetGlyphsY(Value: TabcSpriteGlyphIndex);
begin
  if (Value <> FGlyphsY) then
  begin
    FGlyphsY := Value;
    if FGlyphIndex > FGlyphsX * FGlyphsY then
      FGlyphIndex := 1;
    AdjustSize;
    Change(Self);
  end;
end;

procedure TabcCustomBitmapSprite.SetMaskLinked(Value: Boolean);
begin
  if Value <> FMaskLinked then
  begin
    FMaskLinked := Value;
    BitmapChanged(Self);
  end;
end;

procedure TabcCustomBitmapSprite.SetMask(Value: TBitmap);
begin
  if not FSharedBitmaps and (Value <> FMask) then
  begin
    FMask.Assign(Value);
    FMaskLinked := False;
    BitmapChanged(Self);
  end;
end;

procedure TabcCustomBitmapSprite.SetBitmap(Value: TBitmap);
begin

  if not (csDestroying in ComponentState) then
    if not FSharedBitmaps and (Value <> FBitmap) then
    begin
      if (Win32Platform = VER_PLATFORM_WIN32_WINDOWS) and (Win32MajorVersion <= 4) then
      begin
        FBitmap.Free;
        FBitmap := TBitmap.Create;
        if MaskLinked then
        begin
          FMask.Free;
          FMask := TBitmap.Create;
        end;
      end;
      FBitmap.Assign(Value);
      BitmapChanged(Self);
    end;
end;

{ TabcSharedBitmapSprite }

procedure TabcSharedBitmapSprite.Change(Sender: TObject);
begin
  if Sender = Owner then
    BitmapChanged(Self)
  else
    inherited;
end;

{
---------------------------- TabcPictureSprite ---------------------------------
}

constructor TabcPictureSprite.Create(AOwner: TComponent);
begin
  inherited;
  FPicture := TPicture.Create;
  FPicture.OnChange := PictureChanged;
  FMaskLinked := True;
end;

destructor TabcPictureSprite.Destroy;
begin
  Destroying;
  FPicture.Free;
  inherited;
end;

procedure TabcPictureSprite.PictureChanged(Sender: TObject);
var
  G: TGraphic;
begin
  with Bitmap do
  begin
    Width := 0;
    Height := 0;
    Width := Picture.Width;
    Height := Picture.Height;
    G := FPicture.Graphic;
    if (G <> nil) and not ((G is TMetaFile) or (G is TIcon)) then
      G.Transparent := True;
    Canvas.Brush.Color := ABC_TRANS_COLOR;
    Canvas.Draw(0,0,FPicture.Graphic);
    Transparent := True;
  end;
  abcCreateSpriteMask(Bitmap, Mask);
  AdjustSize;
  Change(Self);
end;

procedure TabcPictureSprite.SetPicture(Value: TPicture);
begin
  if not (csDestroying in ComponentState) then
    FPicture.Assign(Value);
end;

{
-------------------------- TabcImageListSprite ---------------------------------
}

constructor TabcImageListSprite.Create(AOwner: TComponent);
begin
  inherited;
  FChangeLink := TChangeLink.Create;
  FChangeLink.OnChange := ImageListChanged;
end;

destructor TabcImageListSprite.Destroy;
begin
  SetImages(nil);
  FChangeLink.Free;
  inherited;
end;

procedure TabcImageListSprite.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (Operation = opRemove) then
    if (AComponent = FImages) then
      SetImages(nil);
end;

procedure TabcImageListSprite.Animate(Frame: Integer; var oldBounds, newBounds: TRect);
begin
  if (DropFrames > 0) and (Frame mod DropFrames <> 1) then Exit;
  if FAnimateGlyphs then
    if FImages <> nil then
    begin
      FImageIndex := FImageIndex + 1;
      if FImageIndex > (FImages.Count-1) then
        FImageIndex := 0;
      DrawInNextFrame := True;
    end;
  inherited;
end;

procedure TabcImageListSprite.Draw(ACanvas: TCanvas);
begin
  if (FImages <> nil) then
    FImages.Draw(ACanvas, Left, Top, FImageIndex);
end;

procedure TabcImageListSprite.AdjustSize;
begin
  if FImages <> nil then
    SetBounds(Left, Top, FImages.Width, FImages.Height)
  else
    SetBounds(Left, Top, 0, 0);
end;

procedure TabcImageListSprite.SetImageIndex(Value: TImageIndex);
begin
  if (Value <> FImageIndex) then
  begin
    FImageIndex := Value;
    Change(Self);
  end;
end;

procedure TabcImageListSprite.SetImages(Value: TImageList);
begin
  if Value <> FImages then
  begin
    if FImages <> nil then FImages.UnRegisterChanges(FChangeLink);
    FImages := Value;
    if Value <> nil then
    begin
      Value.RegisterChanges(FChangeLink);
      Value.FreeNotification(Self);
    end;
    if not (csDestroying in ComponentState) then ImageListChanged(Self);
  end;
end;

procedure TabcImageListSprite.ImageListChanged(Sender: TObject);
begin
  AdjustSize;
  Change(Self);
end;

{
---------------------------- TabcBitmapSpriteFactory ---------------------------
}

constructor TabcBitmapSpriteFactory.Create(AOwner: TComponent);
begin
  inherited;
  FBitmap := TBitmap.Create;
  FBitmap.OnChange := BitmapChanged;
  FMask := TBitmap.Create;
  FMask.Monochrome := True;
  FGlyphIndex := 1;
  FGlyphsX := 1;
  FGlyphsY := 1;
end;

destructor TabcBitmapSpriteFactory.Destroy;
begin
  Destroying;
  FBitmap.Free;
  FMask.Free;
  inherited;
end;

class function TabcBitmapSpriteFactory.SpriteClass: TabcSpriteClass;
begin
  Result := TabcSharedBitmapSprite;
end;

function TabcBitmapSpriteFactory.CreateSprite: TabcSprite;
begin
  Result := TabcSharedBitmapSprite.CreateShared(Self, FBitmap, FMask);
  with TabcSharedBitmapSprite(Result) do
  begin
    AnimateGlyphs := Self.AnimateGlyphs;
    GlyphsX := Self.GlyphsX;
    GlyphsY := Self.GlyphsY;
    GlyphIndex := Self.GlyphIndex;
    Mover := Self.Mover;
    SetBounds(FSpriteLeft, FSpriteTop, FSpriteWidth, FSpriteHeight);
  end;
  if SpriteList <> nil then
    SpriteList.Add(Result);
end;

procedure TabcBitmapSpriteFactory.AddNSprites(numToCreate: Byte);
var
  i: integer;
begin
  if SpriteList <> nil then
  begin
    SpriteList.BeginUpdate;
    for i := 1 to numToCreate do
      CreateSprite;
    SpriteList.EndUpdate;
  end;
end;

function TabcBitmapSpriteFactory.IsMaskStored: Boolean;
begin
  Result := not MaskLinked;
end;

procedure TabcBitmapSpriteFactory.AdjustSize;
begin
  FSpriteWidth := Bitmap.Width div FGlyphsX;
  FSpriteHeight := Bitmap.Height div FGlyphsY;
end;

procedure TabcBitmapSpriteFactory.BitmapChanged(Sender: TObject);
begin
  AdjustSize;
  if FMaskLinked then
    abcCreateSpriteMask(FBitmap, FMask);
  Change;
end;

procedure TabcBitmapSpriteFactory.SetGlyphIndex(Value: TabcSpriteGlyphIndex);
begin
  if (Value <> FGlyphIndex) and (Value <= FGlyphsX * FGlyphsY) then
  begin
    FGlyphIndex := Value;
  end;
end;

procedure TabcBitmapSpriteFactory.SetGlyphsX(Value: TabcSpriteGlyphIndex);
begin
  if (Value <> FGlyphsX) then
  begin
    FGlyphsX := Value;
    if FGlyphIndex > FGlyphsX * FGlyphsY then
      FGlyphIndex := 1;
    AdjustSize;
  end;
end;

procedure TabcBitmapSpriteFactory.SetGlyphsY(Value: TabcSpriteGlyphIndex);
begin
  if (Value <> FGlyphsY) then
  begin
    FGlyphsY := Value;
    if FGlyphIndex > FGlyphsX * FGlyphsY then
      FGlyphIndex := 1;
    AdjustSize;
  end;
end;

procedure TabcBitmapSpriteFactory.SetMaskLinked(Value: Boolean);
begin
  FMaskLinked := Value;
  if Value then
    abcCreateSpriteMask(FBitmap, FMask);
end;

procedure TabcBitmapSpriteFactory.SetMask(Value: TBitmap);
begin
  FMask.Assign(Value);
  FMaskLinked := False;
end;

procedure TabcBitmapSpriteFactory.SetBitmap(Value: TBitmap);
begin
  FBitmap.Assign(Value);
end;

{
-------------------------------- TabcInputSprite -------------------------------
}

procedure TabcInputSprite.Draw(ACanvas: TCanvas);
begin
  DrawInNextFrame := False;
  if (csDesigning in ComponentState) then
    with ACanvas do
    begin
      with Pen do
      begin
        Color := clBlack;
        Mode := pmCopy;
        Style := psDash;
        Width := 1;
      end;
      Brush.Style := bsClear;
      Rectangle(Left, Top, Left+Width, Top+Height);
    end;
end;

{
-------------------------------- TabcInputSprite -------------------------------
}

constructor TabcInputSprite.Create(AOwner: TComponent);
begin
  FNeverDraw := True;
  inherited;
end;

{
-------------------------------- TabcShapeSprite -------------------------------
}

constructor TabcShapeSprite.Create(AOwner: TComponent);
begin
  inherited;
  FPen := TPen.Create;
  FPen.OnChange := StyleChanged;
  FBrush := TBrush.Create;
  FBrush.OnChange := StyleChanged;
  FFont := TFont.Create;
  FFont.OnChange := StyleChanged;
  FColorHighlight := clBtnHighlight;
  FColorShadow := clBtnShadow;
  Width := 185;
  Height := 41;
  FAlignment := taCenter;
  FAlignVertical := vaCenter;
  BevelInner := bvNone;
  BevelOuter := bvNone;
  BevelWidth := 1;
  Brush.Color := clBtnFace;
  CornerWidth := 20;
end;

destructor TabcShapeSprite.Destroy;
begin
  FFont.Free;
  FPen.Free;
  FBrush.Free;
  inherited;
end;

procedure TabcShapeSprite.Draw(ACanvas: TCanvas);
var
  X, Y, W, H, S, C: Integer;
  TextFormat: Word;
  ARect, ATextRect: TRect;
  TopColor, BottomColor: TColor;
  FontHeight: Integer;

const
  Alignments: array[TAlignment] of Word = (DT_LEFT, DT_RIGHT, DT_CENTER);
  VertAlignments: array[TabcVerticalAlign] of Word = (DT_TOP, DT_VCENTER, DT_BOTTOM);

  procedure AdjustColors(Bevel: TPanelBevel);
  begin
    TopColor := FColorHighlight;
    if Bevel = bvLowered then TopColor := FColorShadow;
    BottomColor := FColorShadow;
    if Bevel = bvLowered then BottomColor := FColorHighlight;
  end;

begin
  with ACanvas do begin
    Pen := self.Pen;
    Brush := self.Brush;
    X := Pen.Width div 2;
    Y := Top + X;
    X := Left + X;
    W := Width - Pen.Width + 1;
    H := Height - Pen.Width + 1;
    if W < H then S := W else S := H;
    if W < H then C := W div 2 else C := H div 2;
    if C > FCornerWidth then C := FCornerWidth;
    if Shape in [astSquare, astRoundSquare, astCircle] then begin
      Inc(X, (W - S) div 2);
      Inc(Y, (H - S) div 2);
      W := S;
      H := S;
    end;
    case Shape of
      astRectangle, astSquare:
        Rectangle(X, Y, X + W, Y + H);
      astRoundRect, astRoundSquare:
        RoundRect(X, Y, X + W, Y + H, 2 * C, 2 * C);
      astCircle, astEllipse:
        Ellipse(X, Y, X + W, Y + H);
    end;
  end;

  ARect := Rect(X + Pen.Width, Y + Pen.Width, X + W - Pen.Width - 1, Y + H - Pen.Width - 1);

  {draw Bevels}
  ACanvas.Pen.Mode := pmCopy;
  ACanvas.Pen.Style := psSolid;

  if BevelOuter <> bvNone then begin
    AdjustColors(BevelOuter);
    case Shape of
      astRectangle, astSquare:
        abcRoundFrame3D(ACanvas, ARect, TopColor, BottomColor, BevelWidth, 0);
      astRoundRect, astRoundSquare:
        abcRoundFrame3D(ACanvas, ARect, TopColor, BottomColor, BevelWidth, C);
      astCircle, astEllipse:
        abcEllipseFrame3D(ACanvas, ARect, TopColor, BottomColor, BevelWidth);
    end;
  end;
  if BorderWidth <> 0 then
    case Shape of
      astRectangle, astSquare:
        abcRoundFrame3D(ACanvas, ARect, Brush.Color, Brush.Color, BorderWidth, 0);
      astRoundRect, astRoundSquare:
        abcRoundFrame3D(ACanvas, ARect, Brush.Color, Brush.Color, BorderWidth, C);
      astCircle, astEllipse:
        abcEllipseFrame3D(ACanvas, ARect, Brush.Color, Brush.Color, BorderWidth);
    end;

  abcRoundFrame3D(ACanvas, ARect, Brush.Color, Brush.Color, BorderWidth, C);
  if BevelInner <> bvNone then begin
    AdjustColors(BevelInner);
    case Shape of
      astRectangle, astSquare:
        abcRoundFrame3D(ACanvas, ARect, TopColor, BottomColor, BevelWidth, 0);
      astRoundRect, astRoundSquare:
        abcRoundFrame3D(ACanvas, ARect, TopColor, BottomColor, BevelWidth, C);
      astCircle, astEllipse:
        abcEllipseFrame3D(ACanvas, ARect, TopColor, BottomColor, BevelWidth);
    end;
  end;

  {draw Text}
  with ARect do begin
    Inc(Left);
    Inc(Top);
    Dec(Right);
    Dec(Bottom);
  end;
  if Shape in [astCircle, astEllipse] then
    {fit text to inside of ellipse}
    with ARect do begin
      W := (Right - Left) div 2;
      H := (Bottom - Top) div 2;
      {calculate dX, dY}
      X := W - Round(W / sqrt(2));
      Y := H - Round(H / sqrt(2));
      Inc(Left, X);
      Inc(Top, Y);
      Dec(Right, X);
      Dec(Bottom, Y);
    end
  else if Shape in [astRoundRect, astRoundSquare] then
    with ARect do begin
      C := Round( C * (1 - 1 / sqrt(2)));
      Inc(Left, C);
      Inc(Top, C);
      Dec(Right, C);
      Dec(Bottom, C);
    end;
  ATextRect := ARect;
  with ACanvas do begin
    Pen := self.Pen;
    Brush := self.Brush;
    Font := self.Font;
    TextFormat := DT_EXPANDTABS or Alignments[Alignment];
    if FWordWrap then begin
      TextFormat := TextFormat or DT_WORDBREAK or DT_CALCRECT;
      DrawText(Handle, PChar(Caption), -1, ATextRect, TextFormat);
      with ATextRect do begin
        if Right > ARect.Right then
          Right := ARect.Right;
        case FAlignVertical of
          vaTop: begin
            Inc(Top);
            Inc(Bottom);
          end;
          vaCenter: begin
            FontHeight := Bottom - Top;
            Top := (Height - FontHeight) div 2;
            Bottom := (Height + FontHeight) div 2;
          end;
          vaBottom: begin
            FontHeight := Bottom - Top;
            Top := ARect.Bottom - FontHeight;
            Bottom := ARect.Bottom;
          end;
        end;
        if Top < ARect.Top then begin
          Bottom := Bottom + (ARect.Top - Top);
          Top := ARect.Top;
        end;
        if Bottom > ARect.Bottom then Bottom := ARect.Bottom;
        {adjust X difference}
        X := (ARect.Right - ARect.Left) - (Right - Left);
        if (X > 0) then
          case Alignment of
            taLeftJustify:
              {do nothing};
            taCenter:
              with ATextRect do begin
                X := X div 2;
                Inc(Left, X);
                Inc(Right, X);
              end;
            taRightJustify:
              with ATextRect do begin
                Inc(Left, X);
                Inc(Right, X);
              end;
          end;
      end;
      TextFormat := TextFormat and not DT_CALCRECT;
      DrawText(Handle, PChar(Caption), -1, ATextRect, TextFormat);
    end
    else begin
      TextFormat := TextFormat or VertAlignments[FAlignVertical] or DT_SINGLELINE;
      DrawText(Handle, PChar(Caption), -1, ATextRect, TextFormat);
    end;
  end;
end;

function TabcShapeSprite.IsFontStored: Boolean;
begin
  with Font do
    Result :=
      (Charset <> DEFAULT_CHARSET) or
      (Color <> clWindowText) or
      (Height <> -11) or
      (Name <> 'MS Sans Serif') or
      (Style <> []);
end;

procedure TabcShapeSprite.SetAlignment(const Value: TAlignment);
begin
  if Value <> FAlignment then
  begin
    FAlignment := Value;
    Change(Self);
  end;
end;

procedure TabcShapeSprite.SetAlignVertical(const Value: TabcVerticalAlign);
begin
  if Value <> FAlignVertical then
  begin
    FAlignVertical := Value;
    if (Value = vaCenter) and WordWrap then
      FWordWrap := False;
    Change(Self);
  end;
end;

procedure TabcShapeSprite.SetBevelInner(const Value: TPanelBevel);
begin
  if Value <> FBevelInner then
  begin
    FBevelInner := Value;
    Change(Self);
  end;
end;

procedure TabcShapeSprite.SetBevelOuter(const Value: TPanelBevel);
begin
  if Value <> FBevelOuter then
  begin
    FBevelOuter := Value;
    Change(Self);
  end;
end;

procedure TabcShapeSprite.SetBevelWidth(const Value: TBevelWidth);
begin
  if Value <> FBevelWidth then
  begin
    FBevelWidth := Value;
    Change(Self);
  end;
end;

procedure TabcShapeSprite.SetBorderWidth(const Value: TBorderWidth);
begin
  if Value <> FBorderWidth then
  begin
    FBorderWidth := Value;
    Change(Self);
  end;
end;

procedure TabcShapeSprite.SetBrush(Value: TBrush);
begin
  FBrush.Assign(Value);
end;

procedure TabcShapeSprite.SetCaption(const Value: string);
begin
  if Value <> FCaption then
  begin
    FCaption := Value;
    Change(Self);
  end;
end;

procedure TabcShapeSprite.SetColorHighlight(const Value: TColor);
begin
  if Value <> FColorHighlight then
  begin
    FColorHighlight := Value;
    Change(Self);
  end;
end;

procedure TabcShapeSprite.SetColorShadow(const Value: TColor);
begin
  if Value <> FColorShadow then
  begin
    FColorShadow := Value;
    Change(Self);
  end;
end;

procedure TabcShapeSprite.SetCornerWidth(const Value: TabcCornerWidth);
var
  CalcWidth: Integer;
begin
  CalcWidth := Value;
  if CalcWidth > Height div 2 then
    CalcWidth := Height div 2;
  if CalcWidth > Width div 2 then
    CalcWidth := Width div 2;
  FCornerWidth := CalcWidth;
  Change(Self);
end;

procedure TabcShapeSprite.SetFont(const Value: TFont);
begin
  FFont.Assign(Value);
end;

procedure TabcShapeSprite.SetPen(Value: TPen);
begin
  FPen.Assign(Value);
end;

procedure TabcShapeSprite.SetShape(Value: TabcShapeType);
begin
  if Value <> FShape then
  begin
    FShape := Value;
    Change(Self);
  end;
end;

procedure TabcShapeSprite.SetWordWrap(const Value: Boolean);
begin
  if Value <> FWordWrap then
  begin
    FWordWrap := Value;
    if Value and (AlignVertical = vaCenter) then
      FAlignVertical := vaTop;
    Change(Self);
  end;
end;

procedure TabcShapeSprite.StyleChanged(Sender: TObject);
begin
  Change(Self);
end;

procedure TabcShapeSprite.SetBounds(ALeft, ATop, AWidth, AHeight: Integer);
begin
  inherited SetBounds(ALeft, ATop, AWidth, AHeight);
  SetCornerWidth(FCornerWidth);
end;

end.
