unit abcsprtx;

(*------------------------------------------------------------------------------

Advanced Business Components
ABC for Delphi
Version 6

Copyright  1995, 2001 Objective Software Technology Pty Limited ACN 068 640 353
Functionality derived from Delphi VCL Source Code
is Copyright  1995, 2001 Borland Software Corporation

------------------------------------------------------------------------------*)

{$I abcdefs.inc}

{:Contains sprite components that display animated text.
This unit contains class for scrolling and rotated text sprites.
}

interface

uses
  SysUtils, Windows, Messages, Classes, Graphics, Controls, StdCtrls,
  abcuspr, abcspr, abclabel, abcutilg;

type

  { TabcTextSprite }

  {::The directions which text can scroll in <See Class = TabcTextSprite>.}

  TabcTextScrollDirection = (atsNone, atsLeft, atsRight, atsUp, atsDown);

  {:A sprite with a text caption and scrolling animation.
  TabcTextSprite draws a <See property=Caption> using its Font.<br>
  The position of the <See property=Caption> is set using the <See property=Alignment>,
  <See property=Layout> and <See property=WordWrap> properties.<br>
  <See property=AutoSize>, or <See property=Width> and <See property=Height> can be used to adjust the sprite size.<br>
  The <See property=Caption> can be drawn raised, lowered or shadowed using the <See property=CaptionStyle> property.<br>
  The sprite can be <See property=Transparent> or filled using <See property=Color>.<br>
  Scrolling can be set using the <See property=Scroll> and <See property=ScrollPixelsPerFrame> properties.
  }

  TabcTextSprite = class (TabcSprite)
  private
    FAdjustingBounds: Boolean;
    FBitmap, FMask: TBitmap;
    FCaption: string;
    FCaptionStyle: TabcCaptionStyle;
    FFont: TFont;
    FScrollOffsetX, FScrollOffsetY: integer;
    FAutoSize: Boolean;
    FWordWrap: Boolean;
    FScrollPixelsPerFrame: Byte;
    FScroll: TabcTextScrollDirection;
    FAlignment: TAlignment;
    FLayout: TTextLayout;
    FTransparent: Boolean;
    FColor: TColor;
    FBiDiMode: TBiDiMode;
    procedure AdjustBounds;
    function IsFontStored: Boolean;
    procedure SetCaption(const Value: string);
    procedure SetCaptionStyle(const Value: TabcCaptionStyle);
    procedure SetFont(const Value: TFont);
    procedure SetAlignment(const Value: TAlignment);
    procedure SetAutoSize(const Value: Boolean);
    procedure SetLayout(const Value: TTextLayout);
    procedure SetWordWrap(const Value: Boolean);
    procedure StyleChanged(Sender: TObject);
    procedure SetColor(const Value: TColor);
    procedure SetTransparent(const Value: Boolean);
    function GetTextFlags: Integer;
    procedure SetBiDiMode(const Value: TBiDiMode);
  protected
    procedure Animate(Frame: Integer; var oldBounds, newBounds: TRect); override;
    procedure AnimateBegin; override;
    procedure AnimateEnd; override;
    procedure Draw(ACanvas: TCanvas); override;
    procedure Loaded; override;
    procedure RefreshBitmap;
    procedure DrawOffsetText(ACanvas: TCanvas; DrawR: TRect; Offset: Integer; Emboss, Shadow: Boolean);
    procedure GetTextBounds(var R: TRect);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure SetBounds(ALeft, ATop, AWidth, AHeight: Integer); override;
    {::This function is overridden to return true only for points within the caption text.}
    function PtInSprite(const APoint: TPoint): Boolean; override;
    {::This method resets the text to the original unscrolled position.}
    procedure ResetScrollPosition;
    {::Offset position to draw the text. This is adjusted automatically when animating.}
    property ScrollOffsetX: Integer read FScrollOffsetX write FScrollOffsetX;
    {::Offset position to draw the text. This is adjusted automatically when animating.}
    property ScrollOffsetY: Integer read FScrollOffsetY write FScrollOffsetY;
  published
    property Width;
    property Height;
    property Mover;
    property Visible;
    {::Alignment of the text.<p>
    TAlignment = (taLeftJustify, taRightJustify, taCenter);<P>
    The default is taCenter.}
    property Alignment: TAlignment read FAlignment write SetAlignment default taCenter;
    {::When true the Width and Height adjust to fit the Caption.}
    property AutoSize: Boolean read FAutoSize write SetAutoSize default True;
    property BiDiMode: TBiDiMode read FBiDiMode write SetBiDiMode default bdLeftToRight;
    {::The text to be displayed.}
    property Caption: string read FCaption write SetCaption;
    {::Additional style effects for raised, lowered, shadowed or embossed text.}
    property CaptionStyle: TabcCaptionStyle read FCaptionStyle write SetCaptionStyle;
    {::The fill color to draw if the label is not Transparent.}
    property Color: TColor read FColor write SetColor default clBtnFace;
    {::The font used to draw the Caption.}
    property Font: TFont read FFont write SetFont stored IsFontStored;
    {::The vertical alignment of the Caption.<P>
    TTextLayout = (tlTop, tlCenter, tlBottom);<P>
    The default is tlTop.}
    property Layout: TTextLayout read FLayout write SetLayout default tlTop;
    {::The direction to automatically scroll the Caption when animating.<P>
    TabcTextScrollDirection = (atsNone, atsLeft, atsRight, atsUp, atsDown).<P>
    The default is atsNone.}
    property Scroll: TabcTextScrollDirection read FScroll write FScroll default atsNone;
    {::The number of pixels to scroll the caption by on each frame.<P>
    This only applies if the Scroll property is set to a direction.}
    property ScrollPixelsPerFrame: Byte read FScrollPixelsPerFrame write FScrollPixelsPerFrame default 1;
    {::Whether the label is transparent. Defaults to true.}
    property Transparent: Boolean read FTransparent write SetTransparent default True;
    {::Set to true to wrap the Caption to multiple lines within a fixed Width and Height.}
    property WordWrap: Boolean read FWordWrap write SetWordWrap default False;
  end;

  { TabcRotateTextSprite }

  {:A sprite with a text caption and rotation.
  TabcRotateTextSprite draws a <See property=Caption> using its Font.<br>
  The text can be rotated using the <See property=Angle> property.<br>
  True type fonts must be used for rotated text.
  }
  TabcRotateTextSprite = class (TabcSprite)
  private
    FAdjustingBounds: Boolean;
    FAngle: TabcDegrees;
    FAngleSin, FAngleCos: Double;
    FDrawTextPt: TPoint;
    FBorderPoints: array[0..4] of TPoint;
    FCaption: string;
    FBorderStyle: TabcBorderStyle;
    FCaptionStyle: TabcCaptionStyle;
    FFont: TFont;
    procedure StyleChanged(Sender: TObject);
    function IsFontStored: Boolean;
    procedure SetAngle(const Value: TabcDegrees);
    procedure SetBorderStyle(const Value: TabcBorderStyle);
    procedure SetCaption(const Value: string);
    procedure SetCaptionStyle(const Value: TabcCaptionStyle);
    procedure SetFont(const Value: TFont);
    procedure ReadDrawBorderPts(Reader: TReader);
    procedure ReadDrawTextPt(Reader: TReader);
    procedure ReadHeight(Reader: TReader);
    procedure ReadWidth(Reader: TReader);
    procedure WriteDrawBorderPts(Writer: TWriter);
    procedure WriteDrawTextPt(Writer: TWriter);
    procedure WriteHeight(Writer: TWriter);
    procedure WriteWidth(Writer: TWriter);
  protected
    procedure DefineProperties(Filer: TFiler); override;
    procedure Draw(ACanvas: TCanvas); override;
    procedure AdjustBounds;
    procedure DrawBorder(ACanvas: TCanvas);
    procedure DrawNormalText(ACanvas: TCanvas);
    procedure DrawOffsetText(ACanvas: TCanvas; Offset: Integer; Emboss, Shadow: Boolean);
    procedure DrawCaptionAt(ACanvas: TCanvas; AColor: TColor; APoint: TPoint);
    function  GetRotatedFont: HFont;
    procedure GetTextBounds(var R: TRect);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure SetBounds(ALeft, ATop, AWidth, AHeight: Integer); override;
    function PtInSprite(const APoint: TPoint): Boolean; override;
    property Width;
    property Height;
  published
    {::The angle to rotate the text.<p>
    TabcDegrees = 0..359;}
    property Angle: TabcDegrees read FAngle write SetAngle default 0;
    {::Draws a border around the text.}
    property BorderStyle: TabcBorderStyle read FBorderStyle write SetBorderStyle;
    {::The text to be drawn.}
    property Caption: string read FCaption write SetCaption;
    {::Additional style effects for raised, lowered, shadowed or embossed text.}
    property CaptionStyle: TabcCaptionStyle read FCaptionStyle write SetCaptionStyle;
    {::The font used to draw the Caption. Note True Type fonts must be used for rotated text.}
    property Font: TFont read FFont write SetFont stored IsFontStored;
    property Mover;
    property Visible;
  end;

implementation

uses Math;

{ TabcTextSprite }

constructor TabcTextSprite.Create(AOwner: TComponent);
begin
  inherited;
  FBitmap := TBitmap.Create;
  FBitmap.PixelFormat := pfDevice;
  FMask := TBitmap.Create;
  FMask.Monochrome := True;
  FCaptionStyle := TabcCaptionStyle.Create;
  FCaptionStyle.OnChange := StyleChanged;
  FFont := TFont.Create;
  FFont.OnChange := StyleChanged;
  FAlignment := taCenter;
  FAutoSize := True;
  FColor := clBtnFace;
  FLayout := tlTop;
  FScrollPixelsPerFrame := 1;
  FTransparent := True;
end;

destructor TabcTextSprite.Destroy;
begin
  FBitmap.Free;
  FMask.Free;
  FCaptionStyle.Free;
  FFont.Free;
  inherited;
end;

procedure TabcTextSprite.Loaded;
begin
  inherited;
  RefreshBitmap;
end;

procedure TabcTextSprite.Animate(Frame: Integer; var oldBounds,
  newBounds: TRect);
begin
  CopyRect(oldBounds, BoundsRect);
  if (DropFrames > 0) and (Frame mod DropFrames <> 1) then Exit;
  Mover.MoveSprite;
  case Scroll of
    atsNone:
      Exit;
    atsLeft:
      begin
        FScrollOffsetX := FScrollOffsetX + ScrollPixelsPerFrame;
        if FScrollOffsetX >= FBitmap.Width then
          FScrollOffsetX := - FBitmap.Width;
      end;
    atsRight:
      begin
        FScrollOffsetX := FScrollOffsetX - ScrollPixelsPerFrame;
        if FScrollOffsetX <= - FBitmap.Width then
          FScrollOffsetX := FBitmap.Width;
      end;
    atsUp:
      begin
        FScrollOffsetY := FScrollOffsetY + ScrollPixelsPerFrame;
        if FScrollOffsetY >= FBitmap.Height then
          FScrollOffsetY := - FBitmap.Height;
      end;
    atsDown:
      begin
      FScrollOffsetY := FScrollOffsetY - ScrollPixelsPerFrame;
        if FScrollOffsetY <= - FBitmap.Height then
          FScrollOffsetY := FBitmap.Height;
      end;
  end;
  DrawInNextFrame := True;
  if Assigned(OnAnimate) then
    OnAnimate(Self, Frame);
  CopyRect(newBounds, BoundsRect);
end;

procedure TabcTextSprite.AnimateBegin;
begin
  // init scroll position to start off screen and scroll into view
  case Scroll of
    atsNone: {do nothing};
    atsLeft:
      FScrollOffsetX := - FBitmap.Width;
    atsRight:
      FScrollOffsetX := FBitmap.Width;
    atsUp:
      FScrollOffsetY := - FBitmap.Height;
    atsDown:
      FScrollOffsetY := FBitmap.Height;
  end;
  inherited;
end;

procedure TabcTextSprite.AnimateEnd;
begin
  ResetScrollPosition;
  inherited;
end;

procedure TabcTextSprite.Draw(ACanvas: TCanvas);
var
  OffsetX, OffsetY: Integer;
  destX, destY, destW, destH, srcX, srcY: integer;
begin
  if not FBitmap.Empty then
  begin
    OffsetX := 0;
    OffsetY := 0;
    if not AutoSize then
    begin
      case Alignment of
        taCenter:
          OffsetX := (FBitmap.Width - Width) div 2;
        taLeftJustify:
          {do nothing};
        taRightJustify:
          OffsetX := (FBitmap.Width - Width);
      end;
      case Layout of
        tlTop:
          {do nothing};
        tlCenter:
          OffsetY := (FBitmap.Height - Height) div 2;
        tlBottom:
          OffsetY := (FBitmap.Height - Height);
      end;
    end;
    OffsetX := OffsetX + FScrollOffsetX;
    OffsetY := OffsetY + FScrollOffsetY;

    if OffsetX < 0 then
    begin
      destX := Left - OffsetX;
      destW := Width + OffsetX;
      srcX := 0;
    end
    else
    begin
      destX := Left;
      destW := FBitmap.Width - OffsetX;
      srcX := OffsetX;
    end;

    if OffsetY < 0 then
    begin
      destY := Top - OffsetY;
      destH := Height + OffsetY;
      srcY := 0;
    end
    else
    begin
      destY := Top;
      destH := FBitmap.Height - OffsetY;
      srcY := OffsetY;
    end;

    destW := Min(destW, FBitmap.Width);
    destH := Min(destH, FBitmap.Height);

    if not AutoSize then
    begin
      destW := Min(destW, Width);
      destH := Min(destH, Height);
    end;

    if Transparent then
      abcTransparentBlt(
        ACanvas.Handle, destX, destY, destW, destH,
        FBitmap.Canvas.Handle, srcX, srcY,
        FMask)
    else
      with ACanvas do
      begin
        Brush.Style := bsSolid;
        Brush.Color := Self.Color;
        FillRect(BoundsRect);
        BitBlt(Handle, destX, destY, destW, destH, FBitmap.Canvas.Handle, srcX, srcY, SRCCOPY);
      end;
  end;
end;

procedure TabcTextSprite.DrawOffsetText(ACanvas: TCanvas; DrawR: TRect; Offset: Integer;
  Emboss, Shadow: Boolean);
var
  R: TRect;
  Flags: integer;
begin
  CopyRect(R, DrawR);
  Flags := GetTextFlags;
  with ACanvas do begin
    if not Shadow then begin
      OffsetRect(R, -Offset, -Offset);
      Font.Color := FCaptionStyle.ColorHighlight;
      DrawText(Handle, PChar(Caption), -1, R, Flags);
      OffsetRect(R, 2*Offset, 2*Offset);
    end
    else
      OffsetRect(R, Offset, Offset);

    Font.Color := FCaptionStyle.ColorShadow;
    DrawText(Handle, PChar(Caption), -1, R, Flags);
    OffsetRect(R, -Offset, -Offset);
    if Emboss then
      if Transparent then
        Font.Color := ABC_TRANS_COLOR
      else
        Font.Color := Self.Color
    else
      Font.Color := Self.Font.Color;
    DrawText(Handle, PChar(Caption), -1, R, Flags);
  end;
end;

function TabcTextSprite.GetTextFlags: Integer;
const
  Alignments: array[TAlignment] of Word = (DT_LEFT, DT_RIGHT, DT_CENTER);
  WordWraps: array[Boolean] of Word = (0, DT_WORDBREAK);
begin
  Result := DT_NOPREFIX or DT_EXPANDTABS or WordWraps[FWordWrap] or Alignments[FAlignment];
  if SysLocale.MiddleEast then
  begin
    if BiDiMode = bdRightToLeft then
      if Result and DT_RIGHT = DT_RIGHT then
        Result := Result and not DT_RIGHT
      else if not (Result and DT_CENTER = DT_CENTER) then
        Result := Result or DT_RIGHT;
    if BiDiMode <> bdLeftToRight then
      Result := Result or DT_RTLREADING;
  end;
end;

procedure TabcTextSprite.GetTextBounds(var R: TRect);
var
  DC: HDC;
  SaveFont: HFont;
  adjustBy: integer;
begin
  DC := GetDC(0);
  SaveFont := SelectObject(DC, Font.Handle);
  try
    DrawText(DC, PChar(Caption), -1, R, GetTextFlags or DT_CALCRECT);
  finally
    SelectObject(DC, SaveFont);
    ReleaseDC(0, DC);
  end;
  if CaptionStyle.Effect <> ceNormal then
  begin
    adjustBy := CaptionStyle.BoundsAdjustment;
    R.Right := R.Right + 2 * adjustBy;
    R.Bottom := R.Bottom + 2 * adjustBy;
  end;
end;

function TabcTextSprite.IsFontStored: Boolean;
begin
  with Font do
    Result :=
      (Charset <> DEFAULT_CHARSET) or
      (Color <> clWindowText) or
      (Height <> -11) or
      (Name <> 'MS Sans Serif') or
      (Style <> []);
end;

function TabcTextSprite.PtInSprite(const APoint: TPoint): Boolean;
var
  clientPos: TPoint;
begin
  Result := inherited PtInSprite(APoint);
  if Result then
  begin
    clientPos.X := APoint.X - Left;
    clientPos.Y := APoint.Y - Top;
    with clientPos, FMask do
      Result :=
        (FMask = nil) or Empty or
        ((X < Width) and (Y < Height) and
        (Canvas.Pixels[X,Y] = clBlack));
    end;
end;

procedure TabcTextSprite.RefreshBitmap;
var
  TextR: TRect;
  Offset: Integer;
  adjustBy: integer;
begin
  //calc bitmap size for text
  TextR := Rect(0,0,Width, Height);
  GetTextBounds(TextR);
  //draw text using CaptionStyle
  with FBitmap do
  begin
    Width := TextR.Right;
    Height := TextR.Bottom;
    with Canvas do
    begin
      if Self.Transparent and (CaptionStyle.Effect in [ceEmbossRaised, ceEmbossLowered]) then
        Brush.Color := ABC_TRANS_COLOR
      else
        Brush.Color := Color;
      Brush.Style := bsSolid;
      FillRect(TextR);

      if Caption <> '' then
      begin
        Brush.Style := bsClear;
        Font := Self.Font;

        Offset := FCaptionStyle.Offset;

        if CaptionStyle.Effect <> ceNormal then
        begin
          adjustBy := CaptionStyle.BoundsAdjustment;
          TextR.Left := adjustBy;
          TextR.Top := adjustBy;
        end;

        case FCaptionStyle.Effect of
          ceNormal:
            DrawText(Handle, PChar(Caption), -1, TextR, GetTextFlags);
          ceRaised:
            DrawOffsetText(Canvas, TextR, Offset, False, False);
          ceLowered:
            DrawOffsetText(Canvas, TextR, -Offset, False, False);
          ceEmbossRaised:
            DrawOffsetText(Canvas, TextR, Offset, True, False);
          ceEmbossLowered:
            DrawOffsetText(Canvas, TextR, -Offset, True, False);
          ceShadowed:
            DrawOffsetText(Canvas, TextR, Offset, False, True);
        end;
      end;
    end;
  end;
  if Transparent then
    abcCreateSpriteMask(FBitmap, FMask);
  AdjustBounds;
  Change(Self);
end;

procedure TabcTextSprite.AdjustBounds;
begin
  if AutoSize then
  begin
    FAdjustingBounds := True;
    if not FBitmap.Empty then
      SetBounds(Left, Top, FBitmap.Width, FBitmap.Height);
    FAdjustingBounds := False;
  end;
end;

procedure TabcTextSprite.ResetScrollPosition;
begin
  FScrollOffsetX := 0;
  FScrollOffsetY := 0;
  Change(Self);
end;

procedure TabcTextSprite.SetAlignment(const Value: TAlignment);
begin
  if Value <> FAlignment then
  begin
    FAlignment := Value;
    StyleChanged(nil);
  end;
end;

procedure TabcTextSprite.SetAutoSize(const Value: Boolean);
begin
  if Value <> FAutoSize then
  begin
    FAutoSize := Value;
    if Value and not (csLoading in ComponentState) then
      RefreshBitmap;
  end;
end;

procedure TabcTextSprite.SetBiDiMode(const Value: TBiDiMode);
begin
  if Value <> FBiDiMode then
  begin
    FBiDiMode := Value;
    StyleChanged(nil);
  end;
end;

procedure TabcTextSprite.SetBounds(ALeft, ATop, AWidth, AHeight: Integer);
var
  Resized: Boolean;
begin
  Resized := (AWidth <> Width) or (AHeight <> Height);
  inherited;
  if not ((csLoading in ComponentState) or FAdjustingBounds) and Resized then
  begin
    FAutoSize := False;
    if WordWrap then
      RefreshBitmap;
  end;
end;

procedure TabcTextSprite.SetCaption(const Value: string);
begin
  if Value <> FCaption then
  begin
    FCaption := Value;
    StyleChanged(nil);
  end;
end;

procedure TabcTextSprite.SetCaptionStyle(const Value: TabcCaptionStyle);
begin
  FCaptionStyle.Assign(Value);
  StyleChanged(nil);
end;

procedure TabcTextSprite.SetFont(const Value: TFont);
begin
  FFont.Assign(Value);
  StyleChanged(nil);
end;

procedure TabcTextSprite.SetLayout(const Value: TTextLayout);
begin
  if Value <> FLayout then
  begin
    FLayout := Value;
    StyleChanged(nil);
  end;
end;

procedure TabcTextSprite.SetWordWrap(const Value: Boolean);
begin
  if Value <> FWordWrap then
  begin
    FWordWrap := Value;
    StyleChanged(nil);
  end;
end;

procedure TabcTextSprite.StyleChanged(Sender: TObject);
begin
  if not (csLoading in ComponentState) then
    RefreshBitmap;
end;

procedure TabcTextSprite.SetColor(const Value: TColor);
begin
  if Value <> FColor then
  begin
    FColor := Value;
    if not (csLoading in ComponentState) and not Transparent then
      RefreshBitmap;
  end;
end;

procedure TabcTextSprite.SetTransparent(const Value: Boolean);
begin
  if Value <> FTransparent then
  begin
    FTransparent := Value;
    StyleChanged(nil);
  end;
end;

{ TabcRotateTextSprite }

constructor TabcRotateTextSprite.Create(AOwner: TComponent);
begin
  inherited;
  FCaptionStyle := TabcCaptionStyle.Create;
  FCaptionStyle.OnChange := StyleChanged;
  FBorderStyle := TabcBorderStyle.Create;
  FBorderStyle.OnChange := StyleChanged;
  FFont := TFont.Create;
  //need a truetype font for rotation
  Font.Name := 'Arial';
  Font.Size := 10;
  FFont.OnChange := StyleChanged;
  FAngleCos := 1; {for initial angle 0 degrees, sin is 0}
end;

destructor TabcRotateTextSprite.Destroy;
begin
  FCaptionStyle.Free;
  FBorderStyle.Free;
  FFont.Free;
  inherited;
end;

procedure TabcRotateTextSprite.AdjustBounds;
//calculate rotated text origin, bounds rect and polygon points for borders
var
  TextRect: TRect;
  NewLeft, NewTop: Integer;
begin
  if [csLoading, csReading] * ComponentState = [] then
  begin
    Hide;
    FAdjustingBounds := True;

    NewLeft := Left + Width div 2;
    NewTop := Top + Height div 2;

    TextRect := Rect(0,0,Width,Height);
    GetTextBounds(TextRect);
    if FCaptionStyle.Effect <> ceNormal then
      InflateRect(TextRect, FCaptionStyle.BoundsAdjustment, FCaptionStyle.BoundsAdjustment);
    if FBorderStyle.Effect <> beNone then begin
      InflateRect(TextRect, FBorderStyle.BoundsAdjustment, FBorderStyle.BoundsAdjustment);
      Dec(TextRect.Left, 2);
    end;

    FDrawTextPt := Point(0, 0);
    abcSetRectPoints(TextRect, FBorderPoints);
    abcRotateWinPolygon(FAngleSin, FAngleCos, FDrawTextPt, FBorderPoints);
    FBorderPoints[4] := FBorderPoints[0];

    TextRect := abcGetPolygonBounds(FBorderPoints);

    {inflate bounds to ensure we don't clip on rotation}
    if (FBorderStyle.Effect <> beNone) then
    begin
      Inc(TextRect.Right, FBorderStyle.BoundsAdjustment);
      Inc(TextRect.Bottom, FBorderStyle.BoundsAdjustment);
    end;

    {adjust center}
    abcOffsetPoint(FDrawTextPt, - TextRect.Left, - TextRect.Top);
    abcOffsetPoints(FBorderPoints, - TextRect.Left, - TextRect.Top);
    OffsetRect(TextRect, - TextRect.Left, - TextRect.Top);
    NewLeft := NewLeft - TextRect.Right div 2;
    NewTop := NewTop - TextRect.Bottom div 2;

    abcOffsetPoint(FDrawTextPt, NewLeft, NewTop);
    abcOffsetPoints(FBorderPoints, NewLeft, NewTop);

    SetBounds(NewLeft, NewTop, TextRect.Right, TextRect.Bottom);

    FAdjustingBounds := False;
  end;
end;

procedure TabcRotateTextSprite.Draw(ACanvas: TCanvas);
begin
  if Caption = '' then Exit;
  with ACanvas do
  begin
    if FBorderStyle.Effect <> beNone then
      DrawBorder(ACanvas);
    Brush.Style := bsClear;
    Font := Self.Font;
    case FCaptionStyle.Effect of
      ceNormal:         DrawNormalText(ACanvas);
      ceRaised:         DrawOffsetText(ACanvas, FCaptionStyle.Offset, False, False);
      ceLowered:        DrawOffsetText(ACanvas, -FCaptionStyle.Offset, False, False);
      ceEmbossRaised:   DrawOffsetText(ACanvas, FCaptionStyle.Offset, True, False);
      ceEmbossLowered:  DrawOffsetText(ACanvas, -FCaptionStyle.Offset, True, False);
      ceShadowed:       DrawOffsetText(ACanvas, FCaptionStyle.Offset, False, True);
    end;
  end;
end;

procedure TabcRotateTextSprite.DrawBorder(ACanvas: TCanvas);
var
  DP: array[0..4] of TPoint;
  I: integer;

  procedure DrawRaisedLoweredLines(TopColor, BottomColor: TColor);
  {draw lines using points depending on Angle}
  begin
    with ACanvas do
      case FAngle of
        0..44, 315..359:
          begin
            Pen.Color := TopColor;
            PolyLine([DP[3], DP[0], DP[1]]);
            Pen.Color := BottomColor;
            PolyLine([DP[1], DP[2], DP[3]]);
          end;
        45..134:
          begin
            Pen.Color := TopColor;
            PolyLine([DP[0], DP[1], DP[2]]);
            Pen.Color := BottomColor;
            PolyLine([DP[2], DP[3], DP[0]]);
          end;
        135..224:
          begin
            Pen.Color := TopColor;
            PolyLine([DP[1], DP[2], DP[3]]);
            Pen.Color := BottomColor;
            PolyLine([DP[3], DP[0], DP[1]]);
          end;
        225..314:
          begin
            Pen.Color := TopColor;
            PolyLine([DP[2], DP[3], DP[0]]);
            Pen.Color := BottomColor;
            PolyLine([DP[0], DP[1], DP[2]]);
          end;
      end;
  end;

begin
  for I := 0 to 4 do DP[I] := FBorderPoints[I];
  with ACanvas do begin
    Pen.Mode := pmCopy;
    Pen.Style := psSolid;
    Pen.Width := FBorderStyle.Width;
    Brush.Style := bsClear;
    case FBorderStyle.Effect of
      beLine:
        begin
          Pen.Color := FBorderStyle.ColorShadow;
          Polygon(DP);
        end;
      beRaised:
        begin
          DrawRaisedLoweredLines(FBorderStyle.ColorHighlight, FBorderStyle.ColorShadow);
        end;
      beLowered:
        begin
          DrawRaisedLoweredLines(FBorderStyle.ColorShadow, FBorderStyle.ColorHighlight);
        end;
      beFrameRaised:
        begin
          Brush.Style := bsClear;
          Pen.Color := FBorderStyle.ColorHighlight;
          Polygon(DP);
          Pen.Color := FBorderStyle.ColorShadow;
          abcOffsetPoints(DP, FBorderStyle.Width, FBorderStyle.Width);
          Polygon(DP);
        end;
      beFrameLowered:
        begin
          Brush.Style := bsClear;
          Pen.Color := FBorderStyle.ColorShadow;
          Polygon(DP);
          Pen.Color := FBorderStyle.ColorHighlight;
          abcOffsetPoints(DP, FBorderStyle.Width, FBorderStyle.Width);
          Polygon(DP);
        end;
    end;
  end;
end;

procedure TabcRotateTextSprite.DrawCaptionAt(ACanvas: TCanvas; AColor: TColor; APoint: TPoint);
var
  SaveFont: HFont;
begin
  with ACanvas do begin
    Font.Color := AColor;
    if Angle = 0 then
      TextOut(APoint.X, APoint.Y, Caption)
    else
    begin
      SaveFont := SelectObject(Handle, GetRotatedFont);
      try
        TextOut(APoint.X, APoint.Y, Caption);
      finally
        SaveFont := SelectObject(Handle, SaveFont);
        DeleteObject(SaveFont);
      end;
    end;
  end;
end;

procedure TabcRotateTextSprite.DrawNormalText(ACanvas: TCanvas);
begin
  DrawCaptionAt(ACanvas, Font.Color, FDrawTextPt);
end;

procedure TabcRotateTextSprite.DrawOffsetText(ACanvas: TCanvas; Offset: Integer; Emboss, Shadow: Boolean);
var
  DrawPt: TPoint;
  DrawColor: TColor;
  AdjustOffsetX, AdjustOffsetY: Integer;
begin
  case Angle of
    0..45, 316..359:
      begin
        AdjustOffsetX := Offset;
        AdjustOffsetY := Offset;
      end;
    46..135:
      begin
        AdjustOffsetX := Offset;
        AdjustOffsetY := -Offset;
      end;
    136..225:
      begin
        AdjustOffsetX := -Offset;
        AdjustOffsetY := -Offset;
      end;
    else {226..315:}
      begin
        AdjustOffsetX := -Offset;
        AdjustOffsetY := Offset;
      end;
  end;
  with ACanvas do begin
    DrawPt := FDrawTextPt;
    if not Shadow then begin
      abcOffsetPoint(DrawPt, -AdjustOffsetX, -AdjustOffsetY);
      DrawColor := FCaptionStyle.ColorHighlight;
      DrawCaptionAt(ACanvas, DrawColor, DrawPt);
      abcOffsetPoint(DrawPt, 2*AdjustOffsetX, 2*AdjustOffsetY);
    end
    else
      abcOffsetPoint(DrawPt, AdjustOffsetX, AdjustOffsetY);

    DrawColor := FCaptionStyle.ColorShadow;
    DrawCaptionAt(ACanvas, DrawColor, DrawPt);
    abcOffsetPoint(DrawPt, -AdjustOffsetX, -AdjustOffsetY);
    DrawColor := Self.Font.Color;
    DrawCaptionAt(ACanvas, DrawColor, DrawPt);
  end;
end;

function TabcRotateTextSprite.GetRotatedFont: HFont;
var
  LogFontRec: TLogFont;
begin
  GetObject(Font.Handle, SizeOf(TLogFont), @LogFontRec);
  LogFontRec.lfEscapement := FAngle * 10;
  Result := CreateFontIndirect(LogFontRec);
end;

procedure TabcRotateTextSprite.GetTextBounds(var R: TRect);
var
  DC: HDC;
  SaveFont: HFont;
begin
  DC := GetDC(0);
  SaveFont := SelectObject(DC, GetRotatedFont);
  try
    DrawText(DC, PChar(Caption), -1, R, DT_CALCRECT or DT_SINGLELINE);
  finally
    SaveFont := SelectObject(DC, SaveFont);
    DeleteObject(SaveFont);
    ReleaseDC(0, DC);
  end;
end;

function TabcRotateTextSprite.IsFontStored: Boolean;
begin
  with Font do
    Result :=
      (Charset <> DEFAULT_CHARSET) or
      (Color <> clWindowText) or
      (Height <> -13) or
      (Name <> 'Arial') or
      (Style <> []);
end;

function TabcRotateTextSprite.PtInSprite(const APoint: TPoint): Boolean;
var
  rgn: HRGN;
begin
  rgn := CreatePolygonRgn(FBorderPoints, 5, WINDING);
  try
    Result := PtInRegion(rgn, APoint.X, APoint.Y);
  finally
    DeleteObject(rgn);
  end;
end;

procedure TabcRotateTextSprite.SetAngle(const Value: TabcDegrees);
begin
  if Value <> FAngle then
  begin
    FAngle := Value;
    FAngleSin := Sin(DegToRad(FAngle));
    FAngleCos := Cos(DegToRad(FAngle));
    AdjustBounds;
  end;
end;

procedure TabcRotateTextSprite.SetBorderStyle(const Value: TabcBorderStyle);
begin
  if Value <> FBorderStyle then
  begin
    FBorderStyle.Assign(Value);
    FBorderStyle.OnChange := StyleChanged;
    AdjustBounds;
  end;
end;

procedure TabcRotateTextSprite.SetBounds(ALeft, ATop, AWidth, AHeight: Integer);
var
  dX, dY: Integer;
begin
  if not FAdjustingBounds and
    ([csLoading, csReading] * ComponentState = []) then
  begin
    dX := ALeft - Left;
    dY := ATop - Top;
    abcOffsetPoint(FDrawTextPt, dX, dY);
    abcOffsetPoints(FBorderPoints, dX, dY);
    //Width and Height are autosized to text
    inherited SetBounds(ALeft, ATop, Width, Height);
  end
  else
    inherited;
end;

procedure TabcRotateTextSprite.SetCaption(const Value: string);
begin
  if Value <> FCaption then
  begin
    FCaption := Value;
    AdjustBounds;
  end;
end;

procedure TabcRotateTextSprite.SetCaptionStyle(const Value: TabcCaptionStyle);
begin
  if Value <> FCaptionStyle then
  begin
    FCaptionStyle.Assign(Value);
    FCaptionStyle.OnChange := StyleChanged;
    AdjustBounds;
  end;
end;

procedure TabcRotateTextSprite.SetFont(const Value: TFont);
begin
  FFont.Assign(Value);
end;

procedure TabcRotateTextSprite.StyleChanged(Sender: TObject);
begin
  AdjustBounds;
end;

procedure TabcRotateTextSprite.DefineProperties(Filer: TFiler);

  function DrawTextPtIsStored: Boolean;
  begin
    if Filer.Ancestor <> nil then
      with TabcRotateTextSprite(Filer.Ancestor) do
        Result := (FDrawTextPt.X <> Self.FDrawTextPt.X) or
                  (FDrawTextPt.Y <> Self.FDrawTextPt.Y)
    else
        Result := (FDrawTextPt.X <> 0) or (FDrawTextPt.Y <> 0);
  end;

  function DrawBorderPtsAreStored: Boolean;
  var
    i: integer;
  begin
    if Filer.Ancestor <> nil then
    begin
      Result := True;
      with TabcRotateTextSprite(Filer.Ancestor) do
      for i := Low(FBorderPoints) to High(FBorderPoints) do
        if (FBorderPoints[i].X <> self.FBorderPoints[i].X) or
           (FBorderPoints[i].Y <> self.FBorderPoints[i].Y) then
          Exit;
      Result := False;
    end
    else
      Result := True;
  end;

  function WidthIsStored: Boolean;
  begin
    if Filer.Ancestor <> nil then
      Result := TabcRotateTextSprite(Filer.Ancestor).Width <> Width
    else
      Result := Width <> 0;
  end;

  function HeightIsStored: Boolean;
  begin
    if Filer.Ancestor <> nil then
      Result := TabcRotateTextSprite(Filer.Ancestor).Height <> Height
    else
      Result := Height <> 0;
  end;

begin
  inherited;
  Filer.DefineProperty('Width', ReadWidth, WriteWidth, WidthIsStored);
  Filer.DefineProperty('Height', ReadHeight, WriteHeight, HeightIsStored);
  Filer.DefineProperty('DrawTextPt', ReadDrawTextPt, WriteDrawTextPt, DrawTextPtIsStored);
  Filer.DefineProperty('DrawBorderPts', ReadDrawBorderPts, WriteDrawBorderPts, DrawBorderPtsAreStored);
end;

procedure TabcRotateTextSprite.ReadDrawTextPt(Reader: TReader);
begin
  Reader.ReadListBegin;
  FDrawTextPt.X := Reader.ReadInteger;
  FDrawTextPt.Y := Reader.ReadInteger;
  Reader.ReadListEnd;
end;

procedure TabcRotateTextSprite.WriteDrawTextPt(Writer: TWriter);
begin
  Writer.WriteListBegin;
  Writer.WriteInteger(FDrawTextPt.X);
  Writer.WriteInteger(FDrawTextPt.Y);
  Writer.WriteListEnd;
end;

procedure TabcRotateTextSprite.ReadDrawBorderPts(Reader: TReader);
var
  i: integer;
begin
  Reader.ReadListBegin;
  for i := Low(FBorderPoints) to High(FBorderPoints) do
  begin
    FBorderPoints[i].X := Reader.ReadInteger;
    FBorderPoints[i].Y := Reader.ReadInteger;
  end;
  Reader.ReadListEnd;
end;

procedure TabcRotateTextSprite.WriteDrawBorderPts(Writer: TWriter);
var
  i: integer;
begin
  Writer.WriteListBegin;
  for i := Low(FBorderPoints) to High(FBorderPoints) do
  begin
    Writer.WriteInteger(FBorderPoints[i].X);
    Writer.WriteInteger(FBorderPoints[i].Y);
  end;
  Writer.WriteListEnd;
end;

procedure TabcRotateTextSprite.ReadWidth(Reader: TReader);
begin
  Width := Reader.ReadInteger;
end;

procedure TabcRotateTextSprite.WriteWidth(Writer: TWriter);
begin
  Writer.WriteInteger(Width);
end;

procedure TabcRotateTextSprite.ReadHeight(Reader: TReader);
begin
  Height := Reader.ReadInteger;
end;

procedure TabcRotateTextSprite.WriteHeight(Writer: TWriter);
begin
  Writer.WriteInteger(Height);
end;

end.
