unit abctabf;

(*------------------------------------------------------------------------------

Advanced Business Components
ABC for Delphi
Version 6

Copyright  1995, 2001 Objective Software Technology Pty Limited ACN 068 640 353
Functionality derived from Delphi VCL Source Code
is Copyright  1995, 2001 Borland Software Corporation

------------------------------------------------------------------------------*)

{$I abcdefs.inc}

{:This unit contains the abcFormTabControl and related collection classes.
}

interface

uses Messages, Windows, SysUtils, Classes, Graphics, Controls, Forms,
     CommCtrl, ComCtrls, StdCtrls, ImgList,
     abcsubfm, abcfx, abcmsg;

type

  { TabcFormTabControl }

  TabcFormTabControl = class;

  { Items and Collection }

  TabcFormTabItems = class;

  {:This class is the collection item for the <See Class=TabcFormTabControl Property=Forms> property.
  This defines a tab and a reference to the class of the form to be created or shown
  when the tab is activated.<br>
  <See property=Caption> defines the tab caption.<br>
  <See property=FormName> defines the class of the form to be loaded.<br>
  <See property=FormActions> define actions to create, destroy or hide the form when the tab is selected or deselected.<br>
  <See property=Transition> defines a transition effect to use when the form is shown.
  }

  TabcFormTabItem = class(TCollectionItem)
  private
    FParentTransition: Boolean;
    FCaption: string;
    FFormName: string;
    FFormActions: TabcSubFormActions;
    FTransition: TabcTransition;
    FChangingTransition: Boolean;
    FForm: TScrollingWinControl;
    FOnFormChanging: TabcFormChangingEvent;
    FOnLoadForm: TabcFormNotifyEvent;
    FOnCreateForm: TabcFormNotifyEvent;
    FOnDestroyForm: TabcFormNotifyEvent;
    FOnFormChange: TNotifyEvent;
    FSaveFormBorderStyle: TFormBorderStyle;
    FSaveFormBorderIcons: TBorderIcons;
    FSaveFormStyle: TFormStyle;
    FSaveFormAlign: TAlign;
    function GetFormCollection: TabcFormTabItems;
    function GetTabControl: TabcFormTabControl;
    function IsTransitionStored: Boolean;
    procedure SetCaption(const Value: string);
    procedure SetFormActions(const Value: TabcSubFormActions);
    procedure SetFormName(const Value: string);
    procedure SetParentTransition(const Value: Boolean);
    procedure SetTransition(const Value: TabcTransition);
    procedure TransitionChanged(Sender: TObject);
    procedure CancelTransition(Sender:TObject; var Cancel: Boolean);
    function GetFormLoaded: Boolean;
    procedure SetFormLoaded(Value: Boolean);
    procedure DoSetFormLoaded(Value: Boolean);
    function CanChangeTab: Boolean;
    function CanChangeForm(const NewFormName: string): Boolean;
    function GetActive: Boolean;
    procedure SetActive(Value: Boolean);
  protected
    FSkipFormPropertyChanges: Boolean;
    function GetDisplayName: string; override;
    procedure SetDisplayName(const Value: string); override;
    procedure SetIndex(Value: Integer); override;
    procedure DestroyForm;
    function CreateForm: TScrollingWinControl;
    function FindExistingForm: TScrollingWinControl;
    procedure SetForm(const Value: TScrollingWinControl);
    procedure DoSetForm(const Value: TScrollingWinControl);
  public
    constructor Create(Collection: TCollection); override;
    destructor Destroy; override;
    procedure Assign(Source: TPersistent); override;
    {::This destroys the current form, if any.}
    procedure FreeForm;
    {::This returns the <See class=TabcFormTabItems> collection.}
    property FormCollection: TabcFormTabItems read GetFormCollection;
    {::This returns the <See class=TabcFormTabControl> the item belongs to.}
    property TabControl: TabcFormTabControl read GetTabControl;
    {::This is the Form or Frame instance that is assigned to the tab item.<P>
     The value can also be nil.<p>
     Read this property to inspect or operate on the form instance in code.<br>
     For example:
<Code>
 if Form is TForm1 then
   with TForm1(Form) do
     ...
</Code>
<P>
     You can set this property to assign an existing form instance to a tab item.}
    property Form: TScrollingWinControl read FForm write SetForm;
    {::Setting this property to True will apply <See property=FormActions> to autocreate or assign a form.<br>
    Setting this property to False will apply <See property=FormActions> to destroy or hide a form.}
    property FormLoaded: Boolean read GetFormLoaded write SetFormLoaded;
    {::Returns True if the tab is the active (selected) tab in the tabcontrol.<br>
     Set to True to make the tab active.<br>
     Setting False when active will select the next tab.}
    property Active: Boolean read GetActive write SetActive;
  published
    {::This is the text that appears on the tab.}
    property Caption: string read FCaption write SetCaption;
    {::The form actions define when to create or destroy forms.<p>
     <code>
TabcSubFormAction = (afaCreate, afaActivate, afaDeactivate, afaHide, afaClose, afaDestroy, afaShowExisting);<br>
TabcSubFormActions = set of TabcSubFormAction;
     </code><p>
    The effect of each value is:
    <UL>
    <LI>afaCreate - a new instance of the form or frame is created when the tab is selected.
    <LI>afaActivate - the form is activated when the tab is selected.
    <LI>afaDeactivate - the form is deactivated when the tab is deselected.
    <LI>afaHide - The form is hidden when the tab is deselected (this is automatic for this control).
    <LI>afaClose - The form is closed when the tab is deselected and the form close event occurs.
    <LI>afaDestroy - The form instance is destroyed when the tab is deselected.
    <LI>afaShowExisting - if a form instance of the class already exists in the application, it is shown, instead of creating a new instance.
    This can be used with projects containing autocreated forms.
    </UL><p>
     The <See property=FormName> property defines the class of the form to be created or shown.<P>
     You can also add forms at runtime, by setting the <See property=Form> property of the collection item; or by
     calling the <See class=TabcformTabControl method=AddForm> or <See class=TabcformTabControl method=CreateForm> methods.<P>
    }
    property FormActions: TabcSubFormActions read FFormActions write SetFormActions
      default [afaCreate, afaActivate, afaDeActivate, afaHide, afaClose, afaDestroy];
    {::This defines the class name of the form to create or show.<br>
    The "T" class prefix is omitted, e.g. "Form1" will create an instance of class TForm1.<P>
    This is used when the <See property=FormActions> property contains values afaCreate or afaShowExisting.<P>
    To create forms indirectly, the form class must be registered.<br>
    For instance, include the following initialization code in your form unit:
<Code>

initialization
  RegisterClass(TForm1);
</Code>
    }
    property FormName: string read FFormName write SetFormName;
    {::Set to True if the <See class=TabcformTabControl property=Transition> is used when showing the form.<br>
    When True, the transition property is not stored.<br>
    Set to False to define a custom transition for this tab only.<br>}
    property ParentTransition: Boolean read FParentTransition write SetParentTransition default True;
    {::Defines a transition effect to show when the tab is selected.}
    property Transition: TabcTransition read FTransition write SetTransition
      stored IsTransitionStored;
    {::This event occurs when a form instance is created.}
    property OnCreateForm: TabcFormNotifyEvent read FOnCreateForm write FOnCreateForm;
    {::This event occurs when a form instance is destroyed.}
    property OnDestroyForm: TabcFormNotifyEvent read FOnDestroyForm write FOnDestroyForm;
    {::This event occurs when a form instance is loaded.}
    property OnLoadForm: TabcFormNotifyEvent read FOnLoadForm write FOnLoadForm;
    {::This event occurs when the active form and tab has changed.}
    property OnFormChange: TNotifyEvent read FOnFormChange write FOnFormChange;
    {::This event occurs before the active form and tab changes.}
    property OnFormChanging: TabcFormChangingEvent read FOnFormChanging write FOnFormChanging;
  end;

  {:This class is the collection for the <See Class=TabcFormTabControl Property=Forms> property.
  This collection defines the set of tabs and linked forms.<br>
  The collection items are of class <See Class=TabcFormTabItem>.
  }

  TabcFormTabItems = class(TCollection)
  private
    FTabControl: TabcFormTabControl;
    FOnUpdate: TNotifyEvent;
    function GetItem(Index: Integer): TabcFormTabItem;
    procedure SetItem(Index: Integer; Value: TabcFormTabItem);
  protected
    function GetOwner: TPersistent; override;
    procedure Update(Item: TCollectionItem); override;
  public
    constructor Create(ATabControl: TabcFormTabControl);
    {::Creates a new collection item.}
    function Add: TabcFormTabItem;
    {::Locates the item index of a particular form instance.}
    function IndexOfForm(AForm: TScrollingWinControl): Integer;
    {::The collection items, indexed 0..Count-1.}
    property Items[Index: Integer]: TabcFormTabItem read GetItem write SetItem; default;
    {::This returns the <See class=TabcFormTabControl> the items belong to.}
    property TabControl: TabcFormTabControl read FTabControl;
    {::This event occurs when an item changes.}
    property OnUpdate: TNotifyEvent read FOnUpdate write FOnUpdate;
  end;


  {:Class type for instantiating embedded forms and frames.}
  TScrollingWinControlClass = class of TScrollingWinControl;

  {:A custom tab control containing embedded forms or frames linked to tabs.
  This tab control is a container class for forms or frames that are linked to each tab.<P>
  Forms and tabs can either be declared at designtime or added/removed dynamically at runtime.<P>
  The <See property=Forms> property defines a collection of tabs and forms.<br>
  <See property=FormIndex> defines the active tab and form. This value can be -1 if there is no active tab.<P>
  The <See class=TabcFormTabItem property=FormActions> and <See class=TabcFormTabItem property=FormName>
  properties define automatic form display actions when a tab is activated or deactivated.<br>
  When a tab becomes active, a new form instance can be created, or an existing form can be shown.<br>
  When a tab is deactivated, the active form instance can be hidden or destroyed.<P>
  The <See property=Transition> property can define a transition effect
  to apply when the visible form changes.<P>
  The <See method=AddForm> or <See method=CreateForm> methods can be used to add forms and tabs at runtime.<P>
  The <See method=RemoveForm> or <See method=RemoveFormAtIndex> methods can be used to remove forms and tabs, and optionally free the form instance.<P>

  This control can also act as a dock site, allowing forms to be dragged onto and off the tab control.<br>
  Some limitations apply to docking behaviour:<br>
  - Only forms that were docked can be undocked, not autocreated or loaded forms.<br>
  - Docking TControl instances is not supported; only descendants of TScrollingWinControl can be docked.<br>
  - No transition effects are shown when docking or undocking.
  <P>

  <strong>See Also</strong><p>
  <A HRef="TabcEffectsImage_Component">TabcEffectsImage</A> for details of transition effects.
  }

  TabcFormTabControl = class(TCustomTabControl)
  private
    FForms: TabcFormTabItems;
    FTransition: TabcTransition;
    FSaveImage: TBitmap;
    FPaintImage: Boolean;
    FLastFormIndex: Integer;
    FCreateFormsOnLoad: Boolean;
    FChanging: Boolean;
    FFormsChanging: Boolean;
    FFormIndex: Integer;
    FNewDockTab: TabcFormTabItem;
    FUndockingTab: TabcFormTabItem;
    FRepaintAfterTransition: Boolean;
    FSettingFormIndex: Boolean;
    FLoaded: Boolean;
    function GetFormCount: Integer;
    procedure SetFormIndex(const Value: Integer);
    procedure SetForms(const Value: TabcFormTabItems);
    procedure SetTransition(const Value: TabcTransition);
    procedure CancelTransition(Sender:TObject; var Cancel: Boolean);
    procedure SavePageImage;
    procedure ClearPageImage;
    procedure AMShowingChanged(var Message: TMessage); message AM_SHOWINGCHANGED;
    procedure CMShowingChanged(var Message: TMessage); message CM_SHOWINGCHANGED;
    procedure CMDockClient(var Message: TCMDockClient); message CM_DOCKCLIENT;
    procedure CMUnDockClient(var Message: TCMUnDockClient); message CM_UNDOCKCLIENT;
    procedure WMLButtonDown(var Message: TWMLButtonDown); message WM_LBUTTONDOWN;
    procedure WMLButtonDblClk(var Message: TWMLButtonDblClk); message WM_LBUTTONDBLCLK;
    procedure WMEraseBkgnd(var Message: TWMEraseBkgnd); message WM_ERASEBKGND;
    procedure WMPaint(var Message: TWMPaint); message WM_PAINT;
  protected
    function CanChange: Boolean; override;
    procedure Change; override;
    procedure DoAddDockClient(Client: TControl; const ARect: TRect); override;
    procedure DockOver(Source: TDragDockObject; X, Y: Integer;
      State: TDragState; var Accept: Boolean); override;
    procedure DoRemoveDockClient(Client: TControl); override;
    procedure GetSiteInfo(Client: TControl; var InfluenceRect: TRect;
      MousePos: TPoint; var CanDock: Boolean); override;

    procedure Loaded; override;

    procedure FormsChanged;
    procedure BeginUpdate;
    procedure EndUpdate;
    function GetTabOfDockClient(Client: TControl): TabcFormTabItem;
    function GetDockClientFromMousePos(MousePos: TPoint): TControl;
    function AddDockedForm(AForm: TScrollingWinControl): TabcFormTabItem;
    procedure TransitionEnd(Sender: TObject);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    {::Adds an existing form or frame instance to the tab control.<br>
     AForm is the form to be added.<br>
     ACaption is the caption to show on the tab.<br>
     Activate is true to select the tab or false to not select it.<P>
     Returns the new item added to the <See property=Forms> collection.
    }
    function AddForm(const AForm: TScrollingWinControl; const ACaption: string; Activate: Boolean): TabcFormTabItem;
    {::Creates a new form instance by name reference and adds it to the tab control.<br>
     AFormName is the name of the form class to be added.<br>
     ACaption is the caption to show on the tab.<br>
     Activate is true to select the tab or false to not select it.<P>
     Returns the new item added to the <See property=Forms> collection.
    }
    function CreateForm(const AFormName, ACaption: string; Activate: Boolean): TabcFormTabItem;
    {::Removes an instance of a form.<br>
     The corresponding item in the <See property=Forms> collection is freed and the tab is removed.<br>
     AForm is the form instance to be removed.<br>
     AFreeForm is true to automatically destroy the form.<br>
    }
    procedure RemoveForm(const AForm: TScrollingWinControl; AFreeForm: Boolean);
    {::Removes an instance of a form at the specified index in the <See property=Forms> collection.<br>
     The corresponding item in the <See property=Forms> collection is freed and the tab is removed.<br>
     AFreeForm is true to automatically destroy the form.<br>
    }
    procedure RemoveFormAtIndex(Index: Integer; AFreeForm: Boolean);
    {::Activates the next form and tab, forward or back from the current active item.}
    procedure SelectNextForm(GoForward: Boolean);
    {::Returns the index of the next forms item, forward or back from the specified index.}
    function FindNextFormIndex(CurIndex: Integer;
      GoForward: Boolean): Integer;
    {::Returns the DisplayRect, i.e. the client rectangle in which forms are displayed.}
    property DisplayRect;
    {::Returns the count of the <See property=Forms> collection.}
    property FormCount: Integer read GetFormCount;
  published
    {::Set to true if you want the form instances of all tabs to be created when loading.<br>
     Requires more memory, and may result in faster tab changes.}
    property CreateFormsOnLoad: Boolean read FCreateFormsOnLoad write FCreateFormsOnLoad default False;
    {::The index of the active form and tab in the <See property=Forms> collection.}
    property FormIndex: Integer read FFormIndex write SetFormIndex default -1;
    {::This collection defines the set of tabs and forms to display in the control.}
    property Forms: TabcFormTabItems read FForms write SetForms;
    {::Defines a wipe, blend or shaped transition effect to display when activating each tab.<P>
    <strong>See Also</strong><p>
    <A HRef="TabcEffectsImage_Component">TabcEffectsImage</A> for details of transition effects.
    }
    property Transition: TabcTransition read FTransition write SetTransition;
    {inherited properties}
    property Align;
    property Anchors;
    property BiDiMode;
    property Constraints;
    property DockSite;
    property DragCursor;
    property DragKind;
    property DragMode;
    property Enabled;
    property Font;
    property HotTrack;
    property Images;
    property MultiLine;
    property OwnerDraw;
    property ParentBiDiMode;
    property ParentColor;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property RaggedRight;
    property ScrollOpposite;
    property ShowHint;
//    property Style; not published, assume always tabs
    property TabHeight;
    property TabOrder;
    property TabPosition;
    property TabStop;
    property TabWidth;
    property Visible;
    property OnChange;
    property OnChanging;
{$IFDEF DELPHI5}
    property OnContextPopup;
{$ENDIF}
    property OnDockDrop;
    property OnDockOver;
    property OnDragDrop;
    property OnDragOver;
    property OnDrawTab;
    property OnEndDock;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnGetImageIndex;
    property OnGetSiteInfo;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnResize;
    property OnStartDock;
    property OnStartDrag;
    property OnUnDock;
  end;

implementation

uses abcconst, abcutilf;

{ TabcFormTabItem }

constructor TabcFormTabItem.Create(Collection: TCollection);
begin
  inherited;
  FFormActions := [afaCreate, afaActivate, afaDeActivate, afaHide, afaClose, afaDestroy];
  FTransition := TabcTransition.Create;
  FTransition.OnChange := TransitionChanged;
  FTransition.OnEffectStep := CancelTransition;
  if TabControl <> nil then
    FTransition.OnEffectEnd := TabControl.TransitionEnd;
  FParentTransition := True;
end;

destructor TabcFormTabItem.Destroy;
var
  tabCtrl: TabcFormTabControl;
begin
  tabCtrl := TabControl;
  if (tabCtrl <> nil) then
    with tabCtrl do
    begin
      if (FUndockingTab = Self) then
        FUndockingTab := nil
      else if (csDestroying in tabCtrl.ComponentState) then
      begin
        if afaDestroy in FormActions then
          DestroyForm;
      end
      else
        Form := nil;
      if not (csDestroying in tabCtrl.ComponentState) and
        (tabCtrl.Forms.UpdateCount = 0) then
      begin
        if Index < FFormIndex then
          Dec(FFormIndex)
        else if (Index = FFormIndex) then
        begin
          Active := False;
          if FFormIndex > 0 then
            Dec(FFormIndex);
        end;
      end;
    end;
  FTransition.Free;
  inherited;
end;

procedure TabcFormTabItem.Assign(Source: TPersistent);
begin
  if Source is TabcFormTabItem then
    with TabcFormTabItem(Source) do
    begin
      Self.FCaption := Caption;
      Self.FFormActions := FormActions;
      Self.FFormName := FormName;
      Self.FTransition.Assign(Transition);
      Self.FParentTransition := ParentTransition;
    end
  else
    inherited Assign(Source);
end;

function TabcFormTabItem.GetDisplayName: string;
begin
  Result := Caption;
end;

function TabcFormTabItem.GetFormCollection: TabcFormTabItems;
begin
  if (Collection <> nil) and (Collection is TabcFormTabItems) then
    Result := TabcFormTabItems(Collection)
  else
    Result := nil;
end;

function TabcFormTabItem.GetTabControl: TabcFormTabControl;
begin
  if (Collection <> nil) and (Collection is TabcFormTabItems) then
    Result := TabcFormTabItems(Collection).TabControl
  else
    Result := nil;
end;

function TabcFormTabItem.IsTransitionStored: Boolean;
begin
  Result := not ParentTransition;
end;

procedure TabcFormTabItem.SetCaption(const Value: string);
begin
  FCaption := Value;
  Changed(True);
end;

procedure TabcFormTabItem.SetDisplayName(const Value: string);
begin
  Caption := Value;
end;

procedure TabcFormTabItem.SetIndex(Value: Integer);
var
  oldIndex: integer;
  oldActiveItem: TabcFormTabItem;
begin
  if (TabControl <> nil) and (FormCollection <> nil) then
  begin
    oldIndex := Index;
    oldActiveItem := FormCollection[TabControl.FormIndex];
    FormCollection.BeginUpdate;
    try
      inherited;
        with TabControl do
          if oldIndex = FormIndex then
          begin
            FFormIndex := Index;
            TabIndex := Index;
          end
          else if Index = FormIndex then
          begin
            FFormIndex := oldActiveItem.Index;
            TabIndex := FFormIndex;
          end;
    finally
       FormCollection.EndUpdate;
    end;
  end
  else
    inherited;
end;

procedure TabcFormTabItem.SetForm(const Value: TScrollingWinControl);
var
  NewFormName: string;
begin
  if Value <> nil then
    NewFormName := Copy(Value.ClassName, 2, 256)
  else
    NewFormName := '';
  if CanChangeForm(NewFormName) then
    DoSetForm(Value);
end;

procedure TabcFormTabItem.DoSetForm(const Value: TScrollingWinControl);
var
  TabCtrl: TabcFormTabControl;
  isActive: Boolean;
begin
  if (Value <> FForm) then
  begin
    TabCtrl := TabControl;
    if TabCtrl = nil then Exit;

    isActive :=
      (TabCtrl.FormIndex = Index) and  //tab is active
      not TabCtrl.FChanging and        //not already changing with tab click
      ([csLoading, csDesigning, csDestroying] * TabCtrl.ComponentState = []) and
      TabCtrl.Showing;

    if FForm <> nil then
    begin
      with FForm do
        if HostDockSite = TabCtrl then
        begin
          if not Visible then Show;
          Align := alNone;
          HostDockSite := nil;
          Dock(nil, Rect(0,0,UndockWidth,UndockHeight));
        end
        else
        begin
          if afaDeActivate in FormActions then Perform(CM_DEACTIVATE,0,0);
          Hide;
          if afaClose in FormActions then
            if FForm is TForm then
              TForm(FForm).Close;
          Parent := nil;
        end;
      if afaDestroy in FormActions then
        DestroyForm
      else if not FSkipFormPropertyChanges then
        if FForm is TForm then
          with TForm(FForm) do
          begin
            BorderStyle := FSaveFormBorderStyle;
            BorderIcons := FSaveFormBorderIcons;
            FormStyle := FSaveFormStyle;
            Align := FSaveFormAlign;
          end;
    end;
    FForm := nil;

    if (Value <> nil) then
    begin
      with Value do
      begin
        if not FSkipFormPropertyChanges then
        begin
          {change the form style}
          if Value is TForm then
            with TForm(Value) do
            begin
              FSaveFormBorderStyle := BorderStyle;
              FSaveFormBorderIcons := BorderIcons;
              FSaveFormStyle := FormStyle;
              FSaveFormAlign := Align;
              BorderStyle := bsNone;
              BorderIcons := [];
              FormStyle := fsNormal;
            end;
          if Align = alNone then
            Align := alClient;
          if Value is TForm then
            TForm(Value).AutoScroll := True;
        end;
        Visible := False;
        Parent := TabCtrl;

      end;
      FForm := Value;
      if Assigned(FOnLoadForm) then FOnLoadForm(Self, Value);
    end;

    if isActive then
      if Value <> nil then
      begin
        Value.BringToFront;
        Value.Visible := True;
        TabCtrl.FLastFormIndex := TabCtrl.TabIndex;
      end
      else
        TabCtrl.Repaint;

    if Assigned(FOnFormChange) then FOnFormChange(Self);
  end;
end;

procedure TabcFormTabItem.SetFormActions(const Value: TabcSubFormActions);
begin
  FFormActions := Value;
end;

procedure TabcFormTabItem.SetFormName(const Value: string);
var
  TabCtrl: TabcFormTabControl;
begin
  if CanChangeForm(Value) then
  begin
    FFormName := Value;
    TabCtrl := TabControl;
    if (TabCtrl <> nil) and ([csLoading, csDesigning] * TabCtrl.ComponentState = []) and
       ([afaCreate, afaShowExisting] * FormActions <> []) and
       TabCtrl.HandleAllocated and Active then
      DoSetFormLoaded(Value <> '');
  end;
end;

procedure TabcFormTabItem.SetParentTransition(const Value: Boolean);
begin
  FParentTransition := Value;
end;

procedure TabcFormTabItem.SetTransition(const Value: TabcTransition);
begin
  FTransition.Assign(Value);
end;

procedure TabcFormTabItem.TransitionChanged(Sender: TObject);
begin
  if not FChangingTransition then
    ParentTransition := False;
end;

procedure TabcFormTabItem.CancelTransition(Sender:TObject; var Cancel: Boolean);
begin
  Cancel := Application.Terminated or
    ((GetOwner is TComponent) and (csDestroying in TComponent(GetOwner).ComponentState));
end;

procedure TabcFormTabItem.DestroyForm;
begin
  if FForm <> nil then
  begin
    if Assigned(FOnDestroyForm) then FOnDestroyForm(Self, FForm);
    if not (csDestroying in FForm.ComponentState) then
      FForm.Free;
    FForm := nil;
  end;
end;

function TabcFormTabItem.CreateForm: TScrollingWinControl;
var
  AClass: TPersistentClass;
  AFormClass: TScrollingWinControlClass;
  TabCtrl: TabcFormTabControl;
begin
  Result := nil;
  TabCtrl := TabControl;
  if TabCtrl = nil then Exit;
  if FFormName = '' then
    raise Exception.Create(sabcErrFormNameExpected);
  AClass := FindClass('T' + FFormName);
  AFormClass := TScrollingWinControlClass(AClass);
  Result := AFormClass.Create(TabCtrl);
  if Assigned(FOnCreateForm) then FOnCreateForm(Self, Result);
end;

function TabcFormTabItem.FindExistingForm: TScrollingWinControl;
var
  AClassName: string;
  i: integer;
begin
  Result := nil;
  if FFormName = '' then
    raise Exception.Create(sabcErrFormNameExpected);
  AClassName := 'T' + FFormName;
  with Application do
    for i := 0 to ComponentCount - 1 do
      if Components[i].ClassName = AClassName then
      begin
        Result := TScrollingWinControl(Components[i]);
        Break;
      end;
end;

procedure TabcFormTabItem.FreeForm;
begin
  DestroyForm;
end;

procedure TabcFormTabItem.SetFormLoaded(Value: Boolean);
var
  NewFormName: string;
begin
  if Value then
    NewFormName := FormName
  else
    NewFormName := '';
  if CanChangeForm(NewFormName) then
    DoSetFormLoaded(Value);
end;

procedure TabcFormTabItem.DoSetFormLoaded(Value: Boolean);
var
  AForm: TScrollingWinControl;
begin
  AForm := nil;
  if Value then
  begin
    if (afaShowExisting in FormActions) then
      AForm := FindExistingForm;
    if (AForm = nil) and (afaCreate in FormActions) then
      AForm := CreateForm;
    DoSetForm(AForm);
  end
  else
    DoSetForm(nil);
end;

function TabcFormTabItem.GetFormLoaded: Boolean;
begin
  Result := FForm <> nil;
end;

function TabcFormTabItem.CanChangeTab: Boolean;
begin
  if (FForm is TForm) and ([afaClose, afaDestroy] * FormActions <> []) then
    Result := TForm(FForm).CloseQuery
  else
    Result := True;
end;

function TabcFormTabItem.CanChangeForm(const NewFormName: string): Boolean;
begin
  if (FForm is TForm) and ([afaClose, afaDestroy] * FormActions <> []) then
  begin
    Result := TForm(FForm).CloseQuery;
    if not Result then Exit;
  end
  else
    Result := True;
  if Assigned(FOnFormChanging) then FOnFormChanging(Self, NewFormName, Result);
end;

function TabcFormTabItem.GetActive: Boolean;
begin
  Result := (TabControl <> nil) and
    (TabControl.FormIndex = Index);
end;

procedure TabcFormTabItem.SetActive(Value: Boolean);
var
  nextIndex: Integer;
  tabctl: TabcFormTabControl;
begin
  if Value <> Active then
  begin
    tabctl := TabControl;
    if (tabctl <> nil) then
      if Value then
        tabctl.FormIndex := Index
      else
      begin
        nextIndex := tabctl.FindNextFormIndex(Index, True);
        if nextIndex = Index then
          nextIndex := -1;
        tabctl.FormIndex := nextIndex;
      end;
  end;
end;

{ TabcFormTabItems }

constructor TabcFormTabItems.Create(ATabControl: TabcFormTabControl);
begin
  inherited Create(TabcFormTabItem);
  FTabControl := ATabControl;
end;

function TabcFormTabItems.Add: TabcFormTabItem;
begin
  Result := TabcFormTabItem(inherited Add);
end;

function TabcFormTabItems.GetItem(Index: Integer): TabcFormTabItem;
begin
  Result := TabcFormTabItem(inherited GetItem(Index));
end;

function TabcFormTabItems.GetOwner: TPersistent;
begin
  Result := FTabControl;
end;

procedure TabcFormTabItems.SetItem(Index: Integer; Value: TabcFormTabItem);
begin
  inherited SetItem(Index, Value);
end;

procedure TabcFormTabItems.Update(Item: TCollectionItem);
begin
  if FTabControl <> nil then
      FTabControl.FormsChanged;
end;

function TabcFormTabItems.IndexOfForm(AForm: TScrollingWinControl): Integer;
begin
  for Result := 0 to Count - 1 do
    if Items[Result].Form = AForm then
      Exit;
  Result := -1;
end;

{ TabcFormTabControl }

constructor TabcFormTabControl.Create(AOwner: TComponent);
begin
  inherited;
  FTransition := TabcTransition.Create;
  FTransition.OnEffectStep := CancelTransition;
  FTransition.OnEffectEnd := TransitionEnd;
  FForms := TabcFormTabItems.Create(Self);
  FLastFormIndex := -1;
  FFormIndex := -1;
end;

destructor TabcFormTabControl.Destroy;
begin
  FTransition.Free;
  ClearPageImage;
  FForms.Free;
  inherited;
end;

type
  TScrollingWinControlAccess = class(TScrollingWinControl);

function TabcFormTabControl.AddForm(const AForm: TScrollingWinControl; const ACaption: string; Activate: Boolean): TabcFormTabItem;
begin
  Result := Forms.Add;
  with Result do
  begin
    FormActions := [afaActivate, afaDeActivate, afaHide];
    if ACaption <> '' then
      Caption := ACaption
    else
      Caption := TScrollingWinControlAccess(AForm).Caption;
    ParentTransition := False;
    Form := AForm;
    ParentTransition := True;
    if Activate then
      FormIndex := Result.Index;
  end;
end;

function TabcFormTabControl.CreateForm(const AFormName, ACaption: string; Activate: Boolean): TabcFormTabItem;
begin
  Result := Forms.Add;
  with Result do
  begin
    Caption := ACaption;
    FormName := AFormName;
    if Activate then
      FormIndex := Result.Index;
  end;
end;

function TabcFormTabControl.AddDockedForm(AForm: TScrollingWinControl): TabcFormTabItem;
begin
  Result := Forms.Add;
  with Result do
  begin
    FormActions := [afaActivate, afaDeActivate, afaHide];
    FSkipFormPropertyChanges := True;
    FForm := AForm;
    Caption := TScrollingWinControlAccess(AForm).Caption;
  end;
end;

procedure TabcFormTabControl.FormsChanged;
var
  i: integer;
begin
  if (csLoading in ComponentState) then Exit;
  FFormsChanging := True;
  {rebuild tab strings from forms collection}
  BeginUpdate;
  try
    Tabs.Clear;
    for i := 0 to FormCount - 1 do
      with Forms[i] do
        Tabs.AddObject(Caption, Forms[i]);
  finally
    EndUpdate;
  end;
  if FLastFormIndex >= FormCount then
    FLastFormIndex := -1;
  if FormIndex >= FormCount then
    if FormCount = 0 then
      FFormIndex := -1
    else
      FFormIndex := 0;
  TabIndex := FormIndex;
  FLastFormIndex := FormIndex;
  FFormsChanging := False;
end;

function TabcFormTabControl.GetFormCount: Integer;
begin
  Result := FForms.Count;
end;

procedure TabcFormTabControl.AMShowingChanged(var Message: TMessage);
begin
  if FLoaded and Showing then
  begin
    FormsChanged;
    Repaint;
    Change;
    FLoaded := False;
  end;
end;

procedure TabcFormTabControl.CMShowingChanged(var Message: TMessage);
begin
  inherited;
  PostMessage(Handle, AM_SHOWINGCHANGED, 0, 0);
end;

procedure TabcFormTabControl.Loaded;
var
  i: integer;
begin
  inherited;
  if not (csDesigning in ComponentState) and CreateFormsOnLoad then
    for i := 0 to FormCount - 1 do
      Forms[i].FormLoaded := True;
  FLoaded := True;
end;

procedure TabcFormTabControl.RemoveForm(const AForm: TScrollingWinControl; AFreeForm: Boolean);
begin
  RemoveFormAtIndex(Forms.IndexOfForm(AForm), AFreeForm);
end;

procedure TabcFormTabControl.RemoveFormAtIndex(Index: Integer; AFreeForm: Boolean);
begin
  if (Index >= 0) and (Index < FormCount) then
    if (Index <> FormIndex) or CanChange then
    begin
      if AFreeForm then
        Forms[Index].FreeForm;
      Forms[Index].Free;
    end;
end;

procedure TabcFormTabControl.SetForms(const Value: TabcFormTabItems);
begin
  FForms := Value;
  FormsChanged;
end;

procedure TabcFormTabControl.SetTransition(const Value: TabcTransition);
begin
  FTransition.Assign(Value);
end;

procedure TabcFormTabControl.CancelTransition(Sender:TObject; var Cancel: Boolean);
begin
  Cancel := Application.Terminated or (csDestroying in ComponentState) or not Visible;
end;

procedure TabcFormTabControl.SavePageImage;
begin
  ClearPageImage;
  FSaveImage := abcCopyWindowFromDesktop(Self, True);
end;

procedure TabcFormTabControl.ClearPageImage;
begin
  if FSaveImage <> nil then
  begin
    FSaveImage.Free;
    FSaveImage := nil;
  end;
end;

procedure TabcFormTabControl.WMEraseBkgnd(var Message: TWMEraseBkgnd);
begin
  if not FPaintImage then
    inherited
  else
    Message.Result := 1;
end;

procedure TabcFormTabControl.WMPaint(var Message: TWMPaint);
var
  DC: HDC;
  PS: TPaintStruct;
begin
  if FPaintImage then
  begin
    DC := Message.DC;
    if DC = 0 then
      DC := BeginPaint(Handle, PS);
    try
      BitBlt(DC, Left, Top, Width, Height, FSaveImage.Canvas.Handle, 0, 0, SRCCOPY);
    finally
      if Message.DC = 0 then EndPaint(Handle, PS);
    end;
    Canvas.Draw(0,0,FSaveImage)
  end
  else
    inherited;
end;

function TabcFormTabControl.CanChange: Boolean;
begin
  Result := FFormsChanging or
   (FLastFormIndex = -1) or
   ((not FChanging and inherited CanChange) and
    ((FormIndex < 0) or Forms[FormIndex].CanChangeTab)
   );
end;

procedure TabcFormTabControl.Change;
var
  DrawTransition: TabcTransition;
  ClientR, OffScreenR: TRect;
begin
  if (ComponentState * [csLoading, csDesigning] <> []) then Exit;
  if FChanging then Exit;
  FChanging := True;
  DrawTransition := nil;
  try
    SavePageImage;
    FPaintImage := True;
    {hide previous form}
    if (FLastFormIndex >= 0) and (Forms[FLastFormIndex].Form <> nil) then
      with Forms[FLastFormIndex], Form do
      begin
        SendToBack;
        if afaDeActivate in FormActions then Perform(CM_DEACTIVATE,0,0);
        Hide;
        if afaClose in FormActions then
          if Form is TForm then
            TForm(Form).Close
          else
            Form.Visible := False;
        if afaDestroy in FormActions then
          DestroyForm;
      end;
    {load or show form}
    if (FLastFormIndex = -1) and (FormCount > 1) and (FFormIndex <> -1) then
      FRepaintAfterTransition := True;
    if FSettingFormIndex then
      FLastFormIndex := FFormIndex
    else
      FLastFormIndex := TabIndex;
    FFormIndex := FLastFormIndex;
    if FFormIndex = -1 then Exit;
    if Forms[FFormIndex].Form = nil then
      Forms[FFormIndex].FormLoaded := True;

    {show form with transition}
    if Forms[FFormIndex].Form <> nil then
      with Forms[FFormIndex], Form do
      begin
        Align := alNone;
        ClientR := DisplayRect;
        OffScreenR := ClientR;
        OffsetRect(OffScreenR, -OffScreenR.Right, -OffScreenR.Top);
        with OffScreenR do SetBounds(Left, Top, Right-Left, Bottom-Top);
        if not Showing then Show;
        BringToFront;
        if ParentTransition then
          DrawTransition := Self.Transition
        else
          DrawTransition := Transition;
//        DrawTransition.OnEffectEnd := TransitionEnd;
        if Self.Showing and DrawTransition.Enabled then
        begin
          DrawTransition.DrawWindow(Form, Self.Canvas, False, ClientR.Left, ClientR.Top);
          if DrawTransition.NoRedraw and DrawTransition.Enabled then
            SetWindowPos(Handle, HWND_TOP, ClientR.Left, ClientR.Top, Width, Height, SWP_NOREDRAW);
        end;
        if afaActivate in FormActions then
        begin
          SetFocus;
          Perform(CM_ACTIVATE,0,0);
        end;
        Align := alClient;
      end
    else
    begin
      {need to clear the page}
      FLastFormIndex := -1;
      FRepaintAfterTransition := True;
    end;
  finally
    if FLoaded or not Showing or (FLastFormIndex = -1) or (DrawTransition = nil) or not DrawTransition.Enabled then
      TransitionEnd(Self);
  end;
  inherited;
end;

procedure TabcFormTabControl.TransitionEnd(Sender: TObject);
begin
  FChanging := False;
  FPaintImage := False;
  ClearPageImage;
  if FRepaintAfterTransition then
  begin
    Repaint;
    FRepaintAfterTransition := False;
  end;
end;

procedure TabcFormTabControl.BeginUpdate;
begin
  if HandleAllocated then
    SendMessage(Handle, WM_SETREDRAW, Ord(False), 0);
end;

procedure TabcFormTabControl.EndUpdate;
begin
  if HandleAllocated then
  begin
    SendMessage(Handle, WM_SETREDRAW, Ord(True), 0);
    Invalidate;
    SendMessage(Handle, WM_SIZE, SIZE_RESTORED,
      Word(Width) or Word(Height) shl 16);
    Realign;
  end;
end;

procedure TabcFormTabControl.SetFormIndex(const Value: Integer);
var
  oldFormIndex: integer;
begin
  if (csLoading in ComponentState) then
    FFormIndex := Value
  else if (Value <> FFormIndex) then
  begin
    if not CanChange then Exit;
    oldFormIndex := FFormIndex;
    if (Value < 0) or (Value >= FormCount) then
      FFormIndex := -1
    else
      FFormIndex := Value;
    if HandleAllocated then
    begin
      TabIndex := FFormIndex;
      if oldFormIndex = -1 then
        Repaint;
      FSettingFormIndex := True;
      Change;
      FSettingFormIndex := False;
    end;
  end;
end;

procedure TabcFormTabControl.DockOver(Source: TDragDockObject; X, Y: Integer;
  State: TDragState; var Accept: Boolean);
var
  R: TRect;
begin
  GetWindowRect(Handle, R);
  Source.DockRect := R;
  DoDockOver(Source, X, Y, State, Accept);
end;

procedure TabcFormTabControl.GetSiteInfo(Client: TControl; var InfluenceRect: TRect;
  MousePos: TPoint; var CanDock: Boolean);
begin
  CanDock := (Client is TScrollingWinControl) and (GetTabOfDockClient(Client) = nil);
  inherited GetSiteInfo(Client, InfluenceRect, MousePos, CanDock);
end;

procedure TabcFormTabControl.CMUnDockClient(var Message: TCMUnDockClient);
var
  Tab: TabcFormTabItem;
begin
  Message.Result := 0;
  Tab := GetTabOfDockClient(Message.Client);
  if (Tab <> nil) then
  begin
    FUndockingTab := Tab;
    Message.Client.Align := alNone;
  end;
end;

procedure TabcFormTabControl.DoRemoveDockClient(Client: TControl);
var
  nextIndex: integer;
begin
  if (FUndockingTab <> nil) and not (csDestroying in ComponentState) then
    with FUndockingTab do
    begin
      nextIndex := FindNextFormIndex(Index, True);
      if nextIndex = Index then
        nextIndex := -1;
      FLastFormIndex := -1;
      FormActions := [];
      ParentTransition := False;
      Transition.Enabled := False;
      FForm := nil;
      Free;
      FUndockingTab := nil;
      Repaint;
      FormIndex := nextIndex;
      Client.BringToFront;
    end;
end;

procedure TabcFormTabControl.CMDockClient(var Message: TCMDockClient);
var
  DockCtl: TScrollingWinControl;
begin
  if (Message.DockSource.Control is TScrollingWinControl) then
  begin
    DockCtl := TScrollingWinControl(Message.DockSource.Control);
    Message.Result := 0;
    FNewDockTab := AddDockedForm(DockCtl);
    with DockCtl do
    try
      try
        Dock(Self, Message.DockSource.DockRect);
      except
        FNewDockTab.Free;
        raise;
      end;
      if Visible then
        FormIndex := FNewDockTab.Index;
      Align := alClient;
    finally
      FNewDockTab := nil;
      Self.Invalidate;
    end;
  end;
end;

procedure TabcFormTabControl.DoAddDockClient(Client: TControl; const ARect: TRect);
begin
  if FNewDockTab <> nil then
    Client.Parent := Self;
end;

function TabcFormTabControl.GetTabOfDockClient(Client: TControl): TabcFormTabItem;
var
  i: integer;
begin
  for i := 0 to FormCount - 1 do
  begin
    Result := Forms[i];
    if (Client = Result.Form) and
       (Client.HostDockSite = Self) then
      Exit;
  end;
  Result := nil;
end;

function TabcFormTabControl.GetDockClientFromMousePos(MousePos: TPoint): TControl;
var
  HitIndex: Integer;
  HitTestInfo: TTCHitTestInfo;
  FormTab: TabcFormTabItem;
begin
  Result := nil;
  if DockSite and not FChanging then
  begin
    HitTestInfo.pt := MousePos;
    HitIndex := SendMessage(Handle, TCM_HITTEST, 0, Longint(@HitTestInfo));
    if (HitIndex >= 0) and (HitIndex < Tabs.Count) then
    begin
      FormTab := TabcFormTabItem(Tabs.Objects[HitIndex]);
      if FormTab <> nil then
        Result := FormTab.Form;
      if (Result <> nil) and (Result.HostDockSite <> Self) then
         Result := nil;
    end;
  end;
end;

procedure TabcFormTabControl.WMLButtonDown(var Message: TWMLButtonDown);
var
  DockCtl: TControl;
begin
  DockCtl := GetDockClientFromMousePos(SmallPointToPoint(Message.Pos));
  if (DockCtl <> nil) and (Style = tsTabs) then DockCtl.BeginDrag(False);
  inherited;
end;

procedure TabcFormTabControl.WMLButtonDblClk(var Message: TWMLButtonDblClk);
var
  DockCtl: TControl;
begin
  inherited;
  DockCtl := GetDockClientFromMousePos(SmallPointToPoint(Message.Pos));
  if DockCtl <> nil then
    DockCtl.ManualDock(nil, nil, alNone);
end;

procedure TabcFormTabControl.SelectNextForm(GoForward: Boolean);
var
  nextIndex: Integer;
begin
  nextIndex := FindNextFormIndex(FormIndex, GoForward);
  if (nextIndex <> -1) and (nextIndex <> FormIndex) and CanChange then
    FormIndex := nextIndex;
end;

function TabcFormTabControl.FindNextFormIndex(CurIndex: Integer;
  GoForward: Boolean): Integer;
begin
  if FormCount <> 0 then
  begin
    if GoForward then
      Result := CurIndex + 1
    else
      Result := CurIndex - 1;
    if Result >= FormCount then
      Result := 0
    else if Result < 0 then
      Result := FormCount - 1;
  end
  else
    Result := -1;
end;

end.
