unit abcuspr;

(*------------------------------------------------------------------------------

Advanced Business Components
ABC for Delphi
Version 6

Copyright  1995, 2001 Objective Software Technology Pty Limited ACN 068 640 353
Functionality derived from Delphi VCL Source Code
is Copyright  1995, 2001 Borland Software Corporation

------------------------------------------------------------------------------*)

{$I abcdefs.inc}

{::Contains utility functions for sprite graphics drawing.
}

interface

uses Windows, SysUtils, Classes, Graphics;

{::Creates a monochrome mask used to transparently draw the Source bitmap.<P>
The mask pixels are black where the Source is solid and white where Source is transparent.
}

procedure abcCreateSpriteMask(const Source, Mask: TBitmap);

{::A transparent version of the BitBlt function.<br>
Mask is a monochrome bitmap with black pixels in the solid parts.<br>
If Mask is nil or empty the procedure defaults to BitBlt.<br>
On WindowsNT this uses MaskBlt.<br>
On Windows this uses the inverted source method, calling 3 x BitBlt.
}

procedure abcTransparentBlt(
  hdcDest: HDC; destLeft, destTop, Width, Height: Integer;
  hdcSrc: HDC; srcLeft, srcTop: Integer;
  Mask: TBitmap);

{::A transparent version of the BitBlt function.<br>
This uses the MaskHandle property of Source bitmap to draw transparently.<br>
On WindowsNT this uses MaskBlt.<br>
On Windows this uses the inverted source method, calling 3 x BitBlt.<P>
<Strong>WARNING:</Strong><p>
Do not use this method when other threads may be drawing on Source.<Br>
Concurrent Canvas locking may result in Source.MaskHandle not being updated.
}
  
procedure abcTransparentBltMasked(
  Dest: TCanvas; destLeft, destTop, Width, Height: Integer;
  Source: TBitmap; srcLeft, srcTop: Integer; IsTransparent: Boolean);

implementation

(* This works for bitmapsprites on win95 machine *)

procedure abcCreateSpriteMask(const Source, Mask: TBitmap);
begin
  with Mask do
  begin
    Assign(Source);
    Mask(Source.TransparentColor);
    Monochrome := True;
  end;
end;

(*
//this version seems better than using Assign, but not tested on win95
procedure abcCreateSpriteMask(const Source, Mask: TBitmap);
var
  BkColor: TColorRef;
  hdcSrc, hdcDest: HDC;
begin
  with Mask do
  begin
    Height := Source.Height;
    Width := Source.Width;
    Transparent := False;
    Canvas.Lock;
    Source.Canvas.Lock;
    try
      hdcSrc := Source.Canvas.Handle;
      hdcDest := Canvas.Handle;
      BkColor := SetBkColor(hdcSrc, ColorToRGB(Source.TransparentColor));
      try
        BitBlt(hdcDest, 0, 0, Width, Height, hdcSrc, 0, 0, SRCCOPY);
      finally
        SetBkColor(hdcSrc, BkColor);
      end;
    finally
      Canvas.Unlock;
      Source.Canvas.Unlock;
    end;
    Mask(Source.TransparentColor);
    Monochrome := True;
  end;
end;
*)

(*
procedure abcCreateSpriteMask(const Source, Mask: TBitmap);
var
  BkColor: TColorRef;
  hdcSrc, hdcDest: HDC;
begin
  with Mask do
  begin
    Height := Source.Height;
    Width := Source.Width;
    Transparent := False;
    Canvas.Lock;
    Source.Canvas.Lock;
    try
      hdcDest := Canvas.Handle;
      hdcSrc := Source.Canvas.Handle;
      BkColor := SetBkColor(hdcSrc, ColorToRGB(Source.TransparentColor));
      try
        BitBlt(hdcDest, 0, 0, Width, Height, hdcSrc, 0, 0, SRCCOPY);
      finally
        SetBkColor(hdcSrc, BkColor);
      end;
    finally
      Canvas.Unlock;
      Source.Canvas.Unlock;
    end;
    Mask(Source.TransparentColor);
    Monochrome := True;
  end;
end;
*)

(* This should be much faster, but does not work for the Windows 95 machine at all
//and has occasional invalid dc problems in bitblt

procedure abcCreateSpriteMask(const Source, Mask: TBitmap);
var
  BkColor: TColorRef;
  hdcSrc, hdcDest: HDC;
begin
  with Mask do
  begin
    Width := 0;
    Height := 0;
    Monochrome := True;
    Width := Source.Width;
    Height := Source.Height;
    Canvas.Lock;
    Source.Canvas.Lock;
    try
      hdcDest := Canvas.Handle;
      hdcSrc := Source.Canvas.Handle;
      BkColor := SetBkColor(hdcSrc, ColorToRGB(Source.TransparentColor));
      try
        BitBlt(hdcDest, 0, 0, Width, Height, hdcSrc, 0, 0, SRCCOPY);
      finally
        SetBkColor(hdcSrc, BkColor);
      end;
    finally
      Source.Canvas.Unlock;
      Canvas.Unlock;
    end;
  end;
end;
*)

procedure abcTransparentBlt(
  hdcDest: HDC; destLeft, destTop, Width, Height: Integer;
  hdcSrc: HDC; srcLeft, srcTop: Integer;
  Mask: TBitmap);
const
  ROP_DstCopy = $00AA0029;
var
  OldBkColor, OldTextColor: TColorRef;
  hdcMask: HDC;
  SavePal: HPALETTE;
begin
  if (Mask = nil) or Mask.Empty then
    BitBlt(hdcDest, destLeft, destTop, Width, Height, hdcSrc, srcLeft, srcTop, SRCCOPY)
  else if (Win32Platform = VER_PLATFORM_WIN32_NT) then
    MaskBlt(
      hdcDest, destLeft, destTop, Width, Height,
      hdcSrc, srcLeft, srcTop,
      Mask.Handle, srcLeft, srcTop,
      MakeRop4(ROP_DstCopy, SRCCOPY))
  else
  begin
    SavePal := 0;
    try
      SavePal := SelectPalette(hdcSrc, SystemPalette16, False);
      SelectPalette(hdcSrc, SavePal, False);
      if SavePal <> 0 then
        SavePal := SelectPalette(hdcDest, SavePal, True)
      else
        SavePal := SelectPalette(hdcDest, SystemPalette16, True);
      RealizePalette(hdcDest);

      hdcMask := Mask.Canvas.Handle;
      OldBkColor := SetBkColor(hdcDest, RGB(255, 255, 255));
      OldTextColor := SetTextColor(hdcDest, RGB(0, 0, 0));
      BitBlt(hdcDest, destLeft, destTop, Width, Height, hdcSrc, srcLeft, srcTop, SRCINVERT);
      BitBlt(hdcDest, destLeft, destTop, Width, Height, hdcMask, srcLeft, srcTop, SRCAND);
      BitBlt(hdcDest, destLeft, destTop, Width, Height, hdcSrc, srcLeft, srcTop, SRCINVERT);
      SetBkColor(hdcDest, OldBkColor);
      SetTextColor(hdcDest, OldTextColor);
    finally
      if SavePal <> 0 then SelectPalette(hdcDest, SavePal, False);
    end;
  end;
end;

procedure abcTransparentBltMasked(
  Dest: TCanvas; destLeft, destTop, Width, Height: Integer;
  Source: TBitmap; srcLeft, srcTop: Integer; IsTransparent: Boolean);
//warning - do not use when concurrent threads are drawing on Source
//the MaskHandle is not updated potentially causing transparency problems
const
  ROP_DstCopy = $00AA0029;
var
  OldBkColor, OldTextColor: TColorRef;
  SavePal: HPALETTE;
  SrcDC, DestDC, MaskDC: HDC;
  MaskHandle: HBITMAP;
begin
  Dest.Lock;
  Source.Canvas.Lock;
  try
    if IsTransparent then
      if (Win32Platform = VER_PLATFORM_WIN32_NT) then
        MaskBlt(
          Dest.Handle, destLeft, destTop, Width, Height,
          Source.Canvas.Handle, srcLeft, srcTop,
          Source.MaskHandle, srcLeft, srcTop,
          MakeRop4(ROP_DstCopy, SRCCOPY))
      else
      begin
        SavePal := 0;
        SrcDC := Source.Canvas.Handle;
        DestDC := Dest.Handle;
        try
          SavePal := SelectPalette(SrcDC, SystemPalette16, False);
          SelectPalette(SrcDC, SavePal, False);
          if SavePal <> 0 then
            SavePal := SelectPalette(DestDC, SavePal, True)
          else
            SavePal := SelectPalette(DestDC, SystemPalette16, True);
          RealizePalette(DestDC);

          MaskHandle := Source.MaskHandle;
          MaskDC := CreateCompatibleDC(0);
          MaskHandle := SelectObject(MaskDC, MaskHandle);
          try
            OldBkColor := SetBkColor(DestDC, RGB(255, 255, 255));
            OldTextColor := SetTextColor(DestDC, RGB(0, 0, 0));
            BitBlt(DestDC, destLeft, destTop, Width, Height, SrcDC, srcLeft, srcTop, SRCINVERT);
            BitBlt(DestDC, destLeft, destTop, Width, Height, MaskDC, srcLeft, srcTop, SRCAND);
            BitBlt(DestDC, destLeft, destTop, Width, Height, SrcDC, srcLeft, srcTop, SRCINVERT);
            SetBkColor(DestDC, OldBkColor);
            SetTextColor(DestDC, OldTextColor);
          finally
            if MaskHandle <> 0 then SelectObject(MaskDC, MaskHandle);
            DeleteDC(MaskDC);
          end;
        finally
          if SavePal <> 0 then SelectPalette(DestDC, SavePal, False);
        end;
      end
    else
      BitBlt(Dest.Handle, destLeft, destTop, Width, Height, Source.Canvas.Handle, srcLeft, srcTop, SRCCOPY);
  finally
    Dest.Unlock;
    Source.Canvas.Unlock;
  end;
end;

end.
