unit abcutilf;

(*------------------------------------------------------------------------------

Advanced Business Components
ABC for Delphi
Version 6

Copyright  1995, 2001 Objective Software Technology Pty Limited ACN 068 640 353
Functionality derived from Delphi VCL Source Code
is Copyright  1995, 2001 Borland Software Corporation

Utility Functions and Procedures for ABC:

1. String Formatting, Printing and Clipboard
2. Control Bitmap Capture and Printing
3. Control Animation
4. File & Directory Processing
5. Form Designer / Utility functions

------------------------------------------------------------------------------*)

{$I abcdefs.inc}


interface

uses SysUtils, Windows, Messages, Classes, Controls, Graphics, Forms;

type
  TabcCharSet = set of Char;

  TabcWindowAnimationStyle = (awaNone,
                              awaMoveDown, awaMoveUp, awaMoveRight, awaMoveLeft,
                              awaMoveRightDown, awaMoveRightUp,
                              awaMoveLeftDown, awaMoveLeftUp,
                              awaSizeDown, awaSizeUp, awaSizeRight, awaSizeLeft,
                              awaSizeRightDown, awaSizeRightUp,
                              awaSizeLeftDown, awaSizeLeftUp);

{
1. String Formatting, Printing and Clipboard
}

function abcLTrim(S: string): string;
  {Trims leading spaces from string}
function abcRTrim(S: string): string;
  {Trims trailing spaces from string}
function abcLRTrim(S: string): string;
  {Trims leading and trailing spaces from string}

function abcReplaceStr(SubStr, ReplaceStr, S: string; MatchCase: Boolean): string;
  {Replaces first occurence of SubStr in ReplaceStr}

function abcReplaceAll(
  const SubStr, ReplaceStr, S: string;
  MatchCase: Boolean; WholeWordsOnly: Boolean): string;
  {Replaces all occurences of SubStr in ReplaceStr}

function abcReplaceAllCount(
  const SubStr, ReplaceStr, S: string; var ResultStr: string;
  MatchCase: Boolean; WholeWordsOnly: Boolean): Integer;
  {Replaces all occurences of SubStr in ReplaceStr and returns count of replacements made}

  {WholeWordsOnly option - Whole words are delimited by these chars:}
var
  abcWordDelimiterChars: set of Char;

  {see initialization - These chars are the ASCII chars on the top line: null, TAB, CR, LF, space,
   plus the Delphi special symbols}


function abcShorten(var str: string; i: Integer): Integer;
  {Removes i chars from end of string and returns length of result}

function abcStripTrailingReturns(const str: string): string;
  {Removes all trailing CRLF #13#10 pairs from end of string}

function abcProperCase(str: string): string;
  {Capitalises the first letter of each word in the string}

procedure abcParseStringToList(S, sep: string; List: TStrings; WantBlanks: Boolean);

function abcLastCharInSet(S: string; chSet: TabcCharSet): Boolean;

function abcLastCharIs(S: string; Ch: char): Boolean;

function abcEncodeASCIItoHex(s: string): string;

function abcDecodeHextoASCII(s: string): string;

procedure abcPrintStrings(Strings: TStrings);
  {Prints TStrings to the Default Printer}

procedure abcPrintStringsFormat(Strings: TStrings; Indent, LinesPerPage: Byte);
  {Prints TStrings to the Default Printer using the format specified}

procedure abcStrToClip(S: string);
  {Copies string to Clipboard}

function abcNoDefaultPrinter: Boolean;
  {Returns true if there is no default printer set up.
   To prevent exception raised by Delphi.}

{
2. Control Bitmap Capture and Printing
}

function abcGetControlBitmap(Ctrl: TWinControl): TBitmap;
  {Returns a bitmap image of a window,
   repaints the window using WM_PAINT and WM_PRINT messages}

function abcCopyWindowFromDesktop(Ctrl: TWinControl; IncludeNonClient: Boolean): TBitmap;
  {Returns a bitmap image of a window,
   copied off the desktop without repainting
   assumes control is visible on the screen}

procedure abcPrintControlImage(Ctrl: TWinControl);
  {Prints a bitmap image of a window}

{
3. Control Animation
}

procedure abcAnimateControl(
  Control: TControl;
  FromBounds: TRect;
  ToBounds: TRect;
  Steps: Byte;
  Duration: integer;
  Animation: TabcWindowAnimationStyle);

procedure abcTimedResizeControl(
  Control: TControl;
  FromBounds: TRect;
  ToBounds: TRect;
  Steps: Byte;
  Duration: integer);

procedure abcCenterInParent(AControl: TControl);

{
4. File & Directory Processing
}

function abcFileNamePart(AFile: TFileName): string;
  {Returns everything up to the last . in the file name}
function abcReplaceFileExt(AFile: TFileName; AFileExt: string): TFileName;
  {Replaces the '.XXX' extension of a AFile with AFileExt.
   Assumes AFileExt is '.XXX' four character format}

function abcGetFileNames(
  StartDir, FileSpecs: string;
  ReturnFullPath, RecurseSubDirs: Boolean;
  List: TStrings): Integer;
{input filespecs can contain multiple wildcards separated by a semi-colon
 returns file names matching spec into list, returns count of found files}

function abcGetOneFileSpec(
  Dir, FileSpec: string;
  ReturnFullPath: Boolean;
  List: TStrings): integer;

function abcRunningFromCDROM: Boolean;
function abcRunningFromRODrive: Boolean;
function abcDriveIsReadOnly(Root: string): Boolean;
function abcRunningDriveFreeBytes: Double;
function abcDriveFreeBytes(Root: string): Double;

{
5. Form Designer etc.
}

procedure abcDesignerModified(AControl: TControl);
  {if designing, set the parent form's modified flag}

function abcFirstFormParentOf(aControl: TControl): TCustomForm;
  {returns the first form if you have a form parenting another form}

function abcDescendsFromClassByName(const AnObject: TObject; AClassName: string): Boolean;

function abcIsFontStored(AFont: TFont): Boolean;

function abcParentFormorFrameOf(AComponent: TPersistent): TWinControl;

function abcFileContentString(filename: string): string;

implementation

uses Printers, Clipbrd, RichEdit, StdCtrls, abcconst;

{
1. String Formatting, Printing and Clipboard
}

function abcLTrim(S: string): string;
begin
  if (S = '') or (S[1] <> ' ') then
    Result := S
  else
    Result := abcLTrim(Copy(S, 2, Length(S)));
end;

function abcRTrim(S: string): string;
begin
  if (S = '') or (S[Length(S)] <> ' ') then
    Result := S
  else
    Result := abcRTrim(Copy(S, 1, Length(S) - 1));
end;

function abcLRTrim(S: string): string;
begin
  Result := abcLTrim(abcRTrim(S));
end;

function abcReplaceStr(SubStr, ReplaceStr, S: string; MatchCase: Boolean): string;
var
  SubPos: Integer;
begin
  if MatchCase then
    SubPos := Pos(SubStr, S)
  else
    SubPos := Pos(UpperCase(SubStr), UpperCase(S));
  if SubPos > 0 then
    Result := Copy(S, 1, SubPos - 1) + ReplaceStr
                + Copy(S, SubPos + Length(SubStr), Length(S))
  else
    Result := S;
end;

function abcReplaceAll(
  const SubStr, ReplaceStr, S: string;
  MatchCase: Boolean; WholeWordsOnly: Boolean): string;
  {Replaces all occurences of SubStr in ReplaceStr}
  {WholeWordsOnly option - see interface definition of abcWordDelimiterChars}
begin
  abcReplaceAllCount(SubStr, ReplaceStr, S, Result, MatchCase, WholeWordsOnly);
end;

function abcReplaceAllCount(
  const SubStr, ReplaceStr, S: string; var ResultStr: string;
  MatchCase: Boolean; WholeWordsOnly: Boolean): Integer;
  {Replaces all occurences of SubStr in ReplaceStr and returns count of replacements made}
  {WholeWordsOnly option - see interface definition of abcWordDelimiterChars}
var
  SubPos, SubLen: Integer;
  ChBefore, ChAfter: Char;
  sTemp, sSub: string;
begin
  sTemp := S;
  ResultStr := '';
  Result := 0;
  if MatchCase then
    sSub := SubStr
  else
    sSub := UpperCase(SubStr);
  SubLen := Length(sSub);
  {find first pos}
  if MatchCase then
    SubPos := Pos(sSub, sTemp)
  else
    SubPos := Pos(sSub, UpperCase(sTemp));
  while SubPos > 0 do
  begin
    if WholeWordsOnly then
    begin
      if SubPos = 1 then
        ChBefore := #0
      else
        ChBefore := sTemp[SubPos-1];
      if Length(sTemp) > (SubPos+SubLen) then
        ChAfter := sTemp[SubPos+SubLen]
      else
        ChAfter := #0;
      ResultStr := ResultStr + Copy(sTemp, 1, SubPos - 1);
      if (ChBefore in abcWordDelimiterChars) and
         (ChAfter in abcWordDelimiterChars) then
      begin
        ResultStr := ResultStr + ReplaceStr;                   {replace whole word}
        Result := Result + 1;
      end
      else
        ResultStr := ResultStr + Copy(sTemp, SubPos, SubLen); {keep original}
      sTemp := Copy(sTemp, SubPos + SubLen, Length(sTemp));
    end
    else
    begin
      ResultStr := ResultStr + Copy(sTemp, 1, SubPos - 1) + ReplaceStr;
      Result := Result + 1;
      sTemp := Copy(sTemp, SubPos + SubLen, Length(sTemp));
    end;
    {find next pos}
    if MatchCase then
      SubPos := Pos(sSub, sTemp)
    else
      SubPos := Pos(sSub, UpperCase(sTemp));
  end;
  ResultStr := ResultStr + sTemp;
end;

function abcProperCase(str: string): string;
{
Capitalises the first letter of each word in the string
}
var
  TempStr: string;
  NewChr: char;
  NewWord: boolean;
  I: integer;
begin
  TempStr := LowerCase(str);
  NewWord := True;
  Result := '';
  for I := 1 to Length(TempStr) do
    case TempStr[I] of
      'a'..'z':
        begin
          if NewWord then begin
            NewWord := False;
            NewChr := Chr(Ord(TempStr[I]) - 32);
            Result := Result + NewChr;
          end
          else
            Result := Result + TempStr[I];
        end;
      ' ':
        begin
          NewWord := True;
          Result := Result + TempStr[I];
        end;
      else
        Result := Result + TempStr[I];
    end;
end;

function abcShorten(var str: string; i: Integer): Integer;
  {Removes i chars from end of string and returns length of result}
var
  l: integer;
begin
  Result := 0;
  if (str <> '') then
  begin
    l := Length(str);
    if l > i then
    begin
      Result := l-i; 
      str := Copy(str, 1, Result);
    end
    else
    begin
      Result := 0;
      str := '';
    end;
  end;
end;

function abcStripTrailingReturns(const str: string): string;
var
  sLast: string;
  iLen: integer;
begin
  {recursive remove last CRLF from str}
  Result := str;
  if Result = '' then Exit;
  iLen := Length(str);
  if iLen >= 2 then
  begin
    sLast := Copy(str, iLen-1, 2);
    if sLast = #13#10 then
      Result := abcStripTrailingReturns(Copy(Result, 1, iLen-2));
  end;
end;

procedure abcParseStringToList(S, sep: string; List: TStrings; WantBlanks: Boolean);
var
  iPos: integer;
  rem, found: string;
begin
  rem := S;
  iPos := Pos(sep, rem);
  while iPos > 0 do
  begin
    found := Copy(rem, 1, iPos-1);
    if (found <> '') or WantBlanks then
      List.Add(found);
    rem := Copy(rem, iPos+1, Length(rem));
    iPos := Pos(sep, rem);
  end;
  if (rem <> '') or WantBlanks then
    List.Add(rem);
end;

type
  charset = set of char;

function abcLastCharInSet(S: string; chSet: TabcCharSet): Boolean;
var
  tmp: Char;
begin
  if S = '' then
    Result := False
  else
  begin
    tmp := S[Length(S)];
    Result := (tmp in chSet);
  end;
end;

function abcLastCharIs(S: string; Ch: char): Boolean;
var
  tmp: Char;
begin
  if S = '' then
    Result := False
  else
  begin
    tmp := S[Length(S)];
    Result := (tmp = Ch);
  end;
end;

function abcEncodeASCIItoHex(s: string): string;
const
  ALPHA_CHARS = ['0'..'9', 'A'..'Z', 'a'..'z'];
var
  i: integer;
  ch: char;
begin
  Result := '';
  for i := 1 to Length(s) do
  begin
    ch := s[i];
    if ch in ALPHA_CHARS then
      Result := Result + ch
    else
      Result := Result + '%' + Format('%2.2x', [Ord(ch)]);
  end;
end;

function abcDecodeHextoASCII(s: string): string;
const
  INT_ERR = -1;
var
  iPos, iHex: integer;
  sRem, sHex: string;
begin
  Result := '';
  sRem := s;
  repeat
    iPos := Pos('%', sRem);
    if iPos <> 0 then
    begin
      Result := Result + Copy(sRem, 1, iPos-1);
      sHex := '$' + UpperCase(Copy(sRem, iPos+1, 2));
      iHex := StrToIntDef(sHex, INT_ERR);
      if iHex >= 0 then
        Result := Result + Char(iHex)
      else
        raise Exception.Create(sabcErrorCannotDecodeHextoAscii);
      sRem := Copy(sRem, iPos+3, Length(sRem) - iPos - 2);
    end
    else
      Result := Result + sRem;
  until (iPos = 0);
end;

procedure abcPrintStrings(Strings: TStrings);
var
  Prn: TextFile;
  i: word;
begin
  AssignPrn(Prn);
  Rewrite(Prn);
  try
    for i := 0 to Strings.Count - 1 do
      writeln(Prn, Strings.Strings[i]);
  finally
    CloseFile(Prn);
  end;
end;

procedure abcPrintStringsFormat(Strings: TStrings; Indent, LinesPerPage: Byte);
{
Prints a TStrings object to the default printer
}
var
  Py: integer;
  PLines: integer;
  I: integer;

  procedure CheckPageEnd;
  begin
    Inc(PLines);
    Inc(Py,100);
    if PLines > LinesPerPage then begin
      Printer.NewPage;
      Py := 100;
      PLines := 1;
    end;
  end;
begin
  Screen.Cursor := crHourGlass;
  with Printer do
    try
      BeginDoc;
      PLines := 0;
      Py := 0;
      for I := 0 to Strings.Count - 1 do begin
        CheckPageEnd;
        Canvas.TextOut(Indent,Py,Strings[I]);
      end;
      EndDoc;
    finally
      Screen.Cursor := crDefault;
    end;
end;

procedure abcStrToClip(S: string);
begin
  Clipboard.AsText := S;
end;

type
  TReadPrinterDevice = class
    Driver, Device, Port: String;
  end;

function abcNoDefaultPrinter: Boolean;
  {Returns true if there is no default printer set up.
   To prevent exception raised by Delphi.}
var
  I: Integer;
  DefaultPrinter: array[0..79] of Char;
  Cur, Device: PChar;

  function FetchStr(var Str: PChar): PChar;
  var
    P: PChar;
  begin
    Result := Str;
    if Str = nil then Exit;
    P := Str;
    while P^ = ' ' do Inc(P);
    Result := P;
    while (P^ <> #0) and (P^ <> ',') do Inc(P);
    if P^ = ',' then
    begin
      P^ := #0;
      Inc(P);
    end;
    Str := P;
  end;

begin
  Result := False;
  GetProfileString('windows', 'device', '', DefaultPrinter,
    SizeOf(DefaultPrinter) - 1);
  Cur := DefaultPrinter;
  Device := FetchStr(Cur);
  with Printer.Printers do
    for I := 0 to Count-1 do
    begin
      if TReadPrinterDevice(Objects[I]).Device = Device then
        Exit;
    end;
  Result := True;
end;

{
2. Control Bitmap Capture and Printing
}

type
  TWinAccess = class(TWinControl);
  TComboAccess = class(TCustomComboBox);

procedure abcPaintRichEditClientTextToDC(RichEdit: TWinControl; DC: HDC);
//workaround for rich text rendering in a TRichEdit
var
  Range: TFormatRange;
  LogX, LogY, OldMapMode, OldColor: Integer;
  ClientR: TRect;
  TopIndex: TPoint;
begin
  OldMapMode := SetMapMode(DC, MM_TEXT);
  OldColor := SetBkColor(DC, ColorToRGB(TWinAccess(RichEdit).Color));
  try
    FillChar(Range, SizeOf(TFormatRange), 0);
    LogX := GetDeviceCaps(DC, LOGPIXELSX);
    LogY := GetDeviceCaps(DC, LOGPIXELSY);
    with Range do
    begin
      hdc := DC;
      hdcTarget := DC;

      RichEdit.Perform(EM_GETRECT, 0, Longint(@ClientR));

      rc.left := ClientR.Left * 1440 div LogX;
      rc.top := ClientR.Top * 1440 div LogY;
      rc.right := ClientR.Right * 1440 div LogX;
      rc.bottom := ClientR.Bottom * 1440 div LogY;
      rcPage := rc;

      TopIndex.X:= 1;
      TopIndex.Y:= 1;
      chrg.cpMin := RichEdit.Perform(EM_CHARFROMPOS, 0, Longint(@TopIndex));
      chrg.cpMax := -1;
    end;
    RichEdit.Perform(EM_FORMATRANGE, 0, 0);
    RichEdit.Perform(EM_FORMATRANGE, 1, Longint(@Range));
    RichEdit.Perform(EM_FORMATRANGE, 0, 0);
  finally
    SetMapMode(DC, OldMapMode);
    SetBkColor(DC, OldColor);
  end;
end;

procedure abcPrintCtrlTo(Ctrl: TWinControl; DC: HDC; X, Y: Integer);
(*
  datetimepicker dropdown calendar or combo box dropdown list not supported
  client-only printing is not supported
*)

  function IsAnimateCtrl: Boolean;
  begin
    Result := abcDescendsFromClassByName(Ctrl, 'TAnimate');
  end;

  function IsCustomRichEdit: Boolean;
  begin
    Result := abcDescendsFromClassByName(Ctrl, 'TCustomRichEdit');
  end;

  function IsABCCustomRichEdit: Boolean;
  begin
    Result := abcDescendsFromClassByName(Ctrl, 'TabcCustomRichEdit');
  end;

  function IsTreeView: Boolean;
  begin
    Result := abcDescendsFromClassByName(Ctrl, 'TCustomTreeView');
  end;

  function IsListView: Boolean;
  begin
    Result := abcDescendsFromClassByName(Ctrl, 'TCustomListView');
  end;

  function IsGrid: Boolean;
  begin
    Result := abcDescendsFromClassByName(Ctrl, 'TCustomGrid');
  end;

  function IsABCToolbarControl: Boolean;
  begin
    Result := abcDescendsFromClassByName(Ctrl, 'TabcCustomToolbar') or
              abcDescendsFromClassByName(Ctrl, 'TabcDock') or
              abcDescendsFromClassByName(Ctrl, 'TabcCustomToolWindow') or
              abcDescendsFromClassByName(Ctrl, 'TabcToolbarEdit') or
              abcDescendsFromClassByName(Ctrl, 'TabcToolbarMaskEdit');
  end;

  function IsABCButtonPage: Boolean;
  begin
    Result := abcDescendsFromClassByName(Ctrl, 'TabcButtonPage');
  end;

  function IsSpinEdit: Boolean;
  begin
    Result := abcDescendsFromClassByName(Ctrl, 'TSpinEdit') or
              abcDescendsFromClassByName(Ctrl, 'TabcSpinEdit');
  end;

  function IsProgressBar: Boolean;
  begin
    Result := abcDescendsFromClassByName(Ctrl, 'TProgressBar');
  end;

  function IsComboBox: Boolean;
  begin
    Result := Ctrl is TCustomComboBox;
  end;

  function IsListBox: Boolean;
  begin
    Result := Ctrl is TCustomListBox;
  end;

  function IsCustomMemo: Boolean;
  begin
    Result := Ctrl is TCustomMemo;
  end;

  function IsScrollingWinControl: Boolean;
  begin
    Result := Ctrl is TScrollingWinControl;
  end;

  function IsCustomForm: Boolean;
  begin
    Result := Ctrl is TCustomForm;
  end;


var
  I, EdgeFlags, BorderFlags, SaveIndex: Integer;
  R: TRect;
  childWin: TWinControl;
  ptOffset: TPoint;
  hasScrollBars: Boolean;

begin
  SaveIndex := SaveDC(DC);
  with Ctrl do
  try
    MoveWindowOrg(DC, X, Y);
    IntersectClipRect(DC, 0, 0, Width, Height);
    SetRect(R, 0, 0, Width, Height);
    BorderFlags := 0;
    EdgeFlags := 0;
    if (GetWindowLong(Handle, GWL_EXSTYLE) and WS_EX_CLIENTEDGE) <> 0 then
    begin
      EdgeFlags := EDGE_SUNKEN;
      BorderFlags := BF_RECT or BF_ADJUST
    end
    else if (GetWindowLong(Handle, GWL_STYLE) and WS_BORDER) <> 0 then
    begin
      EdgeFlags := BDR_OUTER;
      BorderFlags := BF_RECT or BF_ADJUST or BF_MONO;
    end;
    if (BorderFlags <> 0) and not (IsCustomForm or IsCustomMemo or IsABCToolbarControl or IsABCButtonPage) then
    begin
      SetRect(R, 0, 0, Width, Height);
      DrawEdge(DC, R, EdgeFlags, BorderFlags);
      if not (IsScrollingWinControl or IsCustomRichEdit or IsAnimateCtrl or IsTreeView or IsListView or IsGrid or IsListBox or IsComboBox) then
      begin
        MoveWindowOrg(DC, R.Left, R.Top);
        IntersectClipRect(DC, 0, 0, R.Right - R.Left, R.Bottom - R.Top);
      end;
    end
    else if IsABCButtonPage then
    begin
      R.Bottom := R.Bottom - 2;
      MoveWindowOrg(DC, R.Left, R.Top);
      IntersectClipRect(DC, 0, 0, R.Right - R.Left, R.Bottom - R.Top);
    end;

    if IsSpinEdit then
      ptOffset := R.TopLeft;

    if not (IsCustomMemo or IsGrid or IsABCToolbarControl) then
    begin
      Perform(WM_ERASEBKGND, WPARAM(DC), 0);
      Perform(WM_PAINT, WPARAM(DC), 0);
    end;

    hasScrollBars := False;
    if (GetWindowLong(Handle, GWL_STYLE) and WS_HSCROLL) <> 0 then
    begin
      hasScrollBars := True;
      R.Bottom := R.Bottom - GetSystemMetrics(SM_CYHSCROLL);
    end;
    if (GetWindowLong(Handle, GWL_STYLE) and WS_VSCROLL) <> 0 then
    begin
      hasScrollBars := True;
      R.Right := R.Right - GetSystemMetrics(SM_CXVSCROLL);
    end;

    if hasScrollBars
      or IsCustomForm
      or IsCustomRichEdit
      or IsAnimateCtrl
      or IsCustomMemo
      or IsTreeView
      or IsListView
      or IsGrid
      or IsABCToolbarControl
      or IsProgressBar
      or IsListBox
    then
      Perform(WM_PRINT, WPARAM(DC), PRF_CLIENT or PRF_NONCLIENT or PRF_ERASEBKGND);

    if IsComboBox then
    begin
      MoveWindowOrg(DC, 3, 3);
      SendMessage(TComboAccess(Ctrl).EditHandle, WM_PRINT, WPARAM(DC), PRF_CLIENT or PRF_NONCLIENT or PRF_ERASEBKGND);
      Exit;
    end;

    if IsCustomRichEdit and not IsABCCustomRichEdit then
      abcPaintRichEditClientTextToDC(Ctrl, DC);

    if Ctrl is TCustomForm then
      with TCustomForm(Ctrl) do
      begin
        //adjust for form borders
        ptOffset.X := Left;
        ptOffset.Y := Top;
        if Parent = nil then
          Windows.ScreenToClient(Handle, ptOffset)
        else
          ptOffset := Ctrl.ScreenToClient(Parent.ClientToScreen(ptOffset));
        MoveWindowOrg(DC, - ptOffset.X, - ptOffset.Y);
      end;

    if hasScrollBars then
    begin
      MoveWindowOrg(DC, R.Left, R.Top);
      IntersectClipRect(DC, 0, 0, R.Right - R.Left, R.Bottom - R.Top);
    end;

    if IsGrid then
      Perform(WM_PAINT, DC, 0);

    if not IsABCToolbarControl then
      TWinAccess(Ctrl).PaintControls(DC, nil)
    else
      Perform(WM_NCCALCSIZE, 0, integer(@R));

    if not (IsScrollingWinControl or IsCustomMemo or IsGrid) then
    begin
      MoveWindowOrg(DC, R.Left, R.Top);
      IntersectClipRect(DC, 0, 0, R.Right - R.Left, R.Bottom - R.Top);
    end;

    for I := 0 to ControlCount - 1 do
      if Controls[i] is TWinControl then
      begin
        childWin := TWinControl(Controls[I]);
        with childWin do
          if Visible then
            abcPrintCtrlTo(childWin, DC, Left, Top);
      end;
  finally
    RestoreDC(DC, SaveIndex);
  end;

  if IsSpinEdit then
    with Ctrl do
      if (ControlCount > 0) and (Controls[0] is TWinControl) then
      begin
        childWin := TWinControl(Controls[0]);
        abcPrintCtrlTo(childWin, DC, Ctrl.Left + childWin.Left + ptOffset.X, Ctrl.Top + childWin.Top + ptOffset.Y);
    end;

end;

function abcGetControlBitmap(Ctrl: TWinControl): TBitmap;
var
  DC, MemDC: HDC;
  MemBitmap, OldBitmap: HBITMAP;
begin
  Result := nil;
  with Ctrl do
  begin
    DC := GetWindowDC(Handle);
    try
      MemDC := CreateCompatibleDC(DC);
      try
        MemBitmap := CreateCompatibleBitmap(DC, Width, Height);
        try
          OldBitmap := SelectObject(MemDC, MemBitmap);
          try
            abcPrintCtrlTo(Ctrl, MemDC, 0, 0);
          finally
            SelectObject(MemDC, OldBitmap);
          end;

          Result := TBitmap.Create;
          Result.Handle := MemBitmap;
          MemBitmap := 0;

        except
          Result.Free;
          Result := nil;
          if MemBitmap <> 0 then
            DeleteObject(MemBitmap);
          raise;
        end;
      finally
        DeleteDC(MemDC);
      end;
    finally
      ReleaseDC(Handle, DC);
    end;
  end;
end;

function abcCopyWindowFromDesktop(Ctrl: TWinControl; IncludeNonClient: Boolean): TBitmap;
  {Returns a bitmap image of a window,
   copied off the desktop without repainting
   assumes control is visible on the screen}
var
  ScreenDC, MemDC: HDC;
  MemBitmap, OldBitmap: HBITMAP;
  ScreenR: TRect;
begin
  Result := nil;
  with Ctrl do
  begin
    ScreenDC := GetDC(0);
    try
      MemDC := CreateCompatibleDC(ScreenDC);
      try
        MemBitmap := CreateCompatibleBitmap(ScreenDC, Width, Height);
        try
          OldBitmap := SelectObject(MemDC, MemBitmap);
          try
            if IncludeNonClient then
              GetWindowRect(Handle, ScreenR)
            else
            begin
              ScreenR := Rect(0, 0, Width, Height);
              MapWindowPoints(Handle, 0, ScreenR, 2);
            end;
            BitBlt(MemDC, 0, 0, ScreenR.Right-ScreenR.Left, ScreenR.Bottom-ScreenR.Top,
                   ScreenDC, ScreenR.Left, ScreenR.Top, SRCCOPY);
          finally
            SelectObject(MemDC, OldBitmap);
          end;
          Result := TBitmap.Create;
          Result.Handle := MemBitmap;
          MemBitmap := 0;

        except
          Result.Free;
          Result := nil;
          if MemBitmap <> 0 then
            DeleteObject(MemBitmap);
          raise;
        end;
      finally
        DeleteDC(MemDC);
      end;
    finally
      ReleaseDC(0, ScreenDC);
    end;
  end;
end;

procedure abcPrintControlImage(Ctrl: TWinControl);
var
  ScreenDC: HDC;
  PrinterImage: TBitmap;
  Info: PBitmapInfo;
  InfoSize: DWORD;
  Image: Pointer;
  ImageSize: DWORD;
  Bits: HBITMAP;
  DIBWidth, DIBHeight: Longint;
  PrintWidth, PrintHeight: Longint;
begin
  Printer.BeginDoc;
  try
    PrinterImage := abcCopyWindowFromDesktop(Ctrl, True);
    try
      { Paint bitmap to the printer }
      with Printer, Canvas do
      begin
        Bits := PrinterImage.Handle;
        GetDIBSizes(Bits, InfoSize, ImageSize);
        Info := AllocMem(InfoSize);
        try
          Image := AllocMem(ImageSize);
          try
            GetDIB(Bits, 0, Info^, Image^);
            with Info^.bmiHeader do
            begin
              DIBWidth := biWidth;
              DIBHeight := biHeight;
            end;

            {Adjust in proportion of Screen to Printer}
            ScreenDC := GetDC(0);
            try
              PrintWidth := MulDiv(DIBWidth,
                                   GetDeviceCaps(Handle, LOGPIXELSX),
                                   GetDeviceCaps(ScreenDC, LOGPIXELSX));
              PrintHeight := MulDiv(DIBHeight,
                                    GetDeviceCaps(Handle, LOGPIXELSY),
                                    GetDeviceCaps(ScreenDC, LOGPIXELSY));
            finally
              ReleaseDC(0, ScreenDC);
            end;
            StretchDIBits(Canvas.Handle, 0, 0, PrintWidth, PrintHeight, 0, 0,
              DIBWidth, DIBHeight, Image, Info^, DIB_RGB_COLORS, SRCCOPY);
          finally
            FreeMem(Image, ImageSize);
          end;
        finally
          FreeMem(Info, InfoSize);
        end;
      end;
    finally
      PrinterImage.Free;
    end;
  finally
    Printer.EndDoc;
  end;
end;

{
3. Control Animation
}

procedure abcAnimateControl(
  Control: TControl;
  FromBounds: TRect;
  ToBounds: TRect;
  Steps: Byte;
  Duration: integer;
  Animation: TabcWindowAnimationStyle);
var
  AdjustedFromBounds: TRect;
begin
  case Animation of
    awaNone:
      begin
        {special case just size once and exit.}
        AdjustedFromBounds := ToBounds;
        with ToBounds do Control.SetBounds(Left, Top, Right, Bottom);
        Exit;
      end;
    awaMoveDown:
      AdjustedFromBounds := Rect(
        ToBounds.Left,
        ToBounds.Top - ToBounds.Bottom,
        ToBounds.Right,
        ToBounds.Bottom);
    awaMoveUp:
      AdjustedFromBounds := Rect(
        ToBounds.Left,
        ToBounds.Bottom,
        ToBounds.Right,
        ToBounds.Bottom);
    awaMoveRight:
      AdjustedFromBounds := Rect(
        ToBounds.Left - ToBounds.Right,
        ToBounds.Top,
        ToBounds.Right,
        ToBounds.Bottom);
    awaMoveLeft:
      AdjustedFromBounds := Rect(
        ToBounds.Right,
        ToBounds.Top,
        ToBounds.Right,
        ToBounds.Bottom);
    awaMoveRightDown:
      AdjustedFromBounds := Rect(
        ToBounds.Left - ToBounds.Right,
        ToBounds.Top - ToBounds.Bottom,
        ToBounds.Right,
        ToBounds.Bottom);
    awaMoveRightUp:
      AdjustedFromBounds := Rect(
        ToBounds.Left - ToBounds.Right,
        ToBounds.Bottom,
        ToBounds.Right,
        ToBounds.Bottom);
    awaMoveLeftDown:
      AdjustedFromBounds := Rect(
        ToBounds.Right,
        ToBounds.Top - ToBounds.Bottom,
        ToBounds.Right,
        ToBounds.Bottom);
    awaMoveLeftUp:
      AdjustedFromBounds := Rect(
        ToBounds.Right,
        ToBounds.Bottom,
        ToBounds.Right,
        ToBounds.Bottom);
    awaSizeDown:
      AdjustedFromBounds := Rect(
        ToBounds.Left,
        ToBounds.Top,
        ToBounds.Right,
        FromBounds.Bottom);
    awaSizeUp:
      AdjustedFromBounds := Rect(
        ToBounds.Left,
        ToBounds.Bottom - FromBounds.Bottom,
        ToBounds.Right,
        FromBounds.Bottom);
    awaSizeRight:
      AdjustedFromBounds := Rect(
        ToBounds.Left,
        ToBounds.Top,
        FromBounds.Right,
        ToBounds.Bottom);
    awaSizeLeft:
      AdjustedFromBounds := Rect(
        ToBounds.Right - FromBounds.Right,
        ToBounds.Top,
        FromBounds.Right,
        ToBounds.Bottom);
    awaSizeRightDown:
      AdjustedFromBounds := Rect(
        ToBounds.Left,
        ToBounds.Top,
        FromBounds.Right,
        FromBounds.Bottom);
    awaSizeRightUp:
      AdjustedFromBounds := Rect(
        ToBounds.Left,
        ToBounds.Bottom - FromBounds.Bottom,
        FromBounds.Right,
        FromBounds.Bottom);
    awaSizeLeftDown:
      AdjustedFromBounds := Rect(
        ToBounds.Right - FromBounds.Right,
        ToBounds.Top,
        FromBounds.Right,
        FromBounds.Bottom);
    awaSizeLeftUp:
      AdjustedFromBounds := Rect(
        ToBounds.Right - FromBounds.Right,
        ToBounds.Bottom - FromBounds.Bottom,
        FromBounds.Right,
        FromBounds.Bottom);
  end;
  abcTimedResizeControl(Control, AdjustedFromBounds, ToBounds, Steps, Duration);
end;

procedure abcTimedResizeControl(
  Control: TControl;
  FromBounds: TRect;
  ToBounds: TRect;
  Steps: Byte;
  Duration: integer);
var
  NextBounds: TRect;
  dLeft, dTop, dWidth, dHeight: Integer;
  i: Integer;
  NextStep, StepDuration: DWORD;
begin
  if Steps > 0 then
  begin
    StepDuration := Duration div Steps;
    dLeft := (ToBounds.Left - FromBounds.Left) div Steps;
    dTop := (ToBounds.Top - FromBounds.Top) div Steps;
    dWidth := (ToBounds.Right - FromBounds.Right) div Steps;
    dHeight := (ToBounds.Bottom - FromBounds.Bottom) div Steps;
    NextBounds := FromBounds;
    NextStep := GetTickCount + StepDuration;
    with Control do
    begin
      NextBounds := FromBounds;
      for i := 1 to Steps do
      begin
        with NextBounds do
        begin
          Inc(NextBounds.Left, dLeft);
          Inc(NextBounds.Top, dTop);
          Inc(NextBounds.Right, dWidth);
          Inc(NextBounds.Bottom, dHeight);
          SetBounds(Left, Top, Right, Bottom);
        end;
        while GetTickCount < NextStep do
          Application.ProcessMessages;
        NextStep := GetTickCount + StepDuration;
      end;
    end;
  end;
  {final resize in case of rounding error in steps}
  with ToBounds do Control.SetBounds(Left, Top, Right, Bottom);
end;

procedure abcCenterInParent(AControl: TControl);
var
  newL, newT: integer;
begin
  if AControl.Parent = nil then Exit;
  with AControl do
  begin
    newL := (Parent.Width - Width) div 2;
    newT := (Parent.Height - Height) div 2;
    SetBounds(newL, newT, Width, Height);
  end;
end;

{
4. File & Directory Processing
}

function abcFileNamePart(AFile: TFileName): string;
{
Parses just the name part of a file name possibly containing a path and extension
}
var
  DotPos: integer;
begin
  Result := ExtractFileName(AFile);
  DotPos := Pos('.', Result);
  if DotPos > 0 then
    Result := Copy(Result, 1, DotPos - 1);
end;

function abcReplaceFileExt(AFile: TFileName; AFileExt: string): TFileName;
{
Replaces the '.xxx' extension of a AFile with AFileExt.
If there is no extension, the function returns AFile.
NB: AFileExt must include the .
}
var
  DotPos: integer;
begin
  Result := AFile;
  DotPos := Pos('.', AFile);
  if DotPos > 0 then
    Result := Copy(AFile, 1, DotPos - 1) + AFileExt;
end;

function abcGetFileNames(StartDir, FileSpecs: string;
  ReturnFullPath, RecurseSubDirs: Boolean; List: TStrings): Integer;
{input filespecs can contain multiple wildcards separated by a semi-colon
 returns file names matching spec into list, returns count of found files}
var
  Searches: TStringList;
  i: integer;
  F: TSearchRec;
  Path, PathSpec: string;

  procedure ProcessSubDirectory;
  begin
    if ((F.Attr and faDirectory) > 0) and (F.Name <> '.') and (F.Name <> '..') then
      Result := Result + abcGetFileNames(Path + F.Name, FileSpecs, ReturnFullPath, RecurseSubDirs, List);
  end;

begin
  Result := 0;
  Searches := TStringList.Create;
  try
    abcParseStringToList(FileSpecs, ';', Searches, False);
    for i := 0 to Searches.Count - 1 do
      Result := Result + abcGetOneFileSpec(StartDir, Searches[i], ReturnFullPath, List);
  finally
    Searches.Free;
  end;
  if RecurseSubDirs then
  begin
    if abcLastCharInSet(StartDir, ['\','/']) then
      Path := StartDir
    else
      Path := StartDir + '\';
    PathSpec := Path + '*.*';
    if SysUtils.FindFirst(PathSpec, faDirectory, F) = 0 then
    try
      ProcessSubDirectory;
      while SysUtils.FindNext(F) = 0 do
        ProcessSubDirectory;
    finally
      SysUtils.FindClose(F);
    end;
  end;
end;

function abcGetOneFileSpec(Dir, FileSpec: string; ReturnFullPath: Boolean; List: TStrings): integer;
var
  F: TSearchRec;
  Path, PathSpec: string;

  procedure ProcessItem;
  begin
    if ReturnFullPath then
      List.Add(Path + F.Name)
    else
      List.Add(F.Name);
  end;

begin
  Result := List.Count;
  if (Dir = '') or (FileSpec = '') then Exit;
  if abcLastCharInSet(Dir, ['\','/']) then
    Path := Dir
  else
    Path := Dir + '\';
  PathSpec := Path + FileSpec;
  if SysUtils.FindFirst(PathSpec, 0, F) = 0 then
  try
    ProcessItem;
    while SysUtils.FindNext(F) = 0 do
      ProcessItem;
  finally
    SysUtils.FindClose(F);
  end;
  Result := List.Count - Result;
end;

function abcRunningFromCDROM: Boolean;
var
  Root: string;
begin
  Root := ExtractFileDrive(Application.ExeName);
  Result := GetDriveType(PChar(Root)) = DRIVE_CDROM;
end;

function abcRunningFromRODrive: Boolean;
var
  Root: string;
begin
 Root := ExtractFileDrive(Application.ExeName) + '/';
 Result := abcDriveIsReadOnly(Root);
end;

function abcDriveIsReadOnly(Root: string): Boolean;
var
  TempFileName: string;
  TempFileStream: TFileStream;
begin
 TempFileName := Root + Copy(Inttostr(GetTickCount),1,8) + '.tmp';
 while FileExists(TempFileName) do
   TempFileName := Root + Copy(Inttostr(GetTickCount),1,8) + '.tmp';

 TempFileStream := nil;
 try
   TempFileStream := TFileStream.Create(TempFileName, fmCreate);
   Result := False;
 except
   Result := True;
 end;
 TempFileStream.Free;
 if FileExists(TempFileName) then
   DeleteFile(PChar(TempFileName));
end;

function abcDriveFreeBytes(Root: string): Double;
var
 Secs, Bytes, FreeClus, TotClus: DWORD;
begin
  Result := 0.0;
  if GetDiskFreeSpace(PChar(Root), Secs, Bytes, FreeClus, TotClus) then
    Result := Secs * Bytes * FreeClus;
end;

function abcRunningDriveFreeBytes: Double;
var
 Root: string;
begin
 Root := ExtractFileDrive(Application.ExeName) + '/';
 Result := abcDriveFreeBytes(Root);
end;

{
5. Form Designer interaction
}

procedure abcDesignerModified(AControl: TControl);
  {if designing, set the parent form's modified flag}
var
  ParentForm: TCustomForm;
begin
  with AControl do
    if csDesigning in ComponentState then
    begin
      {mark designer modified}
      ParentForm := GetParentForm(AControl);
      if ParentForm <> nil then
        if ParentForm.Designer <> nil then
          ParentForm.Designer.Modified;
    end;
end;

function abcFirstFormParentOf(aControl: TControl): TCustomForm;
  {returns the first form if you have a form parenting another form}
var
  resultWin: TWinControl;
begin
  if aControl = Application.MainForm then
  begin
    Result := TCustomForm(aControl);
    Exit;
  end;
  resultWin := aControl.Parent;
  while (resultWin <> nil) and not (resultWin is TCustomForm) do
    resultWin := resultWin.Parent;
  Result := TCustomForm(resultWin);
end;

function abcDescendsFromClassByName(const AnObject: TObject; AClassName: string): Boolean;
var
  ThisClass: TClass;
begin
  ThisClass := AnObject.ClassType;
  while ThisClass <> nil do
  begin
    if ThisClass.ClassName = AClassName then
    begin
      Result := True;
      Exit;
    end
    else
      ThisClass := ThisClass.ClassParent;
  end;
  Result := False;
end;

function abcIsFontStored(AFont: TFont): Boolean;
begin
  with AFont do
    Result := (Name <> 'MS Sans Serif')
      or (Size <> 8)
      or (Style <> [])
      or (Color <> clBlack)
      or (CharSet <> DEFAULT_CHARSET);
end;

function abcParentFormorFrameOf(AComponent: TPersistent): TWinControl;
begin
  Result := nil;
  if AComponent is TControl then
  begin
    Result := TControl(AComponent).Parent;
{$IFDEF DELPHI5}
    while (Result <> nil) and not (Result is TForm) and not (Result is TFrame) do
{$ELSE}
    while (Result <> nil) and not (Result is TForm) do
{$ENDIF}
      Result := Result.Parent;
  end;
end;

function abcFileContentString(filename: string): string;
var
  MemStream: TMemoryStream;
  StrStream: TStringStream;
  s: string;
begin
  Result := '';
  if FileExists(filename) then
  begin
    MemStream := TMemoryStream.Create;
    StrStream := TStringStream.Create(s);
    try
      MemStream.LoadFromFile(filename);
      MemStream.Seek(0, soFromBeginning);
      StrStream.CopyFrom(MemStream, MemStream.Size);
      StrStream.Seek(0, soFromBeginning);
      Result := StrStream.DataString;
    finally
      MemStream.Free;
      StrStream.Free;
    end;
  end;
end;

initialization
  abcWordDelimiterChars := [
    #0, #9, #10, #13, ' ',
    '#', '$', '&', '''', '(', ')',
    '*', '+', ',', '-', '.', '/', '\', ':', ';',
    '<', '=', '>', '@', '[', ']', '^', '{', '}'
    ];

end.
