unit abcutilg;
 
(*------------------------------------------------------------------------------

Advanced Business Components
ABC for Delphi
Version 6

Copyright  1995, 2001 Objective Software Technology Pty Limited ACN 068 640 353
Functionality derived from Delphi VCL Source Code
is Copyright  1995, 2001 Borland Software Corporation

Utility Graphics Drawing Procedures for ABC:

1. Miscellaneous Point Conversion Routines
2. Polygon Bounding Rectangle
3. Windows Rotation Procedures
4. Cartesian Rotation Procedures
5. 3D round frame drawing routines
6. Edge frame drawing routines
7. Screen & OS Information routines
8. Bitmap Format Conversion, Bit Manipulation and Color Replacement Routines

------------------------------------------------------------------------------*)

{$I abcdefs.inc}

interface

uses Windows, SysUtils, Classes, Controls, Graphics;

{
1. Miscellaneous Point Conversion Routines
}

procedure abcOffsetPoint(var P: TPoint; X, Y: integer);
  {Offsets a point by X,Y}

procedure abcOffsetPoints(var Points: array of TPoint; X, Y: integer);
  {Offsets an array of points by X,Y}

procedure abcSetRectPoints(R: TRect; var Points: array of TPoint);
  {Calculates the points on a rectangle}

{
2. Polygon Bounding Function
}

function abcGetPolygonBounds(Points: array of TPoint): TRect;
  {Calculates smallest rectangle containing the polygon}

{
3. Windows Rotation Procedures
}

procedure abcRotateWinPoint(SinTheta, CosTheta: Double; Centre: TPoint; var Point:TPoint);
  {Rotates a point about a centre using Windows coordinates}

procedure abcRotateWinPolygon(SinTheta, CosTheta: Double; Centre: TPoint; var Points: array of TPoint);
  {Rotates a polygon about a centre using Windows coordinates}

procedure abcRotateWinRect(SinTheta, CosTheta: Double; Centre: TPoint; R: TRect; var Points: array of TPoint);
  {Rotates a rectangle about a centre using Windows coordinates and updates a polygon result.
   Assumes Points array contains 4 items.}

procedure abcRotateWinRectBounds(SinTheta, CosTheta: Double; Centre: TPoint; var R: TRect);
  {Calculates smallest bounding rectangle containing the rectangle,
   rotated in Windows coordinates}

{
4. Cartesian Rotation Procedures
}                                                 

procedure abcRotatePoint(SinTheta, CosTheta: Double; Centre: TPoint; var Point: TPoint);
  {Rotates a point about a centre using Cartesian coordinates}

procedure abcRotatePolygon(SinTheta, CosTheta: Double; Centre: TPoint; var Points: array of TPoint);
  {Rotates a polygon about a centre using Cartesian coordinates}

procedure abcRotateRect(SinTheta, CosTheta: Double; Centre: TPoint; R: TRect; var Points: array of TPoint);
  {Rotates a rectangle about a centre using Cartesian coordinates and updates a polygon result.
   Assumes Points array contains 4 items.}

procedure abcRotateRectBounds(SinTheta, CosTheta: Double; Centre: TPoint; var R: TRect);
  {Calculates smallest bounding rectangle containing the rectangle,
   rotated in Cartesian coordinates}

{
5. 3D round frame drawing routines
}

procedure abcRoundFrame3D(Canvas: TCanvas; var Rect: TRect;
  TopColor, BottomColor: TColor; Width, CornerWidth: Integer);
  {Draws a 3D Frame with rounded corners}

procedure abcEllipseFrame3D(Canvas: TCanvas; var Rect: TRect; TopColor, BottomColor: TColor;
  Width: Integer);
  {Draws an eliptical 3D Frame with shading edge changes at 45 and 225 degrees.}

{
6. Edge frame drawing routines
}

procedure abcEdgeFrame3D(Canvas: TCanvas; var Rect: TRect;
  TopColor, BottomColor: TColor; Width: Integer; Edges: TBevelEdges);
  {Draws a 3D Frame with only the selected edges}

procedure abcDrawExpandBox(Canvas: TCanvas; R: TRect; Expanded: Boolean);
{Draws a 3D button with a + Expanded false or - Expanded true}

{
7. Screen & OS Information routines
}

function abcScreenIsVGA: Boolean;
{ is screen <= 256 Colors ? }

function abcScreenIsHighColor: Boolean;
{ is screen > 256 Colors ? }

function abcScreenColorDepth: Integer;
{ color depth of screen in bits per pixel }

function abcScreenPixelFormat: TPixelFormat;
{ color depth of screen in TPixelFormat }

function abcOSVerIsWin95: Boolean;

{
8. Bitmap Bit Manipulation
}

procedure abcGrayScaleBitmap(Bitmap: TBitmap);

procedure abcGenerateMaskBitmapFromDIB (const MaskBitmap, SourceBitmap: TBitmap;
  const SourceOffset, SourceSize: TPoint; TransColors: array of TColor);
{ This a special procedure meant for generating monochrome masks from
  >4 bpp color DIB sections. Because each video driver seems to sport its own
  interpretation of how to handle DIB sections, a workaround procedure like
  this was necessary. }

procedure abcGenerateMaskBitmap (const MaskBitmap, SourceBitmap: TBitmap;
  const SourceOffset, SourceSize: TPoint; const TransColors: array of TColor);
{ Returns handle of a monochrome bitmap, with pixels in SourceBitmap of color
  TransColor set to white in the resulting bitmap. All other colors of
  SourceBitmap are set to black in the resulting bitmap. This uses the
  regular ROP_DSPDxax BitBlt method. }

procedure abcReplaceBitmapColorsFromMask (const MaskBitmap, DestBitmap: TBitmap;
  const DestOffset, DestSize: TPoint; const ReplaceColor: TColor);

function abcCopyBitmapToDDB (const SourceBitmap: TBitmap): TBitmap;
{ Makes a device-dependent duplicate of SourceBitmap. The color palette,
  if any, is preserved. }

function abcCopyPalette (Palette: HPALETTE): HPALETTE;

function abcLoadBitmapByName(S: String): HBitmap;

function abcLightenCaptionColor(clr: TColor): TColor;

var
  ABC_TRANS_COLOR: TColor = $FF80FF; {RGB(255,128,255)}
  ABC_GLYPHS_SAVE_GDI: Boolean = False;

implementation

{
1. Miscellaneous Point Conversion Routines
}

procedure abcOffsetPoint(var P: TPoint; X, Y: integer);
begin
  P.X := P.X + X;
  P.Y := P.Y + Y;
end;

procedure abcOffsetPoints(var Points: array of TPoint; X, Y: integer);
var
  I: integer;
begin
  for I := 0 to High(Points) do begin
    Points[I].X := Points[I].X + X;
    Points[I].Y := Points[I].Y + Y;
  end;
end;

procedure abcSetRectPoints(R: TRect; var Points: array of TPoint);
begin
  Points[0] := R.TopLeft;
  Points[1] := Point(R.Right, R.Top);
  Points[2] := R.BottomRight;
  Points[3] := Point(R.Left, R.Bottom);
end;

{
2. Polygon Bounding Rectangle
}

function abcGetPolygonBounds(Points: array of TPoint): TRect;
var
  I: integer;
begin
  with Result do begin
    TopLeft := Points[0];
    BottomRight := Points[0];
    for I := 1 to High(Points) do begin
      if Points[I].X < Left then
        Left := Points[I].X
      else if Points[I].X > Right then
        Right := Points[I].X;
      if Points[I].Y < Top then
        Top := Points[I].Y
      else if Points[I].Y > Bottom then
        Bottom := Points[I].Y;
    end;
  end;
end;

{
3. Windows Rotation Procedures
}

procedure abcRotateWinPoint(SinTheta, CosTheta: Double; Centre: TPoint; var Point:TPoint);
{General procedure to rotate a point using Windows coords.
 i.e. y axis is reversed from regular cartesian coords.}
var
  Rotated: TPoint;
begin
  Point.X := Point.X - Centre.X;
  Point.Y := Point.Y - Centre.Y;
  Rotated.X := Round(Point.X * CosTheta + Point.Y * SinTheta);
  Rotated.Y := Round(Point.X * SinTheta - Point.Y * CosTheta);
  Rotated.X := Rotated.X + Centre.X;
  Rotated.Y := - Rotated.Y + Centre.Y;
  Point := Rotated;
end;

procedure abcRotateWinPolygon(SinTheta, CosTheta: Double; Centre: TPoint; var Points: array of TPoint);
var
  I: integer;
begin
  for I := 0 to High(Points) do
    if (Points[I].X <> Centre.X) or (Points[I].Y <> Centre.Y) then
      abcRotateWinPoint(SinTheta, CosTheta, Centre, Points[I]);
end;

procedure abcRotateWinRect(SinTheta, CosTheta: Double; Centre: TPoint; R: TRect; var Points: array of TPoint);
begin
  abcSetRectPoints(R, Points);
  abcRotateWinPolygon(SinTheta, CosTheta, Centre, Points);
end;

procedure abcRotateWinRectBounds(SinTheta, CosTheta: Double; Centre: TPoint; var R: TRect);
var
  Points: array[0..3] of TPoint;
begin
  abcSetRectPoints(R, Points);
  abcRotateWinPolygon(SinTheta, CosTheta, Centre, Points);
  R := abcGetPolygonBounds(Points);
end;

{
4. Cartesian Rotation Procedures
}

procedure abcRotatePoint(SinTheta, CosTheta: Double; Centre: TPoint; var Point:TPoint);
var
  Rotated: TPoint;
begin
  Point.X := Point.X - Centre.X;
  Point.Y := Point.Y - Centre.Y;
  Rotated.X := Round(Point.X * CosTheta - Point.Y * SinTheta);
  Rotated.Y := Round(Point.X * SinTheta + Point.Y * CosTheta);
  Rotated.X := Rotated.X + Centre.X;
  Rotated.Y := Rotated.Y + Centre.Y;
  Point := Rotated;
end;

procedure abcRotatePolygon(SinTheta, CosTheta: Double; Centre: TPoint; var Points: array of TPoint);
var
  I: integer;
begin
  for I := 0 to High(Points) do
    if (Points[I].X <> Centre.X) or (Points[I].Y <> Centre.Y) then
      abcRotatePoint(SinTheta, CosTheta, Centre, Points[I]);
end;

procedure abcRotateRect(SinTheta, CosTheta: Double; Centre: TPoint; R: TRect; var Points: array of TPoint);
begin
  abcSetRectPoints(R, Points);
  abcRotatePolygon(SinTheta, CosTheta, Centre, Points);
end;

procedure abcRotateRectBounds(SinTheta, CosTheta: Double; Centre: TPoint; var R: TRect);
var
  Points: array[0..3] of TPoint;
begin
  abcSetRectPoints(R, Points);
  abcRotatePolygon(SinTheta, CosTheta, Centre, Points);
  R := abcGetPolygonBounds(Points);
end;

{
5. 3D round frame drawing routines
}

procedure abcRoundFrame3D(Canvas: TCanvas; var Rect: TRect; TopColor, BottomColor: TColor;
  Width, CornerWidth: Integer);

  procedure DoRoundRect;
  begin
    with Canvas, Rect do
    begin
      Pen.Color := TopColor;
      MoveTo(Left, Bottom - CornerWidth);
      LineTo(Left, Top + CornerWidth - 1);
      MoveTo(Left + CornerWidth, Top);
      LineTo(Right - CornerWidth, Top);
      if CornerWidth <> 0 then begin
        Arc(Left, Bottom - 2 * CornerWidth, Left + 2 * CornerWidth + 1, Bottom,
          Left, Bottom - CornerWidth, Left + CornerWidth, Bottom);
        Arc(Left, Top, Left + 2 * CornerWidth, Top + 2 * CornerWidth,
          Left + CornerWidth, Top, Left, Top + CornerWidth);
      end;
      Pen.Color := BottomColor;
      MoveTo(Right, Top + CornerWidth);
      LineTo(Right, Bottom - CornerWidth);
      MoveTo(Right - CornerWidth, Bottom);
      LineTo(Left + CornerWidth, Bottom);
      if CornerWidth <> 0 then begin
        Arc(Right + 1, Bottom + 1, Right - 2 * CornerWidth, Bottom - 2 * CornerWidth,
          Right - CornerWidth, Bottom, Right, Bottom - CornerWidth);
        Arc(Right - 2 * CornerWidth - 1, Top, Right + 1, Top + 2 * CornerWidth,
          Right, Top + CornerWidth, Right - CornerWidth, Top);
      end;
    end;
  end;

begin
  Canvas.Pen.Width := 1;
  Dec(Rect.Bottom);
  Dec(Rect.Right);
  while Width > 0 do
  begin
    Dec(Width);
    DoRoundRect;
    InflateRect(Rect, -1, -1);
  end;
  Inc(Rect.Bottom);
  Inc(Rect.Right);
end;

procedure abcEllipseFrame3D(Canvas: TCanvas; var Rect: TRect; TopColor, BottomColor: TColor;
  Width: Integer);

  procedure DoEllipse;
  begin
    with Canvas, Rect do
    begin
      Pen.Color := TopColor;
      Arc(Left, Top, Right + 1, Bottom + 1, Right, Top, Left, Bottom + 1);
      Pen.Color := BottomColor;
      Arc(Left, Top, Right + 1, Bottom + 1, Left, Bottom, Right, Top - 1);
    end;
  end;

begin
  Canvas.Pen.Width := 1;
  Dec(Rect.Bottom); Dec(Rect.Right);
  while Width > 0 do
  begin
    Dec(Width);
    DoEllipse;
    InflateRect(Rect, -1, -1);
  end;
  Inc(Rect.Bottom); Inc(Rect.Right);
end;

{
6. Edge frame drawing routines
}

procedure abcEdgeFrame3D(Canvas: TCanvas; var Rect: TRect;
  TopColor, BottomColor: TColor; Width: Integer; Edges: TBevelEdges);

  procedure DoRect;
  begin
    with Canvas, Rect do begin
      Pen.Color := TopColor;
      MoveTo(Left, Bottom);
      if beLeft in Edges then
        LineTo(Left, Top)
      else
        MoveTo(Left, Top);
      if beTop in Edges then
        LineTo(Right, Top)
      else
        MoveTo(Right, Top);
      Pen.Color := BottomColor;
      Dec(Left);
      if beRight in Edges then
        LineTo(Right, Bottom)
      else
        MoveTo(Right, Bottom);
      if beBottom in Edges then
        LineTo(Left, Bottom)
      else
        MoveTo(Left, Bottom);
      Inc(Left);
    end;
  end;

begin
  Canvas.Pen.Width := 1;
  Dec(Rect.Bottom); Dec(Rect.Right);
  while Width > 0 do
  begin
    Dec(Width);
    DoRect;
    InflateRect(Rect, -1, -1);
  end;
  Inc(Rect.Bottom); Inc(Rect.Right);
end;

procedure abcDrawExpandBox(Canvas: TCanvas; R: TRect; Expanded: Boolean);
var
  DrawR: TRect;
  MidX, MidY, ShrinkX, ShrinkY, WidthR, HeightR: integer;
begin
  DrawR := R;
  WidthR := R.Right - R.Left;
  HeightR := R.Bottom - R.Top;
  MidX := R.Left + (WidthR) div 2;
  MidY := R.Top + (HeightR) div 2;
  ShrinkX := (WidthR) div 4 - 1;
  ShrinkY := (HeightR) div 4 - 1;
  with Canvas do
  begin
    DrawFrameControl(
      Canvas.Handle,
      DrawR,
      DFC_BUTTON,
      DFCS_BUTTONPUSH or DFCS_ADJUSTRECT);
    InflateRect(DrawR, -ShrinkX, -ShrinkY);
    if not Odd(WidthR) then
      DrawR.Right := DrawR.Right + 1;
    if not Odd(HeightR) then
      DrawR.Bottom := DrawR.Bottom + 1;
    Pen.Color := clWindowFrame;
    MoveTo(DrawR.Left, MidY);
    LineTo(DrawR.Right, MidY);
    if not Expanded then
    begin
      MoveTo(MidX, DrawR.Top);
      LineTo(MidX, DrawR.Bottom);
    end;
  end;
end;

{
7. Screen & OS Information routines
}

function abcScreenIsVGA: Boolean;
{ is screen <= 256 Colors ? }
begin
  Result := abcScreenColorDepth <= 8;
end;

function abcScreenIsHighColor: Boolean;
{ is screen > 256 Colors ? }
begin
  Result := abcScreenColorDepth > 8;
end;

function abcScreenColorDepth: Integer;
{ color depth of screen in bits per pixel }
var
  ScreenDC: hDC;
begin
  ScreenDC := GetDC(0);
  Result := GetDeviceCaps(ScreenDC, BITSPIXEL);
  ReleaseDC(0, ScreenDC);
end;

function abcScreenPixelFormat: TPixelFormat;
begin
  case abcScreenColorDepth of
    32: Result := pf32bit;
    24: Result := pf24bit;
    16: Result := pf16bit;
    15: Result := pf15bit;
     8: Result := pf8bit;
     4: Result := pf4bit;
     1: Result := pf1bit;
  else
    Result := pfDevice;
  end;
end;

function abcOSVerIsWin95: Boolean;
{ are we running on Windows 95? }
var
  lpVersionInformation: TOSVersionInfo;
begin
  lpVersionInformation.dwOSVersionInfoSize := SizeOf(TOSVersionInfo);
  GetVersionEx(lpVersionInformation);
  Result := (lpVersionInformation.dwPlatformId = VER_PLATFORM_WIN32_WINDOWS) and
            (lpVersionInformation.dwMajorVersion <= 4);
end;

{
8. Bitmap Bit Manipulation
}

const
  ROP_DSPDxax = $00E20746;
  ROP_PSDPxax = $00B8074A;
  ROP_DSna = $00220326;  { D & ~S }
{  ROPs: array[Boolean] of DWORD = (ROP_PSDPxax, ROP_DSPDxax);  }

procedure abcGrayScaleBitmap(Bitmap: TBitmap);
(* alternate function - a bit slower but may give better grayscale.
  function RgbToGray(RGBColor : TColor) : TColor;
  var  Gray : byte;
  begin
    Gray := Round((0.30 * GetRValue(DWORD(RGBColor))) +
                  (0.59 * GetGValue(DWORD(RGBColor))) +
                  (0.11 * GetBValue(DWORD(RGBColor))));
    Result := RGB(Gray, Gray, Gray);
  end;

var
  x, y: integer;
begin
  with Bitmap do
    for x := 0 to Width - 1 do
      for y := 0 to Height - 1 do
        Canvas.Pixels[x,y] := RgbToGray(Canvas.Pixels[x,y]);
end;
*)
var
  x, y, z: integer;
begin
  with Bitmap do
    for x := 0 to Width - 1 do
      for y := 0 to Height - 1 do
      begin
        z := Canvas.Pixels[x,y] and $00000FF; {and not $0000FFFF is interesting}
        z := z or (z shl 8) or (z shl 16);
        Canvas.Pixels[x,y] := z;
      end;
end;

procedure abcGenerateMaskBitmapFromDIB (const MaskBitmap, SourceBitmap: TBitmap;
  const SourceOffset, SourceSize: TPoint; TransColors: array of TColor);
{ This a special procedure meant for generating monochrome masks from
  >4 bpp color DIB sections. Because each video driver seems to sport its own
  interpretation of how to handle DIB sections, a workaround procedure like
  this was necessary. }
type
  TColorArray = array[0..536870910] of TColorRef;
var
  Info: packed record
    Header: TBitmapInfoHeader;
    Colors: array[0..1] of TColorRef;
  end;
  W, H: Integer;
  I, Y, X: Integer;
  Pixels: ^TColorArray;
  Pixel: ^TColorRef;
  MonoPixels: Pointer;
  MonoPixel, StartMonoPixel: ^Byte;
  MonoScanLineSize, CurBit: Integer;
  DC: HDC;
  MaskBmp: HBITMAP;
begin
  W := SourceBitmap.Width;
  H := SourceBitmap.Height;
  MonoScanLineSize := SourceSize.X div 8;
  if SourceSize.X mod 8 <> 0 then
    Inc (MonoScanLineSize);
  if MonoScanLineSize mod 4 <> 0 then  { Compensate for scan line boundary }
    MonoScanLineSize := (MonoScanLineSize and not 3) + 4;
  MonoPixels := AllocMem(MonoScanLineSize * SourceSize.Y);  { AllocMem is used because it initializes to zero }
  try
    GetMem (Pixels, W * H * 4);
    try
      ZeroMemory(@Info, SizeOf(Info));
      with Info do begin
        with Header do begin
          biSize := SizeOf(TBitmapInfoHeader);
          biWidth := W;
          biHeight := -H;  { negative number makes it a top-down DIB }
          biPlanes := 1;
          biBitCount := 32;
          {biCompression := BI_RGB;}  { implied }
        end;
        {Colors[0] := clBlack;}  { implied }
        Colors[1] := clWhite;
      end;
      DC := CreateCompatibleDC(0);
      GetDIBits (DC, SourceBitmap.Handle, 0, H, Pixels, PBitmapInfo(@Info)^,
        DIB_RGB_COLORS);
      DeleteDC (DC);

      for I := 0 to High(TransColors) do
        if TransColors[I] = -1 then
          TransColors[I] := Pixels[W * (H-1)] and $FFFFFF;
            { ^ 'and' operation is necessary because the high byte is undefined }

      MonoPixel := MonoPixels;
      for Y := SourceOffset.Y to SourceOffset.Y+SourceSize.Y-1 do begin
        StartMonoPixel := MonoPixel;
        CurBit := 7;
        Pixel := @Pixels[(Y * W) + SourceOffset.X];
        for X := 0 to SourceSize.X-1 do begin
          for I := 0 to High(TransColors) do
            if Pixel^ and $FFFFFF = Cardinal(TransColors[I]) then begin
              { ^ 'and' operation is necessary because the high byte is undefined }
              MonoPixel^ := MonoPixel^ or (1 shl CurBit);
              Break;
            end;
          Dec (CurBit);
          if CurBit < 0 then begin
            Inc (Integer(MonoPixel));
            CurBit := 7;
          end;
          Inc (Integer(Pixel), SizeOf(Longint));  { proceed to the next pixel }
        end;
        Integer(MonoPixel) := Integer(StartMonoPixel) + MonoScanLineSize;
      end;
    finally
      FreeMem (Pixels);
    end;

    { Write new bits into a new HBITMAP, and assign this handle to MaskBitmap }
    MaskBmp := CreateBitmap(SourceSize.X, SourceSize.Y, 1, 1, nil);
    with Info.Header do begin
      biWidth := SourceSize.X;
      biHeight := -SourceSize.Y;  { negative number makes it a top-down DIB }
      biPlanes := 1;
      biBitCount := 1;
    end;
    DC := CreateCompatibleDC(0);
    SetDIBits (DC, MaskBmp, 0, SourceSize.Y, MonoPixels, PBitmapInfo(@Info)^,
      DIB_RGB_COLORS);
    DeleteDC (DC);
  finally
    FreeMem (MonoPixels);
  end;

  MaskBitmap.Handle := MaskBmp;
end;

procedure abcGenerateMaskBitmap (const MaskBitmap, SourceBitmap: TBitmap;
  const SourceOffset, SourceSize: TPoint; const TransColors: array of TColor);
{ Returns handle of a monochrome bitmap, with pixels in SourceBitmap of color
  TransColor set to white in the resulting bitmap. All other colors of
  SourceBitmap are set to black in the resulting bitmap. This uses the
  regular ROP_DSPDxax BitBlt method. }
var
  CanvasHandle: HDC;
  SaveBkColor: TColorRef;
  DC: HDC;
  MaskBmp, SaveBmp: HBITMAP;
  I: Integer;
const
  ROP: array[Boolean] of DWORD = (SRCPAINT, SRCCOPY);
begin
  CanvasHandle := SourceBitmap.Canvas.Handle;

  MaskBmp := CreateBitmap(SourceSize.X, SourceSize.Y, 1, 1, nil);
  DC := CreateCompatibleDC(0);
  SaveBmp := SelectObject(DC, MaskBmp);
  SaveBkColor := GetBkColor(CanvasHandle);
  for I := 0 to High(TransColors) do begin
    SetBkColor (CanvasHandle, ColorToRGB(TransColors[I]));
    BitBlt (DC, 0, 0, SourceSize.X, SourceSize.Y, CanvasHandle,
      SourceOffset.X, SourceOffset.Y, ROP[I = 0]);
  end;
  SetBkColor (CanvasHandle, SaveBkColor);
  SelectObject (DC, SaveBmp);
  DeleteDC (DC);

  MaskBitmap.Handle := MaskBmp;
end;

procedure abcReplaceBitmapColorsFromMask (const MaskBitmap, DestBitmap: TBitmap;
  const DestOffset, DestSize: TPoint; const ReplaceColor: TColor);
var
  DestDC: HDC;
  SaveBrush: HBRUSH;
  SaveTextColor, SaveBkColor: TColorRef;
begin
  DestDC := DestBitmap.Canvas.Handle;

  SaveBrush := SelectObject(DestDC, CreateSolidBrush(ColorToRGB(ReplaceColor)));
  SaveTextColor := SetTextColor(DestDC, clBlack);
  SaveBkColor := SetBkColor(DestDC, clWhite);
  BitBlt (DestDC, DestOffset.X, DestOffset.Y, DestSize.X, DestSize.Y,
    MaskBitmap.Canvas.Handle, 0, 0, ROP_DSPDxax);
  SetBkColor (DestDC, SaveBkColor);
  SetTextColor (DestDC, SaveTextColor);
  DeleteObject (SelectObject(DestDC, SaveBrush));
end;

function abcCopyBitmapToDDB (const SourceBitmap: TBitmap): TBitmap;
{ Makes a device-dependent duplicate of SourceBitmap. The color palette,
  if any, is preserved. }
var
  SB: HBITMAP;
  SavePalette: HPALETTE;
  DC: HDC;
  BitmapInfo: packed record
    Header: TBitmapInfoHeader;
    Colors: array[0..255] of TColorRef;
  end;
  Bits: Pointer;
begin
  Result := TBitmap.Create;
  try
    Result.Palette := abcCopyPalette(SourceBitmap.Palette);
    Result.Width := SourceBitmap.Width;
    Result.Height := SourceBitmap.Height;
    SB := SourceBitmap.Handle;
    if SB = 0 then Exit;  { it would have a null handle if its width or height was zero }
    SavePalette := 0;
    DC := CreateCompatibleDC(0);
    try
      if Result.Palette <> 0 then begin
        SavePalette := SelectPalette(DC, Result.Palette, False);
        RealizePalette (DC);
      end;
      BitmapInfo.Header.biSize := SizeOf(TBitmapInfoHeader);
      BitmapInfo.Header.biBitCount := 0;  { instructs GetDIBits not to fill in the color table }
      { First retrieve the BitmapInfo header only }
      if GetDIBits(DC, SB, 0, 0, nil, PBitmapInfo(@BitmapInfo)^, DIB_RGB_COLORS) <> 0 then begin
        GetMem (Bits, BitmapInfo.Header.biSizeImage);
        try
          { Then read the actual bits }
          if GetDIBits(DC, SB, 0, SourceBitmap.Height, Bits, PBitmapInfo(@BitmapInfo)^, DIB_RGB_COLORS) <> 0 then
            { And copy them to the resulting bitmap }
            SetDIBits (DC, Result.Handle, 0, SourceBitmap.Height, Bits, PBitmapInfo(@BitmapInfo)^, DIB_RGB_COLORS);
        finally
          FreeMem (Bits);
        end;
      end;
    finally
      if SavePalette <> 0 then SelectPalette (DC, SavePalette, False);
      DeleteDC (DC);
    end;
  except
    Result.Free;
    raise;
  end;
end;

function abcCopyPalette (Palette: HPALETTE): HPALETTE;
begin
  Result := CopyPalette(Palette);
end;

function abcLoadBitmapByName(S: String): HBitmap;
var
  J: Integer;
begin
  J := 1;
  while J <= Length(S) do
    if ((S[J] >= 'A') and (S[J] <= 'Z')) or (S[J] = '_') then
      Inc(J)
    else if (S[J] >= 'a') and (S[J] <= 'z') then
      Dec(S[J], 32)
    else
      System.Delete(S, J, 1);
  Result := FindResource(hInstance, PChar(S), RT_BITMAP);
  if Result <> 0 then
    Result := LoadBitmap(hInstance, PChar(S));
end;

function abcLightenCaptionColor(clr: TColor): TColor;
var
  R, G, B: Byte;
begin
  R := GetRValue(DWORD(clr));
  G := GetGValue(DWORD(clr));
  B := GetBValue(DWORD(clr));
  //hack special case for windows standard colors
  if (R = 0) and (G = 0) and (B = 128) then
    Result := RGB(16, 132, 208)
  else if (R = 128) and (G = 128) and (B = 128) then
    Result := RGB(184, 180, 184)
  else
    Result := RGB(
      127 + R div 2,
      127 + G div 2,
      127 + B div 2);
end;

end.





