unit abcutilm;

(*------------------------------------------------------------------------------

Advanced Business Components
ABC for Delphi
Version 6

Copyright  1995, 2001 Objective Software Technology Pty Limited ACN 068 640 353
Functionality derived from Delphi VCL Source Code
is Copyright  1995, 2001 Borland Software Corporation

------------------------------------------------------------------------------*)

{$I abcdefs.inc}

{ General purpose utility math functions }

interface

function abcCeil(const X: Extended): Integer;
  {Returns the number rounded up to the next largest integer}

function abcMin(const X, Y: Integer): Integer;
  {Returns the minimum of two integers}

function abcMax(const X, Y: Integer): Integer;
  {Returns the maximum of two integers}

function abcMinArray(const A: array of Integer): Integer;
  {Returns the minimum of n integers}

function abcMaxArray(const A: array of Integer): Integer;
  {Returns the maximum of n integers}

function abcMinMax(const Value, Min, Max: Integer): Integer;
  {Returns Value or the nearest boundary if Value is outside Min..Max}

function abcMinMaxNotDefault(const Value, Min, Max, Default: Integer): Integer;
  {Returns Value or the nearest non-default boundary if Value is outside Min..Max}

function abcVarRecToVariant(const Value: TVarRec): Variant;
  {convert a TVarRec into a variant}

function abcConstArrayToVariantArray(const Values: array of const): Variant;
  {convert a constant array into a variant array}

type
  abcVarRecArray = array of TVarRec;

function abcVariantToVarRecArray(const Value: Variant): abcVarRecArray;
  {convert a constant array into a variant array}

function abcVariantToVarRec(const Value: Variant): TVarRec;

implementation

uses SysUtils, SysConst, Variants 
//     {$IFDEF DELPHI7}, Variants {$ENDIF}
     ;

function abcCeil(const X: Extended): Integer;
begin
  Result := Trunc(X);
  if Frac(X) > 0 then
    Inc(Result);
end;

{ Utility functions for integer compares }

function abcMin(const X, Y: Integer): Integer;
begin
  if X < Y then
    Result := X
  else
    Result := Y;
end;

function abcMax(const X, Y: Integer): Integer;
begin
  if X > Y then
    Result := X
  else
    Result := Y;
end;

function abcMinArray(const A: array of Integer): Integer;
var
  aSize, i: integer;
begin
  aSize := SizeOf(A);
  if aSize > 1 then
  begin
    Result := A[0];
    for i := 1 to aSize - 1 do
      Result := abcMin(Result, A[i]);
  end
  else if aSize = 1 then
    Result := A[0]
  else
    Result := 0;
end;

function abcMaxArray(const A: array of Integer): Integer;
var
  aSize, i: integer;
begin
  aSize := SizeOf(A);
  if aSize > 1 then
  begin
    Result := A[0];
    for i := 1 to aSize - 1 do
      Result := abcMax(Result, A[i]);
  end
  else if aSize = 1 then
    Result := A[0]
  else
    Result := 0;
end;

function abcMinMax(const Value, Min, Max: Integer): Integer;
begin
  if Value < Min then
    Result := Min
  else if Value > Max then
    Result := Max
  else
    Result := Value;
end;

function abcMinMaxNotDefault(const Value, Min, Max, Default: Integer): Integer;
begin
  if (Value < Min) and (Min <> Default) then
    Result := Min
  else if (Value > Max) and (Max <> Default) then
    Result := Max
  else
    Result := Value;
end;

function abcVarRecToVariant(const Value: TVarRec): Variant;
begin
  with Value do
    case VType of
      vtInteger    : Result := VInteger;
      vtBoolean    : Result := VBoolean;
      vtChar       : Result := VChar;
      vtExtended   : Result := VExtended^;
      vtString     : Result := VString^;
      vtPChar      : Result := VPChar^;
      vtAnsiString : Result := string(VAnsiString);
      vtCurrency   : Result := VCurrency^;
      vtVariant    : if not VarIsEmpty(VVariant^) then
                       Result := VVariant^;
{$IFDEF DELPHI7}
      vtInt64:
        Result := VInt64^;
{$ENDIF}
      else
        raise EVariantError.Create(SInvalidVarOp);
    end;
end;

function abcConstArrayToVariantArray(const Values: array of const): Variant;
{convert a variant into a TVarRec}
var
  i, lower, upper: integer;
begin
  lower := Low(Values);
  upper := High(Values);
  Result := VarArrayCreate([lower, upper], varVariant);
  for i := lower to upper do
    Result[i] := abcVarRecToVariant(Values[i]);
end;

function abcVariantToVarRec(const Value: Variant): TVarRec;
begin
  case VarType(Value) of
    varSmallint:
      Result.VInteger := TVarData(Value).VSmallint;
    varInteger:
      Result.VInteger := TVarData(Value).VInteger;
    varSingle:
      Result.VExtended^ := TVarData(Value).VSingle;
    varDouble:
      Result.VExtended^ := TVarData(Value).VDouble;
    varCurrency:
      Result.VCurrency^ := TVarData(Value).VCurrency;
    varDate:
      Result.VExtended^ := TVarData(Value).VDate;
    varOleStr:
      Result.VPWideChar := TVarData(Value).VOleStr;
    varDispatch:
      Result.VPointer := TVarData(Value).VDispatch;
    varError:
      Result.VInteger := TVarData(Value).VError;
    varBoolean:
      Result.VBoolean := TVarData(Value).VBoolean;
    varUnknown:
      Result.VPointer := TVarData(Value).VUnknown;
    varByte:
      Result.VInteger := TVarData(Value).VByte;
    varString:
      Result.VAnsiString := TVarData(Value).VString;
    varAny:
      Result.VPointer := TVarData(Value).VAny;
    varArray:
      Result.VVariant^ := Value;
    varByRef:
      Result.VPointer := TVarData(Value).VPointer;
{$IFDEF DELPHI7}
    varShortInt:
      Result.VInteger := TVarData(Value).VShortInt;
    varWord:
      Result.VInteger := TVarData(Value).VWord;
    varLongWord:
      Result.VInteger := TVarData(Value).VLongWord;
    varInt64:
      Result.VInt64^ := TVarData(Value).VInt64;
{$ENDIF}
  end;
end;

function abcVariantToVarRecArray(const Value: Variant): abcVarRecArray;
{convert a variant into an array of TVarRec}
var
  arrSize, i: Integer;
begin
  if VarIsArray(Value) then
  begin
    arrSize := VarArrayHighBound(Value, 1) - VarArrayLowBound(Value, 1);
    SetLength(Result, arrSize);
    for i := 0 to arrSize - 1 do
      Result[i] := abcVariantToVarRec(Value[i]);
  end
  else
  begin
    SetLength(Result, 1);
    Result[0] := abcVariantToVarRec(Value);
  end;
end;

end.
