unit abcwanim;

(*------------------------------------------------------------------------------

Advanced Business Components
ABC for Delphi
Version 6

Copyright  1995, 2001 Objective Software Technology Pty Limited ACN 068 640 353
Functionality derived from Delphi VCL Source Code
is Copyright  1995, 2001 Borland Software Corporation

------------------------------------------------------------------------------*)

{$I abcdefs.inc}

interface

uses
  Windows, Classes, Controls;

type

  { TabcAnimateWin }

  {::<P>Defines possible AnimateWindow effects:
  <UL>
  <LI>aweNone - no effect to be shown.
  <LI>aweRoll - reveals or hides a window in the specified directions.
  <LI>aweSlide - moves a window into/out of view in the specified directions.
  <LI>aweBlend - reveals or hides a window using a blend effect. The direction is not used.
  <LI>aweCenter - reveals or hides a window, rolling from the center.
  </UL>
  }
  TabcAnimEffect = (aweNone, aweRoll, aweSlide, aweBlend, aweCenter);

  {::The direction of AnimateWindow effects, specified for both the horizontal and vertical axes.}

  TabcAnimDirection = (awdNone, awdPositive, awdNegative);

  {:A persistent class to encapsulate Windows 2000 AnimateWindow effects.
  This class allows AnimateWindow effects to be added to WinControls.<P>
  It is used in the <A HRef="TabcSplash_Component">TabcSplash</A> component.<P>
  Note there are some limitations to using these effects:<P>
  - If the window has a shaped region, only the aweBlend effect is possible.<br>
  - When hiding a window, using the Slide, Center or Roll effects;
    forms under the window being hidden are not repainted during the effect.
  }

  TabcAnimateWin = class(TPersistent)
  private
    FDuration: Word;
    FEffect: TabcAnimEffect;
    FHorz: TabcAnimDirection;
    FVert: TabcAnimDirection;
    FOnChange: TNotifyEvent;
    function GetWinFlags(Showing: Boolean): Integer;
    procedure SetDuration(const Value: Word);
    procedure SetEffect(const Value: TabcAnimEffect);
    procedure SetHorz(const Value: TabcAnimDirection);
    procedure SetVert(const Value: TabcAnimDirection);
  protected
    procedure Changed;
  public
    constructor Create;
    procedure Assign(Source: TPersistent); override;
    {::Hides the specified window using the effect.}
    function Hide(const WinControl: TWinControl): Boolean;
    {::Shows the specified window using the effect.}
    function Show(const WinControl: TWinControl): Boolean;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
  published
    {::Duration of the animation effect in milliseconds.}
    property Duration: Word read FDuration write SetDuration default 200;
    {::The roll, slide, blend or center effect to be painted.}
    property Effect: TabcAnimEffect read FEffect write SetEffect default aweNone;
    {::Direction of the effect on the horizontal axis. This is ignored for Blend effects.}
    property Horz: TabcAnimDirection read FHorz write SetHorz default awdNone;
    {::Direction of the effect on the vertical axis. This is ignored for Blend effects.}
    property Vert: TabcAnimDirection read FVert write SetVert default awdNone;
  end;

implementation

{$IFNDEF DELPHI6}
var
  AnimateWindowProc: function(hWnd: HWND; dwTime: DWORD; dwFlags: DWORD): BOOL; stdcall;
{$ENDIF}

{ TabcAnimateWin }

constructor TabcAnimateWin.Create;
begin
  inherited Create;
  FDuration := 200;
  FEffect := aweNone;
  FHorz := awdNone;
  FVert := awdNone;
end;

procedure TabcAnimateWin.Assign(Source: TPersistent);
begin
  if Source is TabcAnimateWin then
    with TabcAnimateWin(Source) do
    begin
      Self.FDuration := Duration;
      Self.FEffect := Effect;
      Self.FHorz := Horz;
      Self.FVert := Vert;
      Changed;
    end
  else
    inherited;
end;

procedure TabcAnimateWin.Changed;
begin
  if Assigned(OnChange) then
    OnChange(Self);
end;

function TabcAnimateWin.GetWinFlags(Showing: Boolean): Integer;
(*
Note1: if window region is changed using SetWindowRgn
       the blend effect is the only one that works
Note2: There is a bug in windows or the VCL or the window style properties
       When using AW_HIDE the AW_SLIDE, AW_CENTER and roll default effects
       forms under the one hiding fail to repaint
*)
const
  SHOWS: array[Boolean] of integer = (AW_HIDE, 0);
  EFFECTS: array[TabcAnimEffect] of integer = (0, 0, AW_SLIDE, AW_BLEND, AW_CENTER);
  DIRNH: array[TabcAnimDirection] of integer = (0, AW_HOR_POSITIVE, AW_HOR_NEGATIVE);
  DIRNV: array[TabcAnimDirection] of integer = (0, AW_VER_POSITIVE, AW_VER_NEGATIVE);
begin
  Result := SHOWS[Showing] + EFFECTS[FEffect] + DIRNH[FHorz] + DIRNV[FVert];
end;

function TabcAnimateWin.Hide(const WinControl: TWinControl): Boolean;
begin
  Result :=
    (Effect <> aweNone) and
    Assigned(AnimateWindowProc) and
    (WinControl <> nil) and
    AnimateWindowProc(WinControl.Handle, FDuration, GetWinFlags(False));
  ShowWindow(WinControl.Handle, SW_HIDE);
end;

function TabcAnimateWin.Show(const WinControl: TWinControl): Boolean;
begin
  Result :=
    (Effect <> aweNone) and
    Assigned(AnimateWindowProc) and
    (WinControl <> nil) and
    AnimateWindowProc(WinControl.Handle, FDuration, GetWinFlags(True));
  if Result then
    WinControl.Invalidate;
end;

procedure TabcAnimateWin.SetDuration(const Value: Word);
begin
  if Value <> FDuration then
  begin
    FDuration := Value;
    Changed;
  end;
end;

procedure TabcAnimateWin.SetEffect(const Value: TabcAnimEffect);
begin
  if Value <> FEffect then
  begin
    FEffect := Value;
    // [aweRoll, aweSlide] effects won't work without any direction
    if (Value in [aweRoll, aweSlide]) and
       (FHorz = awdNone) and
       (FVert = awdNone) then
      FHorz := awdPositive;
    Changed;
  end;
end;

procedure TabcAnimateWin.SetHorz(const Value: TabcAnimDirection);
begin
  if Value <> FHorz then
  begin
    FHorz := Value;
    if (Value = awdNone) and
       (FVert = awdNone) and
       (Effect in [aweRoll, aweSlide]) then
      FEffect := aweNone;
    Changed;
  end;
end;

procedure TabcAnimateWin.SetVert(const Value: TabcAnimDirection);
begin
  if Value <> FVert then
  begin
    FVert := Value;
    if (Value = awdNone) and
       (FHorz = awdNone) and
       (Effect in [aweRoll, aweSlide]) then
      FEffect := aweNone;
    Changed;
  end;
end;

{$IFNDEF DELPHI6}
procedure InitAnimateProc;
var
  hUser32: HMODULE;
begin
  hUser32 := GetModuleHandle('USER32');
  if hUser32 <> 0 then
    @AnimateWindowProc := GetProcAddress(hUser32, 'AnimateWindow');
end;

initialization
  InitAnimateProc;

{$ENDIF}

end.
