unit abcwav;

(*------------------------------------------------------------------------------

Advanced Business Components
ABC for Delphi
Version 6

Copyright  1995, 2001 Objective Software Technology Pty Limited ACN 068 640 353
Functionality derived from Delphi VCL Source Code
is Copyright  1995, 2001 Borland Software Corporation

------------------------------------------------------------------------------*)

{$I abcdefs.inc}

interface

uses
  SysUtils, Classes, Windows, abcmem;

type

  EabcInvalidWaveOperation = class(Exception);

  { TabcWaveImage }

  TabcWaveImage = class(TabcPersistentMemory)
  protected
    procedure ValidateDataFormat(Stream: TMemoryStream); override;
  public
    procedure CopyToClipboard; override;
    procedure PasteFromClipboard; override;
  end;

  { TabcCustomWave }

  TabcPlayOption = (poAsync, poSync, poLoop);

  TabcCustomWave = class(TComponent)
  private
    FFlags: Word;
    FPlayOption: TabcPlayOption;
    FSoundPtr: pChar;
  protected
    procedure SetPlayOption(Value: TabcPlayOption); virtual;
    property Flags: Word read FFlags write FFlags;               {NB not to be published}
    property SoundPointer: pChar read FSoundPtr write FSoundPtr; {NB not to be published}
    property PlayOption: TabcPlayOption read FPlayOption write SetPlayOption default poAsync;
  public
    constructor Create(AOwner: TComponent); override;
    function Play: Boolean; virtual;
    procedure Stop; virtual;
  end;

  { TabcCustomWaveList }

  TabcCustomWaveList = class(TabcCustomWave)
  private
    FItemIndex: Integer;
    FItems: TStrings;
    procedure SetItems(Value: TStrings);
  protected
    property ItemIndex: Integer read FItemIndex write FItemIndex default -1;
    property Items: TStrings read FItems write SetItems;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function PlayAll: Boolean;
    function PlayItem(Index: Integer): Boolean; virtual; abstract;
    function PlayRange(FromIndex, ToIndex: Integer): Boolean;
    function PlaySelection(const IntArray: array of Integer): Boolean;
  end;

  { TabcWave }

  TabcWave = class(TabcCustomWave)
  private
    FSound: TabcWaveImage;
    procedure SetSound(Value: TabcWaveImage);
    procedure WaveImageChanged(Sender: TObject);
    procedure WaveImageChanging(Sender: TObject);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function Play: Boolean; override;
  published
    property PlayOption;
    property Sound: TabcWaveImage read FSound write SetSound;
  end;

  { TabcWaveList }

  TabcWaveList = class(TabcCustomWaveList)
  private
    function GetCount: Integer;
    function GetItems: TStrings;
    function GetSound(Index: Integer): TabcWaveImage;
    function GetSoundName(Index: Integer): string;
    procedure ReadData(Stream: TStream);
    procedure SetItems(Value: TStrings);
    procedure WaveImageChanging(Sender: TObject);
    procedure WriteData(Stream: TStream);
  protected
    procedure DefineProperties(Filer: TFiler); override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function AddSound(SoundName: string; Sound: TabcWaveImage): Integer;
    procedure Clear;
    procedure Delete(Index: Integer);
    function Play: Boolean; override;
    function PlayItem(Index: Integer): Boolean; override;
    property Count: Integer read GetCount;
    property Sound[Index: Integer]: TabcWaveImage read GetSound;
    property SoundName[Index: Integer]: string read GetSoundName;
  published
    property ItemIndex;
    property Items: TStrings read GetItems write SetItems;
    property PlayOption;
  end;

  { TabcWaveFile }

  TabcWaveFile = class(TabcCustomWave)
  private
    FFileName: TFileName;
    FFileNamePtr: pChar;
    procedure SetFileName(Value: TFileName);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function Play: Boolean; override;
    procedure SetPlayOption(Value: TabcPlayOption); override;
  published
    property FileName: TFileName read FFileName write SetFileName;
    property PlayOption;
  end;

  { TabcWaveFileList }

  TabcWaveFileList = class(TabcCustomWaveList)
  private
    FFileExt: string;
    FFileName: TFileName;
    FFileNamePtr: pChar;
    FFilePath: TFileName;
    procedure SetFileName(Value: TFileName);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function Play: Boolean; override;
    function PlayItem(Index: Integer): Boolean; override;
    procedure SetPlayOption(Value: TabcPlayOption); override;
  published
    property FileExt: string read FFileExt write FFileExt;
    property FilePath: TFileName read FFilePath write FFilePath;
    property ItemIndex;
    property Items;
    property PlayOption;
  end;

  { TabcWaveRes }

  TabcWaveRes = class(TabcCustomWave)
  private
    FResourceDLL: string;
    FResourceDLLHandle: THandle;
    FResourceHandle: THandle;
    FResourceName: string;
    procedure FreeResourceHandle;
    procedure SetResourceHandle;
  protected
  public
    destructor Destroy; override;
    function Play: Boolean; override;
    procedure Stop; override;
  published
    property PlayOption;
    property ResourceDLL: string read FResourceDLL write FResourceDLL;
    property ResourceName: string read FResourceName write FResourceName;
  end;

  { TabcWaveResList }

  TabcWaveResList = class(TabcCustomWaveList)
  private
    FResourceDLL: string;
    FResourceDLLHandle: THandle;
    FResourceHandle: THandle;
    procedure FreeResourceHandle;
    procedure SetResourceHandle(AResName: string);
  public
    destructor Destroy; override;
    function Play: Boolean; override;
    function PlayItem(Index: Integer): Boolean; override;
    procedure Stop; override;
  published
    property ItemIndex;
    property Items;
    property PlayOption;
    property ResourceDLL: string read FResourceDLL write FResourceDLL;
  end;

implementation

uses MMSystem, abcconst;

{ TabcCustomWave }

constructor TabcCustomWave.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FPlayOption := poAsync;
  FFlags := SND_MEMORY or SND_ASYNC or SND_NODEFAULT;
end;

function TabcCustomWave.Play: Boolean;
begin
  if FSoundPtr = nil then
    Result := False
  else
    Result := sndPlaySound(FSoundPtr, FFlags);
end;

procedure TabcCustomWave.SetPlayOption(Value: TabcPlayOption);
begin
  if Value <> FPlayOption then begin
    if not (csDesigning in ComponentState) then Stop;
    case Value of
      poAsync:
        Flags := SND_MEMORY or SND_ASYNC or SND_NODEFAULT;
      poSync:
        Flags := SND_MEMORY or SND_SYNC or SND_NODEFAULT;
      poLoop:
        Flags := SND_MEMORY or SND_ASYNC or SND_NODEFAULT or SND_LOOP;
    end;
    FPlayOption := Value;
  end;
end;

procedure TabcCustomWave.Stop;
begin
  sndPlaySound(nil, 0);
end;

{ TabcCustomWaveList }

constructor TabcCustomWaveList.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FItems := TStringList.Create;
  FItemIndex := -1;
end;

destructor TabcCustomWaveList.Destroy;
begin
  FItems.Free;
  inherited Destroy;
end;

function TabcCustomWaveList.PlayAll: Boolean;
begin
  if FItems.Count = 0 then
    Result := False
  else
    Result := PlayRange(0, FItems.Count - 1);
end;

function TabcCustomWaveList.PlayRange(FromIndex, ToIndex: Integer): Boolean;
var
  I: Integer;
  SaveOption: TabcPlayOption;
begin
  Result := False;
  if FItems.Count = 0 then Exit;
  SaveOption := PlayOption;
  PlayOption := poSync;
  Result := True;
  if FromIndex < ToIndex then
    for I := FromIndex to ToIndex do
      Result := Result and PlayItem(I)
  else
    for I := FromIndex downto ToIndex do
      Result := Result and PlayItem(I);
  PlayOption := SaveOption;
end;

function TabcCustomWaveList.PlaySelection(const IntArray: array of Integer): Boolean;
var
  I: Integer;
  SaveOption: TabcPlayOption;
begin
  Result := False;
  if FItems.Count = 0 then Exit;
  SaveOption := PlayOption;
  PlayOption := poSync;
  Result := True;
  for I := Low(IntArray) to High(IntArray) do
    if IntArray[I] <> -1 then
      if (IntArray[I] > -1) and (IntArray[I] < Items.Count) then
        Result := Result and PlayItem(IntArray[I])
      else
        Result := False;
  PlayOption := SaveOption;
end;

procedure TabcCustomWaveList.SetItems(Value: TStrings);
begin
  if Value <> FItems then
    FItems.Assign(Value);
end;

{ TabcWave }

constructor TabcWave.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FSound := TabcWaveImage.Create;
  FSound.OnChange := WaveImageChanged;
  FSound.OnChanging := WaveImageChanging;
end;

destructor TabcWave.Destroy;
begin
  FSound.Free;
  inherited Destroy;
end;

function TabcWave.Play: Boolean;
begin
  if FSound.Empty then
    Result := False
  else
    Result := inherited Play;
end;

procedure TabcWave.SetSound(Value: TabcWaveImage);
begin
  if Value <> FSound then
    FSound.Assign(Value);
end;

procedure TabcWave.WaveImageChanged(Sender: TObject);
begin
  if FSound.Empty then
    SoundPointer := nil
  else
    SoundPointer := FSound.Memory;
end;

procedure TabcWave.WaveImageChanging(Sender: TObject);
begin
  if not FSound.Empty and (PlayOption = poAsync) then Stop;
end;

{ TabcWaveList }

constructor TabcWaveList.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
end;

destructor TabcWaveList.Destroy;
begin
  Clear;
  inherited Destroy;
end;

function TabcWaveList.AddSound(SoundName: string; Sound: TabcWaveImage): Integer;
begin
  Sound.OnChanging := WaveImageChanging;
  Result := Items.AddObject(SoundName, Sound);
end;

procedure TabcWaveList.Clear;
var
  I: integer;
  Sound: TabcWaveImage;
begin
  for I := Items.Count - 1 downto 0 do begin
    Sound := TabcWaveImage(Items.Objects[I]);
    Items.Delete(I);
    Sound.Free;
  end;
end;

procedure TabcWaveList.DefineProperties(Filer: TFiler);
begin
  inherited DefineProperties(Filer);
  Filer.DefineBinaryProperty('Data', ReadData, WriteData, Items.Count > 0);
end;

procedure TabcWaveList.Delete(Index: Integer);
var
  Sound: TabcWaveImage;
begin
  Sound := TabcWaveImage(Items.Objects[Index]);
  Items.Delete(Index);
  Sound.Free;
end;

function TabcWaveList.GetCount: Integer;
begin
  Result := Items.Count;
end;

function TabcWaveList.GetItems: TStrings;
begin
  Result := inherited Items;
end;

function TabcWaveList.GetSound(Index: Integer): TabcWaveImage;
begin
  Result := TabcWaveImage(Items.Objects[Index]);
end;

function TabcWaveList.GetSoundName(Index: Integer): string;
begin
  Result := Items.Strings[Index];
end;

function TabcWaveList.Play: Boolean;
begin
  Result := PlayItem(ItemIndex);
end;

function TabcWaveList.PlayItem(Index: Integer): Boolean;
begin
  Result := False;
  if (Index < 0) or (Index >= Items.Count) then Exit;
  if (Items.Objects[Index] <> nil)
    and not TabcWaveImage(Items.Objects[Index]).Empty then
  begin
    SoundPointer := TabcWaveImage(Items.Objects[Index]).Memory;
    Result := inherited Play;
  end;
end;

procedure TabcWaveList.ReadData(Stream: TStream);
var
  I: Integer;
  Sound: TabcWaveImage;
begin
  for I := 0 to Items.Count - 1 do begin
    Sound := TabcWaveImage.Create;
    Sound.ReadData(Stream);
    Sound.OnChanging := WaveImageChanging;
    Items.Objects[I] := Sound;
  end;
end;

procedure TabcWaveList.SetItems(Value: TStrings);
var
  I: Integer;
begin
  if Value <> inherited Items then begin
    Clear;
    inherited Items := Value;
    for I := 0 to Items.Count - 1 do
      TabcWaveImage(Items.Objects[I]).OnChanging := WaveImageChanging;
  end;
end;

procedure TabcWaveList.WaveImageChanging(Sender: TObject);
begin
  if (PlayOption = poAsync) then Stop;
end;

procedure TabcWaveList.WriteData(Stream: TStream);
var
  I: Integer;
begin
  for I := 0 to Items.Count - 1 do
    TabcWaveImage(Items.Objects[I]).WriteData(Stream);
end;

{ TabcWaveFile }

constructor TabcWaveFile.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Flags := Flags and not SND_MEMORY;
  FFileNamePtr := nil;
end;

destructor TabcWaveFile.Destroy;
begin
  if FFileNamePtr <> nil then begin
    FreeMem(FFileNamePtr, Length(FFileName) + 1);
    FFileNamePtr := nil;
    SoundPointer := nil;
  end;
  inherited Destroy;
end;

function TabcWaveFile.Play: Boolean;
begin
  if FFileName = '' then
    Result := False
  else
    Result := inherited Play;
end;

procedure TabcWaveFile.SetFileName(Value: TFileName);
begin
  if Value <> FFileName then begin
    if FFileNamePtr <> nil then begin
      FreeMem(FFileNamePtr, Length(FFileName) + 1);
      FFileNamePtr := nil;
    end;
    if Value <> '' then begin
      FFileNamePtr := AllocMem(Length(Value) + 1);
      StrPCopy(FFileNamePtr, Value);
      SoundPointer := FFileNamePtr;
    end
    else
      SoundPointer := nil;
    FFileName := Value;
  end;
end;

procedure TabcWaveFile.SetPlayOption(Value: TabcPlayOption);
begin
  inherited SetPlayOption(Value);
  Flags := Flags and not SND_MEMORY;
end;

{ TabcWaveFileList }

constructor TabcWaveFileList.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Flags := Flags and not SND_MEMORY;
  FFileNamePtr := nil;
  FFileExt := 'WAV';
end;

destructor TabcWaveFileList.Destroy;
begin
  if FFileNamePtr <> nil then begin
    FreeMem(FFileNamePtr, Length(FFileName) + 1);
    FFileNamePtr := nil;
    SoundPointer := nil;
  end;
  inherited Destroy;
end;

function TabcWaveFileList.Play: Boolean;
begin
  Result := PlayItem(ItemIndex);
end;

function TabcWaveFileList.PlayItem(Index: Integer): Boolean;
var
  FullFileName: TFileName;
begin
  Result := False;
  if (Index < 0) or (Index >= Items.Count) then Exit;
  FullFileName := FFilePath + '\' + Items.Strings[Index];
  if FFileExt <> '' then
    FullFileName := FullFileName + '.' + FFileExt;
  SetFileName(FullFileName);
  if FFileName <> '' then
    Result := inherited Play;
end;

procedure TabcWaveFileList.SetFileName(Value: TFileName);
begin
  if Value <> FFileName then begin
    if FFileNamePtr <> nil then begin
      FreeMem(FFileNamePtr, Length(FFileName) + 1);
      FFileNamePtr := nil;
    end;
    if Value <> '' then begin
      FFileNamePtr := AllocMem(Length(Value) + 1);
      StrPCopy(FFileNamePtr, Value);
      SoundPointer := FFileNamePtr;
    end
    else
      SoundPointer := nil;
    FFileName := Value;
  end;
end;

procedure TabcWaveFileList.SetPlayOption(Value: TabcPlayOption);
begin
  inherited SetPlayOption(Value);
  Flags := Flags and not SND_MEMORY;
end;

{ TabcWaveRes }

destructor TabcWaveRes.Destroy;
begin
  if FResourceHandle <> 0 then Stop;
  inherited Destroy;
end;

procedure TabcWaveRes.SetResourceHandle;
{
  Loads DLL if required;
  Loads Resource.
  Sets FResourceDLLHandle and FResourceHandle.
}
var
  hResInfo: THandle;
  ErrorMsg: string;
begin
  if FResourceHandle <> 0 then Stop;
  if FResourceName = '' then Exit;
  try
    if FResourceDLL <> '' then begin
      FResourceDLLHandle := LoadLibrary(PChar(FResourceDLL));
      if FResourceDLLHandle < HINSTANCE_ERROR then begin
        ErrorMsg := Format(sabcErrResDllLoad, [FResourceDLLHandle, FResourceDLL]);
        FResourceDLLHandle := 0;
        raise EabcInvalidWaveOperation.Create(ErrorMsg);
      end;
      hResInfo := FindResource(FResourceDLLHandle, PChar(FResourceName), RT_RCDATA);
    end
    else
      hResInfo := FindResource(HInstance, PChar(FResourceName), RT_RCDATA);
    if hResInfo = 0 then
      raise EResNotFound.Create(Format(sabcErrRCDATAResNotFound, [FResourceName]));
    if FResourceDLLHandle <> 0 then
      FResourceHandle := LoadResource(FResourceDLLHandle, hResInfo)
    else
      FResourceHandle := LoadResource(HInstance, hResInfo);
    if FResourceHandle = 0 then
      raise EResNotFound.Create(Format(sabcErrResLoad, [FResourceName]));
  except
    on Exception do begin
      FreeResourceHandle;
      raise;
    end;
  end;
end;

procedure TabcWaveRes.FreeResourceHandle;
begin
  if FResourceHandle <> 0 then begin
    UnlockResource(FResourceHandle);
    FreeResource(FResourceHandle);
    FResourceHandle := 0;
    SoundPointer := nil;
  end;
  if FResourceDLLHandle <> 0 then begin
    FreeLibrary(FResourceDLLHandle);
    FResourceDLLHandle := 0;
  end;
end;

function TabcWaveRes.Play: Boolean;
begin
  Result := False;
  if FResourceHandle <> 0 then Stop;
  if FResourceName = '' then Exit;
  try
    SetResourceHandle;
    SoundPointer := LockResource(FResourceHandle);
    if SoundPointer = nil then
      raise EabcInvalidWaveOperation.Create(Format(sabcErrResLock, [FResourceName]));
    Result := inherited Play;
    if (PlayOption = poSync) or not Result then FreeResourceHandle;
  except
    on Exception do begin
      FreeResourceHandle;
      raise;
    end;
  end;
end;

procedure TabcWaveRes.Stop;
begin
  inherited Stop;
  FreeResourceHandle;
end;

{ TabcWaveResList }

destructor TabcWaveResList.Destroy;
begin
  if FResourceHandle <> 0 then Stop;
  inherited Destroy;
end;

procedure TabcWaveResList.SetResourceHandle(AResName: string);
{
  Loads DLL if required;
  Loads Resource.
  Sets FResourceDLLHandle and FResourceHandle.
}
var
  hResInfo: THandle;
  ErrorMsg: string;
begin
  if FResourceHandle <> 0 then Stop;
  if AResName = '' then Exit;
  try
    if FResourceDLL <> '' then begin
      FResourceDLLHandle := LoadLibrary(PChar(FResourceDLL));
      if FResourceDLLHandle < HINSTANCE_ERROR then begin
        ErrorMsg := Format(sabcErrResDllLoad, [FResourceDLLHandle, FResourceDLL]);
        FResourceDLLHandle := 0;
        raise EabcInvalidWaveOperation.Create(ErrorMsg);
      end;
      hResInfo := FindResource(FResourceDLLHandle, PChar(AResName), RT_RCDATA);
    end
    else
      hResInfo := FindResource(HInstance, PChar(AResName), RT_RCDATA);
    if hResInfo = 0 then
      raise EResNotFound.Create(Format(sabcErrRCDATAResNotFound, [AResName]));
    if FResourceDLLHandle <> 0 then
      FResourceHandle := LoadResource(FResourceDLLHandle, hResInfo)
    else
      FResourceHandle := LoadResource(HInstance, hResInfo);
    if FResourceHandle = 0 then
      raise EResNotFound.Create(Format(sabcErrResLoad, [AResName]));
  except
    on Exception do begin
      FreeResourceHandle;
      raise;
    end;
  end;
end;

procedure TabcWaveResList.FreeResourceHandle;
begin
  if FResourceHandle <> 0 then begin
    UnlockResource(FResourceHandle);
    FreeResource(FResourceHandle);
    FResourceHandle := 0;
    SoundPointer := nil;
  end;
  if FResourceDLLHandle <> 0 then begin
    FreeLibrary(FResourceDLLHandle);
    FResourceDLLHandle := 0;
  end;
end;

function TabcWaveResList.Play: Boolean;
begin
  Result := PlayItem(ItemIndex);
end;

function TabcWaveResList.PlayItem(Index: Integer): Boolean;
begin
  Result := False;
  if (Index < 0) or (Index >= Items.Count) then Exit;
  if Items.Strings[Index] = '' then Exit;
  SetResourceHandle(Items.Strings[Index]);
  try
    SoundPointer := LockResource(FResourceHandle);
    if SoundPointer = nil then
      raise EResNotFound.Create(Format(sabcErrResLock, [Items.Strings[Index]]));
    Result := inherited Play;
    if (PlayOption = poSync) or not Result then FreeResourceHandle;
  except
    on Exception do begin
      FreeResourceHandle;
      raise;
    end;
  end;
end;

procedure TabcWaveResList.Stop;
begin
  inherited Stop;
  FreeResourceHandle;
end;

{ TabcWaveImage }

procedure TabcWaveImage.CopyToClipboard;
begin
  CopyToClipboardFormat(CF_WAVE);
end;

procedure TabcWaveImage.PasteFromClipboard;
begin
  PasteFromClipboardFormat(CF_WAVE);
end;

procedure TabcWaveImage.ValidateDataFormat(Stream: TMemoryStream);
var
  pIOInfo: PMMIOInfo;
  mmckinfoParent: TMMCKInfo;
  mmioHandle: HMMIO;
begin
  mmioHandle := 0;
  pIOInfo := AllocMem(SizeOf(TMMIOInfo));
  try
    with pIOInfo^ do begin
      fccIOProc := FOURCC_MEM;
      pIOProc := nil;
      cchBuffer := Stream.Size;
      pchBuffer := Stream.Memory;
    end;
    mmioHandle := mmioOpen(nil, pIOInfo, MMIO_READ);
    if mmioHandle = 0 then
      raise EabcInvalidWaveOperation.Create(sabcErrWaveData);
    mmckinfoParent.fccType := mmioStringToFOURCC('WAVE', 0);
    if (mmioDescend(mmioHandle, @mmckinfoParent, nil, MMIO_FINDRIFF) <> 0) then
      raise EabcInvalidWaveOperation.Create(sabcErrWaveFormat);
  finally
    if mmioHandle <> 0 then mmioClose(mmioHandle, 0);
    FreeMem(pIOInfo, SizeOf(TMMIOInfo));
  end;
end;

end.
